# Copyright (c) 2006 Red Hat, Inc. All rights reserved. This copyrighted material 
# is made available to anyone wishing to use, modify, copy, or 
# redistribute it subject to the terms and conditions of the GNU General 
# Public License v.2.
# 
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# Author: Greg Nichols
#


from rhcert.tags import Constants, Tags, Attributes
from rhcert.compatability import rhcert_is_string_or_unicode, rhcert_string_or_unicode
import os

class Device:

    def __init__(self, udiSynonym=None, properties=None):
        self.udi = None
        if properties:
            self.properties = properties
            try:
                self.setUDI(properties[udiSynonym])
            except KeyError:
                pass
        else:
            self.properties = dict()
        self.deviceClass = ""

    def getSource(self):
        # abstract class - subclasses MUST define a getSource
        return None

    def getUDI(self):
        return self.udi

    def setUDI(self, udi):
        self.udi = udi

    def getClass(self):
        return self.deviceClass
    def setClass(self, deviceClass):
        self.deviceClass = deviceClass

    def getBus(self):
        return self.getProperty('bus')

    def getProduct(self):
        # abstract class - subclasses MUST define a getProduct
        return None

    def getProperty(self, name):
        try:
            return self.properties[name]
        except KeyError:
            return ""

    def getPropertyString(self, name):
        """ get the property if it's value is a string, or return "" otherwise """
        value = self.getProperty(name)
        return rhcert_string_or_unicode(value)

    def hasPropertyValue(self, hasValue):
        for (key, value) in self.properties.items():
            if rhcert_is_string_or_unicode(value) and hasValue in value:
                return True
        return False

    def getProperties(self):
        return self.properties

    def dump(self):
            output = "\n%s device: %s" % (self.getSource(), self.getUDI())
            output += "\n------------------------------------------------------\n"
            for (key, value) in self.properties.items():
                output += "%s " % key
                output += value
                output += "\n"
            output += "\n"
            return output

    def write(self, file):
            separator = " = " # hal-style
            file.write("udi=%s\n" % self.getUDI())
            keys = self.properties.keys()
            keys.sort()
            for key in keys:
                file.write("  %s%s" % (key, separator))
                if self.properties[key]:
                    file.write("%s" % self.properties[key])
                file.write("\n")
            file.write("\n\n")

class HalDevice(Device):
    def __init__(self, properties=None):
        Device.__init__(self, "info.udi", properties)

    def getSource(self):
        return Constants.hal

    def getProduct(self):
        # try these attributes, in order of preference
        attributes = ['product', 'info.product']
        for attribute in attributes:
            try:
                return self.properties[attribute]
            except KeyError:
                pass

        #otherwise
        return "" 


class ProcDevice(Device):
    def __init__(self, properties=None):
        Device.__init__(self, 'Sysfs', properties)

    def getSource(self):
        return Constants.proc

    def getProduct(self):
        # try these attributes, in order of preference
        attributes = ['Product', 'model name', 'name']
        for attribute in attributes:
            try:
                return self.properties[attribute]
            except KeyError:
                pass

        #otherwise
        return "" 

class UdevDevice(Device):
    def __init__(self, properties=None):
        Device.__init__(self, "DEVPATH", properties)

    def getSource(self):
        return Constants.udev

    def getClass(self):
        return self.getProperty("ID_CLASS")

    def getProduct(self):
        # try these attributes, in order of preference
        attributes = ['ID_MODEL', 'NAME', 'HID_NAME']
        for attribute in attributes:
            try:
                return self.properties[attribute]
            except KeyError:
                pass

        #otherwise
        return "" 

    def getAttr(self, attr):
        # devpath starts with a / so is considered absolute by os.path.join
        attrpath = os.path.join('/sys' + self.getProperty('DEVPATH'), attr)

        try:
            with open(attrpath) as f:
                return f.read().strip()
        except EnvironmentError as e:
            print("Can't read sysfs attr %s for device %s" % (attr, self.getUDI()))
            print(str(e))
            return None

class LogDevice(Device):
    """ Log devices are generic devices for test run hardware logging """
    def __init__(self, properties=None):
        Device.__init__(self, "id", properties)

    def getSource(self):
        return Constants.log

    def getClass(self):
        return self.getProperty("class")

    def getProduct(self):
        # try these attributes, in order of preference
        attributes = ['product', 'name']
        for attribute in attributes:
            try:
                return self.properties[attribute]
            except KeyError:
                pass

        #otherwise
        return "" 
    def setProperty(self, name, value):
        self.properties[name] = value


# still needed as fv_ tests log hardware with kudzu
class KudzuDevice(Device):
    def __init__(self, properties=None):
        Device.__init__(self, None, properties=properties)
        self.udi = self.getProperty("udi")

    def getSource(self):
        return Constants.kudzu

    def getProduct(self):
        return ""

    def write(self, file):
        separator = " : "
        quote = "\'" 
        file.write("\'class\' : \'%s\'\n" % self.getClass())
        keys = self.properties.keys()
        keys.sort()
        for key in keys:
            file.write("  %s%s%s%s" % (quote, key, quote, separator))
            if self.properties[key]:
                file.write("%s%s%s" % (quote, self.properties[key], quote))
            file.write("\n")
        file.write("\n\n")

def getDevices(element):
    """ return a list of devices (rhcert/device.py) """
    deviceElements = element.getElementsByTagName(Tags.device)
    devices = list()
    for deviceElement in deviceElements:
        properties = dict()
        propertyElements = deviceElement.getElementsByTagName(Tags.property)
        for propertyElement in propertyElements:
            value = getTextNode(propertyElement)
            if value:
                properties[propertyElement.getAttribute(Attributes.name)] = value.data
            else:
                properties[propertyElement.getAttribute(Attributes.name)] = None    
        device = createDevice(deviceElement.getAttribute(Attributes.source), properties)

        devices.append(device)
    return devices

def getTextNode(element):
    for node in element.childNodes:
        if node.nodeType == node.TEXT_NODE:
            return node
    return None

def createDevice(source, properties):
    if source == Constants.hal:
        return HalDevice(properties)
    elif source == Constants.proc:
        return ProcDevice(properties)
    elif source == Constants.udev:
        return UdevDevice(properties)
    elif source == Constants.log:
        return LogDevice(properties)
    # might need to parse kudzu devices via FV guest logging
    elif source == Constants.kudzu:
        return KudzuDevice(properties)

    # should never, ever get here!
    print("Error: unknown device source %s" % source)
    raise Exception("Error: unknown device source %s"% source)
