# Copyright (c) 2006 Red Hat, Inc. All rights reserved. This copyrighted material
# is made available to anyone wishing to use, modify, copy, or
# redistribute it subject to the terms and conditions of the GNU General
# Public License v.2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# Author: Greg Nichols
#

from rhcert.tags import Constants, DeviceClass, SystemFunction, PCIDeviceClass, DeviceNumberMajor

class DeviceClassBase():

    def __init__(self):
        self.registrationProperties = None
        self.name = None
        self.function = None

    def hasDevice(self, device):
        if self.registrationProperties:
            for key in self.registrationProperties.keys():
                if device.getProperty(key) == self.registrationProperties[key]:
                    return True
        return False

    def registerDevice(self, device):
        if self.hasDevice(device):
            if self.getName() != "system":
                device.setClass(self.getName())
            return True
        return False

    def getKey(self):
        return self.getName()
    def getName(self):
        return self.name
    def setName(self, name):
        if name:
            self.name = name
    def getFunction(self):
        return self.function
    def setFunction(self, function):
        if function:
            self.function = function
    def getDevicesFromDetector(self, deviceDetector):
        classDevices = list()
        for device in deviceDetector.getDevices():
            if self.hasDevice(device):
                classDevices.append(device)
        return classDevices

class SystemDeviceClass(DeviceClassBase):

    def __init__(self):
        DeviceClassBase.__init__(self)
        self.setName(DeviceClass.system)
        self.setFunction(SystemFunction.system)

    def hasDevice(self, device):
        # take them all!
        return True

class ProcessorDeviceClass(DeviceClassBase):

    def __init__(self):
        DeviceClassBase.__init__(self)
        self.setName(DeviceClass.processor)
        self.setFunction(SystemFunction.system)

    def hasDevice(self, device):
        if (device.getSource() == Constants.hal
             and ('processor' in device.getProperty("info.category")
                or device.getProperty("info.product") == "Computer")):
            return True
        if (device.getSource() == Constants.udev and device.getProperty("DRIVER") == "processor"):
            return True
        return False

class MemoryDeviceClass(DeviceClassBase):

    def __init__(self):
        DeviceClassBase.__init__(self)
        self.setName(DeviceClass.memory)
        self.setFunction(SystemFunction.system)
        self.registrationProperties = dict()
        self.registrationProperties["class"] = "MEMORY"

class BatteryDeviceClass(DeviceClassBase):

    def __init__(self):
        DeviceClassBase.__init__(self)
        self.setName(DeviceClass.battery)
        self.setFunction(SystemFunction.system)

    def hasDevice(self, device):
        if device.getSource() == Constants.hal and 'battery' in device.getProperty("info.category"):
            return True
        if (device.getSource() == Constants.udev
                 and (device.getProperty('POWER_SUPPLY_TYPE') == "Battery"
                      or device.getProperty("DRIVER") in ["battery", "button"]
                      or device.getProperty("NAME") in ["Lid Switch", "Sleep Button"]
                      or device.getProperty("SUBSYSTEM") == "power_supply")):
            return True

        return False

class USBDeviceClass(DeviceClassBase):

    def __init__(self):
        DeviceClassBase.__init__(self)
        self.setName(DeviceClass.usb)
        self.setFunction(SystemFunction.system)
        self.registrationProperties = dict()
        self.registrationProperties["info.bus"] = "usb"
        self.registrationProperties["class"] = "USB"
        self.registrationProperties["SUBSYSTEM"] = "usb"

class PCCardDeviceClass(DeviceClassBase):

    def __init__(self):
        DeviceClassBase.__init__(self)
        self.setName(DeviceClass.pccard)
        self.setFunction(SystemFunction.system)
        self.registrationProperties = dict()
        self.registrationProperties['pci.device_subclass'] = 7
        self.registrationProperties['pci.device_class'] = 6
        self.registrationProperties["class"] = "SOCKET"

class ExpressCardDeviceClass(DeviceClassBase):

    def __init__(self):
        DeviceClassBase.__init__(self)
        self.setName(DeviceClass.expresscard)
        self.setFunction(SystemFunction.system)
        self.registrationProperties = dict()
        self.registrationProperties["info.linux.driver"] = "pcieport-driver"
        self.registrationProperties["DRIVER"] = "pcieport"

class HardDiskDeviceClass(DeviceClassBase):

    def __init__(self):
        DeviceClassBase.__init__(self)
        self.setName(DeviceClass.hard_disk)
        self.setFunction(SystemFunction.storage)

    def hasDevice(self, device):
        if (device.getSource() == Constants.hal
                and ('storage' in device.getProperty("info.capabilities")
                    or device.getProperty("info.linux.driver") == "cciss"
                    # pick up the controller(s) - 1 is Mass Storage
                    or device.getProperty("pci.device_class") == PCIDeviceClass.massStorage)):
                return True

        if device.getSource() == Constants.udev:
            try:
                if int(device.getProperty("MAJOR")) in [DeviceNumberMajor.scsiBlock, DeviceNumberMajor.ideBlock]:
                    return True
            except ValueError:
                # no match if MAJOR is not an integer
                pass

        #otherwise
        return False

class OpticalDeviceClass(DeviceClassBase):

    def __init__(self):
        DeviceClassBase.__init__(self)
        self.setName(DeviceClass.optical)
        self.setFunction(SystemFunction.storage)

    def hasDevice(self, device):
        if device.getSource() == Constants.hal and 'storage.cdrom' in device.getProperty("info.capabilities"):
            return True
        if (device.getSource() == Constants.udev
                and (device.getProperty("DEVTYPE") == 'disk'  and device.getProperty("ID_CDROM") == "1")):
            return True
        return False


class SequentialDeviceClass(DeviceClassBase):

    def __init__(self):
        DeviceClassBase.__init__(self)
        self.setName(DeviceClass.sequential)
        self.setFunction(SystemFunction.storage)

    def hasDevice(self, device):
        if ('tape' in device.getProperty("storage.drive_type") == 7
                or device.getProperty("info.linux.driver") == 'st'
                or device.getProperty("info.linux.driver")[0:4] == 'tape'):
            return True
        return False

class FloppyDeviceClass(DeviceClassBase):

    def __init__(self):
        DeviceClassBase.__init__(self)
        self.setName(DeviceClass.floppy)
        self.setFunction(SystemFunction.storage)

    def hasDevice(self, device):
        if ('floppy' in device.getProperty("storage.drive_type")
                    or  'FLOPPY' in device.getProperty("class")):
            return True
        return False

class NetworkDeviceClass(DeviceClassBase):

    def __init__(self):
        DeviceClassBase.__init__(self)
        self.setName(DeviceClass.network_interface)
        self.setFunction(SystemFunction.networking)

    def hasDevice(self, device):
        if device.getSource() == Constants.hal and "net" in device.getProperty("info.capabilities"):
            return True
        if device.getSource() == Constants.udev and (
                        device.getProperty("SUBSYSTEM") == "net"
                        # infiniband: x060a bridge to IB or x0c06 IB serial controller
                        or device.getProperty("PCI_CLASS") == "60A00" 
                        or device.getProperty("PCI_CLASS") == "C0600"):
            return True

        # otherwise
        return False

class DisplayDeviceClass(DeviceClassBase):

    def __init__(self):
        DeviceClassBase.__init__(self)
        self.setName(DeviceClass.display)
        self.setFunction(SystemFunction.human_interface)
        self.registrationProperties = dict()
        self.registrationProperties["pci.device_subclass"] = 0
        self.registrationProperties["pci.device_class"] = 3
        self.registrationProperties["class"] = "VIDEO"


class AudioDeviceClass(DeviceClassBase):

    def __init__(self):
        DeviceClassBase.__init__(self)
        self.setName(DeviceClass.audio)
        self.setFunction(SystemFunction.human_interface)
        self.registrationProperties = dict()
        self.registrationProperties['linux.device_file'] = '/dev/audio'
        self.registrationProperties['alsa.type'] = 'playback'
        self.registrationProperties['class'] = 'AUDIO'


class InputDeviceClass(DeviceClassBase):

    def __init__(self):
        DeviceClassBase.__init__(self)
        self.setName(DeviceClass.input)
        self.setFunction(SystemFunction.human_interface)
        self.registrationProperties = dict()
        self.registrationProperties['SUBSYSTEM'] = 'INPUT'


def BuildDeviceClasses():
    deviceClasses = [(DeviceClass.system, SystemDeviceClass),
               (DeviceClass.processor, ProcessorDeviceClass),
               (DeviceClass.memory, MemoryDeviceClass),
               (DeviceClass.battery, BatteryDeviceClass),
               (DeviceClass.usb, USBDeviceClass),
               (DeviceClass.pccard, PCCardDeviceClass),
               (DeviceClass.expresscard, ExpressCardDeviceClass),
               (DeviceClass.hard_disk, HardDiskDeviceClass),
               (DeviceClass.optical, OpticalDeviceClass),
               (DeviceClass.sequential, SequentialDeviceClass),
               (DeviceClass.floppy, FloppyDeviceClass),
               (DeviceClass.network_interface, NetworkDeviceClass),
               (DeviceClass.display, DisplayDeviceClass),
               (DeviceClass.audio, AudioDeviceClass),
               (DeviceClass.input, InputDeviceClass)]
    deviceClassInstances = dict()
    for (name, deviceClass) in deviceClasses:
        deviceClassInstance = deviceClass()
        deviceClassInstances[name] = deviceClassInstance
    return deviceClassInstances



if __name__ == "__main__":
    classes = BuildDeviceClasses()
