#!/usr/bin/env python2
# Copyright (c) 2017 Red Hat, Inc. All rights reserved. This copyrighted material 
# is made available to anyone wishing to use, modify, copy, or 
# redistribute it subject to the terms and conditions of the GNU General 
# Public License v.2.
# 
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# Author: Benjamin Berg
#
# Based on code from evemu.
#
"""
The base module provides classes wrapping shared libraries.
"""
import ctypes
import ctypes.util
import os
import glob
import errno

# Import types directly, so they don't have to be prefixed with "ctypes.".
from ctypes import c_char_p, c_int, c_uint, c_void_p, c_long, c_int32, c_uint16, POINTER

import select
import threading
import fcntl

class ExecutionError(Exception):
    pass

def raise_error_if(raise_error, result, func, args):
    """
    Raise an Error for an unexpected result (raise_error == True).

    The exception message includes the API call (name) plus arguments, the
    unexpected result and, if errno is not zero, text describing the
    error number.
    """
    def get_call_str():
        """ Returns a str 'function_name(argument_values...)'. """
        strargs = []
        for (num, arg) in enumerate(func.argtypes):
            # convert args to str for readable output
            if arg == c_char_p:
                strargs.append('"%s"' % args[num].decode("iso8859-1"))
            elif arg == c_void_p:
                strargs.append(hex(int(args[num])))
            else:
                strargs.append(str(args[num]))
        return "%s(%s)" % (func.__name__, ", ".join(strargs))

    def get_retval_str():
        """ Returns a str with the unexpected return value. """
        return ", Unexpected return value: %s" % result

    def get_errno_str():
        """ Returns a str describing the error number or an empty string. """
        errno = ctypes.get_errno()
        if errno != 0:
            return ", errno[%d]: %s" % (errno, os.strerror(errno))
        else:
            return ""

    if raise_error:
        msg = "%s%s%s" % (get_call_str(), get_retval_str(), get_errno_str())
        raise ExecutionError(msg)
    else:
        # If the errcheck function returns the argument tuple it receives
        # unchanged, ctypes continues the normal processing it does on the
        # output parameters.
        return args


def expect_eq_zero(result, func, args):
    """ Expect 'result' being equal to zero. """
    return raise_error_if(result != 0, result, func, args)


def expect_ge_zero(result, func, args):
    """ Expect 'result' being greater or equal to zero. """
    return raise_error_if(result < 0, result, func, args)


def expect_gt_zero(result, func, args):
    """ Expect 'result' being greater then zero. """
    return raise_error_if(result <= 0, result, func, args)


def expect_not_none(result, func, args):
    """ Expect 'result' being not None. """
    return raise_error_if(result is None, result, func, args)


class LibraryWrapper(object):
    """
    Base class for wrapping a shared library.
    """
    _loaded_lib = None
        # Class variable containing the instance returned by CDLL(), which
        # represents the shared library.
        # Initialized once, shared between all instances of this class.

    def __init__(self):
        super(LibraryWrapper, self).__init__()
        self._load()

    # Prototypes for the API calls to wrap. Needs to be overwritten by sub
    # classes.
    _api_prototypes = {
        #"API_CALL_NAME": {
        #    "argtypes": sequence of ARGUMENT TYPES,
        #    "restype": RETURN TYPE,
        #    "errcheck": callback for return value checking, optional
        #    },
        }

    @classmethod
    def _load(cls):
        """
        Returns an instance of the wrapped shared library.

        If not already initialized: set argument and return types on API
        calls and optionally a callback function for return value checking.
        Add the API call as attribute to the class at the end.
        """
        if cls._loaded_lib is not None:
            # Already initialized, just return it.
            return cls._loaded_lib

        # Get an instance of the wrapped shared library.
        cls._loaded_lib = cls._cdll()

        # Iterate the API call prototypes.
        for (name, attrs) in cls._api_prototypes.items():
            # Get the API call.
            api_call = getattr(cls._loaded_lib, name)
            # Add argument and return types.
            api_call.argtypes = attrs["argtypes"]
            api_call.restype = attrs["restype"]
            # Optionally, add a callback for return value checking.
            if "errcheck" in attrs:
                api_call.errcheck = attrs["errcheck"]
            # Add the API call as attribute to the class.
            setattr(cls, name, api_call)

        return cls._loaded_lib

    @staticmethod
    # @abc.abstractmethod - Would be nice here, but it can't be mixed with
    #                       @staticmethod until Python 3.3.
    def _cdll():
        """ Returns a new instance of the wrapped shared library. """
        raise NotImplementedError



class InputEvent(ctypes.Structure):
    _fields_ = [("sec", c_long),
                ("usec", c_long),
                ("type", c_uint16),
                ("code", c_uint16),
                ("value", c_int32)]

    def matches(self, etype, ecode=None):
        etype, ecode = EVDev._to_ids(etype, ecode)

        return etype == self.type and (ecode is None or ecode == self.code)

class EVDevStruct(ctypes.Structure):
    _fileds_ = []

class LibEvdev(LibraryWrapper):
    """
    Wrapper for API calls to libevdev
    """

    GRAB = 3
    UNGRAB = 4

    PROP_CNT = 0x20
    EV_CNT = 0x20

    @staticmethod
    def _cdll():
        return ctypes.CDLL("libevdev.so.2", use_errno=True)

    _api_prototypes = {
        #struct libevdev* libevdev_new(void);
        "libevdev_new": {
            "argtypes": None,
            "restype": POINTER(EVDevStruct),
            "errcheck": expect_not_none
            },
        #void libevdev_free(struct libevdev *dev);
        "libevdev_free": {
            "argtypes": (POINTER(EVDevStruct),),
            "restype": None
            },
        #int libevdev_grab(struct libevdev *dev, enum libevdev_grab_mode grab);
        "libevdev_grab": {
            "argtypes": (POINTER(EVDevStruct), c_int),
            "restype": c_int,
            "errcheck": expect_eq_zero
        },
        #int libevdev_set_fd(struct libevdev* dev, int fd);
        "libevdev_set_fd": {
            "argtypes": (POINTER(EVDevStruct), c_int),
            "restype": c_int,
            "errcheck": expect_eq_zero
        },
        #int libevdev_has_property(const struct libevdev *dev, unsigned int prop);
        'libevdev_has_property': {
            "argtypes": (POINTER(EVDevStruct), c_int),
            "restype": c_int
        },
        #int libevdev_has_event_type(const struct libevdev *dev, unsigned int type);
        'libevdev_has_event_type': {
            "argtypes": (POINTER(EVDevStruct), c_int),
            "restype": c_int
        },
        'libevdev_has_event_code': {
            "argtypes": (POINTER(EVDevStruct), c_int, c_int),
            "restype": c_int
        },
        #int libevdev_next_event(struct libevdev *dev, unsigned int flags, struct input_event *ev);
        "libevdev_next_event": {
            "argtypes": (POINTER(EVDevStruct), c_int, POINTER(InputEvent)),
            "restype": c_int
            },
        #int libevdev_has_event_pending(struct libevdev *dev);
        "libevdev_has_event_pending": {
            "argtypes": (POINTER(EVDevStruct), ),
            "restype": c_int
            },
        #const char* libevdev_get_name(const struct libevdev *dev);
        "libevdev_get_name": {
            "argtypes": (POINTER(EVDevStruct), ),
            "restype": c_char_p
            },
        #int libevdev_get_id_product(const struct libevdev *dev);
        'libevdev_get_id_product': {
            "argtypes": (POINTER(EVDevStruct), ),
            "restype": c_int
            },
        #int libevdev_get_id_vendor(const struct libevdev *dev);
        'libevdev_get_id_vendor': {
            "argtypes": (POINTER(EVDevStruct), ),
            "restype": c_int
            },
        #int libevdev_get_id_bustype(const struct libevdev *dev);
        'libevdev_get_id_bustype': {
            "argtypes": (POINTER(EVDevStruct), ),
            "restype": c_int
            },
        #int libevdev_get_id_version(const struct libevdev *dev);
        'libevdev_get_id_version': {
            "argtypes": (POINTER(EVDevStruct), ),
            "restype": c_int
            },
        #const char *libevdev_event_type_get_name(unsigned int type);
        "libevdev_event_type_get_name": {
            "argtypes": (c_uint,),
            "restype": c_char_p
            },
        #int libevdev_event_type_from_name(const char *name);
        "libevdev_event_type_from_name": {
            "argtypes": (c_char_p,),
            "restype": c_int
            },
        #const char *libevdev_event_code_get_name(unsigned int type, unsigned int code);
        "libevdev_event_code_get_name": {
            "argtypes": (c_uint, c_uint,),
            "restype": c_char_p
            },
        #int libevdev_event_code_from_name(unsigned int type, const char *name);
        "libevdev_event_code_from_name": {
            "argtypes": (c_uint, c_char_p,),
            "restype": c_int
            },
        #const char *libevdev_property_get_name(unsigned int prop);
        "libevdev_property_get_name": {
            "argtypes": (c_uint,),
            "restype": c_char_p
            },
        #int libevdev_property_from_name(const char *name);
        "libevdev_property_from_name": {
            "argtypes": (c_char_p,),
            "restype": c_int
            },
        #int libevdev_get_event_value(const struct libevdev *dev, unsigned int type, unsigned int code);
        "libevdev_get_event_value": {
            "argtypes": (ctypes.c_void_p, ctypes.c_int, ctypes.c_int),
            "restype": ctypes.c_int
            },
        #int libevdev_get_event_value(const struct libevdev *dev, unsigned int type, unsigned int code);
        "libevdev_get_event_value": {
            "argtypes": (ctypes.c_void_p, ctypes.c_int, ctypes.c_int),
            "restype": ctypes.c_int
            },
        #int libevdev_get_abs_minimum(const struct libevdev *dev, unsigned int code);
        "libevdev_get_abs_minimum": {
            "argtypes": (ctypes.c_void_p, ctypes.c_int),
            "restype": ctypes.c_int
            },
        #int libevdev_get_abs_maximum(const struct libevdev *dev, unsigned int code);
        "libevdev_get_abs_maximum": {
            "argtypes": (ctypes.c_void_p, ctypes.c_int),
            "restype": ctypes.c_int
            },
        #int libevdev_get_abs_fuzz(const struct libevdev *dev, unsigned int code);
        "libevdev_get_abs_fuzz": {
            "argtypes": (ctypes.c_void_p, ctypes.c_int),
            "restype": ctypes.c_int
            },
        #int libevdev_get_abs_flat(const struct libevdev *dev, unsigned int code);
        "libevdev_get_abs_flat": {
            "argtypes": (ctypes.c_void_p, ctypes.c_int),
            "restype": ctypes.c_int
            },
        #int libevdev_get_abs_resolution(const struct libevdev *dev, unsigned int code);
        "libevdev_get_abs_resolution": {
            "argtypes": (ctypes.c_void_p, ctypes.c_int),
            "restype": ctypes.c_int
            },
        #int libevdev_event_type_get_max(unsigned int type);
        "libevdev_event_type_get_max": {
            "argtypes": (ctypes.c_int, ),
            "restype": ctypes.c_int
            },
        }

_libevdev = LibEvdev()

def event_code_from_name(etype, ecode):
    return EVDev._to_ids(etype, ecode)[1]

class EVDev(object):
    def __init__(self, f):
        self._evdev = _libevdev.libevdev_new()

        if isinstance(f, basestring):
            f = open(f, 'r')
        self._device = f
        fcntl.fcntl(f.fileno(), fcntl.F_SETFL, fcntl.fcntl(f.fileno(), fcntl.F_GETFL) | os.O_NONBLOCK)
        _libevdev.libevdev_set_fd(self._evdev, self._device.fileno())

    def grab(self):
        _libevdev.libevdev_grab(self._evdev, _libevdev.GRAB)

    def ungrab(self):
        _libevdev.libevdev_grab(self._evdev, _libevdev.UNGRAB)

    def fileno(self):
        """Return the FD number that is relevant for polling for new results."""
        return self._device.fileno()

    def get_name(self):
        return _libevdev.libevdev_get_name(self._evdev)

    @classmethod
    def _to_ids(cls, etype, ecode=None):
        if isinstance(etype, basestring):
            etype = _libevdev.libevdev_event_type_from_name(etype)
        if ecode is not None and isinstance(ecode, basestring):
            ecode = _libevdev.libevdev_event_code_from_name(etype, ecode)
        return etype, ecode

    @classmethod
    def event_type_get_max(cls, etype):
        etype, _t = cls._to_ids(etype)

        return _libevdev.libevdev_event_type_get_max(etype)

    @classmethod
    def property_get_name(cls, prop):
        res = _libevdev.libevdev_property_get_name(prop)
        if res:
            return res
        else:
            return 'PROP_%d' % prop

    @classmethod
    def event_type_get_name(cls, etype):
        etype, _t = cls._to_ids(etype)
        return _libevdev.libevdev_event_type_get_name(etype)

    @classmethod
    def event_code_get_name(cls, etype, ecode):
        etype, ecode = cls._to_ids(etype, ecode)
        return _libevdev.libevdev_event_code_get_name(etype, ecode)

    def get_event_value(self, etype, ecode):
        etype, ecode = self._to_ids(etype, ecode)

        return _libevdev.libevdev_get_event_value(self._evdev, etype, ecode)

    def get_abs_minimum(self, ecode):
        etype, ecode = self._to_ids('EV_ABS', ecode)

        return _libevdev.libevdev_get_abs_minimum(self._evdev, ecode)

    def get_abs_maximum(self, ecode):
        etype, ecode = self._to_ids('EV_ABS', ecode)

        return _libevdev.libevdev_get_abs_maximum(self._evdev, ecode)

    def get_abs_fuzz(self, ecode):
        etype, ecode = self._to_ids('EV_ABS', ecode)

        return _libevdev.libevdev_get_abs_fuzz(self._evdev, ecode)

    def get_abs_flat(self, ecode):
        etype, ecode = self._to_ids('EV_ABS', ecode)

        return _libevdev.libevdev_get_abs_flat(self._evdev, ecode)

    def get_abs_resolution(self, ecode):
        etype, ecode = self._to_ids('EV_ABS', ecode)

        return _libevdev.libevdev_get_abs_resolution(self._evdev, ecode)

    def has_property(self, prop):
        return _libevdev.libevdev_has_property(self._evdev, prop)

    def has_event_type(self, etype):
        etype, _t = self._to_ids(etype, None)

        return _libevdev.libevdev_has_event_type(self._evdev, etype)

    def has_event_code(self, etype, ecode):
        etype, ecode = self._to_ids(etype, ecode)

        return _libevdev.libevdev_has_event_code(self._evdev, etype, ecode)

    def read_events(self):
        while True:
            event = InputEvent()
            # 2 is normal mode processing
            res = _libevdev.libevdev_next_event(self._evdev, 2, event)
            if res >= 0:
                yield event
            if res < 0 and res == -errno.EAGAIN:
                return
            if res < 0:
                raise IOError('Error reading events from libevdev (error: %s)' % os.strerror(-res))



class LoggedEVDev(EVDev):

    def __init__(self, f, log=None, grab=False):
        """
        Initialize an evemu Device for reading events and logging.

        args:
        f -- a file object or filename string for either an existing input
        device node (/dev/input/eventNN) or an evemu prop file that can be used
        to create a pseudo-device node.
        log -- a filename string where logs should be stored.
        If log is None, then a new log file will be created in the current
        directory.
        grab -- if set to True, the input device will be grabbed during
        recording.
        """
        EVDev.__init__(self, f)

        if log is None:
            i = 0
            while True:
                log = '%s-%03d.evemu' % (os.path.basename(f), i)
                i += 1
                if not os.path.exists(log):
                    break

        self._log = open(log, "w+")
        self._start_time = -1

        # Check that the device is not grabbed by anyone else
        self.grab()

        if not grab:
            self.ungrab()

        # Write the description into the log right away
        self._describe(self._log)
        self._log.write('################################\n')
        self._log.write('#      Waiting for events      #\n')
        self._log.write('################################\n')
        self._log.flush()

    def log_event(self, event):
        time = event.sec * 1000000 + event.usec

        if self._start_time == -1:
            self._start_time = time - 1
            self._last_syn = 0

        time = time - self._start_time

        self._log.write('E: %lu.%06u %04x %04x %04d' % (time / 1000000, time % 1000000, event.type, event.code, event.value))

        etype = self.event_type_get_name(event.type)
        ecode = self.event_code_get_name(event.type, event.code)
        if etype == 'EV_SYN':
            if etype == 'SYN_MT_REPORT':
                self._log.write('\t# ++++++++++++ %s (%d) ++++++++++\n' %
                                (ecode, event.value));
            else:
                # Round the times to ms here
                dt = time / 1000 - self._last_syn;
                self._last_syn = time / 1000

                self._log.write('\t# ------------ %s (%d) ---------- %+ldms\n' %
                                (ecode, event.value, dt));
        else:
            self._log.write('\t# %s / %-20s %d\n' % (etype, ecode, event.value))
        self._log.flush()


    def read_events(self):
        for event in EVDev.read_events(self):
            self.log_event(event)
            yield event

    def _describe(self, f):
        # Write the header
        f.write('# EVEMU 1.3 compatible file created by Red Hat hwcert\n')
        f.write('# Kernel: %s\n' % os.uname()[2])
        try:
            dmi = open('/sys/class/dmi/id/modalias', 'r').read()
        except IOError:
            pass
        finally:
            f.write('# %s' % dmi)

        f.write('N: %s\n' % self.get_name())
        f.write('I: %04x %04x %04x %04x\t# bus, vendor, product, version\n' % (
                _libevdev.libevdev_get_id_bustype(self._evdev),
                _libevdev.libevdev_get_id_vendor(self._evdev),
                _libevdev.libevdev_get_id_product(self._evdev),
                _libevdev.libevdev_get_id_version(self._evdev),
                ))

        props = []
        f.write('P:')
        mask = 0
        for i in xrange(max(8*8, _libevdev.PROP_CNT)):
            if (i % 8) == 0 and i > 0:
                f.write(' %02x' % mask)
                mask = 0

            if self.has_property(i):
                mask |= 1 << (i % 8)
                props.append(self.property_get_name(i))
        else:
            f.write(' %02x' % mask)
        f.write('\n')
        if props:
            f.write('# Props:\n')
        for prop in props:
            f.write('# * %s\n' % prop)

        # EV_SYN/EV_SYN_REPORT hardcoded as in evemu
        f.write('B: 00 0b 00 00 00 00 00 00 00\n')
        for t in xrange(1, _libevdev.EV_CNT):
            count = self.event_type_get_max(t)
            if count == -1:
                continue

            codes = []
            mask = 0
            f.write('B: %02x' % t)
            for c in xrange(count + 1):
                if c % 8 == 0 and c > 0:
                    f.write(' %02x' % mask)
                    mask = 0
                    if c % (8*8) == 0:
                        f.write('\nB: %02x' % t)

                if self.has_event_code(t, c):
                    mask |= 1 << (c % 8)
                    codes.append(c)
            else:
                f.write(' %02x' % mask)
                for c in xrange(8 - (count % 64 + 7) / 8):
                    f.write(' 00')
                f.write('\n')

            if codes:
                f.write('# Event type %i (%s):\n' % (t, self.event_type_get_name(t)))
                for c in codes:
                    f.write('#   Event code %i (%s):\n' % (c, self.event_code_get_name(t, c)))

        # EV_ABS
        count = self.event_type_get_max('EV_ABS')
        for i in xrange(0, count):
            if not self.has_event_code('EV_ABS', i):
                continue

            f.write('A: %02x %d %d %d %d %d\n' %
                    (i,
                    self.get_abs_minimum(i), self.get_abs_maximum(i),
                    self.get_abs_fuzz(i), self.get_abs_flat(i),
                    self.get_abs_resolution(i)
                    ))

        # EV_LED
        count = self.event_type_get_max('EV_LED')
        for i in xrange(0, count):
            if not self.has_event_code('EV_LED', i):
                continue

            state = self.get_event_value('EV_LED', i) != 0
            f.write('L: %02x %d\t# %s\n' % (i, int(state), self.event_code_get_name('EV_LED', i)))

        # EV_SW
        count = self.event_type_get_max('EV_SW')
        for i in xrange(0, count):
            if not self.has_event_code('EV_SW', i):
                continue

            state = self.get_event_value('EV_SW', i) != 0
            f.write('S: %02x %d\t# %s\n' % (i, int(state), self.event_code_get_name('EV_SW', i)))

class RecordedEVDev(LoggedEVDev):
    def __init__(self, f, log=None, grab=False):
        LoggedEVDev.__init__(self, f, log, grab)

        self._stopped = False
        self._thread = None
        self._events = []

    def read_events(self):
        for event in LoggedEVDev.read_events(self):
            self._events.append(event)
            yield event

    def _thread_func(self):
        poll_fds = [self.fileno()]
        while not self._stopped:
            r, w, x = select.select(poll_fds, [], poll_fds, 0.5)
            if x:
                break
            for e in self.read_events():
                pass

    def start_thread(self):
        assert self._thread is None
        self._thread = threading.Thread(target=self._thread_func)
        self._thread.start()

    def stop_thread(self):
        self._stopped = True
        self._thread.join()
        self._thread = None

    def events(self):
        for event in self._events:
            yield event

def list_devices():
    return list(glob.glob('/dev/input/event*'))

if __name__ == '__main__':
    import sys

    args = sys.argv[1:]

    grab = False
    if '-g' in args:
        grab = True
        args.remove('-g')

    assert(len(args) == 2)

    d = DeviceRecorder(args[0], args[1], grab)

    poll_fds = [d.fileno()]
    while True:
        r, w, x = select.select(poll_fds, [], poll_fds)
        if x:
            print('Error!')
            break

        for event in d.read_events():
            pass
            #print str(event)
            #print event.type, evemu.event_get_name(event.type, event.code), event.value



