#!/usr/bin/env python
# Copyright (c) 2017 Red Hat, Inc. All rights reserved. This copyrighted material 
# is made available to anyone wishing to use, modify, copy, or 
# redistribute it subject to the terms and conditions of the GNU General 
# Public License v.2.
# 
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# Author: Benjamin Berg
#

import sys
import select

from rhcert.tags import DeviceClass, TestTag, TestResult
from rhcert.suites.hwcert.hwcert.hardwareTest import HardwareTest
from rhcert.suites.hwcert.hwcert.udevDeviceDetector import UdevDeviceDetector
from rhcert.redhatrelease import RedHatRelease

class EVDevKeysTest(HardwareTest):

    TEST_KEYS = 'special keys'
    FINAL_LIST_HEADER = 'Special keys detected:'
    FINAL_LIST_FOOTER = 'Were all special keys seen with the correct purpose?'

    acpi_keys = [ 'POWER', 'SUSPEND', 'SLEEP' ]
    allow_combinations = True

    modifier_keycodes = {
            "KEY_LEFTALT"      : 'Alt',
            "KEY_LEFTCTRL"     : 'Ctrl',
            "KEY_LEFTMETA"     : 'Meta',
            "KEY_LEFTSHIFT"    : 'Shift',

            "KEY_RIGHTALT"     : 'Alt',
            "KEY_RIGHTCTRL"    : 'Ctrl',
            "KEY_RIGHTMETA"    : 'Meta',
            "KEY_RIGHTSHIFT"   : 'Shift',
        }

    # The meaning of some keys is a bit cryptic, so give them a saner description
    name_override = {
        'F20' : 'F20_MICMUTE',
        'MIN_INTERESTING' : 'MUTE',
        'SLEEP' : 'SLEEP (suspend to ram)',
        'SUSPEND' : 'SUSPEND (suspend to disk)',
    }

    # Combinations that are known to be good
    good_combinations = {
        '<Meta>P' : 'display output switch',
        '<Alt>SYSRQ' : 'system request',
        '<Ctrl>PAUSE' : 'break',
    }

    def __init__(self, name):
        HardwareTest.__init__(self, name)
        self.interactive = True
        self.deviceClass = DeviceClass.input
        self.redHatRelease = RedHatRelease()

    def tags(self):
        return [TestTag.interactive, TestTag.certification]

    def getRequiredRPMs(self):
        return ['libevdev']

    def is_tested(self, scancode):
        raise NotImplemented

    def is_special(self, scancode):
        if evdev.event_code_from_name("EV_KEY", "KEY_MUTE") <= scancode <= evdev.event_code_from_name("EV_KEY", "KEY_POWER"):
            return True
        if evdev.event_code_from_name("EV_KEY", "KEY_SCALE") <= scancode <= evdev.event_code_from_name("EV_KEY", "KEY_SCALE"):
            return True
        if evdev.event_code_from_name("EV_KEY", "KEY_STOP") <= scancode <= evdev.event_code_from_name("EV_KEY", "KEY_EDIT"):
            return True
        if evdev.event_code_from_name("EV_KEY", "KEY_NEW") <= scancode <= evdev.event_code_from_name("EV_KEY", "KEY_REDO"):
            return True
        # F13 - F24
        if 183 <= scancode <= 194:
            return True
        if evdev.event_code_from_name("EV_KEY", "KEY_PLAYCD") <= scancode <= evdev.event_code_from_name("EV_KEY", "KEY_MICMUTE"):
            return True
        if evdev.event_code_from_name("EV_KEY", "KEY_OK") <= scancode <= evdev.event_code_from_name("EV_KEY", "KEY_LOGOFF"):
            return True
        if evdev.event_code_from_name("EV_KEY", "KEY_FRAMEBACK") <= scancode:
            # Note, includes:
            #  - Diagonal movement keys
            #  - Dollar/Euro keys
            #  - Numeric pads (phone like)
            #  - Generic Fn keys
            #  - ...
            return True

        return False

    def is_acpi(self, scancode):
        for key in self.acpi_keys:
            if evdev.event_code_from_name("EV_KEY", 'KEY_' + key) == scancode:
                return True

        return False

    def format_keycode(self, code):
        # Make the names more human readable
        if code in self.name_override:
            code = self.name_override[code]

        # Is this a key combination?
        if code and code[0] == "<":
            if code not in self.good_combinations:
                code = code + ' (unknown combination)'
            else:
                code = code + ' (%s)' % self.good_combinations[code]

        return code

    def process_event(self, device, event, result):
        """Processes an event updating the global detection list and recording
        state. Will return True if anything of interest happened (i.e. the
        detection list was updated)."""

        if not event.matches('EV_KEY'):
            return False

        # Ignore everything other key down and up
        if event.value > 1:
            return False

        keycode = device.event_code_get_name(event.type, event.code)

        # Ignore key events from mouse/touchpad
        if keycode.startswith('BTN_'):
            return False

        # Do some basic logging of information
        print("Saw %s on device %s" % (event, device.get_name()))

        # Stop recording if the user hits ESC
        if 'KEY_ESC' == keycode:
            self.recording = False
            return False

        if keycode in self.modifier_keycodes:
            # Keep track of modifiers
            # key down
            if event.value == 1:
                self.active_modifiers.add(keycode)
            else:
                if keycode in self.active_modifiers:
                    self.active_modifiers.remove(keycode)
                else:
                    print('Ignoring modifier inactivation from {:s}'.format(str(event)))

            # Ignore any modifier key press
            return False

        modifiers = set()
        for mod in self.active_modifiers:
            modifiers.add(self.modifier_keycodes[mod])

        non_modifier_key = self.is_tested(event.code)
        if non_modifier_key:
            modifiers = set()

        if non_modifier_key or (modifiers and self.allow_combinations):
            # Only use the first code from this point onwards
            if keycode.startswith('KEY_'):
                keycode = keycode[4:]

            if modifiers:
                for modifier in modifiers:
                    keycode = ('<%s>' % modifier) + keycode

                if keycode not in self.good_combinations:
                    result.combine(TestResult.REVIEW)

            print('Detected key (combination) %s' % keycode)

            # Make sure the new detection is in the list and at the far bottom.
            if keycode in self.detections:
                self.detections.remove(keycode)
            self.detections.append(keycode)

            return True

        return False

    def record(self):
        result = TestResult()
        evdevs = []
        fds = []
        self.detections = []
        self.active_modifiers = set()
        self.recording = True
        redisplay = True

        try:
            self.statusMessage('Opening input devices')
            for devnode in evdev.list_devices():
                try:
                    device = evdev.LoggedEVDev(devnode)
                    device.devnode = devnode
                except OSError:
                    # This happens e.g. on the X1 Yoga on RHEL 7.4 with the accelerometer
                    print("Warning: Could not open device %s" % device)
                    result.combine(TestResult.WARN)
                    continue

                evdevs.append(device)
                fds = dict([(device.fileno(), device) for device in evdevs ])
                print("Grabbing all input from device %s" % device.get_name())
                device.grab()

            # This loop is stopped from inside process_event
            while self.recording:
                # We don't have an async handler to find new devices,
                # so probe every 0.5 seconds.

                if redisplay:
                    redisplay = False
                    # Got a new detection, inform the user about it
                    msg = 'Keys detected so far are (newest last):\n'
                    for code in self.detections:
                        msg += ' * %s\n' % self.format_keycode(code)
                    msg += '\nPress "Escape" on the SUT to stop the capture. '

                    missing_common_keys = []
                    for c in self.common_keys:
                        if not c in self.detections:
                            missing_common_keys.append(c)
                    if missing_common_keys:
                        msg += 'Please ensure to test the following keys if they exist:\n'
                        for c in missing_common_keys:
                            msg += '* %s\n' % self.format_keycode(c)

                    self.statusMessage(msg)


                r, w, x = select.select(fds.keys(), [], fds.keys(), 0.5)
                for fd in r + x:
                    try:
                        for event in fds[fd].read_events():
                            if self.process_event(fds[fd], event, result):
                                redisplay = True
                    except EnvironmentError as e:
                        print('Exception: %s' % str(e))
                        print('Device %s disappeared' % fds[fd].get_name())
                        evdevs.remove(fds[fd])
                        del fds[fd]

                for devnode in evdev.list_devices():
                    for d in fds.values():
                        if d.devnode == devnode:
                            break
                    else:
                        # FD is not registered
                        try:
                            device = evdev.LoggedEVDev(devnode, grab=True)
                            device.devnode = devnode
                        except OSError:
                            # We probably warned already above, so ignore this here
                            continue

                        print("New device %s detected" % device.get_name())
                        evdevs.append(device)
                        fds[device.fileno()] = device

        except Exception as e:
            import traceback
            print(traceback.format_exc())
            print('Error: %s' % str(e))
            result.combine(TestResult.FAIL)
        finally:
            for device in evdevs:
                try:
                    device.ungrab()
                except:
                    pass

        return result

    def run(self):
        # Import packages requiring libevdev now that the package is known to be
        # installed so that we can use it everywhere.
        global evdev
        from rhcert.suites.hwcert.hwcert import evdev

        result = TestResult()

        # At the core of this test we grab all input from input devices
        # (to prevent e.g. suspend/power down actions to take place). This
        # means that normal input to confirm is also blocked during that
        # timeframe.
        # As it is quite cumbersome to confirm each special key the concept
        # here is to create a list of all seen special keys until the tester
        # presses the "Escape" key on the SUT. Afterwards the entire result
        # needs to be confirmed to be accurate.

        if not self.promptConfirm("Testing %s requires capturing all input events. "
                                  "This means that *no* input event will be processed "
                                  "by the computer until you press the Escape key "
                                  "on the SUT to stop the capture! Continue?" % self.TEST_KEYS):
            result.combine(TestResult.FAIL)
            return result

        # Start the recording
        result.combine(self.record())

        if self.detections:
            msg = self.FINAL_LIST_HEADER + '\n'
            for code in self.detections:
                msg += ' * %s\n' % self.format_keycode(code)

            msg += '\n' + self.FINAL_LIST_FOOTER

            if not self.promptConfirm(msg):
                # Not all missing keys should result in a failure, so don't
                # scare testers by reporting a FAIL.
                result.combine(TestResult.REVIEW)

                self.promptEdit('Please specify the error (i.e. missing key events, wrong assignment)')
        else:
            msg = 'No keys were detected! Are there really no %s including the hardware power button?' % self.TEST_KEYS
            result.combine(TestResult.REVIEW)

            if not self.promptConfirm(msg):
                result.combine(TestResult.FAIL)

        return result


