#!/usr/bin/env python
# Copyright (c) 2020 Red Hat, Inc. All rights reserved. This copyrighted material
# is made available to anyone wishing to use, modify, copy, or
# redistribute it subject to the terms and conditions of the GNU General
# Public License v.2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# Author: Petersingh Anburaj
#

import io
import os
import re
import time
import sys
import xml.etree.ElementTree as ET

from rhcert.compatability import rhcert_make_str
from rhcert.tags import TestTag, Constants
from rhcert.command import Command, HwCertCommandException
from rhcert.suites.hwcert.hwcert.fvtest import FvTest


class FvPassthrough(FvTest):
    def __init__(self, path):
        FvTest.__init__(self, path)
        self.req_passthrough = True
        self.interactive = True
        self.device_xml_file = None
        self.private_key_file = None
        self.guestIPaddr = None
        self.domain = None
        self.bus = None
        self.slot = None
        self.function = None
        self.vendor_id = None
        self.product_id = None
        self.device = None

    def plan(self, planner):
        empty_test = []

        # Execute FV test plan command and get tests
        tests = FvTest.plan(self, planner)

        # Check if OS is RHEL8 and above
        if self.redHatRelease.getVersion() < 8:
            if self.getDebug() != Constants.off:
                print("Passthrough test is not supported for this OS.")
            return empty_test
        return tests

    def tags(self):
        return [TestTag.interactive, TestTag.certification, TestTag.virtualization]

    def getRequiredRPMs(self):
        rpms = FvTest.getRequiredRPMs(self)
        if sys.version_info.major > 2:
            rpms.append("python3-cryptography")
        else:
            rpms.append("python2-cryptography")
        return rpms

    def get_ssh_cmd(self, cmd):
        """
        This function converts the cmd to its corresponding SSH command
        eg:
        cmd: `uptime`
        SSH cmd: `ssh -o 'StrictHostKeyChecking no' -i <private-key-file> root@<vm-ip-address> 'uptime'`
        """
        return "ssh -o 'StrictHostKeyChecking no' -i %s root@%s '%s'" % (self.private_key_file, self.guestIPaddr, cmd)

    def update_pci_device_details(self, device_id):
        """
        :param device_id: device ID eg: 0000:00:1f.6

        Fetch the device details using lspci command
        eg:  lspci -nDs 0000:00:1f.6 => 0000:00:1f.6 0200: 8086:15bd (rev 30)

        (domain) : (bus) : (slot) . (function)   (device class) :  (vendor ID) : (product ID)
          0000   :  00   :   1f   .     6            0200       :      8086    :     15bd

         Update the domain, bus, function, vendor_id, product_id parameters with devices details
        """
        print("Fetching the details of the device ...")
        print("Given device ID is %s" % device_id)
        if not device_id:
            print("Error: PCI device ID not passed !!!")
            return False

        lspci_cmd = Command("lspci -nDs %s" % device_id)
        try:
            lspci_ns_output = lspci_cmd.getString(regex=".*")
        except HwCertCommandException as e:
            print("Error: Failed to get the device details !!!")
            return False

        # lspci_ns_output => `0000:00:1f.6 0200: 8086:15bd (rev 30)`
        # lspci_ns_list => ['0000:00:1f.6', '0200:', '8086:15bd', '(rev', '30)']
        lspci_ns_list = lspci_ns_output.split()

        # domain_bus_slot_func = `0000:00:1f.6`
        domain_bus_slot_func = lspci_ns_list[0]
        # domain => `0000`
        domain = domain_bus_slot_func.split(":")[0]
        # bus = `00`
        bus = domain_bus_slot_func.split(":")[1]

        # slot_func = `1f.6`
        slot_func = domain_bus_slot_func.split(":")[2]
        # slot = `1f`
        slot = slot_func.split(".")[0]
        # func = `6`
        func = slot_func.split(".")[1]

        # vendor_product = `8086:15bd`
        vendor_product = lspci_ns_list[2]
        # vendor_id = `8086`
        vendor_id = vendor_product.split(":")[0]
        # product_id = `15bd`
        product_id = vendor_product.split(":")[1]

        print("Device details: \n  domain (%s), bus (%s), slot (%s), function (%s)" % (domain, bus, slot, func))
        print("  vendor ID (%s), product ID (%s)" % (vendor_id, product_id))

        # Converting the values to proper hex representation and storing it
        # eg: slot = `1f`, int(slot, 16) = 31, hex(int(slot, 16)) = `0x1f`
        self.domain, self.bus = hex(int(domain, 16)), hex(int(bus, 16))
        self.slot, self.function = hex(int(slot, 16)), hex(int(func, 16))
        self.vendor_id, self.product_id = hex(int(vendor_id, 16)), hex(int(product_id, 16))
        return True

    def update_usb_device_details(self, device_name):
        """
        :param device_name: Device name eg: `Bus 003 Device 010: ID 058f:6387 Alcor Micro Corp. Flash Drive`

        Parse the device name and update the parameters bus, device, vendor_id, product_id
        """
        # device_name = `Bus 003 Device 010: ID 058f:6387 Alcor Micro Corp. Flash Drive`
        # bus = `003`
        bus = device_name.split()[1]
        # device = `010`
        device = device_name.split()[3][:-1]
        self.bus, self.device = int(bus), int(device)

        # vendor_product = `058f:6387`
        vendor_product = device_name.split()[5]

        # vendor = `058f`
        vendor = vendor_product.split(":")[0]
        # product = `6387`
        product = vendor_product.split(":")[1]

        # Converting the values to proper hex representation and storing it
        # eg: vendor = `058f`, int(vendor, 16) = 1423, hex(int(vendor, 16)) = `0x58f`
        self.vendor_id, self.product_id = hex(int(vendor, 16)), hex(int(product, 16))
        return True

    def get_usb_device_driver(self):
        """
        This function is use to fetch the driver name for a USB device
        It parses the `lsusb -t` output to get the driver name for USB device
        """

        lsusb_tree_cmd = Command("lsusb -t")
        try:
            lsusb_tree = lsusb_tree_cmd.getStringList(regex=".*")
        except HwCertCommandException:
            print("Error: Failed to fetch USB device driver !!!")
            return None

        driver, device_id = self.parse_usb_device_tree(output=lsusb_tree, bus=self.bus,
                                                       device=self.device, only_driver=True)
        return driver

    def parse_usb_device_tree(self, output, bus, device, only_driver=False):
        """
        :param output: `lsusb -t` output as a string list
        :param bus: bus number
        :param device: device number
        :param only_driver: if set to True will return only driver and not device_id

        This function is used to parse the output of `lsusb -t`
        It returns driver and device_id for given bus and device number
        """
        bus_search = re.compile("Bus (?P<bus>\\d+)\\.")
        driver_search = re.compile("Driver=(?P<driver>.*),")
        port_search = re.compile("Port (?P<port>\\d+):")
        driver = None
        bus_found = False
        rev_list = []
        dev = "Dev %s" % device
        for line in output:
            if 'Bus' in line:
                if bus_found:
                    break
                bus_id = bus_search.search(line).group("bus")
                if bus == int(bus_id):
                    bus_found = True
                    print(line)
                    continue
            if bus_found:
                print(line)
                rev_list.insert(0, line)
                if dev in line:
                    driver = driver_search.search(line).group("driver")
                    break
        if only_driver:
            return driver, None

        old_level = 100
        ports = []
        device_id = None
        for line in rev_list:
            level = len(line) - len(line.strip())
            if level < old_level:
                port = port_search.search(line).group("port")
                ports.insert(0, port)
                old_level = level

        if not ports:
            return driver, device_id
        port_id = ".".join(ports)
        device_id = "%s-%s" % (bus, port_id)
        print("Device ID: %s" % device_id)
        return driver, device_id

    def create_device_xml(self, dev_type="pci"):
        """
        :param dev_type: can be `pci` or `usb`
        :returns: True if XML file created successfully else False

        This function is used to create device XML file in below format:
        For dev_type='pci':
        <hostdev mode="subsystem" type="pci" managed="yes">
            <source>
                <address domain="0x0" bus="0x0" slot="0x1f" function="0x6"/>
            </source>
        </hostdev>

        For dev_type='usb':
        <hostdev mode="subsystem" type="usb" managed="yes">
            <source>
                <vendor id="0x058f"/>
                <product id="0x6387"/>
                <address bus='2' device='3'/>
            </source>
        </hostdev>
        The XML file is created in /tmp/ directory
        """
        print("Creating the device XML file ...")
        file = "/tmp/%s_device.xml" % dev_type
        try:
            hostdev = ET.Element("hostdev")
            hostdev.set("mode", "subsystem")
            hostdev.set("type", dev_type)
            hostdev.set("managed", "yes")

            source = ET.SubElement(hostdev, "source")

            if dev_type == "pci" and self.domain and self.bus and self.slot and self.function:
                address = ET.SubElement(source, "address")
                address.set("domain", self.domain)
                address.set("bus", self.bus)
                address.set("slot", self.slot)
                address.set("function", self.function)

            elif dev_type == "usb" and self.vendor_id and self.product_id and self.bus and self.device:
                vendor = ET.SubElement(source, "vendor")
                vendor.set("id", self.vendor_id)
                product = ET.SubElement(source, "product")
                product.set("id", self.product_id)
                address = ET.SubElement(source, "address")
                address.set("bus", str(self.bus))
                address.set("device", str(self.device))

            else:
                print("Error: Required device information not found !!")
                return False

            # Write XML to file
            print("Writing the device XML file %s" % file)
            file_content = ET.tostring(hostdev)
            print("File Content: \n%s" % rhcert_make_str(file_content))
            with io.open(file, "w") as xml_file:
                xml_file.write(rhcert_make_str(file_content))
        except Exception as e:
            print("Error: Failed to create %s pass-through device XML" % dev_type)
            print("Error: %s" % str(e))
            return False
        self.device_xml_file = file
        return True

    def update_vm_device(self, cmd="attach"):
        """
        :param cmd: can be 'attach' to attach a new device or 'detach' to remove a device from VM config
        :return: True if device attach/detach was successful else False

        This function is used either to attach a new device to VM or remove existing device from VM.
        """
        print("Updating VM configuration with device XML ...")
        if cmd not in ("attach", "detach"):
            print("Error: Wrong command passed !!!")
            return False

        virsh_device = Command("virsh %s-device %s %s --persistent" % (cmd, self.guestName, self.device_xml_file))
        try:
            virsh_device.run()
            print("Successfully %sed device." % cmd)
        except HwCertCommandException as e:
            if cmd == "attach" and "device is already in the domain" in e.command.errors[-1]:
                print("Warning: Device is already attached to domain.")
            elif cmd == "detach" and "device not present in domain" in e.command.errors[-1]:
                print("Warning: Device is not present in the domain.")
            else:
                print("Error: Failed to %s the device !!!" % cmd)
                e.command.printErrors()
                return False
        return True

    def get_passthrough_devices(self, dev_type="pci"):
        """
        :param dev_type: can be `pci` or `usb`
        :return: List of passthrough devices

        This function is used to get the list of available passthrough devices.
        It can be pci devices or usb devices.
        """
        print("Fetching %s passthrough device list ..." % dev_type)
        device_list = []

        if dev_type == "pci":
            device_get_cmd = "lspci -D | grep -v 'PCI bridge'"
        elif dev_type == "usb":
            device_get_cmd = "lsusb | grep -v 'root hub'"
        else:
            print("Error: Wrong device type passed !!!")
            return device_list

        try:
            device_list = Command(device_get_cmd).getStringList(regex=".*")
        except HwCertCommandException as e:
            print("Error: Failed to get the device list !!!")

        return device_list

    def __create_keys(self):
        """
        Create a new pair of private and public key
        :returns: private and public key
        """
        from cryptography.hazmat.primitives import serialization
        from cryptography.hazmat.primitives.asymmetric import rsa
        from cryptography.hazmat.backends import default_backend

        print("Creating private and public SSH keys ...")
        ssh_key = rsa.generate_private_key(backend=default_backend(), public_exponent=65537, key_size=2048)

        private_key = ssh_key.private_bytes(encoding=serialization.Encoding.PEM,
                                            format=serialization.PrivateFormat.PKCS8,
                                            encryption_algorithm=serialization.NoEncryption())
        public_key = ssh_key.public_key().public_bytes(serialization.Encoding.OpenSSH,
                                                       serialization.PublicFormat.OpenSSH)
        print("Keys created successfully !!!")
        return private_key, public_key

    def setup_password_less_ssh_connection(self):
        """
        Configure password-less ssh for VM and update the private_key_file parameter
        """
        print("Setting up password-less SSH for VM ...")
        private_key, public_key = self.__create_keys()

        # Write private key to the file
        print("Writing the private key ...")
        private_key_file = "/tmp/%s_private_key" % self.guestName
        with io.open(private_key_file, "w") as pkf:
            pkf.write(rhcert_make_str(private_key))

        # Write public key to authorized_keys file
        print("Writing the public key ...")
        authorized_keys_file = "/tmp/authorized_keys"
        with io.open(authorized_keys_file, "w") as akf:
            akf.write(rhcert_make_str(public_key))

        # Update permissions of the file
        print("Updating the permission of the key files ...")
        os.chmod(private_key_file, 0o600)
        os.chmod(authorized_keys_file, 0o600)

        print("Copying the public key into VM ...")
        # Create .ssh directory if it is not present
        if not self.virtCommand("mkdir-p /root/.ssh"):
            print("Error: Failed to create SSH directory inside the VM !!!")
            return False

        # Copy authorized_keys file into VM
        if not self.virtCommand("copy-in %s /root/.ssh/" % authorized_keys_file):
            print("Error: Failed to copy SSH key into the VM !!!")
            return False

        # Update selinux context of the file
        if not self.virtCommand(
                "selinux-relabel /etc/selinux/targeted/contexts/files/file_contexts /root/.ssh/authorized_keys"):
            print("Warning: Failed to relabel the authorized_keys file in VM !!!")
            print("Trying to update rc.local for setting SELinux context ...")
            # selinux-relabel command fails for the current RHEL6 VM image,
            # due to a know bug with `selinux-policy-targeted` package.
            # Further details can be sees in CERTPX-5957
            # Adding `restorecon` command in rc.local file for setting correct context for authorized_keys file
            if not self.virtCommand("write /etc/rc.d/rc.local 'restorecon /root/.ssh/authorized_keys'"):
                print("Error: Failed to update rc.local for setting SELinux context !!!")
                return False
            # updating the permission of the rc.local file
            self.virtCommand("chmod 0755 /etc/rc.d/rc.local")
            print("Updated rc.local successfully !!!")

        print("Successfully configured password-less SSH !!! \nPrivate key can be found at: %s" % private_key_file)
        self.private_key_file = private_key_file
        return True

    def start_vm_update_ip_address(self):
        """
        Start the VM, get VMs IP address and update guestIPaddr parameter
        """
        print("Starting the guest VM ...")
        try:
            vm_start_cmd = Command("virsh start %s" % self.guestName)
            vm_start_cmd.run()
        except HwCertCommandException as e:
            print("Error: Failed to start FV guest !!!")
            e.command.printErrors()
            return False

        # Wait for VM to start
        sys.stdout.flush()
        time.sleep(5)

        # Check if VM is in running state
        running_vms = []
        try:
            vm_running_cmd = Command("virsh list --state-running --name")
            running_vms = vm_running_cmd.getStringList(regex=".*")
        except HwCertCommandException as e:
            print("Error: Failed to get running VMs list !!!")
            return False

        if self.guestName in running_vms:
            print("VM started successfully !!!")

        retry = 10
        count = 1
        print("Trying to fetch VM interface IP address ...")
        interface_ip_address = None
        while count <= retry:
            print("Attempt No. %s ..." % count)
            sys.stdout.flush()
            time.sleep(30)
            # Get guest VM MAC addresses
            get_vm_macs = Command("virsh dumpxml %s | grep 'mac address' | awk -F\\' '{ print $2}'" % self.guestName)
            try:
                vm_macs = get_vm_macs.getStringList(regex=".*")
            except HwCertCommandException as e:
                print("Failed to get VM MAC addresses")
                count += 1
                continue

            # Get IP address of the VM
            for vm_mac in vm_macs:
                get_interface_ip = Command("ip neigh show | grep %s | awk '{print $1}'" % vm_mac)
                try:
                    interface_ip_address = get_interface_ip.getString(regex=".*")
                    print("VM interface IP address is: %s" % interface_ip_address)
                    break
                except HwCertCommandException as e:
                    pass

            if not interface_ip_address:
                print("Failed to get interface IP address")
                count += 1
                continue
            break

        if not interface_ip_address:
            print("Error: Failed to get VM interface IP address !!! ")
            return False

        self.guestIPaddr = interface_ip_address
        return True

    def check_vm_ssh_config(self):
        """
        Check if password-less ssh is working for the VM
        This is done by getting the output of `uptime` command from the VM
        :return: True if it is working else False
        """
        print("Checking password-less SSH configuration ...")
        print("Trying to get uptime of the VM ...")
        if not self.private_key_file or not os.path.isfile(self.private_key_file):
            print("Error: Unable to find private key file")
            return False
        retry = 8
        count = 1
        uptime = None
        while count <= retry:
            print("Attempt No. %s ..." % count)
            sys.stdout.flush()
            uptime_cmd = Command(self.get_ssh_cmd("uptime"))
            try:
                uptime = uptime_cmd.getString(regex=".*")
                print("Uptime of the VM is '%s'" % uptime)
                print("Password-less SSH verification was successful !!!")
                break
            except HwCertCommandException as e:
                time.sleep(30)
                print("Failed to get uptime of the VM")
                count += 1
        if not uptime:
            print("Error: Failed to establish SSH connection to the VM !!!")
            return False
        return True

    def stop_vm(self):
        """
        Stop the VM and confirm that it has stopped
        :return: True if VM stopped running else False
        """
        # Stop the VM
        print("Shutting down the VM ...")
        stop_vm_cmd = Command("virsh destroy %s" % self.guestName)
        try:
            stop_vm_cmd.run()
        except HwCertCommandException as e:
            print("Warning: Failed to stop the VM !!!")
            e.command.printErrors()

        # Check if VM is shutdown
        retry = 5
        count = 0
        vm_shutdown = False
        while count <= retry:
            shutdown_vms = []
            try:
                check_shutdown_vms = Command("virsh list --state-shutoff --name")
                shutdown_vms = check_shutdown_vms.getStringList(regex=".*")
            except HwCertCommandException as e:
                pass

            if self.guestName in shutdown_vms:
                print("VM shutdown successfully !!!")
                vm_shutdown = True
                break
            else:
                count += 1
                time.sleep(30)
        if not vm_shutdown:
            print("Error: Failed to shutdown VM !!!")
            return False
        return True
