#!/usr/bin/env python
# Copyright (c) 2009 Red Hat, Inc. All rights reserved. This copyrighted material
# is made available to anyone wishing to use, modify, copy, or
# redistribute it subject to the terms and conditions of the GNU General
# Public License v.2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# Authors: Greg Nichols
#          YangKun <ykun@redhat.com>
#

import io
import os
import re
import shutil
import sys
import time
from distutils.version import LooseVersion
from xml.dom.minidom import parseString

from rhcert.command import Command, HwCertCommandException, HwCertCommandErrorOutput
from rhcert.environment import Environment
from rhcert.model.resultsdocument import ResultsDocument
from rhcert.redhatrelease import RedHatRelease
from rhcert.tags import Constants, DeviceClass, Tags, TestTag, RHEL, Attributes, TestResult

from rhcert.suites.hwcert.hwcert.hardwareTest import HardwareTest
from rhcert.suites.rhcert.platform import CheckHardwareHealth


class FvTest(HardwareTest):

    def __init__(self, path):
        path = "fvtest/" + path
        HardwareTest.__init__(self, path)

        self.deviceClass = DeviceClass.system
        self.hardwarePlatform = Command("uname -i").getString()
        self.guestName = "hwcert-%s" % self.hardwarePlatform
        self.guestImageFile = 'hwcert-' + self.hardwarePlatform + '.img'
        self.guestConfigFile = 'hwcert-' + self.hardwarePlatform + ".xml"
        self.guestNvramFile = 'hwcert-' + self.hardwarePlatform + '_VARS.fd'
        self.loopBackDevice = None
        self.dataImageFile = 'hwcertData.img'
        # backend command expected in guest
        self.backendInGuest = "rhcert-cli"

        # defaults
        self.hwcertCommandInGuest = self.backendInGuest + " run"
        # subclasses use either testToRun OR tagToRun to specify tests
        self.testToRun = ""
        self.tagToRun = ""
        self.req_passthrough = False
        self.environment = Environment()
        self.guestConfigPath = os.path.join(self.environment.getKVMGuestConfigDirectory(), self.guestConfigFile)
        self.guestConfigDirectory = self.environment.getKVMGuestConfigDirectory()
        self.guestImageDirectory = self.environment.getKVMGuestImageDirectory()
        self.guestNvramDirectory = self.environment.getKVMGuestNvramDirectory()
        self.waitTime = self.environment.getFVGuestTimelimit()
        self.proxyURLs = self.environment.getProxyURLs()
        self.redHatRelease = RedHatRelease()
        self.mountPoint = self.environment.getFvImageMountPoint()
        self.enableTargetDevice = False
        self.targetDevice = None
        self.rpmsDir = "/tmp/rhcertRPMs"
        self.guestResultsPath = os.path.join("/var/rhcert", self.environment.getResultsFileName())

        try:
            Command("mkdir -p %s %s" % (self.mountPoint, self.rpmsDir)).echo()
        except HwCertCommandException as e:
            print("Error: can't create directories on host at {} or at {}".format(self.mountPoint, self.rpmsDir))
            raise

    def tags(self):
        return [TestTag.noninteractive, TestTag.certification, TestTag.virtualization]

    def configureHypervisor(self):
        """ configure the test for the kvm hypervisor """
        # RHEL 6.0 onwards kvm is supported and Xen is dropped
        # check cpu flags, just to log the results.
        self.__hasVirtualizationCPUFlags(verbose=True)

        print("Testing kvm virtualization")

        # check and start libvirtd
        try:
            Command("systemctl enable libvirtd.service").echo()
            Command("systemctl start libvirtd.service").echo()
            print("libvirtd is running")
        except HwCertCommandException as e:
            print("Warning: libvirtd may not be running")
            print(e)

        # check the kvm modules
        try:
            kvmModulesScript = "/etc/sysconfig/modules/kvm.modules"
            if os.path.exists(kvmModulesScript):
                Command(kvmModulesScript).echo()
        except HwCertCommandException as e:
            print("Warning: kvm modules may not be running")
            print(e)

    def __isHVMCapableViaLibvirt(self):
        """ determine if system is FV/hvm capable, via virsh from libvirt """

        start_libvirtd = Command("systemctl start libvirtd")
        try:
            start_libvirtd.run()
        except HwCertCommandException:
            print("Warning: libvirtd is not running or not installed")
            return False

        virsh = Command("virsh capabilities")
        try:
            virsh.run()
            parser = parseString("".join(virsh.output))
            elements = parser.getElementsByTagName("guest")
            for element in elements:
                osType = element.getElementsByTagName(Tags.os_type)[0].childNodes[0]
                if osType.data == "hvm":
                    return True
        except Exception as e:
            if self.getDebug() != Constants.off:
                print("Note: could not parse virsh capabilities")
                print(e)

        #otherwise
        return False

    def __isHVMInCapabilitiesFile(self):
        """ try /sys/../capabilities, looking for FV support"""
        try:
            capabilities = '/sys/hypervisor/properties/capabilities'
            capabilityFile = open(capabilities)
            self.hypervisorCapabilities = capabilityFile.readline()
            capabilityFile.close()
            if  Constants.hvm in self.hypervisorCapabilities:
                return True
        except IOError:
            if self.getDebug() != Constants.off:
                print("Note: could not read " + capabilities)

        #otherwise
        return False

    def __hasVirtualizationCPUFlags(self, verbose=False):
        """ check cpuinfo for virtualization flags """
        cpuflags = Command("fgrep flags /proc/cpuinfo")
        try:
            flags = cpuflags.getString(regex="(vmx|smx|svm)", singleLine=False)
            if verbose or self.getDebug() != Constants.off:
                print("/proc/cpuinfo has %s flag" % flags)
            return True
        except HwCertCommandException:
            if verbose or self.getDebug() != Constants.off:
                print("Note: /proc/cpuinfo does not have vmx, smx, or svm flags")

        return False

    def check_passthrough(self):
        """
        Check if pass through is configured
        :return: True/False
        """
        cmdline_cmd = Command("cat /proc/cmdline")
        try:
            cmdline = cmdline_cmd.getString(regex=".*")
        except HwCertCommandException:
            if self.getDebug() != Constants.off:
                print("Warning: Unable to fetch command line parameters")
            return False

        if 'iommu=on' in cmdline and 'iommu=pt' in cmdline:
            return True
        else:
            if self.getDebug() != Constants.off:
                print("Warning: Unable to find iommu parameters in cmdline")
            return False

    def plan(self, planner):
        tests = list()
        osSupport = False
        # if Red Hat and not RT kernel
        if ("Red Hat" in self.redHatRelease.getProduct()
                and self.redHatRelease.getKernelFlavor() != Constants.rt
                and (self.redHatRelease.getArch() == Constants.x86_64
                     or self.redHatRelease.getArch() == Constants.ppc64le
                     or self.redHatRelease.getArch() == Constants.aarch64)
                or self.redHatRelease.getArch() == Constants.s390x):
            osSupport = True
        #  Fedora x86_64
        elif "Fedora" in self.redHatRelease.getProduct() and self.redHatRelease.getArch() == Constants.x86_64:
            osSupport = True

        # If pass-through is required, check if it is configured
        passthrough = True
        if self.req_passthrough:
            passthrough = self.check_passthrough()

        checkHardwareHealth = CheckHardwareHealth("ram_per_core", TestResult.REVIEW)
        # check libvirt, capabilities file, or cpuflags
        if osSupport and passthrough \
                and checkHardwareHealth.ram_per_core() \
                and (self.__isHVMCapableViaLibvirt()
                     or self.__isHVMInCapabilitiesFile()
                     or self.__hasVirtualizationCPUFlags()):
            test = self.makeCopy()
            tests.append(test)
            # Download the guest XML config file if fv_test is supported
            # self.verifyFile(self.guestConfigDirectory, self.guestConfigFile)
        if not tests:
            print("INFO: Could not plan test {}".format(self.path))
        else:
            print("INFO: Successfully planned test {}".format(self.path))
        return tests

    def getRequiredRPMs(self):
        rpms = ["wget", "libvirt", "virt-install", "qemu-kvm", "tar", "libguestfs-tools-c"]

        if sys.version_info.major > 2:
            rpms.append("python3-libvirt")
        else:
            rpms.append("libvirt-python")
        return rpms

    def verifyGuest(self):
        # Check for guest file and download it if required
        asExpected = self.verifyGuestFiles()
        asExpected = asExpected and self._restartService("libvirtd.service")
        try:
            guestName = Command("virsh list").getString(regex=self.guestName, singleLine=False)
            print("Warning: Guest %s is already running" % guestName)
            print("Attempting shutdown:")
            Command("virsh destroy %s" % self.guestName).echo()
            guestName = Command("virsh list").getString(regex=self.guestName, singleLine=False)
            print("Error: could not shutdown guest - {} ".format(guestName))
            return False
        except:
            print("Verified that guest " + self.guestName + " is not running")
            pass

        virtlog = "virtlogd.socket"
        if self._servicePresent(virtlog):
            asExpected = asExpected and self._restartService(virtlog)
        else:
            print("%s not found by systemctl; will not restart it" % virtlog)
        return asExpected

    def _servicePresent(self, serviceName):
        command = Command("systemctl list-units -a | grep %s" % serviceName)
        try:
            #throws HwCertCommandException if no results are found
            command.getString()
        except HwCertCommandException:
            return False
        return True

    def _restartService(self, serviceName):
        print("Restarting %s" % serviceName)
        try:
            Command("systemctl restart %s" % serviceName).echo()
            time.sleep(2)
        except HwCertCommandException as exception:
            print("Error: could not restart %s" % serviceName)
            print(exception)
            return False
        return True

    def verifyGuestFiles(self):
        print("Verifying guest files ...")
        if not self.verifyFile(self.guestImageDirectory, self.dataImageFile):
            return False
        if not self.verifyFile(self.guestImageDirectory, self.guestImageFile):
            return False
        if not self.verifyFile(self.guestConfigDirectory, self.guestConfigFile):
            return False
        if self.redHatRelease.getArch() == Constants.aarch64 and not self.verifyFile(self.guestNvramDirectory, self.guestNvramFile):
            return False

        print("Guest files verified")
        sys.stdout.flush()
        return True

    def verifyFile(self, destinationDirectory, filename):
        # Check if file is already there
        filePath = os.path.join(destinationDirectory, filename)
        if os.path.exists(filePath):
            print("Verified: " + filePath)
            return True
        print("File '%s' not found !!!" % filePath)

        found_file = False
        # For guest Config file or guest Nvram file
        # Check if it is present in DATA image
        if filename in [self.guestConfigFile, self.guestNvramFile]:
            print("Checking the file '%s' in DATA image ..." % filename)
            found_file = self.getFileFromDataImage(destinationDirectory, filename)

        # If not found in DATA image, download it from FTP
        if not found_file:
            # otherwise
            print("Note: '%s' file not found, attempting download ..." % filename)
            return self.installFile(destinationDirectory, filename)
        return True

    def getFileFromDataImage(self, destinationDirectory, filename):
        self.unmountDataImage(ignoreErrors=True)
        # Mount the DATA image
        if not self.mountDataImage():
            print("Error: Failed to mount the data image")
            return False

        file_path = os.path.join(self.mountPoint, filename)
        file_found = False

        # For new aarch64 image nvram file is not required
        # TODO: remove this once all old images are replaced
        package_dir = os.path.join(self.mountPoint, "Packages")
        if os.path.isdir(package_dir) and filename == self.guestNvramFile:
            # if directory exists then it is a new data image
            # nvram file not required, so setting file_found to True
            print("Note: NVRAM file is not required for new aarch64 Image")
            file_found = True
        # Check if file exists
        elif os.path.exists(file_path):
            print("File '%s' found in DATA image" % filename)
            # Copy the file to destination directory
            shutil.copy(file_path, destinationDirectory)
            # Reset the SELinux context of the file
            restorecon = Command("/sbin/restorecon -v " + os.path.join(destinationDirectory, filename))
            restorecon.run(ignoreErrors=True, ignoreWarnings=True)
            file_found = True
        # Un mount the DATA image
        self.unmountDataImage(ignoreErrors=True)
        return file_found

    def installFile(self, destinationDirectory, filename):
        sys.stdout.flush()
        urlsToTry = list()
        tarredFilename = filename + ".tar.bz2"

        partnerServerBaseUrl = os.path.join(self.environment.getPartnerServerURL(), Constants.fv_images, RHEL.major)
        if self.getTestServer() and self.getTestServer() != "unknown":
            testServerBaseUrl ="http://%s" % os.path.join(self.getTestServer(), Constants.store, Constants.transfer, Constants.fv_images, RHEL.major)
            urlsToTry.append(os.path.join(testServerBaseUrl, self.redHatRelease.getArch(), tarredFilename))
            testServerNewBaseUrl = "http://%s" % os.path.join(self.getTestServer(), Constants.rhcert_transfer, Constants.fv_images, RHEL.major)
            urlsToTry.append(os.path.join(testServerNewBaseUrl, self.redHatRelease.getArch(), tarredFilename))
            #to avoid partners re-using existing LTS stored images w/ new arch SUT without accidentally taking old x86_64 if arch specific is available on partner ftp
            urlsToTry.append(os.path.join(partnerServerBaseUrl, self.redHatRelease.getArch(), tarredFilename))
            urlsToTry.append(os.path.join(testServerBaseUrl, tarredFilename))
            urlsToTry.append(os.path.join(testServerNewBaseUrl, tarredFilename))
        else:
            print("Warning: no hwcert test server set - can not look for guest files there")
        urlsToTry.append(os.path.join(partnerServerBaseUrl, self.redHatRelease.getArch(), tarredFilename))
        urlsToTry.append(os.path.join(partnerServerBaseUrl, tarredFilename))

        wget_cmd = "wget "
        if self.proxyURLs:
            # In proxy environment, use proxy of ftp protocol if that is set,
            # else show a warning message and use proxy of http or https protocol.
            # If none of the above protocols found, FAIL the test.
            supported_protocols = ["ftp", "http", "https"]
            proxy = None
            protocol_used = None
            for protocol in supported_protocols:
                if protocol in list(self.proxyURLs.keys()):
                    protocol_used = protocol
                    proxy = self.proxyURLs[protocol]
                    break
            if protocol_used is None:
                print("Error: Proxy settings are missing for ftp protocol, no other valid proxies found. "
                      "Please check settings in /etc/rhcert.xml.")
                print("Error: Failed to download the guest file %s" % tarredFilename)
                return False
            elif protocol_used != "ftp":
                print("Warning: Proxy settings are missing for ftp protocol. Please check settings in /etc/rhcert.xml. "
                      "Using proxy of '%s' protocol" % protocol_used)
            wget_cmd = "wget -e ftp_proxy='%s' " % proxy

        for url in urlsToTry:
            download_cmd = wget_cmd + url
            print("\nTrying to download the guest file using command: %s" % download_cmd)
            if os.system(download_cmd) == 0:
                try:
                    if os.path.exists(tarredFilename):
                        unpackedFilename = Command("tar xmvfj " + tarredFilename).getString()
                        os.remove(filename + ".tar.bz2")
                        # move it to the right place
                        shutil.move(unpackedFilename, os.path.join(destinationDirectory, filename))
                        print("Moved: " + unpackedFilename + " to " + os.path.join(destinationDirectory, filename))
                        restorecon = Command("/sbin/restorecon -v " + os.path.join(destinationDirectory, filename))
                        restorecon.echo()
                        print("Downloaded " + filename + " to " + destinationDirectory)
                        return True
                    else:
                        print("Note: could not download " + url)
                except Exception as e:
                    print("Warning: download " + url + " failed:")
                    print(e)
            else:
                print("Warning: could not download " + url)

        # otherwise
        print("Error: could not find or download guest file " + tarredFilename)
        return False

    def mountDataImageUpdated(self):
        retries = 6
        while True:
            try:
                Command('mount -o loop,ro %s %s' % (os.path.join(self.guestImageDirectory, self.dataImageFile), self.mountPoint)).echo()
                return True
            except HwCertCommandException as e:
                if not retries:
                    print("Warning: could not mount data image")
                    return False
                print("mounting data image failed, trying again...")
                retries = retries - 1

    def mountDataImage(self):
        if self.redHatRelease.getArch() == Constants.ppc64le or self.redHatRelease.getArch() == Constants.aarch64 or self.redHatRelease.getArch() == Constants.s390x:
            return self.mountDataImageUpdated()
        # New image for x86_64 is an iso image and can be mounted directly
        elif self.redHatRelease.getArch() == Constants.x86_64:
            if self.mountDataImageUpdated():
                # If mounting is successful return True
                # else try the old way of mounting the image
                # TODO: old way of mounting can be completely removed once all the images are updated
                return True
        try:
            self.loopBackDevice = Command("losetup -f").getString(regex="^(?P<device>/dev/loop[0-9]+)", regexGroup="device", singleLine=False)
            print("Using loopback device " + self.loopBackDevice + " for guest data image")
            loop = os.path.basename(self.loopBackDevice)
            Command("losetup " + self.loopBackDevice + " " + os.path.join(self.guestImageDirectory, self.dataImageFile)).echo()
            try:
                Command("kpartx -av " + self.loopBackDevice).echo()
            except HwCertCommandErrorOutput as exception:
                print("Warning: %s" % exception)
                exception.command.printErrors()
            # we may need to wait here - see BZ 874715
            waitTime = 5 # sec
            retries = 6
            while True:
                try:
                    time.sleep(waitTime)
                    partition = Command("ls /dev/mapper").getString(regex="%sp[0-9]" % loop, singleLine=False)
                    break
                except HwCertCommandException as e:
                    if not retries:
                        print("Error: could not create loop device %s" % loop)
                        return False
                    # otherwise
                    print("Warning: %s is not ready." % loop)
                    retries = retries - 1

            Command("mount /dev/mapper/" + partition + " " + self.mountPoint).run()
            print("Mounted /dev/mapper/" + partition + " on " + self.mountPoint)

            if self.getDebug()  != Constants.off:
                Command("df -h " + self.mountPoint).echo()
                Command("ls -al " + self.mountPoint).echo()
        except HwCertCommandException as e:
            print("Error: could not mount data image")
            print(e)
            e.command.printErrors()
            return False
        return True

    def unmountDataImage(self, ignoreErrors=False):
        print("Unmounting " + self.mountPoint)
        sys.stdout.flush()
        try:
            Command("umount " + self.mountPoint).echo()
            if self.loopBackDevice:
                print("Deleting loopback device " + self.loopBackDevice + " configuration.")
                Command("kpartx -dv " + self.loopBackDevice).echo()
                Command("losetup -d " + self.loopBackDevice).echo()
        except HwCertCommandException as e:
            if not ignoreErrors:
                print("Error: could not unmount data image.")
                print(e)
                e.command.printErrors()
            return False
        return True

    def virtCommand(self, command):
        """
        :param command: type=string, can be any guestfish command
        :return: True|False
        """
        try:
            Command("guestfish --rw -d {} -i {}".format(self.guestName, command)).echo()
        except HwCertCommandException:
            print("Error: could not execute guestfish command {}".format(command))
            return False
        return True

    def largestVersionDirectory(self, parentDirectory):
        """ Look one level deep and return largest number/. version directory (ex. 5.12 over 5.2)"""
        subdirs = next(os.walk(parentDirectory))[1]
        lvVersionDirs = [LooseVersion(d) for d in subdirs if re.search("^[\d.]+$", d)]
        lvVersionDirs.sort()
        return str(lvVersionDirs[-1])

    def submitTasks(self):
        # mount the data image and then write test info
        newDataImage = False
        try:
            self.unmountDataImage(ignoreErrors=True)
            if not self.mountDataImage():
                print("Error: Failed to mount the data image")
                return False

            # remove old rpm directories from host
            shutil.rmtree(self.rpmsDir, ignore_errors=True)
            # copying rpm directories from data image to host
            shutil.copytree(self.mountPoint, self.rpmsDir)
            # Un mounting the data image
            self.unmountDataImage(ignoreErrors=True)

            # determine what version (and from which directory) of rhcert to install
            print("Creating the script file ...")
            rpmBasePath = os.path.join(self.rpmsDir, "Packages")
            if os.path.isdir(rpmBasePath):
                # if directory exists then it is a new data image
                newDataImage = True
                print("New data image is being used")
            else:
                latestDetectedVersion = self.largestVersionDirectory(self.rpmsDir)
                rpmBasePath = os.path.join(self.rpmsDir, latestDetectedVersion)
                print("Old data image is being used")

            # creating  the script file
            guestScriptFile = "/tmp/rc.local"
            guestScript = open(guestScriptFile, "w")
            guestScript.write("#!/bin/sh\n")
            guestScript.write("# hwcert FV guest script - auto-generated\n")

            self.virtCommand("copy-in %s /tmp" % self.rpmsDir)

            # remove old test results if any exist
            guestScript.write("rm {}\n".format(self.guestResultsPath))

            # remove hwcert and rhcert rpms in case they were still there
            for package in ["redhat-certification-hardware", "redhat-certification-backend"]:
                guestScript.write("rpm -e %s\n" % package)

            dependencies = ['python-chardet', 'python-ordereddict', 'python-six', 'python-backports', 'mod_wsgi',
                            'python-backports-ssl_match_hostname', 'python-urllib3', 'python-requests', 'lsscsi']
            if self.redHatRelease.getArch() == Constants.ppc64le:
                dependencies = ['lmbench', 'dt', 'stress', 'httpd-tools', 'mailcap', 'httpd', 'mod_wsgi']
            dep_dir = "dependencies"
            if newDataImage:
                # new image has the dependencies already installed except for below packages
                dependencies = ['dt', 'lmbench', 'stress']
                dep_dir = ""
            for dependency in dependencies:
                rpmBlob = "{}*.rpm".format(dependency)
                guestScript.write("rpm -ivh --nodeps {}\n".format(os.path.join(rpmBasePath, dep_dir, rpmBlob)))

            backendBlob = "redhat-certification-backend*.rpm"
            hardwareBlob = "redhat-certification-hardware*.rpm"
            guestScript.write("rpm -ivh --nodeps {}\n".format(os.path.join(rpmBasePath, backendBlob)))
            guestScript.write("rpm -ivh --nodeps {}\n".format(os.path.join(rpmBasePath, hardwareBlob)))

            # remove the tmp files
            guestScript.write("echo -e \'y\' | %s clean --mode auto\n" % self.backendInGuest)

            # plan the tests
            guestScript.write("%s plan --mode auto\n" % self.backendInGuest)

            # run the tests
            options = " --mode auto "
            if self.getTestServer():
                options += ' --server %s ' % self.getTestServer()

            if self.enableTargetDevice and self.targetDevice:
                options += " --device %s " % self.targetDevice

            if self.testToRun:
                options += " --test %s " % self.testToRun
                # for device-specific tests, add them to the plan.
                if self.enableTargetDevice and self.targetDevice:
                    guestScript.write("%s plan --add %s\n" % (self.backendInGuest, options))
            elif self.tagToRun:
                options += " --tag %s " % self.tagToRun

            commandToRun = self.hwcertCommandInGuest + options
            guestScript.write(commandToRun + "\n")

            # remove hwcert and rhcert rpms
            for package in ["redhat-certification-hardware", "redhat-certification-backend"]:
                guestScript.write("rpm -e %s\n" % package)

            # poweroff
            guestScript.write("poweroff\n")
            guestScript.close()

            print("Copying the script into the guest image ...")
            os.chmod(guestScriptFile, 0o755)
            if not self.virtCommand("copy-in %s /etc" % guestScriptFile) \
                    or not self.virtCommand("copy-in %s /etc/rc.d" % guestScriptFile):
                print("Error: Failed to transfer executable script to guest.")
                return False

            print("Submitted tests: " + commandToRun)
            sys.stdout.flush()
            return True
        except HwCertCommandException as e:
            print("Error: could not set guest tests to run")
            print(e)
            return False


    def runGuest(self):
        try:
            Command("virsh start %s" % self.guestName).echo()
            print("FV Guest started ...")
        except HwCertCommandErrorOutput as exception:
            print("Warning: %s" % exception)
            exception.command.printErrors()
        except HwCertCommandException as e:
            print("Error: could not start the FV Guest.")
            print(e)
            return False
        print("%s test is running, it takes some time(less than %s minutes), please be patient ..." % (self.getName(), self.waitTime))
        print("You may use \"virsh console %s\" to monitor the testing progress" % self.guestName)
        sys.stdout.flush()

        # try and log the guest console if debug is set
        if self.getDebug() != Constants.off:
            try:
                tty = Command("virsh ttyconsole %s" % self.guestName).getString(regex="^/dev.*$", singleLine=False)
                Command("cat %s > guest_console.log &" % tty).echoIgnoreErrors()
            except HwCertCommandException as e:
                print("Warning: debug could not display virsh console log.")
                print(e)
                e.command.printErrors()
        sys.stdout.flush()
        # check the guest's status in every minute, this test should not consume more than 1 hour
        count = 0
        success = True
        while 1:
            if count == self.waitTime:
                try:
                    print("time out: destroying the guest.")
                    Command("virsh destroy %s" % self.guestName).echo()
                except HwCertCommandException as exception:
                    print("Warning: %s" % exception)
                success = False
                break
            else:
                time.sleep(60)
                try:
                    guestIsRunning = Command("virsh list").getString(regex=self.guestName, singleLine=False)
                    count = count + 1
                except:
                    print("Guest has shutdown")
                    break

        if self.getDebug() != Constants.off:
            try:
                with io.open("guest_console.log", encoding='utf-8', errors='ignore') as log:
                    print("****** guest console log - start ******")
                    for line in log.readlines():
                        print(line)
                    print("****** guest console log - end ******")
            except IOError as e:
                print(e)

        return success

    def getGuestResults(self):
        # copy the result file from the guest to host
        print("Fetching the result file from guest ...")
        if not self.virtCommand("copy-out %s %s" % (self.guestResultsPath, self.getOutputDirectory())):
            # merge the results
            print("Error: could not get result file from guest")
            return False
        return self.getGuestTestRunResult()

    def getGuestTestRunResult(self):
        # open the test results, and look for the run
        results = ResultsDocument()
        print("Parsing the result file to validate the test run ...")
        if not results.load(os.path.join(self.getOutputDirectory(), self.environment.getResultsFileName())):
            print("Error: could not load %s" % os.path.join(self.getOutputDirectory(), self.environment.getResultsFileName()))
            return False
        print("Loaded %s " % os.path.join(self.getOutputDirectory(), self.environment.getResultsFileName()))

        # using recursive version to ignore structural XML change in guest to get all test documents
        tests = results.getTests(names=None, onlyFailures=False, recursive=True)
        if not tests:
            print("Error: no tests in guest results")
            return False
        if self.tagToRun:
            for testDocument in tests:
                print("Checking test %s %s" % (testDocument.getName(), self.getDocDeviceName(testDocument)))
                print("  tags: %s" % testDocument.getTags())
            tests = results.getTaggedTests(tests, [self.tagToRun])

            if not tests:
                print("Error: no test tagged " + self.tagToRun + " in guest results")
                return False

        # if the test is run on multiple devices (storage, network), look for one that passed
        for testDocument in tests:
            print("Checking test %s %s" % (testDocument.getName(), self.getDocDeviceName(testDocument)))
            # have to check names here to be sure
            if self.testToRun and testDocument.getName() != self.testToRun:
                continue
            for runDocument in testDocument.getRuns():
                print("Run %s %s" % (runDocument.getNumber(), runDocument.getSummary()))
                if runDocument.getSummary() == Constants.PASS:
                    if self.enableTargetDevice and self.targetDevice:
                        if self.getDocDeviceName(testDocument) == self.targetDevice:
                            print("Guest test %s on target %s PASSED" % (testDocument.getName(), self.getDocDeviceName(testDocument)))
                            return True
                    else: # default target
                        print("Guest test %s PASSED" % testDocument.getName())
                        return True

        # otherwise
        print("Error: " + self.testToRun + " run on guest missing or failed")
        return False

    def getDocDeviceName(self, testDocument):
        """ this is an adapter to read the old pre-rhcert value that was a test attribute
        it needs to go away when hwcert on the guest is updated to use TestDocument.getTestParameters()"""
        return testDocument.element.getAttribute(Attributes.logical_device)

    def delete_script(self):
        """
        Delete the script from the image
        """
        print("Deleting the script from the image ...")
        self.virtCommand("write /etc/rc.d/rc.local '#!/bin/bash'")
        self.virtCommand("write /etc/rc.local '#!/bin/bash'")

    def run(self):
        # just pass-through the logical device to the guest call
        if self.getLogicalDeviceName() and self.enableTargetDevice:
            self.targetDevice = self.getLogicalDeviceName()

        # determine whether to use xen or kvm virtualization
        self.configureHypervisor()
        # verify the guest, start if necessary
        if not self.verifyGuest():
            return False

        # submit the tasks to the guest
        if not self.submitTasks():
            return False

        # run the guest, wait for results,
        if not self.runGuest():
            self.delete_script()
            return False

        # then copy the results.xml, and get the guest return value
        if not self.getGuestResults():
            self.delete_script()
            return False

        # delete script file so that VM can be accessed for any changes
        self.delete_script()

        return True


if __name__ == "__main__":
    test = FvTest()
    returnValue = test.do(sys.argv)
    sys.exit(returnValue)
