# Copyright (c) 2006 Red Hat, Inc. All rights reserved. This copyrighted material
# is made available to anyone wishing to use, modify, copy, or
# redistribute it subject to the terms and conditions of the GNU General
# Public License v.2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# Author: Greg Nichols
#
# HardwarePlanner: a class to plan HW tests.
#

from rhcert.tags import Constants
from rhcert.planner import Planner

from .udevDeviceDetector import UdevDeviceDetector
from .procDeviceDetector import ProcDeviceDetector
from .deviceclass import BuildDeviceClasses


class HardwarePlanner(Planner):
    """The Planner class analyses a directory of tests, looking
    for rhcert.Test subclasses.  If found, it adds them to it's
    dictionaries, and provides lists of test(s) by various criteria
    """

    def __init__(self, options):
        Planner.__init__(self, options)
        self.devices = list()
        self.__initializeDeviceDetectors()
        self.deviceClasses = BuildDeviceClasses()

    def __initializeDeviceDetectors(self):
        self.deviceDetectors = dict()
        detector = UdevDeviceDetector()
        self.deviceDetectors[detector.getName()] = detector
        detector = ProcDeviceDetector()
        self.deviceDetectors[detector.getName()] = detector

    def getUdevDeviceDetector(self):
        return self.deviceDetectors[Constants.udev]
    def getProcDeviceDetector(self):
        return self.deviceDetectors[Constants.proc]

    def plan(self):
        self.__loadDevices()
        self.plannedTests = self.planTests()
        return self.plannedTests

    def saveNewPlan(self, resultsDocument):
        Planner.saveNewPlan(self, resultsDocument)

        for device in self.getDevices():
            resultsDocument.appendDevice(device)

    def __loadDevices(self):
        udev = UdevDeviceDetector()
        self.devices = udev.getDevices()
        proc = ProcDeviceDetector()
        self.devices.extend(proc.getDevices(ProcDeviceDetector.inputBusPath))

    def getDevices(self):
        return self.devices

    def getDeviceByUDI(self, udi):
        self.__loadDevices()
        for device in self.devices:
            if udi in device.getUDI():
                return device
        return None

    def getDeviceByName(self, name):
        self.__loadDevices()
        for device in self.devices:
            if device.hasPropertyValue(name):
                return device
        return None

    def getDeviceClass(self, deviceClassName):
        return self.deviceClasses[deviceClassName]

