# Copyright (c) 2006 Red Hat, Inc. All rights reserved. This copyrighted material 
# is made available to anyone wishing to use, modify, copy, or 
# redistribute it subject to the terms and conditions of the GNU General 
# Public License v.2.
# 
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# Author: Greg Nichols
#
import syslog

from rhcert.test import Test
from rhcert.tags import Constants, Programs
from rhcert.suites.rhcert.policy import get_policy
from .device import UdevDevice

class HardwareTest(Test):

    def __init__(self, path):
        self.path = "hwcert/" + path
        Test.__init__(self, self.path)
        self.policy = get_policy()
        self.device = None
        self.source = None # either hal, udev, udisks, etc, or manual

    def getSuiteName(self):
        return Programs.hardware


    def add(self, planner):
        """ manually add a test:
            this default implementation requires a logic device name or UDI (substring match  
            subclasses may override if no device is required.  """
        device = None
        deviceID = None
        # respect --device
        if planner and planner.options and planner.options.device:
            deviceNameType = "logical name"
            deviceID = planner.options.device
            device = planner.getDeviceByName(deviceID)
        elif self.promptConfirm("Is this test for a specific device?"):
            deviceNameType = self.ui.prompt("How would you like to specify the device?", answers=["logical name", "UDI"])
            deviceID = self.ui.prompt("Enter the device %s:" % deviceNameType)
            if deviceNameType == "UDI":
                device = planner.getDeviceByUDI(deviceID)
            else:
                device = planner.getDeviceByName(deviceID)

        if not device and deviceID:
            print("Warning: unknown device: %s " % deviceID)
            # create one from scratch
            device = UdevDevice(dict())
            if deviceNameType == "UDI":
                device.setUDI(deviceID)
            elif "logical" in deviceNameType:
                self.setLogicalDeviceName(deviceID)

        if device:
            self.setDevice(device)
            if "logical" in deviceNameType:
                self.setLogicalDeviceName(deviceID)

        if planner.options.property:
            for (key, value) in planner.options.property.items():
                self.testParameters.set(key, value)

    def getDeviceClass(self):
        return self.deviceClass
    def getDeviceBus(self):
        if self.device:
            return self.device.getPropertyString("bus")
        return ""
    def getLogicalDeviceName(self):
        return self.testParameters.get(Constants.device)
    def setLogicalDeviceName(self, name):
        if name:
            self.testParameters.set(Constants.device, name)
    def getUDI(self):
        if self.device: 
            return self.device.getUDI()
        if self.testParameters.get(Constants.udi): 
            return self.testParameters.get(Constants.udi)
        # otherwise
        return ""
    def setUDI(self, udi):
        self.testParameters.set(Constants.udi, udi)
    def getShortUDI(self):
        udi = self.getUDI()
        if not udi: 
            return ""
        # chop everything before and including /devices/
        prefix = "/devices/"
        if udi.index(prefix) >=0:
            udi = udi[udi.index(prefix)+len(prefix):]   
        # take the last 5 nodes
        udi = "/".join(udi.split("/")[-5:])
        return udi

    def setDevice(self, device):
        if device.getUDI():
            self.setUDI(device.getUDI())
        self.device = device
    def getDevice(self):
        return self.device

    def planWith(self, detector, properties, allDevices=True, strict=False):
        """ Plan for devices matching any property.  
        if all is true, multiple devices that match will be planned 
        if strict is true, all properties must match. """
        tests = list()
        devices = detector.getDevices()
        for device in devices:
            match = False
            for key in properties.keys():
                if device.getProperty(key) != properties[key]:
                    if strict:
                        match = False
                        break
                else:
                    match = True
            if match:          
                test = self.makeCopy()
                test.setDevice(device)
                tests.append(test)
                if allDevices:
                    continue
                else:
                    break
        return tests

    def finish(self, planner, runDocument):
        planner = planner.getPlanner(self) # get the HardwarePlanner object
        devices = self.logDevices(planner)
        if devices:
            runDocument.appendDevices(devices)
            return True

        #otherwise
        return False

    def logDevices(self, planner):
        """ default implementation: log all devices found in udev for the 
        test's device class
        """
        devices = list()
        deviceClass = planner.getDeviceClass(self.getDeviceClass())
        if not deviceClass:
            print("Error: could not find device class %s for logging" % self.getDeviceClass())
            return devices

        try:
            devices = deviceClass.getDevicesFromDetector(planner.getUdevDeviceDetector())
        except Exception as e:
            message = "Error: %s test could not log devices %s" % (self.getPath(), e)
            print(message)
            syslog.syslog(message)

        return devices
