# Copyright (c) 2013 Red Hat, Inc. All rights reserved. This copyrighted material
# is made available to anyone wishing to use, modify, copy, or
# redistribute it subject to the terms and conditions of the GNU General
# Public License v.2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# Author: Greg Nichols
#
import os, tempfile, signal, sys, re, time, requests


from rhcert.tags import Constants, DeviceClass, TestTag, RHEL
from rhcert.command import Command, HwCertCommandException
from rhcert.environment import Environment
from rhcert.hash import hashModule
from rhcert.httpUpload import HttpUpload
from rhcert.localTestServer import LocalTestServer
from rhcert.host import Host
from rhcert.compatability import UrllibRequest, urllib_build_opener, rhcert_make_str

from rhcert.suites.hwcert.hwcert.udevDeviceDetector import UdevDeviceDetector
from rhcert.suites.hwcert.hwcert.hardwareTest import HardwareTest

from rhcert.collabrequest import CollabRequest

class NetworkTest(HardwareTest):
    """ NetworkTest is the base class for specific network tests """

    def __init__(self, path):
        path = "network/" + path
        HardwareTest.__init__(self, path)
        self.tcp_threads = [2, 4, 8, 16]
        self.ipAddress = ""
        self.testFile = "/var/www/html/httptest.file"
        self.allOtherInterfaces = list()
        self.interfaceSpeed = 100 # speed in Mb/sec
        self.oldHandler = None
        self.serverMode = False # set True only in EthernetTest (hwcert/tests/network.py)
        self.priority = 2 # set priority high so it runs before longer tests
        self.isNFSRootSystem = False
        self.deviceClass = DeviceClass.network_interface
        self.ignoreInterfaces = re.compile("lo|peth\d+|virbr\d+|vif\d+.\d+|xenbr\d+|wmaster\d+|sit\d+")
        self.environment = Environment()
        self.enforceSpeed = False
        self.bandwidthTarget = 0.8 # 80%
        self.serverPath = os.path.join(self.environment.getServerDirectory(), Constants.store, self.environment.getServerTransfer())
        self.interfaceConnect = "nmcli dev connect"
        self.interfaceDisconnect = "nmcli dev disconnect"
        if RHEL.major == "RHEL7":
            self.interfaceConnect = "ifup"
            self.interfaceDisconnect = "ifdown"

    def tags(self):
        return [TestTag.noninteractive, TestTag.certification, TestTag.network]

    def getRequiredRPMs(self):
        return ["ethtool"]

    def plan(self, planer):
        """ subclasses must provide this """
        return list()

    def getNetworkDevices(self, planner):
        """ returns a dictionary of devices by logical device names that are networking interfaces """
        interfaceDevices = dict()
        devices = planner.getUdevDeviceDetector().getDevices()
        for device in devices:
            logicalDevice = None
            if "net" in device.getProperty("SUBSYSTEM"):
                logicalDevice = device.getProperty("INTERFACE")

            """ if network logical device was found, and not already planned,
            and it's not in the ignore list, add it"""
            if logicalDevice and logicalDevice != "" and not logicalDevice in interfaceDevices and not self.ignoreInterfaces.search(logicalDevice):
                interfaceDevices[logicalDevice] = device
        return interfaceDevices

    def getInterfaceSpeed(self, interface):
        interfaceSpeed = 0
        for interfaceString in (interface, "p%s" % interface):
            try:
                ethtoolCommand = Command("ethtool %s" % interfaceString)
                interfaceSpeed = ethtoolCommand.getInteger(regex="^([ \t]+Speed:[ \t]+)(?P<speed>\d+)(Mb/s)", regexGroup="speed", singleLine=False)
                if interfaceSpeed:
                    break
            except HwCertCommandException:
                pass

        return interfaceSpeed

    def prepareForRun(self):

        if self.getDebug() != Constants.off:
            print("\nStarting network test verification...")

        errors = list()
        self.interface = self.getLogicalDeviceName()
        if not self.interface:
            errors.append("No logical interface name set.")
        else:
            self.ipAddress = self.getIPAddress(self.interface)
            if not self.ipAddress:
                errors.append("Could not determine IP Address.")

        if self.getTestServer() == None or self.getTestServer() == "unknown":
            errors.append("No test server was set.")
        else:
            # Pass instance of Host class, instead of hostname/IP to LocalTestServer
            localTestServer = LocalTestServer(Host(self.getTestServer()))
            localTestServer.verify()
            errors.extend(localTestServer.getErrors())

        self.getAllOtherInterfaces()
        self.checkNFSRootFileSystem()

        if not errors:
            print("\nNetwork test verified for %s at %s" % (self.interface, self.ipAddress))
            return True
        # otherwise
        print("\nNetwork test for %s failed verification:" % self.interface)
        for error in errors:
            print("Error: %s" % error)

    def nfsSignalHandler(self, sig, frame):
        try:
            os.rmdir("%s/%s" % (self.nfsdir, self.ipAddress))
            os.system("umount %s" % self.nfsdir)
            os.rmdir(self.nfsdir)
            print('signal caught; cleaning up NFS mount')
        except:
            return

    def setSignalHandler(self, handler):
        signals = [signal.SIGINT, signal.SIGTERM]
        for sig in signals:
            if handler:
                self.oldHandler = signal.signal(sig, handler)
            else:
                signal.signal(sig, self.oldHandler)

    def getAllOtherInterfaces(self):
        self.allOtherInterfaces = list()
        # Added awk for @ to support rdma interfaces eg: qib_ib0.8006@qib_ib0
        pipe = os.popen("/sbin/ip -o link show | awk -F \": \" '/UP>/ { print $2 }' | awk -F \"@\" '{print $1}'")
        while 1:
            line = pipe.readline()
            if line:
                interface = line.strip()
                if not self.ignoreInterfaces.search(interface):
                    if interface != self.interface:
                        self.allOtherInterfaces.append(interface)
                elif self.getDebug() != Constants.off:
                    print("ignoring interface %s" % interface)
            else:
                break
        pipe.close()

    def downAllOtherInterfaces(self):
        if not self.allOtherInterfaces:
            return True
        print("Shutting down all other network interfaces...")
        for interface in self.allOtherInterfaces:
            print(interface)
            if not self.shutdownInterface(interface):
                return False
        wait = 10 # sec
        sys.stdout.write("waiting %d sec..." % wait)
        sys.stdout.flush()
        time.sleep(wait)
        print("done")
        return True

    def addRoute(self):
        # is the test server specified by IP address or by host name?
        # ipv6 ??
        pattern = re.compile("\d+\.\d+\.\d+\.\d+")
        match = pattern.search(self.getTestServer())
        if match:
            self.testServerSpec = "-net"
        else:
            self.testServerSpec = "-host"

        if os.system("route add %s %s dev %s" % (self.testServerSpec, self.testServer, self.interface )) != 0:
            return False

        return True

    def removeRoute(self, sig=None, frame=None):
        if os.system("route del %s %s dev %s" % (self.testServerSpec, self.testServer, self.interface )) != 0:
            return False

        return True

    def restoreAllOtherInterfaces(self, sig=None, frame=None):
        print("Restoring all interfaces...")
        for interface in self.allOtherInterfaces:
            print("    interface: %s" % interface)
            if os.system("%s %s" % (self.interfaceConnect, interface)) != 0:
                print("Error: could not restore interface %s" % interface)

        # Bring down and up the interface if current interface is VLAN interface e.g: qib_ib0.8006
        # Required because bringing qib_ib0 up causes qib_ib0.8006 to go down: CERTPX-3796
        pattern = re.compile("(.*)\.(\d+)$")
        match = pattern.match(self.interface)
        if match:
            if match.group(1) in self.allOtherInterfaces:
                self.bounceInterface()

    def bounceInterface(self):

        if self.isNFSRootSystem:
            print("It appears that the root file system is NFS mounted")
            if not self.promptConfirm("The test will shutdown and restore interface %s, is this OK?" % self.interface):
                print("Warning: Bounce of interface %s was not tested" % self.interface)
                return True

        success = True
        # shut it down
        if not self.shutdownInterface(self.interface):
            print("Error: could not shut down interface %s" % self.interface)
            success = False

        # Waiting for 2 sec, before we start the interface again
        time.sleep(2)
        if not self.restartInterface(self.interface):
            print("Error: could not restart interface %s" % self.interface)
            success = False

        if not self.pingTestServer(self.interface):
            print("Error: could not ping test server %s " % self.getTestServer())
            success = False

        return success

    def shutdownInterface(self, interface):
        print("Shutting down interface %s" % interface)
        if os.system("%s %s" % (self.interfaceDisconnect, interface)) != 0:
            print("Warning: could not shut down interface %s" % interface)

        # confirm that it really went down (or is all ready down)
        slept=0
        delay=15
        interfaceUpCommand = Command("ip -f inet addr show %s" % interface)
        down = False
        while not down and slept < delay:
            try:
                interfaceUpCommand.getString("inet", singleLine=False)
                print("Interface %s is up" % interface)
                time.sleep(1)
                slept = slept + 1

            except Exception:
                print("interface %s is down" % interface)
                down = True


        if not down:
            print("Error: Unable to shut down interface %s" % interface)

        return down

    def restartInterface(self, interface):
        count = 0
        retry_count = 10
        interface_up = False
        bring_interface_up = Command("%s %s" % (self.interfaceConnect, interface))
        while count < retry_count and not interface_up:
            sys.stdout.flush()
            print("Bringing up interface %s attempt no. %s" % (interface, count))
            try:
                bring_interface_up.run(ignoreWarnings=True)
                interface_up = True
            except HwCertCommandException:
                time.sleep(2)
                count += 1
        if not interface_up:
            print("Error: Failed to bring up the device %s" % interface)
            return False
        sys.stdout.flush()

        # Since there can be a delay between ifup and linkup we need to
        # wait to see the interface running.
        slept=0
        wait = 3
        delay=30
        interfaceUpCommand = Command("ip -f inet addr show %s" % interface)
        while slept < delay:
            try:
                interfaceUpCommand.getString("inet", singleLine=False)
                print("Interface %s is UP" % interface)
                return True
            except Exception:
                time.sleep(wait)
                slept = slept + wait
                sys.stdout.flush()

        return False

    def get_interface_ip(self, interface):
        try:
            ip_addr_command = Command("ip addr show %s" % interface)
            return ip_addr_command.getString(regex="^.*inet (?P<ip_address>.+)\/.*$", regexGroup="ip_address", singleLine=False)
        except HwCertCommandException:
            print("Error: ip addr show %s has no inet configured" % interface)
            return None

    def pingTestServer(self, interface):

        # The next piece of bad news is that not all drivers support "RUNNING"
        # properly so we also need to see if we can reach the other end each time.
        # After a possible 30 seconds total we give up.

        interface_ip = self.get_interface_ip(interface)
        if not interface_ip:
            return False
        wait = 3 # "sec"
        slept=0
        delay=30 # sec
        sys.stdout.write("Checking via ping to %s... " % self.getTestServer())
        print("ping -I %s -Ac1 %s > /dev/null" % (interface_ip, self.getTestServer()))
        while slept < delay:
            if os.system("ping -I %s -Ac1 %s > /dev/null" % (interface_ip, self.getTestServer())):
                time.sleep(wait)
                slept = slept + wait
            else:
                print("done")
                return True

        return False

    def getIPAddress(self, interface):
        # what about ipv6?
        pipe = os.popen("ip -4 addr show %s" % (interface))
        pattern = re.compile("\d+\.\d+\.\d+\.\d+")
        while 1:
            line = pipe.readline()
            match = pattern.search(line)
            if match or not line:
                break
        pipe.close()
        if match:
            ipAddress = match.group()
            return ipAddress

        return None

    def createTestFile(self):
        # if interface speed is 0, try and get it now
        if not self.interfaceSpeed:
            self.interfaceSpeed = self.getInterfaceSpeed(self.interface)

        if not self.interfaceSpeed:
            self.interfaceSpeed = 100 # Mb/s
            print("Warning: Unable to determine interface speed")
            print("Assuming %u Mb/s" % self.interfaceSpeed)
        else:
            print("interface speed is %u Mb/s" % self.interfaceSpeed)

        return self.regenerateTestFile()

    def regenerateTestFile(self):
        try:
            os.remove(self.testFile)
        except OSError:
            pass

        # use interfaceSpeed/8, creating a file that takes very roughly 10 sec to transfer
        count = self.interfaceSpeed/8
        blocksize = 128 # KB

        dd = Command("dd if=/dev/urandom of=%s bs=%uk count=%u" % (self.testFile, blocksize, count))
        try:
            dd.run()
        except HwCertCommandException:
            # dd shows output on stderr, ignore this
            pass

        return True

    # icmpTest uses ping with various packet sizes
    def icmpTest(self):
        # (<bytes in packet>, <percent lost for warning>)
        # prior release packed sizes, with some margins
        # packetSizes = [(0, 0), (1, 0), (240, 10), (256, 15), (1024, 20), (2000, 30), (10000, 40), (40000, 60), (65507, 80)]
        # We now use default packet size, 100% margin, allowing 5 retries (see BZ 677752 )
        retries = 5
        packetCount = 5000
        lossMargin = 0
        while retries > 0:
            try:
                ping = Command("/bin/ping -i 0 -q -c %u %s" % (packetCount, self.getTestServer()))
                print(ping.command)
                packetLoss = ping.getString(regex="^.*, (?P<packetLoss>\d+\.{0,1}\d*)% packet loss.*$", regexGroup="packetLoss", singleLine=False)
                ping.printOutput()
                if float(packetLoss) <= lossMargin:
                    return True
                # otherwise
                print("Note: packet loss of %s%% is greater than %s%% expected" % (packetLoss, lossMargin))
            except HwCertCommandException as exception:
                print("Warning:")
                print(exception)
                ping.printErrors()
            retries = retries -1

        # otherwise
        return False

    def tcpTestLatency(self):
        # lat_tcp and bw_tcp report results on stderr, so
        # we have to catch the HwCertCommandException and print the results
        # for each of these tests.  Also
        print("tcp test to server " + self.getTestServer())
        sys.stdout.flush()
        cycles = 5
        result = True
        print("testing latency...")
        for p in range(cycles):
            try:
                Command("lat_tcp %s" % self.getTestServer()).echo()
            except HwCertCommandException as exception:
                if "TCP latency" not in exception.command.errors[0]:
                    print("Warning: {0}".format(exception.command.errors[0]))
                    result = False

        if result == False:
            print("Error: tcp latency test failed")

        return result

    def tcpTestBandwidth(self):
        print("testing bandwidth to " +  self.getTestServer())
        cycles = 5
        print("testing these number of threads: %s" % str(self.tcp_threads))
        sizeFlag = ""
        messageSize = getattr(self, 'messageSize', None)
        if messageSize:
            sizeFlag = "-m %s" % messageSize

        for numberOfThreads in self.tcp_threads:
            print("Using %u threads:" % numberOfThreads)
            bw_tcp = Command("bw_tcp -P %u %s %s" % (numberOfThreads, sizeFlag, self.getTestServer()))
            print(bw_tcp.command)
            averageBandwidth = 0.0
            for p in range(cycles):
                try:
                    # bw_tcp is not consistent about return values.
                    # it also produces output on stderr in all cases
                    # this test expects output like "0.0324 0.40 MB/sec"
                    # on stderr and fails otherwise
                    bw_tcp.run()
                except HwCertCommandException as exception:
                    pattern = re.compile("^[\.0-9]+ (?P<bandwidth>[\.0-9]+) MB/sec$")
                    match = pattern.search(bw_tcp.errors[0])
                    if len(bw_tcp.errors) is 0 or not match:
                        print("Error: tcp bandwidth test failed")
                        if len(bw_tcp.errors) > 0:
                            print("bw_tcp errors: %s" % bw_tcp.errors[0])
                        print(exception)
                        return False
                    # otherwise - print the output
                    print(bw_tcp.errors[0])
                    bandwidth = float(match.group("bandwidth"))
                    averageBandwidth = (averageBandwidth*p + bandwidth)/(p+1)
            if averageBandwidth*8 > self.interfaceSpeed*self.bandwidthTarget:
                print("Average Bandwidth: %s" % averageBandwidth)
                return True
            # otherwise
            print("Warning: Average Bandwidth %s Mb/sec is less than %s%% of the interface speed of %s Mb/sec" % ( averageBandwidth*8, self.bandwidthTarget*100, self.interfaceSpeed))

        if self.enforceSpeed:
            print("Error: Could not achieve required bandwidth")
            return False
        return True


    def startServer(self):
        # httpd
        try:
            start = Command("systemctl start httpd.service")
            start.echo()
            return True
        except HwCertCommandException as e:
            print("Error: could not start httpd")
            print(e)

        return False

    def stopServer(self):
        # httpd
        try:
            stop = Command("systemctl stop httpd.service")
            stop.echo()
            return True
        except HwCertCommandException as e:
            print("Error: could not stop httpd.")
            print(e)

        return False

    # YK: added statusServer
    def statusServer(self):
        return True

    def udpTest(self):
        print("testing udp latency to %s..." % self.getTestServer())
        sys.stdout.flush()
        try:
            Command("lat_udp %s" % self.getTestServer()).echo()
        except HwCertCommandException as exception:
            if "UDP latency" not in exception.command.errors[0]:
                print("Error: udp latency test failed")
                print(exception.command.errors)
                return False
            else:
                print(exception.command.errors)

        # otherwise
        return True


    def httpTransferTest(self):
        """ transfer a file to the server via POST, then pull it back via GET and compare the files """
        errorCount = 0
        for count in range(1,5):
            self.regenerateTestFile()
            (sentFileCheckSum, uploadIPAddress) = self.httpUpload()
            if not uploadIPAddress:
                errorCount += 1
                continue
            receivedFileCheckSum = self.httpDownload(uploadIPAddress)
            if sentFileCheckSum == receivedFileCheckSum:
                print("Checksums Match")
            else:
                print("Error: checksum %s does not match the file that was sent" % receivedFileCheckSum)
                errorCount += 1
        if errorCount > 0:
            print("Http test had %u errors" % errorCount)
        return (not errorCount)

    def httpUpload(self):
        # getMD5 checksum
        m=hashModule.md5()
        sentFile = open(self.testFile, "rb")
        m.update(sentFile.read())
        sentFile.close()
        sentCheckSum=m.hexdigest()
        request = HttpUpload(self.getTestServer(), "/save")
        if not request.addFile(self.testFile):
            return (None, None)
        # otherwise
        request.addField("server-path", self.serverPath)
        print("sending file %s hash: %s" % (self.testFile, sentCheckSum))
        startTime = time.time()
        size = os.path.getsize(self.testFile)
        curlCommand = Command("curl -i -X POST -F \"file=@%s;type=text/foo\" \"http://%s/save\"" % (self.testFile, self.getTestServer()))
        response = '\n'.join(curlCommand.getStringList(regex=".*", ignoreErrors=True))
        stopTime = time.time()
        if not response:
            # send failed
            return (None, None)
        print("response: ")
        print(response)
        print("Upload took %.2f seconds, %.2f MB/sec"  % (stopTime-startTime, size/((stopTime-startTime)*1000000)))

        #get the requests' IP address as seen by the server
        # might be ipv4 or ipv6
        pattern = re.compile("Saving File From IP: [ \t]*(?P<address>[^\s]*)")

        match = pattern.search(response)
        if match:
            uploadIPAddress = match.group("address")
        else:
            print("Warning: could not read upload IP address from response, assuming device IP address")
            uploadIPAddress = self.ipAddress

        return (sentCheckSum, uploadIPAddress)

    def httpDownload(self, uploadIPAddress):
        # GET the file back
        fileURL = 'http://%s' % os.path.join(self.getTestServer(), "rhcert-transfer", uploadIPAddress, os.path.basename(self.testFile))
        print("getting file %s " % fileURL)
        sys.stdout.flush()
        request = UrllibRequest(fileURL)
        opener = urllib_build_opener()

        startTime = time.time()
        try:
            response = opener.open(request).read()
        except Exception as error:
            print("Error: could not download file")
            print(error)
            return 0
        stopTime = time.time()
        print("Download took %.2f seconds, %.2f MB/sec"  % (stopTime-startTime, len(response)/((stopTime-startTime)*1000000)))

        m=hashModule.md5()
        m.update(response)
        return m.hexdigest()


    def printInfo(self):
        # YK: grab NIC info and running status
        sys.stdout.flush()
        try:
            ethtool = Command("ethtool %s" % self.interface)
            print(ethtool.command)
            ethtool.echoIgnoreErrors()
            print("")
            try:
                ipLinkShow = Command("ip link show %s" % self.interface)
                address = ipLinkShow.getString(regex="^.*link/ether (?P<address>[a-fA-F0-9:]+)", regexGroup="address", singleLine=False)
                print("MAC Address: %s" % address)
            except HwCertCommandException as e:
                print("Warning: could not determine MAC address")
                print(e)
            biosdevname = Command("biosdevname -d %s" % self.interface)
            print("")
            print(biosdevname.command)
            biosdevname.echoIgnoreErrors()
        except HwCertCommandException as exception:
            print("Warning: %s" % exception)
            exception.command.printErrors()

        return True

    def deviceNaming(self):
        """ show device naming and types for all detected interfaces """
        interfaceDevices = dict()
        devices = UdevDeviceDetector().getDevices()
        for device in devices:
            logicalDevice = None
            if "net" in device.getProperty("SUBSYSTEM"):
                logicalDevice = device.getProperty("INTERFACE")
                interfaceDevices[logicalDevice] = device

        # try biosdevname
        biosdevname = Command("biosdevname -d")
        biosDevices = None
        try:
            biosDevices = self.parseRecords(biosdevname, ["Kernel name","BIOS device"])
        except Exception as e:
            print("Note: could not read biosdevname")
            if self.getDebug() != Constants.off:
                print(e)

        for (interfaceName, interfaceDevice) in interfaceDevices.items():
            print("\n%s:\n%s" % (interfaceName, '-' * (len(interfaceName)+1)))

            """1. Verify the interface is under network manager control 
               (psuedo verifying default naming is in play) label each as NM true/false """
            managed = "GENERAL.NM-MANAGED"
            nmcli = Command("nmcli device show %s" % interfaceName)
            attributes = dict()
            try:
                nmcli.start()
                while True:
                    line = nmcli.readline()
                    if not line:
                        break
                    if ":" in line:
                        values = line.split(":")
                        attributes[values[0].strip()] = values[1].strip()
            except Exception as e:
                print("Error: running %s" % nmcli.command)
                print(e)
                return False

            answer = "no"
            if managed in attributes:
                answer = attributes[managed]
            print("Network Manager Controlled: %s" % answer)


            """2. For all of the Firmeware or BIOS provided indexes (including biosdevname) 
                device names label these in the output as "firmware based naming" """
            embedded = None
            removable = None
            if biosDevices and (interfaceName in biosDevices):
                for property in ["BIOS device", "Kernel name"]:
                    print("%s: %s" % (property, biosDevices[interfaceName][property]))
                """ 3. For all of the Firmware based naming options which are not slot based 
                label these as "embedded" """
                pciSlot = "PCI Slot"
                if pciSlot in biosDevices[interfaceName] and "embedded" in biosDevices[interfaceName][pciSlot]:
                    embedded = "yes"
                    removable = "no"
                else:
                    removable = "yes"
                    embedded = "no"
            else:
                """2a. For all others, label these as "non-firmware based naming" """
                print("BIOS device: non-firmware based naming")

            if embedded:
                print("Embedded: %s" % embedded)
            if removable:
                print("Removable: %s" % removable)

        return True

    def configureInterfaces(self):
        sys.stdout.flush()
        if not self.bounceInterface():
            print("Error: could not bounce interface %s" % self.interface)
            return False

        if self.isNFSRootSystem:
            if not self.addRoute():
                print("Error: could not add route for testing")
                return False
            self.setSignalHandler(self.removeRoute)
        else:
            if not self.downAllOtherInterfaces():
                print("Error: could not shutdown all other interfaces")
                return False
            self.setSignalHandler(self.restoreAllOtherInterfaces)

        if not self.createTestFile():
            print("Error: Unable to create test file")
            return False

        return True

    def verify(self):
        return self.prepareForRun()

    @staticmethod
    def __doLMBenchAction(verb):
        if verb not in ['start', 'stop']:
            return False

        result = True

        if verb == 'start':
            # start the bandwidth test server
            sys.stdout.write("Starting lmbench services...")
            sys.stdout.flush()

            if os.system("bw_tcp -s") != 0:
                print(" FAILED")
                result = False
            elif os.system("lat_udp -s") != 0:
                print(" FAILED")
                result = False
            else:
                print(" done.")

        else:
            sys.stdout.write("Stopping lmbench services...")
            sys.stdout.flush()

            for service in ["bw_tcp", "lat_udp"]:
                try:
                    Command("pkill -f -15 %s" % service).echo()
                except Exception as e:
                    print("Error: could not stop %s" % service)
                    print(e)
                    result = False

            print(" done.")

        return result

    @staticmethod
    def startLMBenchServices():
        return NetworkTest.__doLMBenchAction('start')

    @staticmethod
    def stopLMBenchServices():
        return NetworkTest.__doLMBenchAction('stop')

    def __doLMBenchActionOnLTS(self, verb):
        testName = "NetworkTest"
        testPackageTrail = "hwcert.hwcert.networkTest"
        startLMBFunction = "startLMBenchServices"
        stopLMBFunction = "stopLMBenchServices"
        requiresRoot = True

        if verb not in ['start', 'stop']:
            return None

        if verb == 'start':
            collabRequest = CollabRequest(testName, testPackageTrail, startLMBFunction, requiresRoot)
        else:
            collabRequest = CollabRequest(testName, testPackageTrail, stopLMBFunction, requiresRoot)

        try:
            collab_response = self.collabCall(collabRequest)
        except (ValueError, requests.exceptions.HTTPError) as e:
            print("Error: %s" % str(e))
            return False

        print("Response from test server: %s" % rhcert_make_str(collab_response))
        if "True" not in rhcert_make_str(collab_response):
            return False
        return True

    def startLMBenchServicesOnLTS(self):
        return self.__doLMBenchActionOnLTS('start')

    def stopLMBenchServicesOnLTS(self):
        return self.__doLMBenchActionOnLTS('stop')

    def run(self):

        if not self.runSubTest(self.prepareForRun, "Setup", "prepare for network testing"):
            return False

        if not self.runSubTest(NetworkTest.startLMBenchServices, "Start LMBench Services on system-under-test"):
            return False

        if not self.runSubTest(self.startLMBenchServicesOnLTS, "Start LMBench Services on test server"):
            return False

        if not self.runSubTest(self.configureInterfaces, "Interface Configuration"):
            return False

        if not self.runSubTest(self.printInfo, "Info"):
            return False

        if not self.runSubTest(self.deviceNaming, "Device Naming"):
            return False

        # start testing
        success = True

        if not self.runSubTest(self.tcpTestLatency, "TCP-latency", "tcp latency test via lmbench"):
            success = False
        if not self.runSubTest(self.tcpTestBandwidth, "TCP-throughput", "tcp throughput test via lmbench"):
            success = False
        if not self.runSubTest(self.udpTest, "UDP", "udp latency test via lmbench"):
            success = False
        if not self.runSubTest(self.httpTransferTest, "HTTP", "HTTP file transfer test"):
            success = False
        if not self.runSubTest(self.icmpTest, "ICMP"):
            success = False

        if not self.runSubTest(NetworkTest.stopLMBenchServices, "Stop LMBench Services on system-under-test"):
            success = False

        if not self.runSubTest(self.stopLMBenchServicesOnLTS, "Stop LMBench Services on test server"):
            success = False

        self.setSignalHandler(None)
        if self.isNFSRootSystem:
            self.removeRoute()
        else:
            self.restoreAllOtherInterfaces()

        if success:
            return True

        # otherwise
        return False

if __name__ == "__main__":

    networkTest = NetworkTest()
    networkTest.deviceNaming()


