#!/usr/bin/env python
# Copyright (c) 2006 Red Hat, Inc. All rights reserved. This copyrighted material
# is made available to anyone wishing to use, modify, copy, or
# redistribute it subject to the terms and conditions of the GNU General
# Public License v.2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# Author: Greg Nichols
#         YangKun <ykun@redhat.com>
#         Caspar Zhang <czhang@redhat.com>
# This class is a base class for optical disk (CD, DVD, BluRay) testing.

import os
import sys
import shutil
import time
import filecmp


from rhcert.command import Command, HwCertCommandException
from rhcert.tags import Constants
from rhcert.suites.hwcert.hwcert.udevDeviceDetector import UdevDeviceDetector
from rhcert.suites.hwcert.hwcert.hardwareTest import HardwareTest

class OpticalDiskTest(HardwareTest):

    def __init__(self, path):
        path = "optical/" + path
        HardwareTest.__init__(self, path)
        self.mediaTypes = None # subclasses must set this list
        self.supportedMedia = None
        self.deviceClass = "CDROM"

    def plan(self, planner):
        tests = dict()
        # try udev first
        devices = planner.getUdevDeviceDetector().getDevices()
        for device in devices:
            disk = None
            if device.getProperty("ID_CDROM") == "1" and self.hasAnyCDROMType(device):
                disk = self.getDeviceName(device)

            if disk and not disk in tests:
                test = self.makeCopy()
                test.setDevice(device)
                test.setLogicalDeviceName(disk)
                tests[disk] = test

        if tests:
            return tests.values()

    def getRequiredRPMs(self):
        # mkisofs included in genisoimage
        # cdrecord included in wodim
        rpms = [ "dvd+rw-tools", "genisoimage", "wodim", "util-linux" ]
        return rpms

    def hasAnyCDROMType(self, device):
        for media in self.mediaTypes:
            if device.getProperty("ID_CDROM_%s" % self.convertMediaNameUdev(media)) == "1":
                return True
        return False

    def getDeviceName(self, device):
        disk = device.getProperty("DEVNAME")
        """ chop off /dev/ (5 characters) """
        if disk and len(disk) > 5:
            disk = disk[5:]
            return disk
        return None

    def getMediaChoice(self):
        print("\n  Your device /dev/%s supports %s \n" % (self.getLogicalDeviceName(), self.showSupportedMedia()))

        # default to the most-capable media
        self.mediaChoice = self.supportedMedia[0]
        if not self.promptConfirm("Would you like to test %s media?" % self.printMediaName(self.mediaChoice)):
            choice = self.select("Please choose the media to be tested: ", list(map(self.printMediaName, self.supportedMedia)))
            self.mediaChoice = self.convertMediaName(choice)

    def echoTest(self, deviceName, testType, mediaStatus):
        os.system("eject /dev/%s 2>/dev/null" % deviceName)
        print("")
        self.promptContinue("Please insert a %s %s disc into /dev/%s" % (mediaStatus, self.printMediaName(testType), deviceName))
        # try to close the tray
        try:
            Command("eject -t /dev/%s" % deviceName).run()
        except HwCertCommandException as exception:
            # for those slim-line cdrom drives, auto-close can not work, have to close the tray manually
            self.promptContinue("Could not auto-close the tray, please close the tray manually")
        print("  Tray closed ...")
        sys.stdout.flush()
        time.sleep(25)

    def checkMediaType(self, deviceName, mediaType):
        """verify the media type"""
        print("\n  Verifiying media type: %s ... " % self.printMediaName(mediaType))
        sys.stdout.flush()
        try:
            mediaInDrive = Command("dvd+rw-mediainfo /dev/%s " % deviceName).getString(regex="Mounted Media:.*", singleLine=False, ignoreErrors=True).split(":")[1].strip().split(" ")[1].strip()
            mediaInDrive = self.convertMediaName(mediaInDrive)
        except HwCertCommandException as exception:
            print("Error: could not verify the media type.")
            print(exception)
            return False
        if mediaType == mediaInDrive:
            print("  Media Matched: Media in drive now is %s\n" % self.printMediaName(mediaInDrive))
            return True
        elif mediaType == "bd" and (mediaInDrive == "bdrom" or mediaInDrive in self.mediaTypes):
            print("  Required media is %s , media in drive now is %s , test continues ... \n" % (mediaType, self.printMediaName(mediaInDrive)))
            return True
        else:
            print("  Error: media is %s, but %s is required\n" % (self.printMediaName(mediaInDrive), self.printMediaName(mediaType)))
            return False

    def handleEject(self, cdromDevice):
        """testing the eject and reload functions. also some devices (e.g. Philips) need to eject the medium before creating a new disk."""
        deviceFile = "/dev/" + cdromDevice
        print("  Reloading the media ... ")
        sys.stdout.flush()
        try:
            Command("eject %s" % deviceFile).run()
            print("  tray ejected.")
            sys.stdout.flush()
        except HwCertCommandException as exception:
            # or the tray is already auto-ejected, so continue
            pass

        try:
            Command("eject -t %s" % deviceFile).run()
            print("  tray auto-closed.\n")
            sys.stdout.flush()
        except HwCertCommandException as exception:
            self.promptContinue("Could not auto-close the tray, please close the tray manually")
        time.sleep(25)
        return True

    def cmpTree(self, dir1, dir2):
        """compare regular files in dir1 and dir2; returns true if all match."""
        for dirpath, dirnames, filenames in os.walk(dir1):
            for name in filenames:
                file1 = os.path.join(dirpath, name)
                file2 = file1.replace(dir1, dir2, 1)
                if os.path.isfile(file1) and not os.path.islink(file1):
                    if filecmp.cmp(file1, file2):
                        continue
                    else:
                        print("  +++ Error: File comparison failed on file : %s\n" % file1)
                        return False
                else:
                    continue
        return True

    def getWriteMode(self):
        """return -tao if the drive supports TAO recording, or
           return -sao if supports SAO, use default otherwise"""
        deviceFile = "/dev/" + self.getLogicalDeviceName()
        writeMode="-sao"
        try:
            command = Command("cdrecord dev=/dev/%s -checkdrive" % self.getLogicalDeviceName())
            # setting ignoreErrors=True because cdrecord writes bogus info to stderr
            modes = command.getString(regex="^Supported modes[^:]*:(?P<modes>.*$)", regexGroup="modes", singleLine=False, ignoreErrors=True)
            if "TAO" in modes:
                writeMode="-tao"
            if "SAO" in modes:
                writeMode="-sao"

            # setting ignoreErrors=True because cdrecord writes bogus info to stderr
            flags = command.getString(regex="^Driver flags[^:]*:(?P<flags>.*$)", regexGroup="flags", singleLine=False, ignoreErrors=True)
            if "BURNFREE" in flags:
                writeMode += " driveropts=burnfree"

        except HwCertCommandException as exception:
            print("Warning: " + exception.message)

        return writeMode

    def getSupportedMedia(self):
        """configure the cdrom based on Udev info, return all supported features.
           we check CD/CD_R/CD_RW/DVD/DVD_R/DVD_RW/DVD_PLUS_R/DVD_PLUS_RW/BD/BD_R/BD_RE"""
        self.supportedMedia = list()
        if self.getUDI():
            cdromDevice = UdevDeviceDetector().getDevice(self.getUDI())
            if not cdromDevice:
                print("Error: could not find device with udi = %s" % self.getUDI())
                return # return empty list

            for mediaType in self.mediaTypes:
                if cdromDevice.getProperty("ID_CDROM_%s" % self.convertMediaNameUdev(mediaType)):
                    self.supportedMedia.append(mediaType) # return the most-preferred (first)

            # always add least-preferred one (last).
            if not mediaType in self.supportedMedia:
                self.supportedMedia.append(mediaType)

    def showSupportedMedia(self):
        """ pretty-print supported media list """
        mediaString = ""
        for media in self.supportedMedia:
            if len(mediaString) > 0:
                mediaString += ", "
            mediaString += self.printMediaName(media)
        return mediaString

    def printMediaName(self, media):
        mediaName = media.upper()
        if "PLUS" in mediaName:
            mediaName = mediaName.replace("PLUS", "+")
        elif "DVD" in mediaName and mediaName != "DVD":
            mediaName = mediaName.replace("DVD", "DVD-")
        elif "CD" in mediaName and mediaName != "CD":
            mediaName = mediaName.replace("CD", "CD-")
        elif mediaName == "CD":
            mediaName = "CDROM"
        elif "BD" in mediaName and mediaName != "BD":
            mediaName = mediaName.replace("BD", "BD-")
        return mediaName.lower()

    # cdrw      -> CD_RW
    # cdr       -> CD_R
    # cdrom     -> CD
    # dvdrw     -> DVD_RW
    # dvdplusrw -> DVD_PLUS_RW
    # dvdr      -> DVD_R
    # dvdplusr  -> DVD_PLUS_R
    # dvd       -> DVD
    # bdre      -> BD_RE
    # bdr       -> BD_R
    # bd        -> BD
    def convertMediaNameUdev(self, name):
        mediaName = name.upper()
        if "PLUS" in mediaName:
            mediaName = mediaName.replace("PLUS", "PLUS_")
        if "CD" in mediaName and mediaName != "CDROM":
            mediaName = mediaName.replace("CD", "CD_")
        elif mediaName == "CDROM":
            mediaName = "CD"
        elif "DVD" in mediaName and mediaName != "DVD":
            mediaName = mediaName.replace("DVD", "DVD_")
        elif "BD" in mediaName and mediaName != "BD":
            mediaName = mediaName.replace("BD", "BD_")
        return mediaName

    def convertMediaName(self, name):
        name = name.lower()
        name = name.replace("-", "")
        name = name.replace(":", "")
        name = name.replace("+", "plus")
        return name

    def checkMediaSize(self, size):
        print("Error: abstract OpticalDiskTest.checkMediaSize call")
        return False

    def reader(self):
        """handles CD-ROM or DVD-ROM or BD"""
        cdromDevice = self.getLogicalDeviceName()
        if os.path.exists("tmp"):
            shutil.rmtree("tmp")
        if os.path.exists("device"):
            shutil.rmtree("device")
        os.mkdir("tmp")
        os.mkdir("device")
        try:
            print("Mounting media ...")
            Command("umount /dev/%s" % cdromDevice).echoIgnoreErrors()
            Command("mount -o ro /dev/%s ./device" % cdromDevice).echo()
            print("Media mounted !")

            # check mediaSize
            # we'll take a guess that anything bigger than 800MB is a DVD
            mediaSize = Command("df /dev/%s | tail -n1 | awk '{print $3}'" % cdromDevice).getInteger()
            if not self.checkMediaSize(mediaSize):
                return False

            if os.path.exists("deviceCopy"):
                shutil.rmtree("deviceCopy")
            # "shutil.copytree" can not handle file copying here,
            # use "cp" directly
            print("  Copying files ... ")
            sys.stdout.flush()
            Command("cp -dpR ./device ./deviceCopy").echo()

            print("  Comparing files, this may take some time ... ")
            sys.stdout.flush()
            if self.cmpTree("device", "deviceCopy"):
                print("  +++ Compare data succeeded ! \n")
                returnValue = True
            else:
                print("  +++ Compare data failed ! \n")
                returnValue = False
            Command("umount ./device").echoIgnoreErrors()
            return returnValue
        except HwCertCommandException as exception:
            print(exception)
            return False

    def compare(self, deviceType, cdromDevice):
        Command("umount /dev/%s/" % cdromDevice).echoIgnoreErrors()
        try:
            if os.path.exists("media"):
                print("removing media tree")
                shutil.rmtree("media")
            os.mkdir("media")
        except Exception as exception:
            print(exception)
            print("Error: could not create \"media\" mount point")
            return False

        print("Mounting /dev/%s to ./media " % cdromDevice)
        sys.stdout.flush()
        tries = 0
        maxTries = 10
        mounted = False
        waitBeforeRetry = 5 # sec
        mount = "mount -o ro /dev/%s ./media " % cdromDevice
        while not mounted and (tries < maxTries):
            try:
                print(mount)
                sys.stdout.flush()
                Command(mount).echo()
                mounted = True
            except HwCertCommandException as exception:
                tries += 1
                time.sleep(waitBeforeRetry)
                print("Mount failed, Retrying...")
                sys.stdout.flush()

        if tries >= maxTries:
            print("Error: could not mount media")
            sys.stdout.flush()
            return False

        print("  Start to compare data, this may take some time, please be patient ... ")
        sys.stdout.flush()
        if self.cmpTree(self.fileSourceDir, "media"):
            print("  +++ Compare data succeeded ! \n")
            returnValue = True
        else:
            print("  +++ Compare data failed ! \n")
            returnValue = False
        Command("umount ./media").echo()
        os.rmdir("media")
        return returnValue

    def rewriter(self, deviceType):
        """handles CD-RW or DVD-RW or DVD+RW or BD-RE"""
        cdromDevice = self.getLogicalDeviceName()
        # Unmount the drive first in case it's mounted
        Command("umount /dev/%s" % cdromDevice).echoIgnoreErrors()

        try:
            # cd-rw or dvd-rw
            if deviceType == "cdrom" or deviceType == "dvd":
                # Blank the disk first
                print("\n  +++ Blanking the disk, this may take some time, please be patient ...")
                sys.stdout.flush()
                blankCommand = Command("cdrecord -silent dev=/dev/%s blank=fast > /dev/null 2>&1" % cdromDevice)
                if self.getDebug() != Constants.off:
                    print(" Running command:\n%s" % blankCommand.command)
                blankCommand.echoIgnoreErrors()
                if blankCommand.returnValue == 0:
                    print("  +++ Blank disk done ! \n")
                else:
                    print("  +++ Error: blanking failed ! \n")
                    return False
                # otherwise
                self.handleEject(cdromDevice)
                sys.stdout.flush()
                # re-write now
                returnValue = self.writer(deviceType)
            # dvd+rw
            elif deviceType == "dvdplus":
                # bz 332582 - added to close tray before blanking
                self.handleEject(cdromDevice)
                # blank the disk
                print("\n  +++ Formatting ... ")
                sys.stdout.flush()
                Command("dvd+rw-format -force /dev/%s 2>/dev/null" % cdromDevice).echo()
                print("  +++ Formatting done ! \n")
                self.handleEject(cdromDevice)
                sys.stdout.flush()
                # re-write now
                returnValue = self.writer(deviceType)
            # bdre
            elif deviceType == "bd":
                # blank the disk (if not done before)
                status_command = Command("dvd+rw-mediainfo /dev/%s" %cdromDevice)
                status = status_command.getString(ignoreErrors=True, singleLine=False, regex="(?<= Disc status:).*").strip()
                print("\n  +++ Formatting ... This may take a long time, please be patient ... ")
                sys.stdout.flush()
                # for some reason, the below command does not work without -ssa=default
                Command("dvd+rw-format -format=full -ssa=default /dev/%s 2>/dev/null" % cdromDevice).echo()
                print("  +++ Formatting done ! \n")
                self.handleEject(cdromDevice)
                sys.stdout.flush()
                # re-write now
                returnValue = self.writer(deviceType)
            else:
                print("\n  +++ Error: Un-supported media .\n")
                return False
            return returnValue
        except HwCertCommandException as exception:
            print(exception)
            return False

    def turnOffAutoMountAndAutoBrowse(self):
        try:
            Command("gconftool-2 --type bool --set /desktop/gnome/volume_manager/automount_media false").echo()
            Command("gconftool-2 --type bool --set /desktop/gnome/volume_manager/autobrowse false").echo()
            Command("gconftool-2 --type bool --set /apps/nautilus/preferences/media_automount False").echo()
            Command("gconftool-2 --type bool --set /apps/nautilus/preferences/media_automount_open False").echo()
        except HwCertCommandException as exception:
            print(exception)
            print("Warning: Can not turn off the automount and/or autobrowse for removable media.")

    def initializeDeviceInfo(self):
        self.getSupportedMedia()
        if not self.supportedMedia:
            print("Error: No cdrom device found")
            return False
        self.turnOffAutoMountAndAutoBrowse()
        return True

    def testDisk(self, message, readWriteMode, type):
        print("\n  +++ Start testing ...")
        sys.stdout.flush()
        if readWriteMode == "read":
            return self.reader()
        if readWriteMode == "write":
            return self.writer(type)
        if readWriteMode == "rewrite":
            return self.rewriter(type)

        print("  Error: unknown read/write mode %s" % readWriteMode)
        return False

if __name__ == "__main__":
    test = OpticalDiskTest()
    returnValue = test.do(sys.argv)
    sys.exit(returnValue)
