#!/usr/bin/env python
#
# Author: Greg Nichols

from rhcert.suites.hwcert.hwcert.device import ProcDevice
from rhcert.suites.hwcert.hwcert.deviceDetector import DeviceDetector
from rhcert.tags import Constants
from rhcert.command import Command

class ProcDeviceDetector(DeviceDetector):
    "A class for detecting devices via the /proc file system"
    # some well-known paths into proc
    inputBusPath = "/proc/bus/input/devices"
    procCPUInfo = "/proc/cpuinfo"
    procMeminfo = "/proc/meminfo"
    # pciBusPath = "/proc/bus/pci/devices" - different syntax, nothing to parse this yet.  See lspci

    def __init__(self):
        DeviceDetector.__init__(self, name=Constants.proc)
        self.nextKey = None
        self.devices = None
        self.path = None

    def loadDevices(self, path):
        self.path = path
        self.devices = list()
        self.__loadDevices()
        return self.devices

    def __loadDevices(self):
        # treat the file as a bunch of attrubute=value strings
        # seperated by blank lines for each device
        try:
            pipe = Command("cat " + self.path)
            pipe.start()
            properties = dict()
            while 1:
                line = pipe.readline()

                # blank or empty line (EOF)
                if not line or line.strip() == "":
                    # save the device
                    if len(properties) > 0:
                        self.devices.append(ProcDevice(properties))
                        properties = dict()

                    if not line:
                        break
                    else:
                        continue

                # otherwise, read the line and add it to the device

                # cpuinfo and meminfo are just <attribute> : <value>
                if self.path == ProcDeviceDetector.procCPUInfo or self.path == ProcDeviceDetector.procMeminfo:
                    tokens = line.split(":", 1)
                    if len(tokens) == 2:
                        properties[tokens[0].strip()] = tokens[1].strip()
                    continue

                # otherwise, more complex parsing ala /proc/bus/input/devices
                tokens = line.split(":", 1)[1]
                if tokens.count("=") == 1:
                    self.__addKeyValue(tokens, properties)
                else:
                    keyvalues = tokens.split(" ")
                    for keyvalue in keyvalues:
                        self.__addKeyValue(keyvalue, properties)
        except Exception as e:
            print("Warning: could not read devices from proc")
            print(e)
            pass

    def __addKeyValue(self, keyvalue, properties):
        keyvalue = keyvalue.split("=", 1)
        key = ""
        try:
            key = keyvalue[0].strip()
        except IndexError:
            return

        try:
            value = keyvalue[1].strip()
        except IndexError:
            value = ""

        if key != "":
            if value != "":
                properties[key] = value
                return None
            elif self.nextKey:
                properties[self.nextKey] = key # really a value
                self.nextKey = None
            else:
                self.nextKey = key

    def getDevices(self, path):
        self.loadDevices(path)
        return self.devices

    def getDevice(self, udi, path):
        self.loadDevices(path)
        if self.devices:
            for device in self.devices:
                if device.hasPropertyValue(udi):
                    return device
        # otherwise
        return None

    def dump(self):
        for device in self.devices:
            print(device.getProperties())

    def dumpDevice(self, className=None, busName=None, logicalDeviceName=None):
        output = ""
        if not className and not busName and not logicalDeviceName:
            return ""

        for device in self.devices:
            if logicalDeviceName:
                if device.hasPropertyValue(logicalDeviceName):
                    output += device.dump()
            if className:
                if device.getClass() == className:
                    output += device.dump()
            if busName:
                if device.getBus() == busName:
                    output += device.dump()
        return output

    def getProduct(self, device):
        return device.getProperty("Product")


if __name__ == "__main__":
    detector=ProcDeviceDetector()
    detector.loadDevices(ProcDeviceDetector.procMeminfo)
    print(detector.dump())
