#!/usr/bin/env python2
#
# Copyright (c) 2016 Red Hat, Inc. All rights reserved. This copyrighted material 
# is made available to anyone wishing to use, modify, copy, or 
# redistribute it subject to the terms and conditions of the GNU General 
# Public License v.2.
# 
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# Author: Benjamin Berg <bberg@redhat.com>
#

"""
The purpose of this module is to allow creating a new session with an X server
that correctly handles required VT switches and is properly registered with
logind.

The core of the issue is that the VT switching is a bit racy, as logind might
still have devices assigned to the main session when the VT switch has happened.
This can result in X not picking up the displays and other weird issues.

By going through logind, registering the session and waiting for logind to
announce that our new session is active, we can ensure that the above race
condition is fixed. The code will fall back to work without logind and doing
plain old VT switches with extra wait period to still work around any issues
like the above.
"""

from __future__ import with_statement

import os
import sys
import fcntl
import struct
import random
import signal
import atexit
import time
import threading
import errno

# Ensure that a sigterm handler is registered
def _sigterm_handler(*args):
    if _orig_sigterm:
        _orig_sigterm(*args)

_orig_sigterm = signal.signal(signal.SIGTERM, _sigterm_handler)

# IOCTLs for VT handling
VT_OPENQRY=0x5600 # Find available VT
VT_GETSTATE=0x5603 # Get global VT state info
VT_ACTIVATE=0x5606
VT_WAITACTIVE=0x5607
TIOCSCTTY=0x540E  # Open the VT

_system_bus = None
_systemd_manager_iface = None
_login_manager_iface = None
try:
    import dbus
    _system_bus = dbus.SystemBus()

    _systemd_manager_proxy = _system_bus.get_object('org.freedesktop.systemd1', '/org/freedesktop/systemd1')
    _systemd_manager_iface = dbus.Interface(_systemd_manager_proxy, dbus_interface='org.freedesktop.systemd1.Manager')

    _login_manager_proxy = _system_bus.get_object('org.freedesktop.login1', '/org/freedesktop/login1')
    _login_manager_iface = dbus.Interface(_login_manager_proxy, dbus_interface='org.freedesktop.login1.Manager')
except:
    pass


def _set_cloexec(fd):
    flags = fcntl.fcntl(fd, fcntl.F_GETFD)
    flags |= fcntl.FD_CLOEXEC
    fcntl.fcntl(fd, fcntl.F_SETFD, flags)

def _waitpid(pid, options):
    while True:
        try:
            pid, res = os.waitpid(pid, options)
            if pid:
                return res
            else:
                return None
        except OSError as e:
            if e.errno not in [errno.EINTR, errno.ECHILD]:
                print('Received unexpected error while waiting on pid:')
                print(e)
                raise e


class Session(object):

    xorg_args = [ '{display}', '-seat', 'seat0', '-nolisten', 'tcp', 'vt{vtnr}', '-novtswitch', ]

    # List on the class of objects that need killing if the interpreter exits
    _kill_atexit = []

    def __init__(self, terminate=False):
        # vtnr will be returned when creating the session
        self.vtnr = 0
        self.service = 'test-session'
        self.desktop = 'none'
        # Hardcoded for now
        self.display = ':10'
        self.session_name = None
        self.tty = None
        self.leader_pid = -1
        self._returncode = None
        self.session_fd = -1
        self.terminate = terminate

    def _get_active_vt(self):
        devnode = open('/dev/tty0')
        res = fcntl.ioctl(devnode, VT_GETSTATE, struct.pack('HHH', 0, 0, 0))
        del devnode

        # Struct contains:
        # short: active vt
        # short: signal to send
        # short: bitmask of used VTs

        return struct.unpack('HHH', res)[0]

    def __enter__(self):
        self.run()

    def __exit__(self, exc_type, exc_value, traceback):
        self.kill()

    def _allocate_vt(self):
        devnode = open('/dev/tty0')
        assert devnode >= 0

        # Ask the kernel for a free VT
        res = fcntl.ioctl(devnode, VT_OPENQRY, struct.pack('i', 0))
        del devnode
        self.vtnr = struct.unpack('i', res)[0]

        while True:
            try:
                self.tty = '/dev/tty%d' % self.vtnr
                self.tty_file = open(self.tty, 'rw')
                # Param: force
                fcntl.ioctl(self.tty_file, TIOCSCTTY, struct.pack('i', False))

                return
            except IOError as e:
                print('Error:', e)
                # We could not allocate the VT, try a few more (up to 15)
                self.vtnr += 1
                if self.vtnr >= 16:
                    raise AssertionError('Could not find free VT')

    def _activate(self, session_oid, vtnr):
        vtnr = int(vtnr)
        # Ensure vtnr is an integer and not something else for the ioctl
        if session_oid:
            proxy = _system_bus.get_object('org.freedesktop.login1', session_oid)
            session_iface = dbus.Interface(proxy, dbus_interface='org.freedesktop.login1.Session')

            try:
                session_iface.Activate()
                prop_iface = dbus.Interface(proxy, dbus_interface='org.freedesktop.DBus.Properties')
                # Yeah, this should just listen for the Active change signal
                timeout = 100
                while timeout > 0:
                    if prop_iface.Get('org.freedesktop.login1.Session', 'Active'):
                        return

                    time.sleep(0.01)
                    timeout -= 1
            except dbus.exceptions.DBusException as e:
                 if e.get_dbus_name() != 'org.freedesktop.DBus.Error.NotSupported':
                     raise e

        # Fallback. Switch VT, wait for the switch, then wait a bit longer to
        # ensure logind has done its magic.
        devnode = open('/dev/tty0')

        done = False
        def activate_func():
            timeout = 20
            while not done and timeout > 0:
                fcntl.ioctl(devnode, VT_ACTIVATE, vtnr)
                timeout -= 1
                time.sleep(0.1)

        # Really try hard to switch back. Otherwise we might end up waiting
        # for a switch that is never going to happen.
        thread = threading.Thread(target=activate_func)
        thread.start()
        try:
            res = fcntl.ioctl(devnode, VT_WAITACTIVE, vtnr)
        except IOError as e:
            print('Error waiting for VT activation')
            print(e)
            pass
        done = True
        thread.join(timeout=1)
        time.sleep(0.5)

    def _create_transient_scope(self):
        """Create a temporary scope. This ensures that the process is not
        running under any user session anymore."""

        name = 'create-session-%d.scope' % random.randint(100000, 999999)
        mode = 'fail'
        properties = dbus.Array(signature='(sv)')
        properties.append(dbus.Struct(('Description', 'Transient scope to start graphical session'), signature="sv"))
        pids = dbus.Array(signature='u')
        pids.append(self.leader_pid)
        properties.append(dbus.Struct(('PIDs', pids), signature="sv"))

        aux_units = dbus.Dictionary(signature='sa(sv)')

        self.scope = _systemd_manager_iface.StartTransientUnit(name, mode, properties, aux_units)

    def _create_session(self, cseat='seat0'):
        uid = os.getuid()

        _type='x11'
        _class='user'

        remote = False
        remote_user = ''
        remote_host = ''

        res = _login_manager_iface.CreateSession(
                uid, self.leader_pid,
                self.service, _type, _class, self.desktop,
                cseat, 0, self.tty, self.display,
                remote, remote_user, remote_host,
                dbus.Dictionary(signature='sv'))

        self.session_name = str(res[0])
        self.session_oid = str(res[1])
        self.runtime_dir = str(res[2])
        # We just need to keep this open
        self.session_fd = res[3].take()
        _set_cloexec(self.session_fd)

        session_seat = str(res[5])
        assert session_seat == cseat
        self.vtnr = int(res[6])

    def _run_x(self):
        args = self.xorg_args[:]
        for i in xrange(len(args)):
            args[i] = args[i].replace('{display}', self.display)
            args[i] = args[i].replace('{vtnr}', str(self.vtnr))
            args[i] = args[i].replace('{tty}', self.tty)

        if self.terminate:
            args.append('-terminate')

        # Setup a sigusr1 receive handler to know when the X server is ready
        r, w = os.pipe()
        def _sigusr1_handler(*args):
            os.write(w, ' ')

        orig_signal = signal.signal(signal.SIGUSR1, _sigusr1_handler)

        pid = os.fork()
        if pid == 0:
            os.close(r)
            os.close(w)
            self.tty_file.close()

            # This means we will get a signal back to the parent
            signal.signal(signal.SIGUSR1, signal.SIG_IGN)
            os.execv('/usr/bin/X', ['X',] + args)

        m = ''
        while not m:
            try:
                m = os.read(r, 1)
            except OSError as e:
                # Catch the interrupt
                if e.errno != errno.EINTR:
                    raise e

        os.close(r)
        os.close(w)

        # The X server may send a signal sometimes, so ignore from now on.
        signal.signal(signal.SIGUSR1, signal.SIG_IGN)

        return pid

    def setup_leader(self):
        """
        This function is called in the session leader process after the
        TTY has been allocated and before it is activated.
        You can use this to modify the enviornment both of the started display
        server and all other applications.

        Note that changing the uid/gid is not possible anymore at this point
        as the outside process would not pick up on these changes.
        """
        pass

    def session_leader(self, server_pid):
        """
        This function is called after the X server has been started. The server
        pid is passed in as the first argument. If this function returns an
        os._exit(1) will be executed immediately.

        Unless you know what you are doing, any implementation of this function
        should end with an exeve call.
        """
        res = _waitpid(server_pid, 0)
        if res is not None:
            os._exit(res)
        os._exit(0)

    def _session_leader(self, r, w):
        """
        This function will be called in the forked processes. It should
        never return, if it does, then the outside scope will do a
        os._exit(1).

        To change the behaviour, you can override the setup_leader function
        of the object. This will be called an you can e.g. customize the
        environment at will.
        Some setup to the environment will already be done. In particular the
        DISPLAY and WAYLAND variables are set/cleared correctly.

        The arguments are two file desciptors. Into the first the allocated
        TTY is written, and the second will read a single byte once the
        TTY has been activated correctly.
        """

        # Make ourself the session leader
        os.setsid()

        self._allocate_vt()

        # Signal the VT we got back to the outside process
        os.write(w, self.tty)

        # And wait for the outside process to set things up and do the VT
        # switch. Note that the outside process waits longer than just for
        # VT activation, as logind needs to change the device ACLs.
        os.read(r, 1024)
        os.close(r)

        # Should we redirect stdin/stdout/stderr?
        #os.close(0)
        #os.close(1)
        #os.close(2)
        #os.dup2(self.tty_file.fileno(), 0)
        #os.dup2(self.tty_file.fileno(), 1)
        #os.dup2(self.tty_file.fileno(), 2)

        # Set DISPLAY and clear WAYLAND
        os.environ['DISPLAY'] = self.display
        os.environ['WAYLAND'] = ''

        x_pid = self._run_x()
        os.write(w, str(x_pid))
        os.close(w)
        self.session_leader(x_pid)


    def run(self):
        """Creates a new session on a new VT and runs X.

        A future extension might also run a client or python callback
        instead of just the X server.
        """

        self._orig_vt = self._get_active_vt()

        r1, w1 = os.pipe()
        r2, w2 = os.pipe()

        self.leader_pid = os.fork()
        if self.leader_pid == 0:
            os.close(r1)
            os.close(w2)

            try:
                self._session_leader(r2, w1)
            except Exception as e:
                print('Error: Session leader raised an exception but should never return.')
                print(e)
            finally:
                os._exit(1)

        # Setup the forked process as session leader
        try:
            self._create_transient_scope()
        except:
            pass

        os.close(w1)
        os.close(r2)

        self.tty = os.read(r1, 1024)
        assert(self.tty)

        try:
            self._create_session()
        except AssertionError as e:
            raise e
        except Exception as e:
            print('Falling back due to exception')
            print(e)
            # Session activation through dbus failed, we won't have a VT number
            self.vtnr = int(self.tty[8:])

        # We need to switch to the newly allocated VT so that the X server
        # can initialise properly, as the permissions are revoked otherwise.
        self._activate(self.session_oid, self.vtnr)
        os.write(w2, "1")
        os.close(w2)

        self.x_pid = int(os.read(r1, 1024))
        self.x_alive_fd = r1

        # Register an atexit handler to ensure we kill the session at some point
        self._kill_atexit.append(self)

        return self.leader_pid

    def wait(self, nohang=False):
        if self._returncode is not None:
            return self._returncode

        flags = 0
        if nohang:
            flags = os.WNOHANG
        res = _waitpid(self.leader_pid, flags)
        if res is not None:
            self._returncode = res
        return self._returncode

    @property
    def running(self):
        return self.wait(nohang=True) == None

    @property
    def server_alive(self):
        # Try reading a byte
        mode = fcntl.fcntl(self.x_alive_fd, fcntl.F_GETFL)
        fcntl.fcntl(self.x_alive_fd, fcntl.F_SETFL, mode | os.O_NONBLOCK)
        try:
            os.read(self.x_alive_fd, 1)
        except OSError as e:
            if e.errno == errno.EWOULDBLOCK:
                return True
        finally:
            fcntl.fcntl(self.x_alive_fd, fcntl.F_SETFL, mode)

        return False

    def _wait_x_stop(self):
        # Ensure the X server has shut down
        while True:
            try:
                # Read will block unless X is dead
                os.read(self.x_alive_fd, 1)
                return
            except OSError as e:
                if e.errno != errno.EBADF:
                    raise e

    def kill(self):
        if not self.leader_pid:
            # This can e.g. happen if one uses a try:/finally: in the wrong way
            print("Warning: Cleanup handler called without a leader. Maybe the fork'ed session leader allowed python to do cleanups?")
            return

        if self in self._kill_atexit:
            self._kill_atexit.remove(self)

        print('Killing session %s' % self.session_name)

        try:
            if _login_manager_iface:
                _login_manager_iface.KillSession(self.session_name, 'all', signal.SIGTERM)
                self._wait_x_stop()
                _login_manager_iface.KillSession(self.session_name, 'all', signal.SIGKILL)
        except dbus.exceptions.DBusException as e:
            # Session might be gone already, file not found happens if the session
            # has died on its own accord.
            if e.get_dbus_name() not in ['org.freedesktop.DBus.Error.FileNotFound', ]:
                print('Unexpected exception:', e)

        # Kill the X server should it still be alive
        try:
            if self.server_alive:
                # Kill the server if it is still alive
                os.kill(self.server_pid, signal.SIGTERM)
        except OSError:
            pass

        # Kill the process group in case the above failed (fallback for RHEL 6)
        # Try a bit nicer if there is no manager iface
        try:
            if not _login_manager_iface:
                os.killpg(self.leader_pid, signal.SIGTERM)
                self._wait_x_stop()

            os.killpg(self.leader_pid, signal.SIGKILL)
        except OSError:
            pass

        # Clean up the leader process
        self.wait()
        # Close the FD after killing everything
        if self.session_fd >= 0:
            os.close(self.session_fd)

        # Try activating the original session through DBus
        try:
            orig_session = _login_manager_iface.GetSessionByPID(0)
        except dbus.exceptions.DBusException as e:
            if e.get_dbus_name() not in ['org.freedesktop.login1.NoSessionForPID']:
                print('Unexpected error trying to activate the user session')
            orig_session = None

        self._activate(orig_session, self._orig_vt)

    @classmethod
    def _atexit_handler(cls):
        for obj in cls._kill_atexit:
            obj.kill()

# Try to ensure we reap our children again
atexit.register(Session._atexit_handler)

class SessionExec(Session):
    def __init__(self, path, args, *initargs, **initkwargs):
        self.path = path
        self.args = args

        Session.__init__(self, *initargs, **initkwargs)

    def session_leader(self, server_pid):
        os.execv(self.path, self.args)

    @staticmethod
    def find_in_path(cmd):
        """
        Tries to find the executable in the path. Will *not* throw a
        proper error if it is not found.
        """
        # If there is no path component, then start searching
        if not os.path.dirname(cmd):
            for p in os.environ['PATH'].split(os.pathsep):
                # Might not allow symlinks?
                newcmd = os.path.join(p, cmd)
                if os.path.isfile(newcmd) and os.access(newcmd, os.X_OK):
                    return newcmd

        return cmd

class SessionFunc(Session):
    """Run the given function as session leader. Please see the caveats in
    the session_leader documentatino of the Session class."""
    def __init__(self, func, args=[], *initargs, **initkwargs):
        self.func = func
        self.args = args

        Session.__init__(self, *initargs, **initkwargs)

    def session_leader(self, server_pid):
        self.func(server_pid, *self.args)

    @staticmethod
    def find_in_path(cmd):
        """
        Tries to find the executable in the path. Will *not* throw a
        proper error if it is not found.
        """
        # If there is no path component, then start searching
        if not os.path.dirname(cmd):
            for p in os.environ['PATH'].split(os.pathsep):
                # Might not allow symlinks?
                newcmd = os.path.join(p, cmd)
                if os.path.isfile(newcmd) and os.access(newcmd, os.X_OK):
                    return newcmd

        return cmd

if __name__ == "__main__":

    if sys.argv[1:]:
        args = sys.argv[1:]

        session = SessionExec(SessionExec.find_in_path(args[0]), args, terminate=True)
    else:
        session = Session()

    with session:
        session.wait()

