#!/usr/bin/env python
# Copyright (c) 2006 Red Hat, Inc. All rights reserved. This copyrighted material
# is made available to anyone wishing to use, modify, copy, or
# redistribute it subject to the terms and conditions of the GNU General
# Public License v.2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# Author: Greg Nichols
#         YangKun <ykun@redhat.com>
#
import re, os, string, sys, random
import subprocess
import json

from rhcert.test import TestTag
from rhcert.command import Command, HwCertCommandException
from rhcert.tags import DeviceClass, Constants
from rhcert.compatability import rhcert_get_status_output

from rhcert.suites.hwcert.hwcert.hardwarePlanner import HardwarePlanner
from rhcert.suites.hwcert.hwcert.device import Device
from rhcert.suites.hwcert.hwcert.udevDeviceDetector import UdevDeviceDetector
from rhcert.suites.hwcert.hwcert.hardwareTest import HardwareTest


class StorageTest(HardwareTest):

    def __init__(self, path):
        path = "storage/" + path
        HardwareTest.__init__(self, path)
        self.hostPatternString = "(host|css|cciss|hspa|mmc|nvme|virtio|vbd-)[0-9]+"
        self.controllerPatternString = ""
        self.controllerPatternNotString = ""
        self.planning_methods = list()
        self.req_planning_methods = list()
        # Min and max blocksizes, in bytes. For each test, we loop through block sizes
        # starting with minBs and doubling until we hit maxBs. Both should be
        # powers of 2 in the range [1024-65536].
        self.defaultMinBs = 1024
        self.defaultMaxBs = 65536
        # Maximum size of the test area, in KB. (defaults to 1GB)
        self.maxSize = 1048576
        self.deviceClass = DeviceClass.hard_disk
        self.fileSystemType = "xfs"

    def getRequiredRPMs(self):
        return ["xfsprogs"]

    def plan(self, planner):
        """Plan the the test for the host as the logical device, with the udev device for the 1st disk """
        tests = list()
        disksByHost = self.getDisks(planner)
        for (host, devices) in disksByHost.items():
            if self.shouldPlan(host, devices, planner):
                test = self.makeCopy()
                test.setLogicalDeviceName(host)
                test.setDevice(devices[0])
                tests.append(test)
        return tests

    def shouldPlan(self, host, devices, planner):
        for planning_method in self.req_planning_methods:
            if not planning_method(host, devices, planner):
                # if any one planning method returns False, don't plan the test
                return False

        if len(self.planning_methods) == 0:
            return True

        for planning_method in self.planning_methods:
            if planning_method(host, devices, planner):
                # if any one of the planning method returns True, plan the test
                return True
        return False

    def planByDeviceUnique(self, host, devices, planner):
        """ plan the default test if another storage test has been planned for some other device """
        if hasattr(planner, "allTestsSoFar"):
            testsSoFar = planner.allTestsSoFar
            for test in testsSoFar:
                if TestTag.storage in test.tags():
                    # if some kind of storage test has already been planned for the same device
                    # do not plan the default storage test
                    if hasattr(test, "getLogicalDeviceName") and test.getLogicalDeviceName() == host:
                        print("INFO: For device {0} found another storage test planned, "\
                              "so we will skip the generic one".format(test.getLogicalDeviceName()))
                        return False
        return True #no storage test planned so far for the same device

    def planByUDI(self, host, devices, planner):
        """plan the test if the UDI of the device includes the UDI pattern string"""
        for device in devices:
            # ideally if UDI of one device of the host includes the pattern then all the devices
            # of that host will contain the same pattern
            if self.UDIPatternString in device.getUDI():
                return True

    def planByStorageM2(self, host, devices, planner):
        """
        TODO: Implement logic to plan M2 storage.
        Currently it will always return false, if required, we need to plan it manually.
        """
        return False

    def planByStorageU2(self, host, devices, planner):
        """
        TODO: Implement logic to plan U2 storage.
        Currently it will always return false, if required, we need to plan it manually.
        """
        return False

    def planByController(self, host, devices, planner):
        """plan the test if the controller name of the device includes the controller pattern string """
        diskControllers = self.getDiskAndControllerNames(devices)
        if (self.controllerPatternString in diskControllers[0]['controller']) and\
           (self.controllerPatternNotString not in diskControllers[0]['controller']):
            return True
        return False

    def planByLsscsi(self, host, devices, planner):
        """plan the test if the transport/interface type of the host includes the lsscsi pattern string """
        match = re.search(r"[0-9]+", host)
        # this planning method is currently used by StorageSATATest and StorageSASTest
        # where the host is of the pattern host[0-9]+
        if match:
            host_number = match.group(0)
            try:
                lsscsi_cmd = Command("lsscsi -H -t %s"%(host_number))
                lsscsi_cmd.run()
            except HwCertCommandException as error:
                print("Error in running command 'lsscsi -H -t %s'"%(host_number))
                print(error)
            else:
                if lsscsi_cmd.output[0].find(self.lsscsiPatternString) != -1:
                    return True
        return False

    def planByStorageSSD(self, host, devices, planner):
        """ plan the test if the device is a SSD"""
        for device in devices:
            if self.isStorageDeviceSSD(device):
                return True
        return False

    def logInfoStorageSSD(self):
        """print block and request size information, if block device is a SSD"""
        disksByHost = self.getDisks()
        for (host, devices) in disksByHost.items():
            for device in devices:
                device_path = device.getProperty("DEVNAME")
                if self.isStorageDeviceSSD(device):
                    lsblk_cmd = Command("lsblk -S %s -o NAME,PHY-SEC,LOG-SEC,MIN-IO,OPT-IO,ALIGNMENT" % device_path)
                    try:
                        lsblk_cmd.run()
                    except HwCertCommandException as error:
                            print("Warning: cannot run command '%s'\n" % lsblk_cmd)
                            print(error.message)
                    else:
                        print(lsblk_cmd.output)
        return True

    def isStorageDeviceSSD(self, device):
        """check if disk is SSD: SSD: 0, normal HDD:1"""
        device_name = device.getProperty("DEVNAME").lstrip("/dev/")
        rotational = "/sys/block/%s/queue/rotational" % device_name
        if not os.path.isfile(rotational):
            return False
        try:
            rotational_cmd = Command("cat %s" % rotational)
            rotational_cmd.run()
            if rotational_cmd.output[0] == "0":
                return True
        except HwCertCommandException as error:
            if self.getDebug() != Constants.off:
                print("Warning: cannot run command 'cat %s'\n" % rotational)
                print(error.message)
        return False

    def planByStorageNVDIMM(self, host, devices, planner):
        """ plan the test if the device is NVDIMM"""
        cmd = Command("ndctl list")
        try:
            cmd.run()
            namespaces = cmd.get_json_output()

            for device in devices:
                if device.getProperty("DEVTYPE") == 'disk':
                    device_name = device.getProperty("DEVNAME").lstrip("/dev/")
                    break
            for namespace in namespaces: # allow only sector mode (e.g. pmem0s.1) block dev
                if re.search('pmem\d+s(\.\d+)?', namespace.get('blockdev')) and device_name :
                    return True

        except HwCertCommandException as error:
            if self.getDebug() != Constants.off:
                print("Warning: cannot run command '%s'\n" % cmd)
                print(error.message)

        return False

    def getDisks(self, planner=None):
        """ build a dictionary of disks by hostname """
        disks = dict()
        # planner will be None when called during test execution
        if planner:
            deviceDetector = planner.getUdevDeviceDetector()
        else:
            deviceDetector = UdevDeviceDetector()
        # add manual device
        if planner and hasattr(planner,'options') and hasattr(planner.options, 'device'):
            self.setLogicalDeviceName(planner.options.device)
        devices = deviceDetector.getDevices()
        hostPattern = re.compile("\/(?P<host>%s)" % self.hostPatternString)
        for device in devices:
            # find disks that are not either partitions or virtual block devices
            if ((device.getProperty("ID_TYPE") and device.getProperty("ID_TYPE") == "disk"
                 or not device.getProperty("ID_TYPE") and device.getProperty("DEVTYPE") == "disk")):

                match = hostPattern.search(device.getProperty("DEVPATH"))
                if match:
                    host = match.group("host")

                    if not host in disks:
                        if self.getDebug() != Constants.off:
                            print("    Found host bus adapter: %s" % host)
                        disks[host] = list()
                    disks[host].append(device)
                elif self.getDebug() == Constants.high:
                    print("Couldn't find host for disk %s" % device.getProperty("DEVPATH"))

        return disks

    def getDiskNames(self, diskDevices):
        return [Device.getProperty(device, "DEVNAME").split('/')[-1] for device in diskDevices if device]

    def getDiskAndControllerNames(self, diskDevices):
        '''
        Ex DEVPATH: /devices/pci0000:00/0000:00:11.0/0000:02:00.0/host2/port-2:0/end_device-2:0/target2:0:0/2:0:0:0/block/sda
        last part of the path that looks like a PCI slot is the controller's slot
        Here 0000:02:00.0 is the controller's slot
        0000:02:00.0 Domain number:bus number:device number.function number

        Look up the controller name using its slot from the output of lspci command
        "lspci -D" command output:
        Ex: 0000:02:00.0 Serial Attached SCSI controller: Intel Corporation C604/X79 series chipset 4-Port SATA/SAS Storage Control Unit (rev 05)
        Here,
        0000:02:00.0 - Domain number:bus number:device number.function number
        Serial Attached SCSI controller - Device class
        Intel Corporation - Vendor
        C604/X79 series chipset 4-Port SATA/SAS Storage Control Unit - Name
        rev 05 - Revision
        '''
        disks = list()
        if not hasattr(self, 'lspci'):
            self.lspci = list()
            try:
                lspci_cmd = Command("lspci -D")
                lspci_cmd.run()
            except HwCertCommandException as error:
                print("Error in running command 'lspci -D'")
                print(error)
            else:
                self.lspci = lspci_cmd.output
        for diskDevice in diskDevices:
            disk = dict()
            disk['name'] = Device.getProperty(diskDevice, "DEVNAME").split('/')[-1]
            devpath = Device.getProperty(diskDevice, "DEVPATH")
            controller_pattern = re.compile("\/([0-9a-fA-F]{4}\:[0-9a-fA-F]{2}\:[0-9a-fA-F]{2}\.[0-9a-fA-F]{1,2})")
            controller_path = re.findall(controller_pattern, devpath)
            if controller_path:
                controller = controller_path[-1]
                for pci_device in self.lspci:
                    if pci_device.find(controller) != -1:
                        disk['controller'] = pci_device
                        break
            else:
                disk['controller'] = "Cannot find Controller"
            disks.append(disk)
        return disks

    def doDt(self, minBs, maxBs, size, options):
        """Helper function: run dt with various block sizes and reasonable defaults."""
        records = size * 1024 / maxBs
        bs = minBs
        while  bs <= maxBs:
            # dt will default enable verify
            (returnValue, junk) = rhcert_get_status_output("dt pattern=0xDEADBEEF bs=%d records=%d %s" % (bs, records, options))
            print(junk)
            print("returns %s" % returnValue)
            sys.stdout.flush()
            if returnValue != 0:
                print("Error: dt error !")
                return False
            bs = bs * 2
        return True

    def dtTest(self, minBs, maxBs, file, size, otherOptions):
        """Do a few passes of read/write testing on the given device."""
        # if debugging and interactive, prompt before diving into the testing
        if self.getDebug() != Constants.off and self.getMode() == Constants.normal:
            if not self.promptContinue("Starting dt testing"):
                print("Error: dt test skipped via prompt.")
                return False
        seq_opt = "of=%s iotype=sequential dispose=keep %s" % (file, otherOptions)
        # random passes
        # generate a random seed first
        seed = random.randint(0, 32767)
        if "direct" in otherOptions:
            otherOptions = otherOptions + " rlimit=%s" % (size * 1024)
        ran_opt = "of=%s iotype=random dispose=keep enable=verify ralign=%s rseed=%s dlimit=%s %s" % (file, minBs, seed, size, otherOptions)
        # start
        print("--- sequential write/read")
        sys.stdout.flush()
        returnValue1 = self.doDt(minBs, maxBs, size, seq_opt)
        print("--- random write/read")
        sys.stdout.flush()
        returnValue2 = self.doDt(minBs, maxBs, size, ran_opt)
        if returnValue1 and returnValue2:
            returnValue = True
        else:
            returnValue = False
        return returnValue

    def devSize(self, testdevice):
        # Use 1GB or the device size whichever is smaller.
        size = self.maxSize
        (status, output) = rhcert_get_status_output("cat /sys/block/%s/size" % testdevice)
        # Why is the /sys size value 2x what /proc shows?
        if status == 0:
            output = int(output) / 2
            if output < self.maxSize:
                size = output
            print("Testing device size %u" % size)
        else:
            print("Warning: can not get device size from /sys/block/%s/size, defaulting to %u. " % (testdevice, size))
        sys.stdout.flush()
        return size

    def dev_sector_size(self, testdev):
        size = 0
        (status, output) = rhcert_get_status_output("cat /sys/block/%s/queue/hw_sector_size" % testdev)
        if not status:
            size = int(output)
        else:
            print("Warning: can't get device sector size from /sys/block/%s/queue/hw_sector_size" % testdev)
        return size

    def deviceIsLvm(self, key, Device):
        """Return true if the given device is actually an LV, and that LV is only on
           the device we're trying to test."""
        tmp = '{print $5"\\\\s\\\\+"$6}'
        status, deviceNumber = rhcert_get_status_output("ls -l %s | tr -d , | awk '%s'" % (key, tmp))
        lvs = "lvs -okernel_major,kernel_minor,devices"
        if os.system("%s | grep -e '%s' | grep -q %s && [ $(%s | grep -e '%s' | wc -l) == 1 ]" % (lvs, deviceNumber, Device, lvs, deviceNumber)) == 0:
            return True
        else:
            return False

    def unmount(self, testdev):
        """Unmount, if the device is mounted"""
        mountFileTmp = open("/proc/mounts", "r")
        mountFile = mountFileTmp.read()
        mountFileTmp.close()
        deviceFile = "/dev/" + testdev
        if not deviceFile in mountFile and self.getDebug() != Constants.off:
            print("Warning: device %s not in /proc/mounts" % deviceFile)
        umount = "umount %s" % deviceFile
        print(umount)
        if os.system(umount) == 0:
            return True
        else:
            return False

    def getSwapLabel(self, swapdev):
        """try to get the swap label"""
        swapLabel = None
        if swapdev:
            try:
                blkidCommand = Command("blkid /dev/%s" % swapdev)
                swapString = blkidCommand.getString().split(" ")
                for swapStringItem in swapString:
                    if swapStringItem[:5] == "LABEL":
                        swapLabel = swapStringItem[7:-1]
                        print("----swap device has label '%s'----" % swapLabel)
            except HwCertCommandException as exception:
                # not really an error - just didn't find a label, carry on
                pass
        return swapLabel

    def restoreSwap(self, swapdev, swapLabel):
        """restore a swap device, given its short name (e.g. 'hda')"""
        swapFileTmp = open("/proc/swaps", "r")
        swapFile = swapFileTmp.read()
        swapFileTmp.close()
        if swapdev not in swapFile:
            swaponDevice = "/dev/%s" % swapdev
            if swapLabel:
                swaponDevice = "LABEL=%s" % swapLabel
                mkswapOption = "-L %s" % swapLabel
                print("restoring swap label '%s' ..." % swapLabel)
            else:
                mkswapOption = " "
                print("swap device has no label.")
            # restore the swap device
            print("restoring swap device /dev/%s ..." % swapdev)
            try:
                Command("umount /dev/%s" % swapdev).echo()
            except HwCertCommandException as e:
               #ignore errors
               if self.getDebug() != Constants.off:
                   print("umount failed:")
                   print(e)
               pass

            try:
                Command("mkswap %s /dev/%s" % (mkswapOption, swapdev)).run()
            except HwCertCommandException as exception:
                print("Warning: mkswap may have failed.")
                print(exception)
                exception.command.printErrors()
            try:
                Command("swapon %s" % swaponDevice).run()
                print("done.")
            except HwCertCommandException as exception:
                print("Error: could not restore swap")
                print(exception)
                exception.command.printErrors()
                return False
        if self.getDebug() != Constants.off:
            print("swap does not need restoring, it is in /proc/swaps")
        return True

    def grabHardwareInfo(self):
        """Gather some Hardware information for reviewers."""
        print("+-------- System Info Section Start --------+")
        Command("uname -a").echo(ignoreErrors=True)
        Command("cat /proc/partitions").echo(ignoreErrors=True)
        Command("cat /proc/swaps").echo(ignoreErrors=True)
        Command("cat /proc/diskstats").echo(ignoreErrors=True)
        Command("cat /proc/mounts").echo(ignoreErrors=True)
        Command("cat /etc/sysconfig/hwconf").echo(ignoreErrors=True)
        Command("fdisk -l").echo(ignoreErrors=True)
        Command("lsmod").echo(ignoreErrors=True)
        try:
            lspci_cmd = Command("lspci -D")
            lspci_cmd.echo()
        except HwCertCommandException as error:
            print("Error in running command 'lspci -D'")
            print(error)
        else:
            self.lspci = lspci_cmd.output
        print("+-------- System Info Section End --------+\n")
        return True

    def getTestDevice(self, storageDevice, swapdev, isLvm):
        """find the test device"""
        if swapdev:
            # Found a swap device!
            print("----swapdev is %s----" % swapdev)
            print("----lvm is %s----" % isLvm)
            print("Using swap device %s for testing." % swapdev)
            sys.stdout.flush()
            # Turn off swapping.
            if os.system("swapoff /dev/%s" % swapdev) is 0:
                testdev = swapdev
            else:
                print("Error: can not swapoff %s" % swapdev)
                return False
        else:
            # No swap, but if the disk is unused, we'll test with it.
            # Is it in use by LVM?
            if os.system("pvs | grep -q '/dev/%s' " % storageDevice) == 0:
                print("Error: %s is currently in use by LVM." % storageDevice)
                print("It cannot be tested. You may need to reinstall.")
                return False
            # Is it part of a RAID set?
            mdstatFileTmp = open("/proc/mdstat", "r")
            mdstatFile = mdstatFileTmp.read()
            mdstatFileTmp.close()
            if storageDevice in mdstatFile:
                print("Error: %s is part of a RAID set." % storageDevice)
                print("It cannot be tested. You may need to reinstall.")
                return False
            # Is there an active filesystem on it?
            mountsFileTmp = open("/proc/mounts", "r")
            mountsFile = mountsFileTmp.read()
            mountsFileTmp.close()
            mtabFileTmp = open("/etc/mtab", "r")
            mtabFile = mtabFileTmp.read()
            mtabFileTmp.close()
            if ("/dev/%s" % storageDevice) in mountsFile or ("/dev/%s" % storageDevice) in mtabFile:
                print("Error: %s has active filesystems on it." % storageDevice)
                print("It cannot be tested. You may need to reinstall.")
                return False
            testdev = storageDevice
            # You can't use a dasd without a partition table..
            if testdev[0:4] == "dasd" and testdev[0:5] == testdev:
                try:
                    Command("fdasd -a /dev/%s; sync" % testdev)  # sync avoids the udev race
                    Command("stat /dev/%s" % testdev)
                    testdev = testdev + "1"
                except HwCertCommandException as exception:
                    print("Error: /dev/%s1 was not created" % testdev)
                    return False
        print("----test device is %s----" % testdev)
        return testdev

    def testVFS(self, storageDevice, testdev, minBs, maxBs):
        """VFS (buffered filesystem) testing."""
        print("*** Testing buffered filesystem (%s) performance on %s" % (self.fileSystemType, testdev))
        # Create a new filesystem and mount it.
        oldMountdir = storageDevice + ".XXXXXX"
        # replace '/' with '-' for cciss devices
        # e.g. 'cciss/c0d0' -> 'cciss-c0d0'
        oldMountdir = oldMountdir.replace('/', '-')
        (status, mountdir) = rhcert_get_status_output("mktemp -d %s" % oldMountdir)
        if status == 0:
            dtTestFile = mountdir + "/dtTestFile"
        else:
            print("Error: can not make the mount dir.")
            return False
        if os.system("mkfs -t %s -f /dev/%s" % (self.fileSystemType, testdev)) == 0:
            if os.system("mount -t %s /dev/%s %s" % (self.fileSystemType, testdev, mountdir)) == 0:
                print("mount %s ok." % testdev)
            else:
                print("Error: mount %s failed." % testdev)
                return False
        else:
            print("Error: mkfs on %s failed." % testdev)
            return False
        (success, junkSize) = rhcert_get_status_output("df -Pk %s | awk '{print $4}' | tail -n1" % mountdir)
        if success == 0:
            size = self.maxSize
            junkSize = int(junkSize)
            if junkSize < self.maxSize:
                size = junkSize
            otherOptions = " enable=aio aios=8"
            if self.dtTest(minBs, maxBs, dtTestFile, size, otherOptions):
                if self.unmount(testdev):
                    print("umount ok")
                    return True
                else:
                    print("Error: umount error !")
                    return False
                print("filesystem test passed !")
            else:
                self.unmount(testdev)  # try to umount anyway
                print("Error: filesystem test failed !")
                return False
        else:
            print("Error: can not get mountdir size.")
            return False

    def testRawIO(self, storageDevice, testdev, minBs, maxBs):
        """Test raw I/O. """
        print("*** Testing raw I/O performance on %s" % testdev)
        size = self.devSize(testdev)
        rawTestFile = "/dev/" + testdev
        # 2.6.* kernels use O_DIRECT
        otherOptions = " flags=direct"
        if self.dtTest(minBs, maxBs, rawTestFile, size, otherOptions):
            print("done.")
            return True
        else:
            print("Error: raw I/O test failed !")
            return False

    def logLVM(self):
        try:
            # these commands write to standard error for no reason
            Command("pvdisplay").echoIgnoreErrors()
            Command("vgdisplay").echoIgnoreErrors()
            Command("lvdisplay").echoIgnoreErrors()
        except Exception as exception:
            print("Warning: could not log LVM info")
            print(exception)

        return True

    def runStorage(self):
        """Overall test strategy: Instruct the tester to install RHEL *only* on
           a single disk. We can use the swap partition for testing that disk, and
           every other (unused) disk we can test directly."""

        # check for dt first
        if os.path.exists("/usr/sbin/dt"):
            print("Found dt package.")
        else:
            print("Error: No dt package found !")
            return False

        # run each disk per user's choice
        disksByHost = self.getDisks()
        diskDevices = None
        # assume our device is a host:
        if self.getLogicalDeviceName() in disksByHost:
            diskDevices = disksByHost[self.getLogicalDeviceName()]
        # else, is it a disk?
        else:
            for disks in disksByHost.values():
                for disk in disks:
                    if self.getLogicalDeviceName() == disk.getProperty("DEVNAME").split('/')[-1]:
                        diskDevices = [disk]

        if not diskDevices and self.getLogicalDeviceName():
            diskDevices = [HardwarePlanner(None).getDeviceByName(self.getLogicalDeviceName())]
            if not diskDevices:
                print("Error: could not find disk or host matching \"%s\"" % self.getLog)
                return False

        disk = None
        disks = self.getDiskNames(diskDevices)
        disksAndControllers = self.getDiskAndControllerNames(diskDevices)
        '''
        All disks of same hosts(host6 & host7) should have the same controller (0000:08:00.4 & 0000:08:00.5 respectively)
        Example:
        [root@dell-per320-02]# udevadm info --export-db | grep host6 | grep sd | grep DEVPATH
        E: DEVPATH=/devices/pci0000:00/0000:00:03.0/0000:08:00.4/host6/rport-6:0-2/target6:0:0/6:0:0:0/block/sda
        E: DEVPATH=/devices/pci0000:00/0000:00:03.0/0000:08:00.4/host6/rport-6:0-3/target6:0:1/6:0:1:0/block/sdb
        E: DEVPATH=/devices/pci0000:00/0000:00:03.0/0000:08:00.4/host6/rport-6:0-4/target6:0:2/6:0:2:0/block/sdc
        E: DEVPATH=/devices/pci0000:00/0000:00:03.0/0000:08:00.4/host6/rport-6:0-5/target6:0:3/6:0:3:0/block/sdd

        [root@dell-per320-02]# udevadm info --export-db | grep host7 | grep sd | grep DEVPATH
        E: DEVPATH=/devices/pci0000:00/0000:00:03.0/0000:08:00.5/host7/rport-7:0-2/target7:0:0/7:0:0:0/block/sde
        E: DEVPATH=/devices/pci0000:00/0000:00:03.0/0000:08:00.5/host7/rport-7:0-3/target7:0:1/7:0:1:0/block/sdf
        E: DEVPATH=/devices/pci0000:00/0000:00:03.0/0000:08:00.5/host7/rport-7:0-4/target7:0:2/7:0:2:0/block/sdg
        E: DEVPATH=/devices/pci0000:00/0000:00:03.0/0000:08:00.5/host7/rport-7:0-6/target7:0:3/7:0:3:0/block/sdh

        Display a tree view to select disk:
        controller name -
        - disk1
        - disk2 etc..
        '''
        if len(disksAndControllers) >= 1:
            message = ""
            try:
                controller_info = disksAndControllers[0]['controller']
                controller = re.match(r"([0-9a-fA-F]{4}\:[0-9a-fA-F]{2}\:[0-9a-fA-F]{2}\.[0-9a-fA-F]{1,2})\s(.*)",
                                      controller_info)
                if controller:
                    message = "For Disk Controller,\n'" + controller.group(2) + " [" + controller.group(1) + "]' ,\n"
            except (KeyError, ValueError, IndexError) as err:
                print("Cannot fetch controller details")

        if self.getMode() != Constants.auto and len(disks) > 1:
            print("Host bus adapter %s has storage devices %s" % (self.getLogicalDeviceName(), ", ".join(disks)))
            disks.append("all")
            disk = self.promptEdit(message + "which disk would you like to test: ", disks[0], disks)
            if disk == "all":
                success = True
                for disk in disks[:-1]:
                    self.storageDevice = disk
                    if not self.runSubTest(self.runDisk, "Test storage I/O for %s" % disk):
                        success = False
                return success

        # otherwise, only one disk on the host, or mode auto should do the 1st one
        elif len(disks) == 1 or (self.getMode() == Constants.auto and len(disks) >= 1):
            disk = disks[0]
        else:
            print("Error: could not find storage devices for host %s" % self.getLogicalDeviceName())
            return False

        # otherwise, only one disk, or defaulted to the first, or user picked one
        self.storageDevice = disk
        return self.runSubTest(self.runDisk, "Test storage I/O for %s" % disk)


    def runDisk(self):
        # initialize
        size = 0
        isLvm = "false"
        # Device sanity check: does it exist?
        partitionFileTmp = open("/proc/partitions", "r")
        partitionFile = partitionFileTmp.read()
        partitionFileTmp.close()
        if self.storageDevice in partitionFile:
            print("Found %s in /proc/partitions." % self.storageDevice)
        else:
            print("Error: Can not find %s in /proc/partitions !" % self.storageDevice)
            return False

        # If there's a swap device, use that for testing
        # turn 'em all on first, just in case
        os.system("swapon -a 2>/dev/null")
        swapFielTmp = open("/proc/swaps", "r")
        swapFile = swapFielTmp.read()
        swapFielTmp.close()
        (success, output1) = rhcert_get_status_output("sort -rgk3 /proc/swaps | grep -m1 '/dev/%s' | awk '{print $1}' | colrm 1 5 " % self.storageDevice)
        if success == 0:
            swapdev = output1
        else:
            swapdev = None
        if self.storageDevice not in swapFile:
            # if we find a swap device that's an LVM LV, and that LV
            # is restricted to the target device, then we can use that
            lines = swapFile.split("\n")
            for line in lines:
                if line.strip():
                    items = line.split()
                    key = items[0]
                    if key.find('/dev/') == 0:
                        if self.deviceIsLvm(key, self.storageDevice):
                            swapdev = key[5:]  # cut off the leading "/dev/"
                            isLvm = "true"
                            break
        # check if the swap has label and record it here
        swapLabel = self.getSwapLabel(swapdev)
        # check the test device
        testdev = self.getTestDevice(self.storageDevice, swapdev, isLvm)
        if not testdev:
            return False
        # Set device-appropriate blocksizes and other special stuff.
        if self.storageDevice[0:4] == "dasd":
            minBs = 4096
            maxBs = self.defaultMaxBs
        else:
            minBs = self.defaultMinBs
            maxBs = self.defaultMaxBs
        # Check set min block size to its sector size
        size_tmp = self.dev_sector_size(testdev)
        if size_tmp:
            minBs = size_tmp
        # start to torture the disks
        try:
            success = self.testVFS(self.storageDevice,testdev,minBs,maxBs)
            if success:
                print("VFS (buffered filesystem) testing passed.\n")
            if not self.testRawIO(self.storageDevice,testdev,minBs,maxBs):
                success = False
            else:
                print("Raw I/O testing passed.\n")
        finally:
            # Clean up, try to restore swap if needed.
            if testdev == swapdev:
                if self.restoreSwap(swapdev, swapLabel):
                    print("restored the swap device.")
                else:
                    print("Error: restore swap error !")
                    success = False
            # try to remove temp directories
            os.system("rmdir %s.* 2>/dev/null" % self.storageDevice)

        if success:
            result = "passed"
        else:
            result = "failed"
        print("\nStorage test on device %s %s" % (self.storageDevice, result))
        return success

    def run(self):

        if not self.runSubTest(self.logLVM, "LVM", "Log LVM info"):
            return False

        if not self.runSubTest(self.grabHardwareInfo, "Hardware", "Log Hardware info"):
            return False

        if not self.runSubTest(self.logInfoStorageSSD, "SSD", "Log SSD Block and Request Size info"):
            return False

        if not self.runStorage():
            return False

        return True

if __name__ == "__main__":
    test = StorageTest("storage")
    returnValue = test.do(sys.argv)
    sys.exit(returnValue)

