#!/usr/bin/env python
# udevDeviceDetector.py - a Python wrapper for udev info
# Author: Greg Nichols <gnichols@redhat.com>


from rhcert.suites.hwcert.hwcert.device import UdevDevice
from rhcert.suites.hwcert.hwcert.deviceDetector import DeviceDetector
from rhcert.tags import Constants
from rhcert.model.documentbase import DocumentWrapper
from rhcert.command import Command, HwCertCommandException

class UdevDeviceDetector(DeviceDetector):
    "A class for working with udev."
    def __init__(self):
        DeviceDetector.__init__(self, name=Constants.udev)
        self.enabled = True
        self.recordSeperator = "\n"
        self.allDevicesCommand = "udevadm info --export-db"
        self.getDeviceCommand = "udevadm info --query=all --path %s"

    def getDevice(self, udi, quiet=False):
        "Return a Device object for the given UDI."
        properties = None
        try:
            udevadm = Command(self.getDeviceCommand % udi)
            lines = udevadm.getStringList(regex=".*")
            stringNumber=0
            properties = dict()
            for line in lines:
            	stringNumber = self.__getProperty(line, properties, stringNumber)
        except HwCertCommandException as e:
            if not quiet:
                print("Error: could not find udev device %s" % udi)
                print(str(e))
            return None

        if properties:
            return UdevDevice(properties)
        # otherwise
        return None

    def getParent(self, device):
        "Return a Device object for the given UDI."
        udi = device.getUDI()
        parent = None

        while True:
            udi = udi.rsplit('/', 1)[0]
            if '/' not in udi:
                print("Error: could not find parent for udev device %s" % device.getUDI())
                return None

            parent = self.getDevice(udi, quiet=True)
            if parent:
                return parent

    def __getProperty(self, line, properties, stringNumber):
        docWrapper = DocumentWrapper()
        line = docWrapper.removeBadCharacters(line)
        property = line.split(":", 1)
        if len(property) == 2:
            type = property[0].strip('\ \'\n')
            attribute = property[1].strip('\ \'\n')
            # assume name an path show up as attributes, so Ignore P and N
            if type == "E":
                keyvalue = attribute.split("=", 1)
                if len(keyvalue) == 2:
                    properties[keyvalue[0]]= keyvalue[1]
            elif type == "S":
                properties[stringNumber] = attribute
                stringNumber += 1
            elif type == "P":
                properties[Constants.info_dot_udi] = attribute

        return stringNumber

    def getDevices(self):
        self.devices = list()
        try:
            pipe = Command(self.allDevicesCommand)
            pipe.start()
            properties = dict()
            stringNumber = 0
            while 1:
                line = pipe.readline()
                if line:
                    if line == self.recordSeperator:
                        if len(properties) > 0:
                            device = UdevDevice(properties)
                            # only return devices that have a UDI (AKA udev path).
                            if device.getUDI() != "":
                                self.devices.append(device)
                            properties = dict()
                            stringNumber = 0
                    else:
                        stringNumber = self.__getProperty(line, properties, stringNumber)
                else:
                    break
        except Exception as e:
            if self.debug != Constants.off:
                print("Warning: udev device detector could not load devices")
                print(str(e))
            pass
        return self.devices

if __name__ == "__main__":
    udev = UdevDeviceDetector()
    for device in  udev.getDevices():
        print(device.dump())
