#!/usr/bin/env python
# Copyright (c) 2006 Red Hat, Inc. All rights reserved. This copyrighted material
# is made available to anyone wishing to use, modify, copy, or
# redistribute it subject to the terms and conditions of the GNU General
# Public License v.2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# Author: Greg Nichols
#

from rhcert.tags import DeviceClass, TestTag
from rhcert.suites.hwcert.hwcert.udevDeviceDetector import UdevDeviceDetector
from rhcert.suites.hwcert.hwcert.hardwareTest import HardwareTest

class USBTestBase(HardwareTest):

    def __init__(self, path):
        path = "usbbase/" + path
        HardwareTest.__init__(self, path)
        self.pluggedInDevices = list()
        self.deviceClass = DeviceClass.usb
        self.deviceDetector = None

    def tags(self):
        return [TestTag.interactive, TestTag.certification, TestTag.usb]

    def plan(self, planner):
        # plan only subclasses that define plan
        return list()

    def getUSBDevices(self):
        ''' in fact, this function fetches all devices. '''
        devices = list()
        if not self.deviceDetector:
            self.deviceDetector = UdevDeviceDetector()
        devices = self.deviceDetector.getDevices()

        return devices

    def checkAllUSBPortsUnplugged(self):
        devices = self.getUSBDevices()
        if not devices:
            return None
        self.pluggedInDevices = list()
        for dev in devices:
            found = self.getUSBDetails(dev)
            if found:
                (busID, devNUM, name, deviceID) = found
                self.pluggedInDevices.append(deviceID)
                if busID and devNUM:
                    print("    %s appears to be plugged into bus %s port %s" % (name, busID, devNUM))

    def addNewPluggedUSBPorts(self):
        ''' should run checkAllUSBPortsUnplugged() first
            returns list of all newly plugged ports '''
        items = []
        devices = self.getUSBDevices()
        if not devices:
            print("Could not find any usb devices")
            return None
        for dev in devices:
            found = self.getUSBDetails(dev)
            if found:
                (busID, devNUM, name, deviceID) = found
                if (not deviceID in self.pluggedInDevices) and busID and devNUM:
                    print("    %s appears to be plugged into bus %s port %s" % (name, busID, devNUM))
                    items.append((busID, devNUM, name, deviceID))
        return items

    def addNewPluggedUSBPort(self):
        ''' should run checkAllUSBPortsUnplugged() first
            old, only returns first newly plugged port :( '''
        devices = self.getUSBDevices()
        if not devices:
            print("Could not find any usb devices")
            return None
        for dev in devices:
            found = self.getUSBDetails(dev)
            if found:
                (busID, devNUM, name, deviceID) = found
                if (not deviceID in self.pluggedInDevices) and busID and devNUM:
                    print("    %s appears to be plugged into bus %s port %s" % (name, busID, devNUM))
                    return (busID, devNUM, name, deviceID)
        return None

    def confirmUSBPortUnplugged(self, searchID):
        ''' should run checkAllUSBPortsUnplugged() first '''
        devices = self.getUSBDevices()
        if not devices:
            return True
        for dev in devices:
            found = self.getUSBDetails(dev)
            if found:
                (busID, devNUM, name, deviceID) = found
                if deviceID == searchID:
                    return False

        return True

    def getUSBDetails(self, device):
        if (device.getProperty("SUBSYSTEM") != "usb" or
            device.getProperty("DEVTYPE") != "usb_device" or
            device.getProperty("ID_BUS") != "usb" or
            device.getProperty("BUSNUM") == "" or
            device.getProperty("DEVNUM") == ""):
                return None

        deviceID = device.getProperty("DEVPATH")
        busID = device.getProperty("BUSNUM")
        devNUM = device.getProperty("DEVNUM")
        name = "%s %s" % (device.getProperty("ID_VENDOR"), device.getProperty("ID_MODEL"))

        return (busID, devNUM, name, deviceID)
