# Copyright (c) 2013 Red Hat, Inc. All rights reserved. This copyrighted material
# is made available to anyone wishing to use, modify, copy, or
# redistribute it subject to the terms and conditions of the GNU General
# Public License v.2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# Author: Greg Nichols
#
import subprocess, re
from rhcert.command import Command, HwCertCommandException

class Wireless:
    """ parser for the output of wireless commands 
        uses nmcli for a list of wireless interfaces 
        uses iw for link parameters """

    def __init__(self):
        self.interfaces = list()
        self.parameters = dict() # parameter[interface][name]
        self.output = None
        self.__parse()

    def __parse(self, searchInterface=None):
        nmcli = "nmcli device"
        """
        expected output:
            NetworkManager-0.9.9.0
            DEVICE             TYPE      STATE
            wlp3s0             wifi      connected
            0C:71:5D:F6:AD:A3  bt        disconnected
            em1                ethernet  unavailable
            lo                 loopback  unmanaged
            virbr0-nic         tap       unmanaged
            tun0               tun       unmanaged'

            NetworkManager-0.9.9.1-13:
            DEVICE   TYPE      STATE                                  CONNECTION           
            enp0s25  ethernet  connected                              enp0s25              
            wlp3s0   wifi      connected                              Auto Metropolis 5GHz 
            virbr0   bridge    connecting (getting IP configuration)  virbr0               
            lo       loopback  unmanaged                              --                   
            tun0     tun       unmanaged                              --           
        """

        try: # python3
            pipe = subprocess.Popen(nmcli, shell=True,
                        stdin=subprocess.PIPE,
                        stdout=subprocess.PIPE,
                        stderr=subprocess.PIPE,
                        encoding='utf8')
        except TypeError: # python2
            pipe = subprocess.Popen(nmcli, shell=True,
                        stdin=subprocess.PIPE,
                        stdout=subprocess.PIPE,
                        stderr=subprocess.PIPE,)

        (output, errors) = pipe.communicate()
        if output:
            self.output = output.splitlines()
            interfacePatten = re.compile("(?P<interface>[^\s]+)\s+(?P<type>[^\s]+)\s+(?P<state>[^\s]+)")
            for line in self.output:
                match = interfacePatten.search(line)
                if match:
                    if match.group("type") == "wifi":
                        self.interfaces.append(match.group("interface"))

        for interface in self.interfaces:
            iw =  "iw %s link" % interface
            """
            Connected to 00:18:f8:70:8b:ee (on wlp3s0)
                SSID: AtomicSpatula
                freq: 2462
                RX: 886341 bytes (2184 packets)
                TX: 82203 bytes (473 packets)
                signal: -59 dBm
                tx bitrate: 54.0 MBit/s
                bss flags:    short-slot-time
                dtim period:    0
                beacon int:    100
            """
            try: # python3
                pipe = subprocess.Popen(iw, shell=True,
                        stdin=subprocess.PIPE,
                        stdout=subprocess.PIPE,
                        stderr=subprocess.PIPE,
                        encoding='utf8')
            except TypeError: # python2
                pipe = subprocess.Popen(iw, shell=True,
                        stdin=subprocess.PIPE,
                        stdout=subprocess.PIPE,
                        stderr=subprocess.PIPE,)

            (output, errors) = pipe.communicate()
            if output:
                self.output = output.splitlines()
                for line in self.output:
                    self.__addParameters(interface, line)

    def __addParameters(self, interface, parameterString):
            pair = parameterString.split(":")
            if len(pair) == 2:
                if not interface in self.parameters:
                    self.parameters[interface] = dict()
                self.parameters[interface][pair[0].strip()] = pair[1].strip()

    def getConnectedSpeed(self, interface):
        if interface in self.interfaces and interface in self.parameters:
            if "tx bitrate" in self.parameters[interface]:
                return self.parameters[interface]["tx bitrate"]
        return None

    def getPhyList(self):
        rfkill_command = Command("rfkill list all")
        phy = ""
        try:
            return rfkill_command.getStringList(regex="^\d: (?P<wifi>.*):.*Wireless LAN", regexGroup="wifi")

        except HwCertCommandException as e:
            print("Warning: no wifi device found")
            return []

    def get_product_name(self, logical_name):
        """
        """
        lshw_command_string_list = Command("lshw -c network | grep -B5 'logical name: {}'".format(logical_name)).getStringList(regex="^.*")
        products = [x.strip().replace('product: ','') for x in lshw_command_string_list if 'product' in x]
        if products:
            return products[0]
        return str()

    def get_phy(self, logical_name):
        """
        This method returns phy of device from
        the logical_name
        Parameters:
        logical_name (str):  Logical Name of Device

        Returns:
        char: 'phy' of Device
        """
        iw_dev_cmd = Command("iw dev")
        try:
            dev_cmd_info = iw_dev_cmd.getStringList('^.*')
        except HwCertCommandException as e:
            return None

        phy = str()
        for attr in dev_cmd_info:
            try:
                if attr.startswith("phy"):
                    phy = attr.replace('#','')
                elif 'Interface' in attr:
                    interface_name = attr.strip().split(' ')[1]
                    if interface_name == logical_name:
                        return phy
            except KeyError as e:
                continue
        return None

    def is_ax_device(self, logical_name):
        """
        This method identifies whether the device has
        AX Support based on the Product Name.

        Parameters:
        logical_name (str):  Logical Name of Device

        Returns:
        bool: True if the device passes the AX Test
        parameters else False.
        """
        # Check product name that includes 'AX' or 'Wifi 6'
        product_name = self.get_product_name(logical_name)
        filtered_product_name = ''.join(e.lower() for e in product_name if e.isalnum())
        if 'ax' in filtered_product_name or 'wifi6' in filtered_product_name:
            return True
        return False

    def test_interface_speed(self, logical_name, test_speed=0):
        """
        This method performs interface speed test.

        Parameters:
        logical_name (str):  Logical Name of Device
        test_speed (int):  Assuming integer values as test speed.
        Applicable for float type also.

        Returns:
        bool: True if the device passes the Speed Test
        parameters else False.
        """
        # Check Interface Speed (Connection Speed)
        print("INFO: Performing AX Wifi Connection Speed Test")
        connection_speed = self.getConnectedSpeed(logical_name)
        try:
            connection_speed_value = connection_speed.split('MBit/s')[0].strip()
            #Check if the interface speed is greater than or equal to 1200
            if float(connection_speed_value) >= test_speed:
                print("INFO: AX Wifi Connection Speed Test Passed.")
                return True
            else:
                print("INFO: AX Wifi Connection Speed Test Failed.")
                return False
        except:
            print("INFO: Could Not Perform Connection Speed Test.")
            return False

    def getType(self, logical_name=None):
        phy_list = self.getPhyList()
        for phy in phy_list:
            # Added condition to match phy identified from the logical name
            if phy != self.get_phy(logical_name):
                continue

            iw_command = Command("iw %s info" % phy)
            try:
                all_the_info = iw_command.getStringList(regex="^.*")
            except HwCertCommandException as e:
                continue
            vht_capabilities = [x for x in all_the_info if 'VHT Capabilities' in x]
            short_GI80 = [x for x in all_the_info if 'short GI (80 MHz)' in x]
            ht20_ht40 = [x for x in all_the_info if 'HT20/HT40' in x]
            fifty_four_Mbps = [x for x in all_the_info if '54.0 Mbps' in x]

            if self.is_ax_device(logical_name) and self.test_interface_speed(logical_name, 1200):
                return "ax"
            elif vht_capabilities and short_GI80 and self.test_interface_speed(logical_name, 1200):
                return "ax"
            elif vht_capabilities and short_GI80:
                return "ac"
            elif ht20_ht40 and not vht_capabilities:
                return "n"
            elif fifty_four_Mbps and not ht20_ht40 and not vht_capabilities:
                return "g"
            else:
                return "b"
        return None

    def getInterfaces(self): return self.interfaces

    def getParameter(self, interface, parameter):
        if interface in self.parameters and parameter in self.parameters[interface]:
            return self.parameters[interface][parameter]
        # otherwise
        return None

    def dump(self):
        for interface in self.interfaces:
            print("%s type %s" % (interface, self.getType()))
            for (name, value) in self.parameters[interface].items():
                print("    %s: %s" % (name, value))

if __name__ == "__main__":
    test = Wireless()
    test.dump()
