# Copyright (c) 2006-14 Red Hat, Inc. All rights reserved. This copyrighted material
# is made available to anyone wishing to use, modify, copy, or
# redistribute it subject to the terms and conditions of the GNU General
# Public License v.2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# Author: Greg Nichols
import os, sys, shutil, time, glob, re

from rhcert.tags import Constants, DeviceClass
from rhcert.command import Command

from rhcert.suites.hwcert.hwcert.hardwareTest import HardwareTest
from rhcert.suites.hwcert.hwcert.udevDeviceDetector import UdevDeviceDetector

class AudioTest(HardwareTest):

    def __init__(self):
        HardwareTest.__init__(self, "audio")
        self.deviceClass = DeviceClass.audio
        self.cardNumber = 0
        self.captureSupport = False
        self.hdmi = False

    def plan(self, planner):
        """
        Plans tests for audio test.  Can handle multiple variations of cards and
        functionality including one solitary card supporting hdmi and internal sound.
        First, tests are planned by matching devices to a set of criteria, assigning
        one test per sound card matching it, where if hdmi is supported, that will be
        the test run on that card.  This is enough for where hdmi and internal sound
        are not supported on the same card.  In order to handle the case where they
        are on the same card, we must check and see if there is an Internal sound test
        planned and if not, plan one if possible, even if the card is already planned
        for an hdmi test.
        """
        properties = dict()
        properties['SUBSYSTEM'] = 'sound'
        properties['SOUND_INITIALIZED'] = '1'
        properties['ID_BUS'] = 'pci'
        tests = self.planWith(planner.getUdevDeviceDetector(), properties, strict=True)
        for test in tests:
            self.set_logical_name(test)
        if not len(tests):
            return tests
        try:
            if not [test for test in tests if "Internal" in test.getLogicalDeviceName()]:
                list_aplay_command = Command("aplay -l")
                card_number = list_aplay_command.getString(regex="(?<=card )\d+(?=.*Analog)", singleLine=False)
                if card_number:
                    # add new test with logical name internal
                    device_command = Command("udevadm info --export-db | grep -i -B 10 audio | grep card%s" % card_number)
                    device = device_command.getString(regex="(?<=P: ).*", singleLine=False)
                    test = self.makeCopy()
                    test.setDevice(planner.getDeviceByName(device))
                    test.setLogicalDeviceName("Internal")
                    tests.append(test)
        except:
            pass
        
        return tests

    def set_logical_name(self, test):
        self.cardNumber = test.getDevice().getUDI()[test.getDevice().getUDI().rfind('/')+5:]
        pipe = os.popen("amixer -c %s controls" % self.cardNumber)
        if 'HDMI/DP,pcm=' in ' '.join(pipe.readlines()):
            test.setLogicalDeviceName("HDMI/DP")
        else:
            test.setLogicalDeviceName("Internal")

    def getRequiredRPMs(self):
        return ["alsa-utils"]

    def getUDI(self):
        """ Override to change the heirarchy of source for UDI for this test.
            Will revisit getUDI on hardwareTest.py later.
        """
        if self.testParameters.get(Constants.udi):
            return self.testParameters.get(Constants.udi)
        if self.device:
            return self.device.getUDI()
        return ""

    def logSoundCardInfo(self):
        if self.getUDI():
            self.device = UdevDeviceDetector().getDevice(self.getUDI())
            if self.device:
                print("\nSound Card:")
                print(self.device.dump())
                self.cardNumber = self.getUDI()[self.getUDI().rfind('/')+5:]
                for infopath in glob.glob("/proc/asound/card%s/pcm*/info" % self.cardNumber):
                    print(infopath)
                    Command("cat %s" % infopath).echo()
                    print("")
                self.getDeviceNumber()
                return True
        # otherwise
        return False

    def getDeviceNumber(self):
        self.deviceNumber = -1
        numidPattern = re.compile("numid=(?P<id>[0-9]+)")
        pcmPattern = re.compile("pcm=(?P<id>[0-9]+)")
        pipe = os.popen("amixer -c %s controls" % self.cardNumber)
        for line in pipe.readlines():
            if "Capture" in line:
                self.captureSupport = True
            if 'HDMI/DP,pcm=' in line and "Internal" not in self.getLogicalDeviceName():
                self.hdmi = True
                match = numidPattern.search(line)
                if match:
                    id = match.group("id")
                    match = pcmPattern.search(line)
                    if match:
                        pcm = match.group("id")
                        checkStatusCommand = Command("amixer -c %s cget numid=%s" %(self.cardNumber, id))
                        try:
                            if checkStatusCommand.getString("values=on", singleLine=False):
                                if self.getDebug() != Constants.off:
                                    print("device numid = %s" % id)
                                    print("device pcm = %s" % pcm)
                                self.deviceNumber = int(pcm)
                                break
                        except Exception as e:
                            if self.getDebug() != Constants.off:
                                print("Exception searching for device id:")
                                print(str(e))


    def run(self):
        #set hdmi to false here because the class is re-used for every audio test
        self.hdmi = False
        if not self.logSoundCardInfo():
            print("Error: could not log sound card information")
            return False

        if self.deviceNumber < 0 and self.hdmi:
            print("Error: No active HDMI/DP audio devices found")
            print("       Please make sure the cable is connected and device turned on")
            return False

        if self.getMode() == Constants.auto:
            print("Error: Test Mode: %s: The audio test can not be run in auto mode since audio functions must be verified by the tester." % self.getMode())
            return False

        # otherwise
        message = """This test plays a sound sample and records it to another file.
                    Please use the Volume Control application and insure that the 
                    Capture settings will record the sound."""
        if not self.promptConfirm(message + " Continue?"):
            self.result = Constants.FAIL
            return False


        if self.captureSupport:
            waveFileDuration = 8 # sec. total
            recordedWaveFile="./test.wav"
            print("starting recording while playing demo sound")
            if os.system("arecord -r 44100 -d %s -D plughw:%s %s &" % (waveFileDuration, self.cardNumber, recordedWaveFile)) != 0:
                print("Error: arecord command failed")
                self.result = Constants.FAIL
                return False
        else:
            print("Note: No Capture Support")

        waveFiles=["/usr/share/sounds/alsa/Front_Right.wav",
                   "/usr/share/sounds/alsa/Front_Left.wav"]
        for waveFile in waveFiles:
            if os.path.exists(waveFile):
                playCommand = "aplay %s -D plughw:%s" % (waveFile, self.cardNumber)
                if self.deviceNumber >= 0:
                    playCommand = "aplay %s -D plughw:%s,%s" % (waveFile, self.cardNumber, self.deviceNumber)
                print(playCommand)
                if os.system(playCommand) != 0:
                    print("Error: aplay command failed")
                    self.result = Constants.FAIL
                    return False
                time.sleep(1)
        sys.stdout.flush()
        if not self.promptConfirm("Did you hear the played sound?"):
            self.result = Constants.FAIL
            return False

        if self.captureSupport:
            print("playing recorded sound")
            playCommand = "aplay %s -D plughw:%s" % (recordedWaveFile, self.cardNumber)
            if self.deviceNumber >= 0:
                playCommand = "aplay %s -D plug:hdmi:%s,%s" % (recordedWaveFile, self.cardNumber, self.deviceNumber)
            print(playCommand)
            if os.system(playCommand) != 0:
                print("Error: aplay command failed")
                self.result = Constants.FAIL
                return False

            if not self.promptConfirm("Did you hear the recorded sound?"):
                self.result = Constants.FAIL
                return False

        self.result = Constants.PASS
        return True

if __name__ == "__main__":
    test = AudioTest()
    returnValue = test.do(sys.argv)
    sys.exit(returnValue)
