#!/usr/bin/env python
# Copyright (c) 2006 Red Hat, Inc. All rights reserved. This copyrighted material
# is made available to anyone wishing to use, modify, copy, or
# redistribute it subject to the terms and conditions of the GNU General
# Public License v.2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# Author: Greg Nichols
#         YangKun <ykun@redhat.com>
#         Caspar Zhang <czhang@redhat.com>

import sys, string

from rhcert.suites.hwcert.hwcert.opticalDiskTest import OpticalDiskTest
from rhcert.tags import Constants, DeviceClass
from rhcert.command import Command, HwCertCommandException

class CDRomTest(OpticalDiskTest):

    def __init__(self):
        OpticalDiskTest.__init__(self, "cdrom")
        self.mediaTypes = [ "cdrw", "cdr", "cdrom"] # in order of preference
        self.deviceClass = DeviceClass.optical
        self.fileSourceDir = "/usr/bin"   # ranges from 100MB-270MB

    # since any drive supports cdrom read-only, override
    # OpticalDiskTest.hasAnyCDROMType to always respond true
    def hasAnyCDROMType(self, ignored):
        return True

    def checkMediaSize(self, size):
        if size > 800000 :
            print("Error: %s appears to be a DVD !" % self.getLogicalDeviceName())
            return False
        # otherwise
        return True

    def writer(self, deviceType):
        cdromDevice = self.getLogicalDeviceName()
        print("\n  +++ Start writing ...")
        sys.stdout.flush()

        try:
            # Unmount the drive first in case it's mounted
            Command("umount /dev/%s" % cdromDevice).echoIgnoreErrors()

            # check the write mode
            writeOpts = self.getWriteMode()
            # support burnfree ?

            try:
                mkisofsSize = Command("mkisofs -quiet -R -print-size %s " % self.fileSourceDir)
                cdblocks= mkisofsSize.getInteger()
            except HwCertCommandException as exception:
                # try work-around for RHEL/mkisofs bug 193916
                # remove the return character first
                item = mkisofsSize.output.pop().replace('\n', '')
                if item.isdigit():
                    cdblocks = int(item)
                    print("Warning: mkisofs had errors")
                    print(exception)
                else:
                    print("Error: could not determine block size")
                    print(exception)
                    return False

            # write the directory to disk
            recordCommand = Command("mkisofs -quiet -R %s | cdrecord -silent %s dev=/dev/%s fs=32M tsize=%ss - > /dev/null 2>&1" % (self.fileSourceDir, writeOpts, cdromDevice, cdblocks))
            recordCommand.echoIgnoreErrors()
            if recordCommand.returnValue == 0:
                print("  +++ Write disk succeeded ! \n")
                self.handleEject(cdromDevice)
            else:
                print("  +++ Error: could not write to disk ! \n")
                sys.stdout.flush()
                return False

            # write done, start to compare
            return self.compare(deviceType, cdromDevice)

        except HwCertCommandException as exception:
            print("Error:")
            print(exception)
            return False

    def run(self):
        """main function"""
        if not self.initializeDeviceInfo():
            return False

        if self.getMode() == Constants.auto:
            print("Error: the CDROM test can not be run in auto mode")
            return False

        self.getMediaChoice()
        if self.mediaChoice == "cdrw":
            description = "good"
            mode = "rewrite"
        elif self.mediaChoice == "cdr":
            description = "new"
            mode = "write"
        elif self.mediaChoice == "cdrom":
            description = "good data"
            mode = "read"
        else:
            print("Error: could not determine media type")
            return False

        self.echoTest(self.getLogicalDeviceName(), self.mediaChoice, description)
        if self.testDisk(message="%s %s" %(description, self.printMediaName(self.mediaChoice)), readWriteMode=mode, type="cdrom"):
            return True
        # OTHERWISE
        return False

if __name__ == "__main__":
    test = CDRomTest()
    returnValue = test.do(sys.argv)
    sys.exit(returnValue)
