#!/usr/bin/env python
# Copyright (c) 2006 Red Hat, Inc. All rights reserved. This copyrighted material 
# is made available to anyone wishing to use, modify, copy, or 
# redistribute it subject to the terms and conditions of the GNU General 
# Public License v.2.
# 
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# Author: Greg Nichols
#
import sys

from rhcert.tags import DeviceClass, TestTag
from rhcert.command import Command, HwCertCommandException
from rhcert.redhatrelease import RedHatRelease
from rhcert.controller import Controller

from rhcert.suites.hwcert.hwcert.procDeviceDetector import ProcDeviceDetector
from rhcert.suites.rhcert.platform import CheckHardwareHealth
from rhcert.suites.hwcert.hwcert.hardwareTest import HardwareTest

class CoreTest(HardwareTest):

    def __init__(self):
        HardwareTest.__init__(self, "core")
        self.deviceClass = DeviceClass.processor
        self.redHatRelease = RedHatRelease()

    def tags(self):
        return [TestTag.noninteractive, TestTag.certification]

    def plan(self, planner):
        tests = list()
        checkHardwareHealth = CheckHardwareHealth("ram_per_core")
        if checkHardwareHealth.ram_per_core():
            tests.append(self.makeCopy())
        if not tests:
            print("INFO: Could not plan test {}".format(self.path))
        else:
            print("INFO: Successfully planned test {}".format(self.path))
        return tests

    def getRequiredRPMs(self):
        rpms = list()
        if "Red Hat" in self.redHatRelease.getProduct() and self.redHatRelease.getVersion() >= 6 and self.redHatRelease.getArch() == "ppc64":
            rpms.append("tree") # for tree command
        return rpms

    def logDevices(self, planner):
        devices = list()
        proc = ProcDeviceDetector()
        devices.extend(proc.getDevices(ProcDeviceDetector.procCPUInfo))
        # call the default to log udev devices
        devices.extend(HardwareTest.logDevices(self, planner))
        return devices

    def getClockInfo(self):
        print("Clock Info: ------------------------------------------")
        controller = Controller()
        clockInfo = controller.getSystemLogSinceBoot()
        tsc = False
        clocksource = None
        for line in clockInfo:
            if "TSC" in line or "clocksource" in line or " tsc " in line:
                print(line[line.find("kernel"):].strip())
                if "tsc" in line and "clocksource" in line:
                    tsc = True
                    clocksource = "TSC"
                elif "clocksource" in line: # clocksource set to some other clock
                    tsc = False
                    clocksource = line.split()[-1]

        print("")
        if clocksource:
            print("Clock Source per system log: " + clocksource) 
        else:
            print("Warning: could not determine clocksource from system log.")

        sysCurrentClockSource = "/sys/devices/system/clocksource/clocksource*/current_clocksource"
        sys = Command("cat " + sysCurrentClockSource)
        clocksource = sys.getString()
        print("Clock Source in " + sysCurrentClockSource + ": " + clocksource)
        print("")
        return tsc

    def isIntel(self):
        try:
            intel = Command("cat /proc/cpuinfo")
            vendor =  intel.getString(regex="^vendor_id[^:]*:\ *(?P<vendor>[^\ ]*$)", regexGroup="vendor", singleLine=False)
            print("CPU Vendor: " + vendor)
            sys.stdout.flush()
            if vendor in ["Intel", "GenuineIntel"]:
                return True
            return False
        except HwCertCommandException as e:

            return False

    def runClockTest(self):
        try:
            # run clock tests - jitter and direction
            usingTSC = self.getClockInfo()
            if self.isIntel() and not usingTSC:
                print("Warning: Intel processor(s) not using TSC")
                # warn only for now
                # result = False

            print("Running clock tests")
            sys.stdout.flush()
            clocktest = Command("./clocktest")
            clocktest.echo()
            sys.stdout.flush()
            return True

        except HwCertCommandException as e:
            print("Error:")
            print(e)
            e.command.printErrors()
            return False

    def runStress(self):

        try:
            limit = 10 # min.
            numberOfProcesses = 12
            processSize = 128 # MB
            # check memory limits
            if self.freeMemory < (numberOfProcesses * processSize):
                processSize = int(self.freeMemory/numberOfProcesses)
                print("Note: scaling back %u processes at %u MB for memory limit of %u MB" %(numberOfProcesses, processSize, self.freeMemory))

            print("Running stress for %u min." % limit)
            sys.stdout.flush()
            stress = Command("stress --cpu %u --io %u --vm %u --vm-bytes %uM --timeout %um" % (numberOfProcesses, numberOfProcesses, numberOfProcesses, processSize, limit))
            print(stress.command)
            stress.echo()
            return True
        except HwCertCommandException as e:
            print("Error:")
            print(e)
            e.command.printErrors()
            return False

        return False

    def run(self):
        result = True

        if not self.runSubTest(self.getMemoryInfo, name="Limits", description="Get test parameters based on hardware"):
            print("Error: could not determine memory limits, using default settings")
            result = False

        if not self.runSubTest(self.runClockTest, name="clocktest", description="Clock tests"):
            result = False

        if not self.runSubTest(self.runStress, name="Stress"):
            result = False

        return result

if __name__ == "__main__":
    test = CoreTest()
    returnValue = test.do(sys.argv)
    sys.exit(returnValue)



