#!/usr/bin/env python
# Copyright (c) 2008 Red Hat, Inc. All rights reserved. This copyrighted material 
# is made available to anyone wishing to use, modify, copy, or 
# redistribute it subject to the terms and conditions of the GNU General 
# Public License v.2.
# 
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#

import struct



class EffectiveFrequency:
    """
    Simple mechanism to use the Effective Frequency Interface
    """

    def __init__(self, cpunum):
        """
        @param cpunum: Number of the CPU as used by the OS
        @type  cpunum: int
        """
        self.__cpunum = int(cpunum)
        self.__aperf = -1
        self.__mperf = -1

    def has_support(self):
        """
        Check for Effective Frequency Interface support

        This information is stored in CPUID function 0000_0006,
        register ECX, Bit 0.

        @return: Effective Frequency Interface support (0/1)
        @rtype : int
        """
        support = 0
        cpuidregs = self.__read_cpuid(0x00000006)
        if cpuidregs:
            support = cpuidregs[2] & 0x1
        return support

    @staticmethod
    def __read_file(path, offset, size):
        """
        Read a specified amount of bytes from a file starting at some offset

        @param path  : The full path to the file
        @type  path  : str
        @param offset: The byte offset at which the read should start
        @type  offset: int
        @param size  : The amount of bytes to be read
        @param size  : int
        @return      : The chunk of the file
        @rtype       : str
        """
        filedesc = None
        try:
            filedesc = open(path, "rb")
            filedesc.seek(offset, 1)
            value = filedesc.read(size)
        except IOError:
            value = ""
        if filedesc:
            filedesc.close()
        if len(value) != size:
            value = ""
        return value

    def __read_msr(self, msr):
        """
        Read the content of a MSR

        Returns the 64-bit value of the given MSR, or -1 in case of errors.

        @param msr: MSR number
        @type  msr: int
        @return   : 64-bit value of the MSR
        @rtype    : int
        """
        path = "/dev/cpu/%d/msr" % (self.__cpunum, )
        value = self.__read_file(path, msr, 8)
        if len(value) == 8:
            value = struct.unpack("Q", value)[0]
        else:
            value = -1
        return value

    def __read_cpuid(self, function, index=0):
        """
        Read the register values for a CPUID function

        Returns a tuple with values for the registers EAX, EBX, ECX
        and EDX, or an empty tuple in case of errors.

        @param function: CPUID instruction function
        @type  function: int
        @param index   : CPUID instruction index (value of ECX)
        @type  index   : int
        @return        : Values for registers EAX, EBX, ECX, and EDX
        @rtype         : tuple
        """
        path = "/dev/cpu/%d/cpuid" % (self.__cpunum, )
        function &= 0xffffffff
        index &= 0xffffffff
        offset = function | (index << 32)
        value = self.__read_file(path, offset, 16)
        if len(value) == 16:
            value = struct.unpack("4I", value)
        else:
            value = tuple()
        return value

    def __read_aperf_mperf(self):
        """
        Read APERF and MPERF MSRs

        Returns a tuple with the values of the APERF and MPERF MSRs.
        Both values will be set to -1 in case of errors or if EFRI
        is not supported by the CPU.

        @return: Values of APERF and MPERF MSRs
        @rtype : tuple
        """
        aperf, mperf = (-1, -1)
        if self.has_support():
            aperf = self.__read_msr(0x000000E8)
            mperf = self.__read_msr(0x000000E7)
            if -1 in (aperf, mperf):
                aperf, mperf = (-1, -1)
        return aperf, mperf

    def __get_max_frequency(self):
        """
        Get the highest non-boosted frequency

        The highest non-boosted frequency is the frequency of the
        P-State 0, using the software P-State numbering.

        @return: Highest non-boosted frequency in kHz, or -1 on error
        @rtype : int
        """
        sysfs = "/sys/devices/system/cpu/cpu%d/" \
                "cpufreq/scaling_available_frequencies" % (self.__cpunum, )
        try:
            filedesc = open(sysfs, "r")
            freqs = filedesc.readline()
            filedesc.close()
            frequency = max([int(freq) for freq in freqs.split()])
        except (IOError, ValueError):
            frequency = -1
        return frequency

    def start_sampling(self):
        """
        Start the frequency sampling period

        Read APERF and MPERF MSRs and store the results as initial
        values for use as offsets in later calculations.

        @return: Success of the reading
        @rtype : bool
        """
        self.__aperf, self.__mperf = self.__read_aperf_mperf()
        if -1 in (self.__aperf, self.__mperf):
            return False
        return True

    def stop_sampling(self):
        """
        Stop the sampling period and return the effective frequency

        Reads the APERF and MPERF MSRs once again, does the required
        calculations, and returns the effective frequency during the
        sampling period.

        @return: Effective frequency in kHz, or -1 on error
        @rtype : int
        """
        frequency = -1
        aperf, mperf = self.__read_aperf_mperf()
        maxfreq = self.__get_max_frequency()
        if -1 not in (self.__aperf, self.__mperf, aperf, mperf, maxfreq):
            aperfdiff = aperf - self.__aperf
            mperfdiff = mperf - self.__mperf
            if struct.unpack("L", struct.pack("l", -1))[0] / 100 < aperfdiff:
                shiftcnt = 7
                aperfdiff >>= shiftcnt
                mperfdiff >>= shiftcnt
            perfpercent = (aperfdiff * 100) / mperfdiff
            frequency = maxfreq * perfpercent / 100
        self.__aperf, self.__mperf = (-1, -1)
        return frequency