#!/usr/bin/env python
# Copyright (c) 2014 Red Hat, Inc. All rights reserved. This copyrighted material 
# is made available to anyone wishing to use, modify, copy, or 
# redistribute it subject to the terms and conditions of the GNU General 
# Public License v.2.
# 
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# Author: Greg Nichols <gnichols@redhat.com>

""" CPUPower: a python interface to the cpupower command from kernel-tools 

   Note: all frequency parameters in and out are integer KHz

    Definitions:
        Maximum Frequency: The maximum frequency supported (ignoring boost state or "turbo")
        Minimum Frequency: The minimum frequency supported unloaded
        Maximum Model Frequency: The frequency stated in for the cpu 

        Minimum Boost Frequency: The minimum boost frequency for a partially or singly loaded core
        Maximum Boost Frequency: The maximum boost frequency for a partially or singly loaded core

        SupportedGovernors: the list of supported governors

        Current Policy: a governor combined with min/max frequency settings

        Current Frequency: Current reported clock frequency

        Driver: The cpufrequency driver reported by cpupower

        Vendor: The processor vendor
"""

import re

from decimal import Decimal

from rhcert.command import Command, HwCertCommandException

from rhcert.suites.hwcert.hwcert.procDeviceDetector import ProcDeviceDetector

class CPUPower():
    def __init__(self, debug=False):
        self.debug = debug
        self.cpuNumbers = None
        self.minimumFrequency = dict()
        self.maximumFrequency = dict()
        self.minimumBoostFrequency = dict()
        self.maximumBoostFrequency = dict()
        self.maximumModelFrequency = None
        self.supportedGovernors = dict() 
        self.procCPUinfoCPUs = None
        self.possibleGovernors = ["performance", "powersave", "userspace", "conservative", "ondemand"]
        self.vendor = None
        self.vendors = ["Intel", "AMD", "IBM"]
        self.model = None
        self.driver = None
        self.driverPattern = re.compile("driver:\s+(?P<driver>.+)")

        self.__getSupportInformation()

    def __getSupportInformation(self):
        # get the online cpus
        self.cpuNumbers = list()
        cpu_nos_cpupower = Command("grep 'processor' /proc/cpuinfo")
        try:
            cpu_nos = cpu_nos_cpupower.getStringList(regex='.*:\\s+(?P<cpuno>\\d+)',
                                                        regexGroup='cpuno')
            self.cpuNumbers = [int(cpu_no) for cpu_no in cpu_nos]
        except HwCertCommandException as e:
            print("Warning: Unable to fetch cpu numbers using /proc/cpuinfo command.")
            print(e)
            return

        # determine cpu vendor
        proc = ProcDeviceDetector()
        self.procCPUinfoCPUs = proc.getDevices(ProcDeviceDetector.procCPUInfo)
        if self.procCPUinfoCPUs:
            self.model = self.procCPUinfoCPUs[0].getProperty("model name")
            if self.model:
                for vendor in self.vendors:
                    if vendor in self.model:
                        self.vendor = vendor
                if self.vendor and ("Intel" in self.vendor):
                    if not self.__parseCPUPowerIntel():
                        return False
        # else AMD or unknown
        if not self.__parseCPUPowerUnkown():
            return False

        # check /proc/cpuinfo model string for maxFreq
        if self.model:
            maxFreq = None
            pattern = re.compile("(?P<max>[0-9\.]+)(?P<units>(MHz|GHz))")
            match = pattern.search(self.model)
            if match:
                maxFreq = int(Decimal(match.group("max"))*1000000) # to KHz
                if match.group("units") == "MHz":
                    maxFreq = int(maxFreq/1000)
            if maxFreq:
                self.maximumModelFrequency = maxFreq

        return True


    def __parseCPUPowerIntel(self):
        """
        Example: intel_pstate:

          analyzing CPU 0:
          driver: intel_pstate
          CPUs which run at the same hardware frequency: 0
          CPUs which need to have their frequency coordinated by software: 0
          maximum transition latency: 0.97 ms.
          hardware limits: 1.20 GHz - 3.60 GHz
          available cpufreq governors: performance, powersave
          current policy: frequency should be within 1.20 GHz and 3.60 GHz.
                          The governor "powersave" may decide which speed to use
                          within this range.
          current CPU frequency is 1.71 GHz (asserted by call to hardware).
          boost state support:
            Supported: yes
            Active: yes
            3400 MHz max turbo 4 active cores
            3400 MHz max turbo 3 active cores
            3400 MHz max turbo 2 active cores
            3600 MHz max turbo 1 active cores

        Example: acpi_cpufreq (Intel)

           analyzing CPU 0:
              driver: acpi-cpufreq
              CPUs which run at the same hardware frequency: 0
              CPUs which need to have their frequency coordinated by software: 0
              maximum transition latency: 10.0 us.
              hardware limits: 1.20 GHz - 2.60 GHz
              available frequency steps: 2.60 GHz, 2.60 GHz, 2.50 GHz, 2.40 GHz, 2.30 GHz, 2.20 GHz, 2.10 GHz, 2.00 GHz, 1.90 GHz, 1.80 GHz, 1.70 GHz, 1.60 GHz, 1.50 GHz, 1.40 GHz, 1.30 GHz, 1.20 GHz
              available cpufreq governors: conservative, userspace, powersave, ondemand, performance
              current policy: frequency should be within 1.20 GHz and 2.60 GHz.
                              The governor "ondemand" may decide which speed to use
                              within this range.
              current CPU frequency is 1.20 GHz (asserted by call to hardware).
              boost state support:
                Supported: yes
                Active: yes
                3400 MHz max turbo 4 active cores
                3400 MHz max turbo 3 active cores
                3500 MHz max turbo 2 active cores
                3600 MHz max turbo 1 active cores
        """

        # Get the min/max supported frequencies, governors
        frequencyPattern = re.compile("hardware limits: (?P<min>[0-9\.]+) (?P<minunits>(MHz|GHz)) - (?P<max>[0-9\.]+) (?P<maxunits>(MHz|GHz))")
        coreFrequencyPattern = re.compile("(?P<max>[0-9\.]+) (?P<activeunits>(MHz|GHz)) max (?P<boost>[a-zA-Z]+) (?P<active>[0-9\.]+) active cores")
        governorPattern = re.compile("available cpufreq governors:")
        for cpu in self.cpuNumbers:
            maxActiveCores = 0
            cpupower = Command("cpupower -c %s frequency-info -n" % cpu)
            try:
                cpupower.start()
                while True:
                    line = cpupower.readline()
                    if not line:
                        break
                    match = frequencyPattern.search(line)
                    if match:
                        # assume GHz
                        freq = int(Decimal(match.group("min"))*1000000) # to KHz
                        if match.group("minunits") == "MHz":
                            freq = freq/1000
                        self.minimumFrequency[cpu] = freq
                        freq = int(Decimal(match.group("max"))*1000000) # to KHz
                        if match.group("maxunits") == "MHz":
                            freq = freq/1000
                        self.maximumFrequency[cpu] = freq
                        continue
                    match = coreFrequencyPattern.search(line)
                    if match:
                        freq = int(Decimal(match.group("max"))*1000000) # to KHz
                        if match.group("activeunits") == "MHz":
                            freq = freq/1000
                        cores = int(match.group("active"))
                        if not maxActiveCores:
                            self.minimumBoostFrequency[cpu] = freq
                            maxActiveCores = cores
                        elif cores < maxActiveCores:
                            self.maximumBoostFrequency[cpu] = freq
                            maxActiveCores = cores
                        continue
                    if governorPattern.search(line):
                        for governor in self.possibleGovernors:
                            if governor in line:
                                if not cpu in self.supportedGovernors:
                                    self.supportedGovernors[cpu] = list()
                                self.supportedGovernors[cpu].append(governor)
                    match = self.driverPattern.search(line)
                    if match:
                        self.driver = match.group("driver")

                # the max hardware speed is often the max boost freq, so set max to min boost (fully loaded)
                if cpu in self.minimumBoostFrequency:
                    self.maximumFrequency[cpu] = self.minimumBoostFrequency[cpu]

            except Exception as e:
                print("Error: could not determine intel cpu support")
                print(e)
                self.cpuNumbers = None
                return False

        return True

    def __parseCPUPowerUnkown(self):
        """
        Example: AMD - Unknown:

        analyzing CPU 0:
          no or unknown cpufreq driver is active on this CPU
          boost state support:
            Supported: yes
            Active: yes
            Boost States: 2
            Total States: 8
            Pstate-Pb0: 3000MHz (boost state)
            Pstate-Pb1: 2400MHz (boost state)
            Pstate-P0:  2100MHz
            Pstate-P1:  1900MHz
            Pstate-P2:  1700MHz
            Pstate-P3:  1500MHz
            Pstate-P4:  1400MHz
            Pstate-P5:  500MHz
            
        AMD - acpi-cpufreq driver:
        
        analyzing CPU 0:
          driver: acpi-cpufreq
          CPUs which run at the same hardware frequency: 0
          CPUs which need to have their frequency coordinated by software: 0
          maximum transition latency: 4.0 us.
          hardware limits: 1.40 GHz - 4.00 GHz
          available frequency steps: 4.00 GHz, 3.40 GHz, 2.80 GHz, 2.10 GHz, 1.40 GHz
          available cpufreq governors: conservative, userspace, powersave, ondemand, performance
          current policy: frequency should be within 4.00 GHz and 4.00 GHz.
                          The governor "ondemand" may decide which speed to use
                          within this range.
          current CPU frequency is 4.00 GHz (asserted by call to hardware).
          boost state support:
            Supported: yes
            Active: yes
            Boost States: 2
            Total States: 7
            Pstate-Pb0: 4200MHz (boost state)
            Pstate-Pb1: 4100MHz (boost state)
            Pstate-P0:  4000MHz
            Pstate-P1:  3400MHz
            Pstate-P2:  2800MHz
            Pstate-P3:  2100MHz
            Pstate-P4:  1400MHz

           
        """
        # Get the min/max supported frequencies, governors
        frequencyPattern = re.compile("(?P<state>Pstate-P[b0-9]+):\s+(?P<freq>[0-9\.]+)(?P<units>(MHz|GHz))")
        hardwareLimitsPattern = re.compile("hardware limits: (?P<min>[0-9\.]+) (?P<minunits>(MHz|GHz)) - (?P<max>[0-9\.]+) (?P<maxunits>(MHz|GHz))")
        governorPattern = re.compile("available cpufreq governors:")
        for cpu in self.cpuNumbers:
            frequencies = list()
            boostFrequencies = list()
            hardwareMinimumFrequency = 0
            hardwareMaximumFrequency = 0
            cpupower = Command("cpupower -c %s frequency-info" % cpu)
            try:
                cpupower.start()
                while True:
                    line = cpupower.readline()
                    if not line:
                        break
                    match = hardwareLimitsPattern.search(line)
                    if match:
                        # assume GHz
                        freq = int(Decimal(match.group("min"))*1000000) # to KHz
                        if match.group("minunits") == "MHz":
                            freq = freq/1000
                        hardwareMinimumFrequency = freq
                        freq = int(Decimal(match.group("max"))*1000000) # to KHz
                        if match.group("maxunits") == "MHz":
                            freq = freq/1000
                        hardwareMaximumFrequency = freq
                        continue
                    match = frequencyPattern.search(line)
                    if match:
                        # assume GHz
                        freq = int(Decimal(match.group("freq"))*1000000) # to KHz
                        if match.group("units") == "MHz":
                            freq = freq/1000
                        if "boost state" in line:
                            boostFrequencies.append(freq)
                        else:
                            frequencies.append(freq)
                        continue
                    match = governorPattern.search(line)
                    if match:
                        for governor in self.possibleGovernors:
                            if governor in line:
                                if not cpu in self.supportedGovernors:
                                    self.supportedGovernors[cpu] = list()
                                self.supportedGovernors[cpu].append(governor)
                        continue

                    match = self.driverPattern.search(line)
                    if match:
                        self.driver = match.group("driver")

                # got all the lines parsed - now sort the freqs
                if frequencies:
                    frequencies.sort()
                    self.minimumFrequency[cpu] = frequencies[0]
                    self.maximumFrequency[cpu] = frequencies[-1]
                elif hardwareMinimumFrequency and hardwareMaximumFrequency: # no list of pstates, use the hardware limits
                    self.minimumFrequency[cpu] = hardwareMinimumFrequency
                    self.maximumFrequency[cpu] = hardwareMaximumFrequency
                else:
                    print("Error: could not determine minimum and maximum frequencies.")
                if boostFrequencies:
                    boostFrequencies.sort()
                    self.maximumBoostFrequency[cpu] = boostFrequencies[-1]
                    self.minimumBoostFrequency[cpu] = boostFrequencies[0]

                # no supported governors???  guessing performance and powersave
                if cpu not in self.supportedGovernors:
                    self.supportedGovernors[cpu] = ["performance", "powersave"]

            except Exception as e:
                print("Error: could not determine unknown cpu support")
                print(e)
                self.cpuNumbers = None
                return False

        return True

    def dumpOutput(self):
        for cpu in self.cpuNumbers:
            cpupower = Command("cpupower -c %s frequency-info" % cpu)
            cpupower.echo()

    def printInformation(self):
        if self.procCPUinfoCPUs:
            print("\nProcessor Model: %s" % self.model)
        if self.maximumModelFrequency:
            print("Maximum Model Frequency: %s MHz" % int(self.maximumModelFrequency/1000))
        else:
            print("Note: no stated maximum frequency for the model")
        if self.driver:
            print("cpufreq driver: %s" % self.driver)
        else:
            print("cpufreq driver: unknown")
        self.dumpOutput()
        for cpu in self.cpuNumbers:
            print("\nCPU: %s\n----------" % cpu)
            print("           min: %s MHz" % int(self.minimumFrequency[cpu]/1000))
            print("           max: %s MHz" % int(self.maximumFrequency[cpu]/1000))
            if self.minimumBoostFrequency:
                print("     min boost: %s MHz" % int(self.minimumBoostFrequency[cpu]/1000))
            if self.maximumBoostFrequency:
                print("     max boost: %s MHz" % int(self.maximumBoostFrequency[cpu]/1000))
            print("     governors: " + ", ".join(self.supportedGovernors[cpu]))

    def isSupported(self):
        """ return True if cpuscaling is supported in hardware """
        if self.cpuNumbers:
            for cpu in self.cpuNumbers:
                if cpu in self.supportedGovernors and self.supportedGovernors[cpu]:
                    return True
        if self.debug:
            print("cpuscaling is NOT supported")
        return False

    def isValid(self):
        """ return True if info looks correct and reasonable """
        if not self.isSupported():
            return False
        # otherwise
        min = None
        max = None
        maxBoost = None
        minBoost = None
        governors = list()
        try:
            for cpu in self.cpuNumbers:
                if cpu not in self.minimumFrequency:
                    print("Error: no min frequency for cpu %s" % cpu)
                    return False
                if not min:
                    min = self.minimumFrequency[cpu]
                elif min != self.minimumFrequency[cpu]:
                    print("Error: minimum frequency mismatch")
                    return False

                if cpu in self.maximumBoostFrequency:
                    if not maxBoost:
                        maxBoost = self.maximumBoostFrequency[cpu]
                    elif maxBoost != self.maximumBoostFrequency[cpu]:
                        print("Error: maximum boost frequency mismatch")
                        return False
                if cpu in self.minimumBoostFrequency:
                    if not minBoost:
                        minBoost = self.minimumBoostFrequency[cpu]
                    elif minBoost != self.minimumBoostFrequency[cpu]:
                        print("Error: minimum boost frequency mismatch")
                        return False

                if cpu not in self.maximumFrequency:
                    print("Error: no max frequency for cpu %s" % cpu)
                    return False
                if not max:
                    max = self.maximumFrequency[cpu]
                elif max != self.maximumFrequency[cpu]:
                    print("Error: maximum frequency mismatch")
                    return False
                if not governors:
                    governors = self.supportedGovernors[cpu]
                else:
                    for governor in governors:
                        if not governor in self.supportedGovernors[cpu]:
                            print("Error: governor support mismatch")
        except KeyError as e:
            print("Error: missing cpu data")
            print(e)
            return False

        return True

    def getCPUs(self):
        return self.cpuNumbers

    def getGovernors(self):
        """ return a list of supported governors """
        return self.supportedGovernors[self.cpuNumbers[0]]

    def getMinimumFrequency(self):
        """ get the minimum supported frequency, in int(MHz) """
        return self.minimumFrequency[self.cpuNumbers[0]]

    def getMaximumFrequency(self):
        """ get the maximum supported frequency stated for the model, or all cores active in int(MHz)  """
        if self.maximumModelFrequency:
            return self.maximumModelFrequency
        # otherwise
        return self.maximumFrequency[self.cpuNumbers[0]]

    def getMaximumModelFrequency(self):
        """ get the maximum stated frequency for the processor model (from /proc/cpuinfo) """
        return self.maximumModelFrequency

    def getMaximumBoostFrequency(self):
        """ get the maximum supported frequency, in int(MHz)  """
        if self.maximumBoostFrequency:
            return self.maximumBoostFrequency[self.cpuNumbers[0]]
        return 0

    def getMinimumBoostFrequency(self):
        """ get the minimum supported frequency, in int(MHz)  """
        if self.minimumBoostFrequency:
            return self.minimumBoostFrequency[self.cpuNumbers[0]]
        else:
            return 0

    def getCurrentFrequency(self, cpu):
        """ get the current frequency """
        cpupower = Command("cpupower -c %s frequency-info -w -n" % cpu)
        pattern = re.compile("[0-9]+")
        try:
            cpupower.start()
            line = cpupower.readline() # dump the line: analyzing CPU 0:
            line = cpupower.readline()
            if line:
                match = pattern.search(line)
                if match:
                    return int(match.group())
        except Exception as e:
            print("Error: could not determine current frequency")
            print(e)

        return None


    def setFrequency(self, frequency):
        """ set the frequency """
        cpupower = Command("cpupower -c all frequency-set --freq %s" % frequency)
        try:
            cpupower.echo()
        except Exception as e:
            print("Error: can't set the frequency:")
            print(e)
            return False
        return True

    def getCurrentPolicyThreeLine(self, cpu):
        cpupower = Command("cpupower -c %s frequency-info -p -n" % cpu)
        outputLines = cpupower.getStringList(regex=".*")
        pattern = re.compile("(?P<min>[0-9.]+)[^0-9.]*(?P<max>[0-9.]+)")
        governorPattern = re.compile("governor\s+\"(?P<governor>.*?)\"")
        min = max = governor = None
        for line in outputLines:
            match = pattern.search(line)
            if match:
                min = match.group("min")
                max = match.group("max")
                if "." in min:
                    min = float(min)*1000
                if "." in max:
                    max = float(max)*1000
                min = int(min)
                max = int(max)
                if min >= 10000:
                   min = min/1000
                if max >= 10000:
                   max = max/1000
            match = governorPattern.search(line)
            if match:
                governor = match.group("governor")
        return (min, max, governor)


    def getCurrentPolicy(self, cpu):
        """ get the current frequency """
        cpupower = Command("cpupower -c %s frequency-info -p -n" % cpu)
        pattern = re.compile("(?P<min>[0-9]+)\s(?P<max>[0-9]+)\s(?P<gov>[a-z]+)")
        min = max = governor = None
        try:
            cpupower.start()
            line = cpupower.readline() # dump the line: analyzing CPU 0:
            line = cpupower.readline()
            if line:
                match = pattern.search(line)
                if match:
                    min = int(match.group("min"))
                    max = int(match.group("max"))
                    governor = match.group("gov")
        except Exception as e:
            print("Error: could not determine current policy")
            print(e)

        if min and max and governor:
            return (min, max, governor)
        else:
            return self.getCurrentPolicyThreeLine(cpu)

    def getCurrentGovernor(self, cpu):
        """ get the current governor """
        policy = self.getCurrentPolicy(cpu)
        if policy:
            (min, max, governor) = policy
            return governor
        return None

    def setGovernor(self, governor):
        """ set the governor """
        cpupower = Command("cpupower -c all frequency-set --governor %s" % governor)
        try:
            cpupower.echo()
        except Exception as e:
            print("Error: can't set the governor:")
            print(e)
            return False
        return True

    def setPolicy(self, min, max, governor):
        """ set the governor """
        cpupower = Command("cpupower -c all frequency-set --governor %s --min %s --max %s" % (governor, min, max))
        try:
            cpupower.echo()
        except Exception as e:
            print("Error: can't set the governor:")
            print(e)
            return False
        return True

    def getSingleCPUTestGovernor(self):
        if self.driver and self.driver == "acpi-cpufreq" and "ondemand" in self.getGovernors():
            return "ondemand"
        # otherwise
        if "powersave" in self.getGovernors():
            return "powersave"
        # otherwise
        return None

    def get_CPPC_support(self):
        return self.driver == "cppc_cpufreq"

    def get_POWERNV_support(self):
        return self.driver == "powernv-cpufreq"

if __name__ == "__main__":

    cpuPower = CPUPower()
    cpuPower.printInformation()
    if cpuPower.isSupported():
        print("IS supported")
    if cpuPower.isValid():
        print("IS valid")
    print("Current Frequency: %s" % cpuPower.getCurrentFrequency(0))
    if cpuPower.getCurrentPolicy(0):
        (min, max, gov) = cpuPower.getCurrentPolicy(0)
        print("policy: %s %s %s" % (min, max, gov))
    print("Current Governor; %s" % cpuPower.getCurrentGovernor(0))
    if cpuPower.setGovernor("powersave"):
        print("governor set to powersave")
    if cpuPower.setPolicy(2000000, 2500000, "powersave"):
        print("set policy")
    policy = cpuPower.getCurrentPolicy(0)
    if policy:
        (min, max, gov) = policy
        print("policy: %s %s %s" % (min, max, gov))
    else:
        print("could not get policy")
    if cpuPower.setPolicy(1200000, 3600000, "powersave"):
        print("set policy")
    policy = cpuPower.getCurrentPolicy(0)
    if policy:
        (min, max, gov) = policy
        print("policy: %s %s %s" % (min, max, gov))
    else:
        print("could not get policy")
