#!/usr/bin/env python
# Copyright (c) 2008 Red Hat, Inc. All rights reserved. This copyrighted material 
# is made available to anyone wishing to use, modify, copy, or 
# redistribute it subject to the terms and conditions of the GNU General 
# Public License v.2.
# 
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# Author: Greg Nichols <gnichols@redhat.com>

import os, re, string, sys, time, decimal, filecmp
from rhcert.tags import TestResult

if __name__ == "__main__":
    from rhcert.suites.hwcert.hwcert.testingimport import TestingImport
    sys.meta_path = [TestingImport(__file__, 4)]

from rhcert.tags import Constants, DeviceClass, TestTag, TestResult
from rhcert.command import Command, HwCertCommandException
from rhcert.redhatrelease import RedHatRelease

from rhcert.suites.hwcert.hwcert.procDeviceDetector import ProcDeviceDetector
from rhcert.suites.hwcert.hwcert.hardwareTest import HardwareTest

from .aperfmperf import EffectiveFrequency
from .cpupower import CPUPower
from .syscpufreq import SysCPUFreq

class MyFormatter(string.Formatter):
    """ String formatter which does not throw error on failed conversion to float

    >>> fmt = MyFormatter()
    >>> fmt.format("{value:.2f}", value=5.435)
    '5.43'
    >>> fmt.format("{value:.2f}", value=None)
    "[ERROR CONVERTING: 'None' to Float]"
    """
    def format_field(self, value, format_spec):
        if re.search("\.\d+f", format_spec):
            try:
                float(value)
            except:
                return "[ERROR CONVERTING: '{}' to Float]".format(str(value))
        return super(MyFormatter, self).format_field(value, format_spec)
fmt = MyFormatter()

class CPUScalingTest(HardwareTest):
    """
    Tolerance Calculation, workload time checks:
    
    p = predicted speedup = (max freq / min freq)
    s = measured speedup = (workload time, min freq / workload time, max freq)
    t = tolerance 

    * CERTPX-238 minimum ignored by setting lower tolerance to 5Ghz
     when 'hwp' cpu flag is present
    
    p - (p - 1.0)*t < s < p + (p - 1.0)
    
    example: p = 2.0, s = 1.9, t = 0.5
    2.0 - (2.0 - 1.0)*0.5 < 1.9 < 2.0 + (2.0-1.0)*.5
    evaluates to:
    1.5 < 1.9 < 2.5  = True (PASS)
    """

    def __init__(self):
        HardwareTest.__init__(self, "cpuscaling")
        self.speedUpTolerance = 50.0 # percent

        self.lowerTolerance = 50 # MHz
        self.higherTolerance = 100 # MHz
        self.minimumCheckLowerTolerance = 200
        self.minimumCheckHigherTolerance = 200

        self.retryLimit = 5
        self.retryTolerance = 5.0 # percent
        self.idaFlag = "ida"
        self.hwpFlag = "hwp"
        self.aperfmperfFlag = "aperfmperf"
        self.idaSpeedupFactor = 8.0 # percent  
        self.selfTestData = [10.0, 10.0, 10.0, 10.0] # min, max, ondemand, performance
        self.selfTestStep = 0
        self.deviceClass = DeviceClass.processor
        self.workloads = ["minimum", "maximum", "powersave", "performance"]
        self.workloadNames = {"minimum": "User Space (min)", "maximum": "User Space (max)", "ondemand": "On Demand", "performance": "Performance", "powersave": "Power Save"}
        self.workloadTime = dict() # workload run time list indexed by [workload][cpu]
        self.workloadFreq = dict() # measured cpu freq indexed by [workload][cpu]
        self.workloadEffectiveFreq = dict() # measured effective cpu freq indexed by [workload][cpu]
        self.workloadErrors = dict() # error log indexed by [workload]
        self.currentWorkload  = "minimum"
        # what RHEL are we testing?
        self.redHatRelease = RedHatRelease()
        self.cpuPower = None # wait until the test gets planned
        self.sysCPUFreq = SysCPUFreq()
        self.cpuNumbers = None

    def getRequiredRPMs(self):
        rpms = ["kernel-tools", "make"] # for cpupower
        return rpms

    def dumpTables(self):
        print("\nDumping Data Tables:")
        for workload in self.workloads:
            print("\nworkload: " + workload)
            print("-----------------------------------------------------------------------------------")
            if workload in self.workloadTime:
                for (cpu, times) in self.workloadTime[workload].items():
                    sys.stdout.write("cpu%s: " % cpu)
                    for time in times:
                        sys.stdout.write(fmt.format("  {time:.2f}s", time=time))
                    print("")
            if workload in self.workloadFreq:
                for (cpu, freqs) in self.workloadFreq[workload].items():
                    sys.stdout.write("cpu%s: " % cpu)
                    for freq in freqs:
                        sys.stdout.write("  %4u MHz" % int(freq/1000))
                    print("")
            if workload in self.workloadEffectiveFreq:
                for (cpu, freqs) in self.workloadEffectiveFreq[workload].items():
                    sys.stdout.write("cpu%s: " % cpu)
                    for freq in freqs:
                        sys.stdout.write("  %4u MHz" % int(freq/1000))
                    print("")

    def setWorkloadTime(self, workload, cpu, time):
        if not workload in self.workloadTime:
            self.workloadTime[workload] = dict()
        if not cpu in self.workloadTime[workload]:
            self.workloadTime[workload][cpu] = list()
        self.workloadTime[workload][cpu].append(time)

    def setWorkloadFreq(self, workload, cpu, freq):
        if not workload in self.workloadFreq:
            self.workloadFreq[workload] = dict()
        if not cpu in self.workloadFreq[workload]:
            self.workloadFreq[workload][cpu] = list()
        self.workloadFreq[workload][cpu].append(freq)

    def setWorkloadEffectiveFreq(self, workload, cpu, freq):
        if not workload in self.workloadEffectiveFreq:
            self.workloadEffectiveFreq[workload] = dict()
        if not cpu in self.workloadEffectiveFreq[workload]:
            self.workloadEffectiveFreq[workload][cpu] = list()
        self.workloadEffectiveFreq[workload][cpu].append(freq)

    def logWorkloadError(self, errorMessage):
        print(errorMessage)
        if not self.currentWorkload in self.workloadErrors:
            self.workloadErrors[self.currentWorkload] = list()
        self.workloadErrors[self.currentWorkload].append(errorMessage)

    def getWorkloadTime(self, workload, cpu):
        try:
            return self.workloadTime[workload][cpu][-1] # get the last time
        except KeyError:
            return None

    def getWorkloadFreq(self, table, workload, cpu):
        try:
            return table[workload][cpu][-1] # get the last freq
        except KeyError:
            return None

    def getWorkloadTimes(self, workload, cpu):
        try:
            return self.workloadTime[workload][cpu] # get the whole list of times
        except KeyError:
            return None
    def getWorkloadTimesLength(self, workload):
        """ return the max number of entries of any cpu """
        max = 0
        for cpu in self.workloadTime[workload]:
            if len(self.workloadTime[workload][cpu]) > max:
                max = len(self.workloadTime[workload][cpu])
        return max

    def speedup(self, cpu, workload="maximum"):
        try:
            return self.getWorkloadTime("minimum", cpu)/self.getWorkloadTime(workload, cpu)
        except Exception as e:
            print("Error: could not determine speedup for workload %s" % workload)
            print(e)
            return None

    def checkFrequency(self, expected, actual, lowerTolerance=None, higherTolerance=None):
        """Check that expected-lowerTolerance <= actual <= expected+higherTolerance

        >>> c_test = CPUScalingTest()
        >>> c_test.lowerTolerance = 50
        >>> c_test.higherTolerance = 100
        >>> c_test.checkFrequency(100*1000, 12*1000)
        False
        >>> c_test.checkFrequency(100*1000, 50*1000)
        True
        >>> c_test.checkFrequency(100*1000, 201*1000)
        False
        >>> c_test.checkFrequency(100*1000, 200*1000)
        True
        """

        if not lowerTolerance:
            lowerTolerance = self.lowerTolerance

        if not higherTolerance:
            higherTolerance = self.higherTolerance

        # expected and actual int, in KHz
        lowest_tolerated = int(expected) - lowerTolerance*1000
        highest_tolerated = int(expected) + higherTolerance*1000
        # take floor at MHz level granularity
        actual = int(actual)/1000*1000
        return lowest_tolerated <= actual and highest_tolerated >= actual

    def checkFrequencyGreaterThan(self, freq1, freq2):
        # freq1 and freq2 int, in KHz
        # check if freq1+tolerance >= freq2
        return int(freq1) + self.higherTolerance*1000 > int(freq2)

    def checkSpeedup(self, speedup, warn=False, maximumLimit=True, expectSpeedup=True):
        if not self.predictedSpeedup and expectSpeedup:
            self.logWorkloadError("Error: could not predict speedup, unable to evaluate measured speedup")
            return False

        message = None
        tag = "Error:"
        if warn:
            tag = "Warning:"

        if speedup < self.minimumRequiredSpeedup():
            if expectSpeedup:
                message = fmt.format("{tag:} measured speedup {speedup:.2f} less than the minimum speedup of {min:.2f}. ", tag=tag, speedup=speedup, min=self.minimumRequiredSpeedup())
            else:
                print(fmt.format("Correctly held speedup to  {speedup:.2f}.", speedup=speedup))
        if speedup > self.maximumRequiredSpeedup():
            if expectSpeedup:
                # if no maximumLimit, downgrade to warn
                if not maximumLimit:
                    warn = True
                    tag = "Warning:"
                message = fmt.format("{tag:} measured speedup {speedup:.2f} greater than the maximum speedup of {max:.2f}", tag=tag, speedup=speedup, max=self.maximumRequiredSpeedup())
            else:
                message = fmt.format("{tag:} measured speedup {speedup:.2f} when no speedup was expected", tag=tag, speedup=speedup)

        if message:
            if warn:
                print(message)
            else:
                self.logWorkloadError(message)
                return False

        return True

    def normalizeForBug(self, frequency):
        """Adapt in response to bug in cpupower command line utility"""

        if frequency > 10000000: #KHz
            return frequency / 1000
        else:
            return frequency

    def getMinimumFrequency(self):
        if self.sysCPUFreq.getMinimumFrequency():
            return self.normalizeForBug(self.sysCPUFreq.getMinimumFrequency())
        # otherwise
        return self.normalizeForBug(self.cpuPower.getMinimumFrequency())

    def getMaximumFrequency(self):
        if self.sysCPUFreq.getMaximumFrequency():
            return self.normalizeForBug(self.sysCPUFreq.getMaximumFrequency())
        # otherwise
        return self.normalizeForBug(self.cpuPower.getMaximumFrequency())

    def minimumRequiredSpeedup(self):
        return self.predictedSpeedup - ((self.predictedSpeedup - 1.00)*self.speedUpTolerance/100.0)

    def maximumRequiredSpeedup(self):
        return self.predictedSpeedup + ((self.predictedSpeedup - 1.00)*self.speedUpTolerance/100.0)

    def tags(self):
        return [TestTag.noninteractive, TestTag.certification]

    def plan(self, planner):
        # if cpu scaling is supported, just schedule one test.
        tests = list()

        # for x86 architectures, check if either C-state/cpuidle is supported or
        # if P-state/frequency scaling is supported
        if self.getCPUarch().find("x86") == 0:
            if self.sysCPUFreq.isCStateSupported() or self.sysCPUFreq.isFrequencyScalingSupported():
                test = self.makeCopy()
                tests.append(test)
        elif self.sysCPUFreq.isFrequencyScalingSupported():
            test = self.makeCopy()
            tests.append(test)
        return tests

    def verify(self):
        if not self.cpuPower:
            self.cpuPower = CPUPower()
        return self.sysCPUFreq.isValid() and self.cpuPower.isValid()

    def logDevices(self, planner):
        devices = list()
        proc = ProcDeviceDetector()
        devices.extend(proc.getDevices(ProcDeviceDetector.procCPUInfo))
        # call the default to log udev devices
        devices.extend(HardwareTest.logDevices(self, planner))
        return devices

    def runLoadTest(self, cpu=None):
        """ run package load test, cpu default None means for all cpus """
        sys.stdout.write("Running CPU load test - ")
        if cpu != None:
            print("for only cpu %s" % cpu)
        else:
            print("for all cpus in the package")

        runTime = None
        tries = 0
        while tries < self.retryLimit:
            sys.stdout.flush()
            averageProcessTime = self.__runSubprocessLoadTest(singleCpu=cpu)
            if averageProcessTime:
                print(fmt.format("average worker process time: {avg:.2f} seconds", avg=averageProcessTime))
                if not runTime:
                    runTime = averageProcessTime
                else:
                    thisTime = averageProcessTime
                    if (abs(thisTime-runTime)/runTime)*100 < self.retryTolerance:
                        return thisTime
                    else:
                        runTime = thisTime
            tries += 1

        print(fmt.format("Warning: could not repeat load test times within {tol:.1f}%", tol=self.retryTolerance))
        return runTime    

    def __runSubprocessLoadTest(self, singleCpu=None):
        print("Running load test for package %s" % self.currentPackage)
        self.loadProcesses = dict()

        # if cpu is set, just load that one
        if singleCpu != None:
            print("Single CPU Test: Loading only cpu%s" % singleCpu)
            loadProcess = LoadProcess(self, singleCpu, aperfmperfSupport=self.effectiveFrequencySupport, armCPPCSupport=self.armCPPCSupport, powerPCSupport=self.powerPCSupport)
            loadProcess.start()
            self.loadProcesses[singleCpu] = loadProcess
        else: # load all cpus in the package
            for cpu in self.sysCPUFreq.getCPUs(self.currentPackage):
                loadProcess = LoadProcess(self, cpu, aperfmperfSupport=self.effectiveFrequencySupport, armCPPCSupport=self.armCPPCSupport, powerPCSupport=self.powerPCSupport)
                loadProcess.start()
                self.loadProcesses[cpu] = loadProcess

        print("waiting for load processes...")
        totalProcessTime = 0
        numberOfProcesses = len(self.loadProcesses)
        while len(self.loadProcesses) > 0 :
            for (cpu, process) in list(self.loadProcesses.items()):
                if process.isDone():
                    if process.returnCode != 0:
                        self.logWorkloadError("Error: load process terminated with return code %s" % process.returnCode)
                        return None    
                    runTime = process.getRunTime()
                    frequency = self.__getCurrentFrequency(cpu)
                    effectiveFrequency = process.getEffectiveFrequency()
                    if not runTime:
                        self.logWorkloadError("Error: could not find process run time")
                        return None
                    if not frequency:
                        self.logWorkloadError("Error: could not determine cpu frequency")
                        return None
                    # otherwise - collect info
                    print(fmt.format("process for cpu {cpu:} is done in {sec:.2f} seconds, at {freq:d} MHz", cpu=cpu, sec=runTime, freq=int(int(frequency)/1000)))
                    self.setWorkloadTime(self.currentWorkload, cpu, runTime)
                    self.setWorkloadFreq(self.currentWorkload, cpu, frequency)
                    if effectiveFrequency:
                        print("process effective frequency: %u MHz" % (effectiveFrequency/1000))
                        self.setWorkloadEffectiveFreq(self.currentWorkload, cpu, effectiveFrequency)   
                    elif not self.armCPPCSupport and not self.powerPCSupport:
                        # For load processes on ARM and power pc we cannot determine the effective frequency
                        print("Warning: could not determine effective frequency")
                    totalProcessTime += runTime
                    del self.loadProcesses[cpu]

            time.sleep(1) # poll once per second
            # if doing a single cpu, log its current freq while waiting
            if singleCpu != None:
                runningFreq = self.__getCurrentFrequency(cpu)

        print("processes complete")
        return totalProcessTime/numberOfProcesses

    def __getCurrentFrequency(self, cpu):
        if self.cpuPower.vendor: #for x86 Intel
            return self.cpuPower.getCurrentFrequency(cpu)
        else:
            return self.sysCPUFreq.getCurrentFrequency(cpu)

    def done(self, cpu):
        print("process for cpu %s is done. (caught SIGCHLD)" % cpu)
        try:
            del self.loadProcesses[cpu]
        except KeyError:
            self.logWorkloadError("Error: no cpu %s in load processes" % cpu)

    def verifyMinimumFrequency(self):
        waitTime = 5
        sys.stdout.write("Waiting %d seconds..." % waitTime)
        sys.stdout.flush()
        time.sleep(waitTime)
        sys.stdout.write(" done.\n")
        currentFrequency = self.__getCurrentFrequency(self.sysCPUFreq.getCPUs(self.currentPackage)[0])
        if self.getDebug() != Constants.off:
            print("%s vs. %s (min.)" % (currentFrequency, self.getMinimumFrequency()))

        if not self.getMinimumFrequency() or not currentFrequency or not self.checkFrequency(self.getMinimumFrequency(), currentFrequency, self.minimumCheckLowerTolerance, self.minimumCheckHigherTolerance):
            # CERTPX-238
            if self.hwpFlag not in self.cpuFlags:
                return False

        # otherwise
        return True

    def getSystemCapabilities(self):
        if not self.cpuPower.isValid():
            print("Error: could not control cpuscaling via cpupower (from kernel-tools)")
            return False
        if not self.sysCPUFreq.isValid():
            print("Error: could not determine cpuscaling capabilities via the %s file system" % self.sysCPUFreq.sysCPUDirectory)
            return False

        print("")
        print("System Capabilites:")
        print("-------------------------------------------------")

        self.logCPUInfo()

        print("System has %u cpus"% len(self.cpuPower.getCPUs()))

        # 1.5 get core/package topology
        self.sysCPUFreq.showCapabilities()

        # 2. ensure that all cpu's scale the same amount; if not bail
        # 3. verify all cpu's support the same control methods
        if self.cpuPower.isValid():
            print("Verified cpus are uniform in supported frequencies and governors")
        else:
            print("Error: non-uniform support of frequencies and governors")
            return False
        # 4. Log the testable govs. which I suggest to be "ondemand", "userspace", and "performance"; each test below is as applicable 
        self.cpuPower.printInformation()
        self.originalGovernor = self.cpuPower.getCurrentGovernor(self.cpuPower.getCPUs()[0])


        print("\nCPU Flags:")
        self.getCPUFlags()
        self.turboBoostSupport = False
        support = " NOT"
        if self.idaFlag in self.cpuFlags:
            support = ""
            self.turboBoostSupport = True
        print("    %s: Turbo Boost is%s supported" % (self.idaFlag, support))
        support = " NOT"
        if self.aperfmperfFlag in self.cpuFlags:
            support = ""
        print("    %s: aperf/mperf is%s supported"% (self.aperfmperfFlag, support))
        sys.stdout.flush()

        print("\n Measurement Method:")
        # use ./aperfmperf.py to assess effective frequency support, as cpuflags are not reliable.
        efi = EffectiveFrequency(0)
        self.armCPPCSupport = self.cpuPower.get_CPPC_support()
        self.powerPCSupport = self.cpuPower.get_POWERNV_support()
        if efi.has_support():
            self.effectiveFrequencySupport = True
            print("   /dev/cpu/0/cpuid indicates aperf/mperf is supported")
            print("   Using C sqrt load processes and aperf/mperf effective freq. measurement (aperf.c)")
        
        else:
            self.effectiveFrequencySupport = False
            if self.armCPPCSupport:
                print("   cppc_cpufreq is the supported driver")
                print("   Using C sqrt load processes.")
            elif self.powerPCSupport:
                print("   powernv-cpufreq is the supported driver")
                print("   Using C sqrt load processes.")
            else:
                print("   /dev/cpu/0/cpuid indicates aperf/mperf is NOT supported")
                print("   Using Python pi load processes with /sys freq. values, (no aperf/mperf support).")
        print("")
        sys.stdout.flush()    

        return True

    def getCPUFlags(self):
        # for now, assume all cpus have the same flags.
        self.cpuFlags = ""
        try:
            flagsCommand = Command("cat /proc/cpuinfo")
            self.cpuFlags = flagsCommand.getString(regex="^flags[^:]+:(?P<flags>.*)", regexGroup="flags", singleLine=False)
        except HwCertCommandException:
            try:
                # for rhel-alt ARM, flags are reported as features in command output
                self.cpuFlags = flagsCommand.getString(regex="^Features[^:]+:(?P<Features>.*)", regexGroup="Features", singleLine=False)
            except HwCertCommandException as exception:
                print("Warning: could not read CPU flags")
                print(exception)

    def getCPUarch(self):
        self.cpuArch = None
        try:
            self.cpuArch = Command("uname -m").getString()
        except HwCertCommandException as exception:
            print("Warning: could not find the CPU architecture")
            print(exception)
        return self.cpuArch

    def logCPUInfo(self):
        try:
            cpuinfo = Command("cat /proc/cpuinfo")
            cpuinfo.run()
            model = None
            cpu = None
            for line in cpuinfo.output:
                if line.startswith("processor"):
                    cpu = line.split(":")[1]
                if line.startswith("model name"):
                    thisModel = line.split(":")[1]
                    if model != thisModel:
                        print("CPU %s Model: %s" % (cpu, thisModel))
                        model = thisModel
        except Exception as exception:
            print("Warning: could not read CPU Info")
            print(exception)

        return True

    def turboBoostSupported(self):
        return self.turboBoostSupport

    def checkFrequencyResults(self, requiredFrequency, unlimitedMaximum=False, lowerTolerance=None, higherTolerance=None):
        """ requiredFrequency: int MHz """

        if not lowerTolerance:
            lowerTolerance = self.lowerTolerance

        if not higherTolerance:
            higherTolerance = self.higherTolerance


        success = True
        if self.effectiveFrequencySupport:
            freqTable = self.workloadEffectiveFreq
        else: # use /sys values 
            freqTable = self.workloadFreq

        for cpu in self.sysCPUFreq.getCPUs(self.currentPackage):
            frequency = self.getWorkloadFreq(freqTable, self.currentWorkload, cpu)
            if not frequency:
                self.logWorkloadError("Error: could not determine the measured frequency for cpu %s under workload %s" % (cpu, self.currentWorkload))
                success = False
                continue

            if unlimitedMaximum:
                if not self.checkFrequencyGreaterThan(frequency, requiredFrequency):
                    self.logWorkloadError("Error: cpu%s in package %s has a measured frequency of %u MHz vs. a requirement of %u MHz (+ %u MHz tolerance)"
                                            % (cpu, self.currentPackage, int(frequency)/1000, int(requiredFrequency/1000), higherTolerance))
                    success = False
            else:
                if not self.checkFrequency(requiredFrequency, frequency, lowerTolerance, higherTolerance):
                    self.logWorkloadError("Error: cpu%s in package %s has a measured frequency of %u MHz vs. a requirement of %u MHz - %u/+ %u MHz"
                                            % (cpu, self.currentPackage, int(frequency/1000), int(requiredFrequency/1000), lowerTolerance, higherTolerance))
                    success = False

        return success

    def runUserSpaceTests(self):
        # cpupower can't seem to set frequencies in userspace for acpi-cpufreq driver
        # user powersave policies for now
        if not "userspace" in self.cpuPower.getGovernors():
            print("Note: The userspace governor is not supported")
        else:
            print("Note: The userspace governor is supported")

        print("Using powersave min/max policies to get min and max performance workloads")
        print("")
        print("On Min/Max Power Save Governor Test:")
        print("-------------------------------------------------")
        governor = "powersave"

        sys.stdout.flush()
        # 5. Set the govr.(*) to "userspace" for each cpu via a command similar to:
        # 6. Verify the govr. setting took by checking ~/scaling_governor
        self.differenceSpeedUp = None
        success = True
        print("Setting governor to %s" % governor)
        if not self.cpuPower.setPolicy(self.getMinimumFrequency(), self.getMaximumFrequency(), governor):
            success = False

        # 7. Set the the cpu speed to it's lowest value
        currentFrequency = self.__getCurrentFrequency(self.sysCPUFreq.getCPUs(self.currentPackage)[0])
        if currentFrequency:
            print("Changing cpu frequency from %u to %u MHz" % (int(currentFrequency/1000), int(self.getMinimumFrequency()/1000)))
        else:
            print("Setting cpu frequency to %u MHz" % (int(self.getMinimumFrequency()/1000)))
        if governor == "userspace":
            if not self.cpuPower.setFrequency(self.getMinimumFrequency()):
                success = False
        else: # powersave - set min and max to hardware min.
            if not self.cpuPower.setPolicy(self.getMinimumFrequency(), self.getMinimumFrequency(), "powersave"):
                success = False
        self.currentWorkload = "minimum"
        sys.stdout.flush()

        # 8. Verify the speed is set to the lowest value by comparing ~/scaling_min_freq to ~/scaling_cur_freq
        # if not, it's a warning - the frequency will be checked again after running the load.
        currentFrequency = self.__getCurrentFrequency(self.sysCPUFreq.getCPUs(self.currentPackage)[0])
        if not self.getMinimumFrequency() or not currentFrequency or not self.checkFrequency(self.getMinimumFrequency(), currentFrequency, self.minimumCheckLowerTolerance, self.minimumCheckHigherTolerance):
            print("Warning: Could not verify that cpu frequency is set to the minimum value of %s KHz" % self.getMinimumFrequency())
        # calculate predicted speedup for the package
        try:
            self.predictedSpeedup = float(self.getMaximumFrequency())/float(self.getMinimumFrequency())
        except Exception as e:
            print("Error: could not determine predicted speedup")
            print(e)
            self.predictedSpeedup = None
            success = False

        # 9. Run and time non-cacheable work(?) test for each CPU
        # 10. Repeat and time until execution times are within self.retryTolerance of each other 
        # run to verify the base line execution time, bail if self.retryTolerance margin cannot be acheived in self.retryLimit runs.
        self.minimumFrequencyTestTime = self.runLoadTest()
        if not self.minimumFrequencyTestTime:
            return False
        print(fmt.format("Minumum frequency average load test time: {min:.2f}", min=self.minimumFrequencyTestTime))

        # 10.5 check that measured frequency is near minimum
        if not self.checkFrequencyResults(self.getMinimumFrequency(), self.minimumCheckLowerTolerance, self.minimumCheckHigherTolerance):
            success = False

        # 11. Set the cpu speed to it's highest value as above.
        currentFrequency = self.__getCurrentFrequency(self.sysCPUFreq.getCPUs(self.currentPackage)[0])
        if governor == "userspace":
            print("Changing cpu frequency from %u to %u MHz" % (int(currentFrequency/1000), int(self.getMaximumFrequency()/1000)))
            if not self.cpuPower.setFrequency(self.getMaximumFrequency()):
                success = False
        else: # powersave - set min and max to hardware max.
            print("Changing cpu frequency from %u to %u MHz, performance governor" % (int(currentFrequency/1000), int(self.getMaximumFrequency()/1000)))
            if not self.cpuPower.setPolicy(self.getMaximumFrequency(), self.getMaximumFrequency(), "performance"):
                success = False
        self.currentWorkload = "maximum"

        currentFrequency = self.__getCurrentFrequency(self.sysCPUFreq.getCPUs(self.currentPackage)[0])
        if not self.getMaximumFrequency() or not currentFrequency or not self.checkFrequency(self.getMaximumFrequency(), currentFrequency):
            print("Warning: Could not verify that cpu frequency %s MHzis set to the maximum value of %s MHz" % (int(currentFrequency/1000), int(self.getMaximumFrequency()/1000)))

        # 12. Repeat workload test, record timing
        self.maximumFrequencyTestTime = self.runLoadTest()
        if not self.maximumFrequencyTestTime:
            return False
        print(fmt.format("Maximum frequency average load test time: {max:.2f}", max=self.maximumFrequencyTestTime))

        # 12.5 check the measured max measured freqency is >= stated max
        if not self.checkFrequencyResults(self.getMaximumFrequency(), unlimitedMaximum=True):
            success = False

        # 13. Verify MHz increase is comparable to time % decrease( eg. slow MHz/fast MHz ~= fast time/slow time; 
        # again with self.speedUpTolerance margin)


        # for each cpu in the package
        for cpu in self.sysCPUFreq.getCPUs(self.currentPackage):
            measuredSpeedup = self.speedup(cpu)
            print("")
            print(fmt.format("CPU Frequency Speed Up: {predicted:.2f}", predicted=self.predictedSpeedup))
            print(fmt.format("CPU {cpu:} Measured Speed Up: {measured:.2f}", cpu=cpu, measured=measuredSpeedup))
            self.checkSpeedup(measuredSpeedup, warn=True)

        return success


    def runPerformanceTests(self):
        if not "performance" in self.cpuPower.getGovernors():
            print("Note: The performance governor is not supported")
            return True

        print("")
        print("Performance Governor Test:")
        print("-------------------------------------------------")
        self.differencePerformanceVsMaximum = None
        # 20. Set and verify the govr. to "performance"
        governor = "performance"
        success = True

        print("Setting governor to %s" % governor)
        if not self.cpuPower.setGovernor(governor):
            success = False
        self.currentWorkload = "performance"

        # 21. Verify the current speed is the same as scaling_max_freq
        currentFrequency = self.__getCurrentFrequency(self.sysCPUFreq.getCPUs(self.currentPackage)[0]) # KHz

        if not self.getMaximumFrequency() or not currentFrequency or not self.checkFrequency(self.getMaximumFrequency(), currentFrequency):
            print("Warning: Current cpu frequency of %s KHz is not set to the maximum value of %s KHz" % (currentFrequency, self.getMaximumFrequency()))

        # 22. Repeat workload test, record timing, also verify frequency does not drop during run
        self.performanceTestTime = self.runLoadTest()
        if not self.performanceTestTime:
            return False
        # otherwise
        print(fmt.format("Performance load test time: {time:.2f}", time=self.performanceTestTime))

        # 22.5 check EFI freq vs. max
        if not self.checkFrequencyResults(self.getMaximumFrequency(), unlimitedMaximum=True):
            success = False

        # 23. Compare the timing to the max results for a self.speedUpTolerance delta
        for cpu in self.sysCPUFreq.getCPUs(self.currentPackage):
            performanceSpeedup = self.speedup(cpu, "performance")
            if performanceSpeedup:
                print(fmt.format("CPU {cpu:} Performance Speedup: {speedup:.2f}", cpu=cpu, speedup=performanceSpeedup))
                self.checkSpeedup(performanceSpeedup, warn=True)
            else:
                print("ERROR: Performance Speedup is None")
                success = False

        return success

    def runPowerSaveTests(self):
        """ powersave AKA ondemand  But not on ARM!"""
        singleCPUGovernor = self.cpuPower.getSingleCPUTestGovernor()
        if not singleCPUGovernor:
            print("Note: The no governor found (powersave or ondemand) for single cpu test.")
            return True

        """ test one cpu in the package while others are idle, and verify that its speedup is greater than when all 
            cpus in the package are loaded """
        print("")
        print("Single CPU %s Test:" % self.workloadNames[singleCPUGovernor])
        print("-------------------------------------------------")
        # 30. Set and verify the govr.
        success = True

        isPowerSave = singleCPUGovernor == "powersave"

        # set the policy for max boost freq if available
        if self.getDebug() != Constants.off:
            print("Max: %s MHz vs. Max Boost: %s MHz" % (self.getMaximumFrequency(), self.cpuPower.getMaximumBoostFrequency()))
        maximumFrequency = self.getMaximumFrequency()
        if self.cpuPower.getMaximumBoostFrequency() and self.cpuPower.getMaximumBoostFrequency() > self.getMaximumFrequency():
            print("Using maximum boost frequency.")
            maximumFrequency = self.cpuPower.getMaximumBoostFrequency()

        print("Setting governor to %s   (min: %s MHz, max: %s MHz)" % ( singleCPUGovernor,self.getMinimumFrequency()/1000, maximumFrequency/1000))
        if not self.cpuPower.setPolicy(self.getMinimumFrequency(), maximumFrequency, singleCPUGovernor):
            success = False
        self.currentWorkload = singleCPUGovernor

        # 16. Wait a fixed period of time, then verify current speed is the slowest in as before
        if not self.verifyMinimumFrequency():
            print("Warning: Could not verify that cpu frequency has settled to the minimum value")

        # 32 Pick a cpu other than cpu0 to load
        singleCpu = self.getSingleCpu()
        print("Using cpu %s to test Single CPU On Demand." % singleCpu)


        # 33. Repeat workload test, loading a single cpu
        self.onDemandTime = self.runLoadTest(cpu=singleCpu)
        if not self.onDemandTime:
            return False
        # otherwise
        print(fmt.format("Single CPU load test time: {ondemand:.2f}", ondemand=self.onDemandTime))

        # 23. Compare the timing to the max results for a self.speedUpTolerance delta
        powersaveSpeedup = self.speedup(singleCpu, singleCPUGovernor)
        # speedup should not happen for ARM if it's 'powersave' governor
        if not powersaveSpeedup:
            return False
        print(fmt.format("Single CPU {gov:} Speedup: {speedup:.1f}", gov=self.workloadNames[singleCPUGovernor], speedup=powersaveSpeedup))

        if self.armCPPCSupport and isPowerSave:
            success = self.checkSpeedup(powersaveSpeedup, warn=False, maximumLimit=False, expectSpeedup=False)
        else:
            success = self.checkSpeedup(powersaveSpeedup, warn=False, maximumLimit=False)

        return success

    def getSingleCpu(self):
        # 32 Pick a cpu other than cpu0 to load
        for cpu in self.sysCPUFreq.getCPUs(self.currentPackage):
            if cpu != 0:
                break
        return cpu

    def restoreGovernors(self):
        print("Restoring original governor to %s" % (self.originalGovernor))
        self.cpuPower.setGovernor(self.originalGovernor)

    def runSummary(self):
        for package in self.sysCPUFreq.getPackages():
            self.printSummary(package)

        if self.workloadErrors:
            print("\nError Summary:")
            print("----------------")
            for workload in self.workloads:
                if workload in self.workloadErrors:
                    for error in self.workloadErrors[workload]:
                        print("%s: %s" % (self.workloadNames[workload], error))
            return False

        return True


    def printSummary(self, package):

        if self.getDebug() != Constants.off:
            self.dumpTables()

        if self.effectiveFrequencySupport:
            freqTable = self.workloadEffectiveFreq
            self.printFreqSummary("CPU Effective Frequency", package, freqTable)
        else:
            freqTable = self.workloadFreq
            self.printFreqSummary("CPU Frequency Test (No Effective Frequency Support)", package, freqTable)

        for governor in ["powersave", "ondemand"]:
            if governor in self.workloadTime:
                print("\nCPU Workload Test:\n")
                print(fmt.format("Expected Speedup: {predicted:.2f}", predicted=self.predictedSpeedup))
                print(fmt.format("Allowable Speedup: {min:.2f} to {max:.2f}", min=self.minimumRequiredSpeedup(), max=self.maximumRequiredSpeedup()))
                print("")
                print("           %s" % self.workloadNames[governor])
                print("-------- -------------- ")

                #actual values, one row per cpu
                workload = governor
                for cpu in self.sysCPUFreq.getCPUs(package):
                    workloadTime = self.getWorkloadTime(workload, cpu)
                    if workloadTime:
                        sys.stdout.write("cpu %-3s  " % cpu)
                        sys.stdout.write(fmt.format(" {speedup:.2f}", speedup=self.speedup(cpu, workload)))
                        sys.stdout.write((fmt.format(" ({freq:4d} MHz, {time:.2f}s)      ", freq=int(self.getWorkloadFreq(freqTable, workload, cpu)/1000), time=workloadTime)[:10]))
                        print("")

                print("")

    def printFreqSummary(self, title, package, freqTable):
        print("")
        print("Summary for Package %s:" % package)
        print("\n%s:\n" % title)
        print("           User Min       User Max      Performance")
        print("-------- -------------- -------------- --------------")

        # expected values
        minimumFrequency = int(self.getMinimumFrequency())/1000
        maximumFrequency = int(self.getMaximumFrequency())/1000
        print("expected %5u MHz      %5u MHz      %5u MHz" % (minimumFrequency, maximumFrequency, maximumFrequency))

        #actual values, one row per cpu
        for cpu in self.sysCPUFreq.getCPUs(package):
            sys.stdout.write("cpu %-3s  " % cpu)
            for workload in self.workloads:
                if workload in self.workloadTime and workload != "powersave" and workload != "turbo":
                    freq = self.getWorkloadFreq(freqTable, workload, cpu)
                    if freq:
                        sys.stdout.write(" %4u" % int(freq/1000))
                        sys.stdout.write((fmt.format(" ({time:.2f}s)      ", time=self.getWorkloadTime(workload, cpu))[:10]))
            print("")

        print("")

    def printWorkloadTimeSummary(self, package):
        sys.stdout.write("Workload run times for package %s:\n" % package)
        sys.stdout.write("-------------------\n")
        workloads = self.workloads
        workloads.append("turbo")
        for workload in workloads:
            if workload in self.workloadTime:
                sys.stdout.write("%-13s" % workload)
                run = 1
                for run in range(1, self.getWorkloadTimesLength(workload)+1):
                    sys.stdout.write(" %7u " % run)
                    run += 1
                sys.stdout.write("\n")
                for cpu in self.sysCPUFreq.getCPUs(package):
                    if self.getWorkloadTimes(workload, cpu):
                        sys.stdout.write("    cpu %3s: " % cpu)
                        for time in self.getWorkloadTimes(workload, cpu):
                            if time:
                                sys.stdout.write(fmt.format("{time:9.2f}", time=time))
                            else:
                                sys.stdout.write("    -    ")
                        print("")
                print("")
        print("\n")

    def adjustMinimumCheckTolerances(self):
        """ Adjusts minimum tolerances.

        Expectation is that self.cpuFlags has already been populated.
        """
        if self.hwpFlag in self.cpuFlags:
            self.minimumCheckLowerTolerance = self.lowerTolerance
            self.minimumCheckLowerTolerance = self.higherTolerance

    def enableAllCStates(self):
        print("Enabling all cstates")
        for cpu in range(0, self.numCPUs):
            for state in range(1, self.numCStates-1):
                Command("echo 0 > /sys/devices/system/cpu/cpu%s/cpuidle/state%s/disable" % (cpu, state)).echo()

    def disableCState(self, cState):
        """disable c-state provided (and all higher) on all cpus
        """
        print("Disabling c-state %s and higher" % cState)
        for cpu in range(0, self.numCPUs):
            for state in range(cState, self.numCStates-1):
                Command("echo 1 > /sys/devices/system/cpu/cpu%s/cpuidle/state%s/disable" % (cpu, state)).echo()

    def outputCStateData(self, cState, filePath):
        Command("echo > %s" % filePath).echo()
        for cpu in range(0, self.numCPUs):
            timeVal = Command("cat /sys/devices/system/cpu/cpu%s/cpuidle/state%s/time" % (cpu, cState)).getString()
            Command('echo "%s %s" >> %s' % (cpu, timeVal, filePath)).echo()

    def runCstate(self):
        #Run the cstate test first for x86
        result = TestResult.WARN

        cpuArch = self.getCPUarch()
        if cpuArch and self.getCPUarch().find("x86") == 0:
            try:
                # get max number of cpus (zero based)
                self.numCPUs = int(Command("nproc").getString())

                # get max number of cstates (zero based)
                numCStates = 0
                cStatesCommand = Command("ls -1 /sys/devices/system/cpu/cpu0/cpuidle | wc -l")
                try:
                    numCStates = int(cStatesCommand.getString())
                except HwCertCommandException:
                    print("Warning - could not determine the number of cstates using command:\n%s" % cStatesCommand.command)

                try:
                    numCStates = int(numCStates)
                except ValueError:
                    print("Warning - could not convert %s to integer value" % numCStates)

                self.numCStates = numCStates

                # reset all cstates
                self.enableAllCStates()

                # iterate over c-states, disable cstate, wait a second to get values from all
                # cpus, wait another 30 seconds, get values again, and diff the results.
                # It is expected the values be consistent.  c-state zero cannot be disabled.
                # NOTE - if for some reason there are no c states detected, but process still gets to here,
                # self.numCStates should be 0 and the below for loop will not break, it just won't execute the inner block
                for x in range(1, self.numCStates-1):
                    self.disableCState(x)
                    time.sleep(1)
                    self.outputCStateData(x, "/tmp/start")
                    time.sleep(30)
                    self.outputCStateData(x, "/tmp/end")
                    if filecmp.cmp("/tmp/start", "/tmp/end"):
                        print("CPU c-state %s passed" % x)
                        result = TestResult.PASS
                    else:
                        print("CPU c-state %s failed" % x)
                        print("starting values:")
                        Command("cat /tmp/start").echo()
                        print("do not match ending values:")
                        Command("cat /tmp/end").echo()
                    self.enableAllCStates()

            except HwCertCommandException as exception:
                print("Warning: Cstate subtest did not pass, thus continuing testing further subtests.")

        return result


    def run(self):
        if self.runSubTest(self.runCstate, "Cstate") == TestResult.PASS:
            return True # If the Cstate test has passed do not execute the Pstate test

        #If the cstate test has failed run the pstate test
        if not self.cpuPower:
            self.cpuPower = CPUPower()

        if not self.runSubTest(self.getSystemCapabilities, "Capabilities"):
            return False

        # CERTPX-1425
        self.adjustMinimumCheckTolerances()
        # CERTPX-238
        if self.hwpFlag in self.cpuFlags:
            self.lowerTolerance = 5000 # Mhz

        result = TestResult()
        for package in self.sysCPUFreq.getPackages():

            print("Testing CPU Package {}".format(package))
            self.currentPackage = package

            self.waitForLull()

            if not self.runSubTest(self.runUserSpaceTests, "User Space, package {}".format(package)):
                result.combine(TestResult.FAIL)

            if not self.runSubTest(self.runPowerSaveTests, "Power Save, package {}".format(package)):
                result.combine(TestResult.FAIL)

            if not self.runSubTest(self.runPerformanceTests, "Performance, package {}".format(package)):
                result.combine(TestResult.FAIL)

            if self.originalGovernor:
                self.restoreGovernors()

        if not self.runSubTest(self.runSummary, "Summary"):
            result.combine(TestResult.FAIL)

        # CERTPX-5818
        total_cpus = self.get_total_cpus()
        online_cpus = len(self.cpuPower.cpuNumbers)
        offline_cpus = total_cpus - online_cpus

        # Generate Summary of Online/Offline CPUs.
        print("Total CPUs found: {}".format(total_cpus))
        print("Total Online CPUs found: {}".format(online_cpus))
        print("Total Offline CPUs found: {}".format(offline_cpus))
        if offline_cpus:
            result.combine(TestResult.WARN)

        planned_required_cpus = self.getTestParameters().get("required-cpus")
        if planned_required_cpus:
            planned_required_cpus = int(planned_required_cpus)
        else:
            planned_required_cpus = online_cpus

        if planned_required_cpus > online_cpus :
            print("Warning: Could not execute test on planned required-cpus.")
            print("Info: Detected {} online cpus whereas {} were planned."
                .format(online_cpus, planned_required_cpus))
            result.combine(TestResult.FAIL)

        return result

    def get_total_cpus(self):
        """
        This method execute 'lscpu' command
        to identify total no. of cpu cores.
        return - total_cpus (type Integer)
        """
        total_cpus = int()
        total_cpus_cmd = Command('lscpu')
        try:
            total_cpus = total_cpus_cmd.getInteger(regex='CPU\\(s\\):\\s+(?P<cpus>\\d+)',
                                                    singleLine=False,
                                                    regexGroup="cpus")
        except HwCertCommandException as e:
            print("Error: Could not determine total cpus.")
        return total_cpus


class LoadProcess:
    def __init__(self, owner, cpu, aperfmperfSupport, armCPPCSupport, powerPCSupport):
        self.owner = owner
        self.cpu = cpu
        self.workProcess  = None
        self.returnCode = None
        self.start_elapsed_time = None
        self.runTime = None
        self.effectiveFrequency = None
        self.aperfmperfSupport = aperfmperfSupport
        self.armCPPCSupport = armCPPCSupport
        self.powerPCSupport = powerPCSupport
        # signal.signal(signal.SIGCHLD, self.sigChldHandler)
        self.command = os.path.dirname(os.path.abspath( __file__ )) + "/cpuscaling.py work "  # Use the Effective Frequency Interface and pi() calculation for a load process
        if self.aperfmperfSupport:
            self.command = "./aperf"               # Use aperf.c, which uses TSC/aperf/mperf and a square root calculation from libm
        elif self.armCPPCSupport or self.powerPCSupport:
            self.command = "./cppc"

    def start(self):
        try:
            print("starting process for cpu %s" % self.cpu)
            self.workProcess = Command("%s %s" % (self.command, self.cpu))
            print("using work process: %s" % self.workProcess.command)
            self.workProcess.start()
            taskset = Command("taskset -pc %s %s" % (self.cpu, self.workProcess.getPID()))
            if self.owner.getDebug() != Constants.off:
                taskset.echo()
            else:
                taskset.run()
        except HwCertCommandException as exception:
            print("Error: could not start load process")
            print(exception)
            print(exception.command.printErrors())
            return False
        return True

    def getRunTime(self):
        return self.runTime

    def getEffectiveFrequency(self):
        return self.effectiveFrequency

    def sigChldHandler(self, signalNumber, stackFrame):
            print("Got signal %d" % signalNumber)
            self.__stopTimer()
            self.owner.done(self.cpu)

    def isDone(self):
        if self.workProcess:
            self.returnCode = self.workProcess.poll()
            if self.returnCode is not None:
                while True:
                    line = self.workProcess.readline()
                    if line:
                        try:
                            if line.startswith("runTime: "):
                                self.runTime = float(line.split()[1])
                            elif line.startswith("freq: "):
                                self.effectiveFrequency = int(line.split()[1])
                        except (ValueError, IndexError):
                                pass # keep trying
                    else:
                        break
                return True
        # otherwise
        return False

    def getReturnCode(self):
        return self.returnCode

def pi():
        decimal.getcontext().prec = 500
        s = decimal.Decimal(1)
        h = decimal.Decimal(3).sqrt()/2
        n = 6
        for i in range(170):
            A = n*h*s/2  # A ... area of polygon
            s2 = ((1-h)**2+s**2/4)
            s = s2.sqrt()
            h = (1-s2/4).sqrt()
            n = 2*n
        # print i,":",A
        return True
        
if __name__ == "__main__":
    import doctest
    doctest.testmod(exclude_empty=True)
    sys.exit(0)

    # this load process uses the Effective Frequency Interface (see aperfmperf.py)
    if "work" in sys.argv:
        effectiveFrequencyInterface = None
        effectiveFrequency = 0
        if len(sys.argv) > 2 and sys.argv[2].isdigit():
            cpuNumber = int(sys.argv[2])
            effectiveFrequencyInterface = EffectiveFrequency(cpuNumber)
            effectiveFrequencyInterface.start_sampling()
        (start_utime, start_stime, start_cutime, start_cstime, start_elapsed_time) = os.times()
        pi()
        (stop_utime, stop_stime, stop_cutime, stop_cstime, stop_elapsed_time) = os.times()
        print("") # make sure data is on its own line
        if effectiveFrequencyInterface:
            effectiveFrequency = effectiveFrequencyInterface.stop_sampling()
            if effectiveFrequency > 0:
                print("freq: %s" % effectiveFrequency)
        runTime = stop_elapsed_time - start_elapsed_time 
        print("runTime: %s" % (runTime, ))
        sys.stdout.flush()
        sys.exit(0)

    test = CPUScalingTest()
    returnValue = test.do(sys.argv)
    sys.exit(returnValue)