#!/usr/bin/env python
# Copyright (c) 2006 Red Hat, Inc. All rights reserved. This copyrighted material
# is made available to anyone wishing to use, modify, copy, or
# redistribute it subject to the terms and conditions of the GNU General
# Public License v.2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# Author: Greg Nichols
#         YangKun <ykun@redhat.com>
#         Caspar Zhang <czhang@redhat.com>
import sys, string

from rhcert.tags import Constants, DeviceClass
from rhcert.command import Command, HwCertCommandException
from rhcert.compatability import rhcert_get_status_output

from rhcert.suites.hwcert.hwcert.opticalDiskTest import OpticalDiskTest

class DVDTest(OpticalDiskTest):

    def __init__(self):
        OpticalDiskTest.__init__(self, "dvd")
        self.mediaTypes = ["dvdrw", "dvdplusrw", "dvdr", "dvdplusr", "dvd"]
        self.deviceClass = DeviceClass.optical
        self.fileSourceDir = "/usr/share" # ranges from 970MB-3.5GB

    def checkMediaSize(self, size):
        if size < 800000 :
            print("Error: %s appears NOT to be a DVD !\n" % self.getLogicalDeviceName())
            return False
        # otherwise
        return True

    def writer(self, deviceType):
        cdromDevice = self.getLogicalDeviceName()
        """handles DVD-R or DVD+R writing a chunk of the filesystem to the media and then checking the
           file sums"""
        print("\n  +++ Start writing ...")
        sys.stdout.flush()

        try:
            # Unmount the drive first in case it's mounted
            Command("umount /dev/%s" % cdromDevice).echoIgnoreErrors()

            # DVD-R
            if deviceType == "dvd":
                # check the write mode
                writeOpts = self.getWriteMode()
                # support burnfree ?

                try:
                    mkisofsSize = Command("mkisofs -quiet -R -print-size %s " % self.fileSourceDir)
                    cdblocks= mkisofsSize.getInteger()
                except HwCertCommandException as exception:
                    # try work-around for RHEL/mkisofs bug 193916
                    # remove the return character first
                    item = mkisofsSize.output.pop().replace('\n', '')
                    if item.isdigit():
                        cdblocks = int(item)
                        print("Warning: mkisofs had errors")
                        print(exception)
                    else:
                        print("Error: could not determine block size")
                        print(exception)
                        return False

                # write the directory to disk
                recordCommand = Command("mkisofs -quiet -R %s | cdrecord -silent %s dev=/dev/%s fs=32M tsize=%ss - > /dev/null 2>&1" % (self.fileSourceDir, writeOpts, cdromDevice, cdblocks))
                #recordCommand = Command("ls")
                recordCommand.echoIgnoreErrors()
                if recordCommand.returnValue == 0:
                    print("  +++ Write disk succeeded ! \n")
                    self.handleEject(cdromDevice)
                else:
                    print("  +++ Error: could not write to disk ! \n")
                    sys.stdout.flush()
                    return False
            # DVD+R
            elif deviceType == "dvdplus":
                # write the directory to disk
                (status, junk) = rhcert_get_status_output("growisofs -Z /dev/%s -quiet -R %s > /dev/null 2>&1" % (cdromDevice, self.fileSourceDir))
                print(junk)
                if status == 0:
                    print("  +++ Write DVD+ disk succeeded ! \n")
                else:
                    print("  +++ Error: Write DVD+ disk failed !\n")
                    return False
                self.handleEject(cdromDevice)
                sys.stdout.flush()
            else:
                print("  +++ Error: Un-supported media .\n")
                return False

            # write done, start to compare
            return self.compare(deviceType, cdromDevice)

        except HwCertCommandException as exception:
            print(" +++ Error:")
            print(exception)
            return False

    def run(self):
        """main function"""
        if not self.initializeDeviceInfo():
            return False

        if self.getMode() == Constants.auto:
            print("Error: The DVD test can not be run in auto mode.")
            return False

        self.getMediaChoice()
        # test DVD+ only
        if self.mediaChoice == "dvdplusrw":
            description = "good"
            mode = "rewrite"
            testtype = "dvdplus"
        elif self.mediaChoice == "dvdplusr":
            description = "new"
            mode = "write"
            testtype = "dvdplus"
        # test DVD- only
        elif self.mediaChoice == "dvdrw":
            description = "good"
            mode = "rewrite"
            testtype = "dvd"
        elif self.mediaChoice == "dvdr":
            description = "new"
            mode = "write"
            testtype = "dvd"
        elif self.mediaChoice == "dvd":
            description = "good data"
            mode = "read"
            testtype = "dvd"
        else:
            print("Error: could not determine media type")
            return False

        self.echoTest(self.getLogicalDeviceName(), self.mediaChoice, description)
        if testtype == "dvdplus" and not self.checkMediaType(self.getLogicalDeviceName(), self.mediaChoice):
            return False

        if self.testDisk(message="%s %s" %(description, self.printMediaName(self.mediaChoice)), readWriteMode=mode, type=testtype):
            return True
        # Otherwise
        return False

if __name__ == "__main__":
    test = DVDTest()
    returnValue = test.do(sys.argv)
    sys.exit(returnValue)
