#!/usr/bin/env python
# Copyright (c) 2006 Red Hat, Inc. All rights reserved. This copyrighted material
# is made available to anyone wishing to use, modify, copy, or
# redistribute it subject to the terms and conditions of the GNU General
# Public License v.2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# Author: YangKun <ykun@redhat.com>
#

import os, sys, re, time, platform
import subprocess

from rhcert.tags import Constants, DeviceClass
from rhcert.command import Command, HwCertCommandException

from rhcert.suites.hwcert.hwcert.usbTest import USBTestBase

class ExpressCardTest(USBTestBase):

    def __init__(self):
        USBTestBase.__init__(self, "expresscard")
        self.deviceClass = DeviceClass.expresscard
        self.lspciCommand = "lspci -D"

    def getRequiredRPMs(self):
        return ["pciutils"]

    def plan(self, planner):
        tests = list()
        hasUSB = False
        hasBattery = False
        hasPCIE = False
        devices = planner.getUdevDeviceDetector().getDevices()
        for device in devices:
            if device.getProperty("SUBSYSTEM") == "usb":
                hasUSB = True
            if device.getProperty("SUBSYSTEM") == "power_supply":
                hasBattery = True
            if device.getProperty("DRIVER") == "pcieport":
                hasPCIE = True

        # lifted from harness.py
        hasExpressCard = False
        arch = platform.machine()
        if arch in ['i386','i586','i686','x86_64']:
            dmidecode = os.popen("/usr/sbin/dmidecode")

            for line in dmidecode:
                if line.find(':') == -1:
                    continue
                (key,val)=[t.strip() for t in line.split(':',1) if line.find(':')]

                # We should revisit this to see if dmidecode can provide
                # the number of card slots as well, my box only has one
                if key == "Designation" and val == "ExpressCard Slot":
                    hasExpressCard = True
                    break

            dmidecode.close()

        if (hasUSB and hasBattery and hasPCIE) or hasExpressCard:
            test = self.makeCopy()
            tests.append(test)

        return tests

    def getPCIDevices(self):
        devices = list()
        pattern = re.compile("(?P<id>[0-9a-f]{4}:[0-9a-f]{2}:[0-9a-f]{2}\.[0-9a-f]) (?P<name>.*$)")
        try:
            pipe = Command(self.lspciCommand)
            pipe.start()
            while 1:
                line = pipe.readline()
                if not line or line.strip() == "":
                    break
                match = pattern.match(line)
                # every list is a tuple containing information like:
                # ('00:00.0','Host bridge: Intel Corporation Core Processor DRAM Controller (rev 02)')
                dev = match.groups()
                devices.append(dev)
        except HwCertCommandException as e:
            print("Error: could not read PCI device from lspci command")
            print(e)
            return None

        return devices

    def checkAllPCIUnplugged(self):
        devices = self.getPCIDevices()
        if not devices:
            return None
        self.pluggedInPCIDevices = list()
        for dev in devices:
            (id, name) = dev
            if id:
                self.pluggedInPCIDevices.append(id)
                print("  %s appears to be attached on PCI bus: %s" % (name, id))

    def addNewPCIPlugged(self, verifiedDevices):
        devices = self.getPCIDevices()
        if not devices:
            return None
        for dev in devices:
            (id, name) = dev
            # should execute checkAllPCIUnplugged() first
            if id and not id in self.pluggedInPCIDevices:
                if id in verifiedDevices:
                    print("\n  Note: Same device been plugged into the same slot.\n")
                    return None
                else:
                    print("\n  Device attached to PCI bus: \
                           \n      %s appears to be attached on PCI bus: %s" % (name, id))
                    return id
        return None


    def confirmPCIUnplugged(self, searchID):
        devices = self.getPCIDevices()
        if not devices:
            return True
        for dev in devices:
            (id, name) = dev
            if id == searchID:
                return False

        return True

    def pciInfo(self):
        """return detailed PCI info"""
        pciinfo = subprocess.getoutput("lspci -vvvv")
        return pciinfo

    def udevInfo(self):
        """return current udev info"""
        udevinfo = subprocess.getoutput("udevadm info --export-db")
        return udevinfo

    def contextDiff(self, context1, context2):
        """compare two "files" and return the diff parts"""
        tmpFile1 = "/tmp/hwcertExpressCardTmpFile1"
        tmpFile2 = "/tmp/hwcertExpressCardTmpFile2"
        file(tmpFile1, mode='w').write(context1)
        file(tmpFile1).close()
        file(tmpFile2, mode='w').write(context2)
        file(tmpFile2).close()
        diffpart = subprocess.getoutput("diff %s %s" % (tmpFile1, tmpFile2))
        os.system("rm -f %s %s" % (tmpFile1, tmpFile2))
        return diffpart

    def run(self):

        verifiedDevs = list()

        if self.getMode() == Constants.auto:
            print("Error: the expresscard test can not be run in auto mode.")
            return False

        print("\nExpressCard Test:")

        if not self.promptContinue("Please remove all ExpressCard devices"):
            result = Constants.ABORT
            return False

        # scan both the USB bus and PCI bus first
        print ("\n---- Current USB devices ----")
        self.checkAllUSBPortsUnplugged()

        print ("\n---- Current PCI devices ----")
        self.checkAllPCIUnplugged()

        # ask user how many ExpressCard slots there are
        while True:
            i = self.promptInteger("\nHow many ExpressCard slots are there ? ")
            if i > 0:
                self.numSlots = i
                break
            print("Please enter a positive integer.")

        while i > 0:
            print("\nTesting slot %s of %s..." % (self.numSlots - i + 1, self.numSlots))

            isUSB = False

            # gather HW info
            pciInfo1  = self.pciInfo()
            udevInfo1 = self.udevInfo()

            if not self.promptContinue("Please plug in an ExpressCard device"):
                result = Constants.ABORT
                return False

            time.sleep(5)
            found = self.addNewPluggedUSBPort()
            if found:
                isUSB = True
            else:
                found = self.addNewPCIPlugged(verifiedDevs)
                if not found:
                    print("Error: Device not found !\n" "Please retry")
                    continue

            # gather HW info a second time
            pciInfo2  = self.pciInfo()
            udevInfo2 = self.udevInfo()
            
            if not self.promptContinue("Please unplug the ExpressCard device."):
                result = Constants.ABORT
                return False

            time.sleep(5)
            if isUSB:
                (busID, portID, name, deviceID) = found
                if self.confirmUSBPortUnplugged(deviceID):
                    print("Unplugged USB device\n    %s (%s)" % (name, deviceID))
                    verifiedDevs.append(deviceID)
                    self.pluggedInDevices.append(deviceID)
                else:
                    print("Error: unplug failed as the device still recognized.")
                    return False

            else:
                deviceID = found
                if self.confirmPCIUnplugged(deviceID):
                    print("Unplugged PCIE device\n    %s" % deviceID)
                    verifiedDevs.append(deviceID)
                    self.pluggedInPCIDevices.append(deviceID)
                else:
                    print("Error: unplug failed as the device still recognized.")
                    return False


            # print out the detailed HW info for reviewers
            pciDiff  = self.contextDiff(pciInfo1, pciInfo2)
            udevDiff = self.contextDiff(udevInfo1, udevInfo2)
            print("\n========== Detailed HW info ==========   \
                   \n----------      lspci       ----------   \
                   \n%s                                       \
                   \n----------      udev        ----------   \
                   \n%s                                       \
                   \n======================================\n"\
                     % (pciDiff, udevDiff))
               
            i -= 1
            
        print("ExpressCard test PASSED")
        return True

if __name__ == "__main__":
    test = ExpressCardTest()
    returnValue = test.do(sys.argv)
    sys.exit(returnValue)
