#!/usr/bin/env python
# Copyright (c) 2020 Red Hat, Inc. All rights reserved. This copyrighted material
# is made available to anyone wishing to use, modify, copy, or
# redistribute it subject to the terms and conditions of the GNU General
# Public License v.2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# Author: Aniket Khisti <akhisti@redhat.com>
#

import time
import sys

from rhcert.tags import TestTag, Constants
from rhcert.command import Command, HwCertCommandException
from rhcert.suites.hwcert.hwcert.fvtest import FvTest

class FvCPUPinning(FvTest):
    def __init__(self):
        FvTest.__init__(self, "fv_cpu_pinning")
        self.req_passthrough = False
        self.interactive = False
        self.guest_name = self.guestName
        self.guest_vcpu_count = 1
        self.guest_vcpu_list = [cpu_no for cpu_no in range(self.guest_vcpu_count)]
        self.guest_vcpu_num = self.guest_vcpu_list[-1]


    def tags(self):
        return [TestTag.noninteractive, TestTag.certification, TestTag.virtualization]


    def plan(self, planner):
        empty_test = []

        # Execute FV test plan command and get tests
        tests = FvTest.plan(self, planner)

        # Check if OS is RHEL8 and above
        if self.redHatRelease.getVersion() < 8:
            if self.getDebug() != Constants.off:
                print("FV CPU Pinning test is not supported for this OS.")
            return empty_test
        return tests


    def getRequiredRPMs(self):
        rpms = FvTest.getRequiredRPMs(self)
        rpms.append("tuna")
        return rpms


    def __vm_is_already_in_state(self, state):
        """
        :param state: running, shutoff
        :return: True if it is in the same state or else False
        """
        try:
            vm_running_cmd = Command("virsh list --state-{} --name".format(state))
            current_state_vms = vm_running_cmd.getStringList(regex=".*", ignoreErrors=True) if vm_running_cmd else []
        except HwCertCommandException:
            return False

        if self.guest_name in current_state_vms:
            print("The guest VM is already in the same state")
            return True
        return False


    def start_vm(self):
        """
        This function starts the stopped guest VM if not already running
        :return: True if the guest VM is running else False
        """
        print("Starting the guest VM...")

        # Check if the guest VM is already in running state
        if self.__vm_is_already_in_state("running"):
            return True

        vm_start_cmd = Command("virsh start {}".format(self.guest_name))
        try:
            vm_start_cmd.run()
        except HwCertCommandException as e:
            print(str(e))
            print("Error: Failed to start the guest VM")
            return False

        # Wait for the guest VM to start
        sys.stdout.flush()
        time.sleep(5)
        return True


    def stop_vm(self):
        """
        This function stops the running guest VM if not already stopped
        :return: True if the guest VM is not running else False
        """
        print("Shutting down the guest VM...")

        # Check if the guest VM is in shutdown state
        if self.__vm_is_already_in_state("shutoff"):
            return True

        stop_vm_cmd = Command("virsh destroy {}".format(self.guest_name))
        try:
            stop_vm_cmd.run()
        except HwCertCommandException as e:
            print(str(e))
            print("Error: Failed to stop the guest VM")
            return False
        return True

    def get_affinity_from_proc(self):
        """
        This function calculates the CPU affinity by mapping running process ids from /proc directory
        :return : dict containing CPU affinity

        # format of expected_affinity = {0: 3}
        """
        print("Getting CPU affinity from /proc directory")
        cpu_affinity = dict()
        try:
            print(" - Getting the VM domain pid and VCPU pid")
            domain_pid = Command("cat /var/run/libvirt/qemu/hwcert*.pid").getInteger()

            vcpu_pid = Command("grep pid /var/run/libvirt/qemu/hwcert*.xml").getInteger(
                       regex='\\s+<vcpu\\sid=\'\\d+\'\\spid=\'(?P<pid>\\d+)', singleLine=False, regexGroup='pid')

            print(" - Getting assigned/allowed host physical CPU list for the particular VCPU of the guest VM")
            output = Command("grep -i cpus_allowed_list /proc/{}/task/{}/status".format(domain_pid, vcpu_pid)
                     ).getInteger(regex='Cpus_allowed_list:\\s+(?P<cpu>\\d+)', singleLine=False, regexGroup='cpu')

            cpu_affinity[self.guest_vcpu_num] = output
        except HwCertCommandException as e:
            print(str(e))
            print("Error: Unable to verify CPU affinity from /proc directory")
            return

        return cpu_affinity


    def get_affinity_from_vcpuinfo(self):
        """
        This function calculates the CPU affinity from virsh vcpuinfo command output
        :return : dict containing CPU affinity

        # virsh vcpuinfo hwcert-x86_64 --pretty
        VCPU:           0                 <<<<<<< VCPU
        CPU:            3
        State:          running
        CPU time:       33.5s
        CPU Affinity:   3 (out of 4)      <<<<<<< CPU Affinity

        # format of expected_affinity = {0: 3}
        """
        print("Getting CPU affinity from 'virsh vcpuinfo' command")
        cpu_affinity = dict()
        try:
            output = Command("virsh vcpuinfo {} --pretty".format(self.guest_name)).getInteger(
                     regex='CPU\\sAffinity:\\s+(?P<cpu>\\d+)', singleLine=False, regexGroup='cpu')

            cpu_affinity[self.guest_vcpu_num] = output
        except HwCertCommandException as e:
            print(str(e))
            return

        return  cpu_affinity

    def get_affinity_from_vcpupin(self):
        """
        This function calculates the CPU affinity from virsh vcpupin command output
        :return : dict containing CPU affinity

        # virsh vcpupin hwcert-x86_64
        VCPU: CPU Affinity
        ----------------------------------
           0: 3              <<<<<<< VCPU : CPU

        format of expected_affinity = {0: 3}
        """
        cpu_affinity = dict()
        try:
            print("Getting CPU affinity from 'virsh vcpupin' command")
            output = Command("virsh vcpupin {}".format(self.guest_name)).getInteger(
                     regex='\\s+\\d+:?\\s+(?P<cpu>\\d+)', singleLine=False, regexGroup='cpu')

            cpu_affinity[self.guest_vcpu_num] = output
        except HwCertCommandException as e:
            print(str(e))
            return

        return cpu_affinity


    def set_current_state_vcpu(self, count):
        """
        This function applies the given number of VCPU to the guest VM
        :param count: integer number of VCPU
        :return: True if it works else False
        """
        print("Changing the count of current state of guest VM VCPU to '{}'".format(count))
        set_current_vcpu = Command("virsh setvcpus {} --count {} --current".format(self.guest_name, count))
        try:
            set_current_vcpu.run()
        except HwCertCommandException as e:
            print(str(e))
            print("Error: Failed to configure the count of current state of the guest VM VCPU")
            return False
        return True


    def set_max_vcpu(self, count):
        """
        This function configures the maximum allowed number of VCPU of the guest VM
        :param count: integer number of VCPU
        :return: True if it works else False
        """
        print("Configuring the maximum allowed guest VM VCPU count to '{}'".format(count))
        set_max_vcpu = Command("virsh setvcpus {} --count {} --maximum --config".format(self.guest_name, count))
        try:
            set_max_vcpu.run()
        except HwCertCommandException as e:
            print(str(e))
            print("Error: Failed to configure the maximum allowed guest VM VCPU count")
            return False
        return True


    def verify_guest_vm_vcpu_pin(self, expected_configured_affinity):
        """
        This function compares the expected configured CPU affinity with
            - actually found CPU affinity from vcpupin command
            - actually found CPU affinity from vcpuinfo command
            - actually found CPU affinity from /proc directory
         :param expected_configured_affinity: dict containing CPU affinity
        :return: True if actual vs expected matches or else False
        """
        actual_affinity_from_vcpupin = self.get_affinity_from_vcpupin()
        actual_affinity_from_vcpuinfo = self.get_affinity_from_vcpuinfo()
        actual_affinity_from_proc = self.get_affinity_from_proc()

        print("\nINFO: Expected VCPU:CPU affinity configured   - {}\n".format(expected_configured_affinity))
        print("INFO: Actual VCPU:CPU affinity from vcpupin   - {}".format(actual_affinity_from_vcpupin))
        print("INFO: Actual VCPU:CPU affinity from vcpuinfo  - {}".format(actual_affinity_from_vcpuinfo))
        print("INFO: Actual VCPU:CPU affinity from /proc dir - {}".format(actual_affinity_from_proc))

        if not (expected_configured_affinity == actual_affinity_from_vcpupin
                == actual_affinity_from_vcpuinfo == actual_affinity_from_proc):
            print("\nError: Inconsistent CPU affinity found in the actual observations compared to the configured\n")
            return False
        print("\nSuccessfully verified the configured vs actually observed CPU affinity\n")
        return True


    def pin_guest_vm_vcpu(self, vcpu, cpu):
        """
        Perform the pinning of guest VM VCPU to host physical CPU
        :param vcpu: guest VM VCPU
        :param cpu: host physical CPU
        :return: dict containing CPU affinity
        """
        print("Pinning guest VM VCPU '{}' to the host physical CPU '{}'".format(vcpu, cpu))
        vcpu_pin_cmd = Command("virsh vcpupin {} {} {}".format(self.guest_name, vcpu, cpu))
        try:
            vcpu_pin_cmd.run()
            time.sleep(1)
        except HwCertCommandException as e:
            print(e)
            print("Error: Unable to pin the guest VM VCPU '{}' to host CPU '{}'".format(vcpu, cpu))
            return False
        expected_affinity = {vcpu: cpu}
        return expected_affinity


    def isolate_cpu(self, cpu):
        """
        This function isolates the host physical CPU from the process scheduler
        :param cpu: host physical CPU to isolate
        :return: True if it works else False
        """
        print("Isolating the host physical CPU '{}' from the OS scheduler".format(cpu))
        isolcpus_cmd = Command("tuna --cpus {} --isolate".format(cpu))
        try:
            isolcpus_cmd.run()
        except HwCertCommandException as e:
            print(str(e))
            print("Error: Failed to isolate the host physical CPU '{}' from the OS scheduler during runtime".format(cpu))
            return False if not self.check_isolcpus_boot_param(cpu) else True

        return True


    def check_isolcpus_boot_param(self, cpu):
        """
        This function checks if isolcpus flag is configured in boot cmdline or not
        :return: True if found or else False
        """
        print("Checking if the cmdline parameter is configured in order to isolate the CPU manually")
        cmdline_cmd = Command("cat /proc/cmdline")
        try:
            cmdline = cmdline_cmd.getString(regex=".*")
        except HwCertCommandException:
            print("Error: Unable to fetch boot command line parameters")
            return False

        if 'isolcpus={}'.format(cpu) in cmdline:
            return True

        print("Error: Unable to find isolcpus={} parameter in /proc/cmdline, check again and rerun the test".format(cpu))
        return False


    def get_cpu_list(self):
        """
        This function calculates the total number of host physical CPU
        :return: list of host physical CPU
        """
        try:
            print("Getting the total number of host physical CPU")
            cpu_count = Command('lscpu').getInteger(regex='CPU\\(s\\):\\s+(?P<cpus>\\d+)',
                                                   singleLine=False, regexGroup="cpus")
            self.host_cpu_list = [cpu_no for cpu_no in range(cpu_count)]
            self.host_cpu_num = self.host_cpu_list[-1]
            self.host_cpu_count = len(self.host_cpu_list)
        except HwCertCommandException as e:
            print(str(e))
            print("Error: Could not calculate the physical CPU in the host system")
        return self.host_cpu_list


    def configure_guest(self):
        """
        This function modifies the guest VM configuration for further progress
        :return: True if it is working else False
        """
        # Check hypervisor configuration
        self.configureHypervisor()

        # Verify guest and guest files
        if not self.verifyGuest():
            return False

        # Wait for few secs as above functions will restart the libvirtd service
        time.sleep(2)

        # stop old running guest VM sessions if any
        if not self.stop_vm():
            return False

        # increase the vcpu limit & apply it to current guest VM state
        if not (self.get_cpu_list()  # calculate the host CPU
                and self.isolate_cpu(self.host_cpu_num)
                and self.set_max_vcpu(self.guest_vcpu_count)
                and self.set_current_state_vcpu(self.guest_vcpu_count)):
            return False

        # start new guest VM fresh session with new configuration
        if not self.start_vm():
            return False

        return True


    def fv_cpu_pinning(self):
        # pin guest VM VCPU to host physical CPU
        expected_affinity = self.pin_guest_vm_vcpu(self.guest_vcpu_num, self.host_cpu_num)
        if not expected_affinity:
            return False

        # Verify the pinning of guest VM VCPU
        if not self.verify_guest_vm_vcpu_pin(expected_affinity): # Check for expected vs actual affinity
            return False

        # restore the guest VM config changes back to default
        if not self.stop_vm():
            return False

        return True


    def run(self):
        print("\nFull Virt CPU Pinning Test")

        # run initial configuration subtest
        if not self.runSubTest(self.configure_guest, "Setup guest VM VCPU",
                               "Configure the maximum and current count of VCPU for the guest VM"):
            return False

        # run fv cpu pin unpin subtest
        if not self.runSubTest(self.fv_cpu_pinning, "Perform FV CPU Pinning",
                               "Configure and verify the pinning of guest VM VCPU to host physical CPU"):
            return False

        return True
