#!/usr/bin/env python
# Copyright (c) 2009 Red Hat, Inc. All rights reserved. This copyrighted material 
# is made available to anyone wishing to use, modify, copy, or 
# redistribute it subject to the terms and conditions of the GNU General 
# Public License v.2.
# 
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# Authors: Greg Nichols 
#          YangKun <ykun@redhat.com>
#          Petersingh Anburaj
#
import sys

from rhcert.suites.hwcert.hwcert.fvtest import FvTest
from rhcert.command import Command, HwCertCommandException


class FvMemoryTest(FvTest):
    
    def __init__(self):
        FvTest.__init__(self, "fv_memory")
        self.testToRun = 'memory'

    def run(self):
        # verify the guest, and check guest files
        if not self.verifyGuest():
            return False

        # Get the page size of the system
        print("Checking the required RAM for the VM ...")
        page_size_cmd = Command("getconf PAGESIZE")
        try:
            page_size = page_size_cmd.getInteger()
        except HwCertCommandException:
            print("Unable to get the PAGESIZE, using default value is 4K")
            page_size = 4096

        # Calculate required RAM by VM
        req_ram = int(page_size / 8)
        print("Required RAM for the VM is {} MB".format(req_ram))

        # Get current RAM allocate to the VM
        cur_ram_cmd = Command("virsh dumpxml {} | grep '<memory'".format(self.guestName))
        try:
            cur_ram_kb = cur_ram_cmd.getInteger(regex='.*\>(?P<ram>\d+)\<', regexGroup="ram")
            cur_ram_mb = int(cur_ram_kb / 1024)
        except HwCertCommandException:
            print("Unable to get current RAM for the VM, using default value as 2560 MB")
            cur_ram_mb = 2560
        print("Current RAM allocated to VM is {} MB".format(cur_ram_mb))

        # Check if VM has the required RAM
        # if not update the VM config
        ram_updated = True
        if cur_ram_mb < req_ram:
            print("Updating the RAM of the VM to {} MB".format(req_ram))
            ram_updated = self.update_vm_ram(req_ram, type="maxmem") and self.update_vm_ram(req_ram, type="mem")

        # Run the FV test if RAM update was successful
        success = False
        if ram_updated:
            success = FvTest.run(self)
        else:
            print("Error: Failed to update the RAM for the VM !!!")

        # Revert the VM RAM config changes
        if cur_ram_mb < req_ram:
            print("Reverting the RAM of the VM to {} MB".format(cur_ram_mb))
            self.update_vm_ram(cur_ram_mb, type="maxmem")
            self.update_vm_ram(cur_ram_mb, type="mem")

        return success

    def update_vm_ram(self, ram, type="maxmem"):
        """
        Set the mem / max mem of the VM to the given value
        :param ram: Amount of RAM in MBs
        :param type: string, can be `maxmem` or `mem`
        :return: True if success else False
        """
        if type == "maxmem":
            cmd = "setmaxmem"
        elif type == "mem":
            cmd = "setmem"
        else:
            print("Error: Wrong type passed !!!")
            return False

        virsh_mem_cmd = Command("virsh {} {} {}M --config".format(cmd, self.guestName, ram))
        try:
            virsh_mem_cmd.run()
        except HwCertCommandException as e:
            print("Command execution failed: {}".format(e.command.command))
            e.command.printErrors()
            return False
        return True


if __name__ == "__main__":
    test = FvMemoryTest()
    returnValue = test.do(sys.argv)
    sys.exit(returnValue)
