#!/usr/bin/env python
# Copyright (c) 2020 Red Hat, Inc. All rights reserved. This copyrighted material
# is made available to anyone wishing to use, modify, copy, or
# redistribute it subject to the terms and conditions of the GNU General
# Public License v.2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# Author: Petersingh Anburaj
#

import time

from rhcert.suites.hwcert.hwcert.fvPassthrough import FvPassthrough
from rhcert.command import Command, HwCertCommandException


class FvPCIePassthrough(FvPassthrough):

    def __init__(self):
        FvPassthrough.__init__(self, "fv_pcie_passthrough")

    def configure_guest(self):
        """
        Subtest to configure the guest VM for PCIe Passthrough
        """
        # Check hypervisor configuration
        self.configureHypervisor()

        # Verify guest and guest files
        if not self.verifyGuest():
            return False

        # Wait for few secs as above functions will restart the libvirtd service
        time.sleep(2)
        # Setup password-less ssh for the guest
        if not self.setup_password_less_ssh_connection():
            return False

        # Get PCI devices
        pci_devices = self.get_passthrough_devices(dev_type="pci")
        if not pci_devices:
            print("Error: unable to get PCI devices !!!")
            return False

        # Allow user to select a PCI device
        msg = "Select the PCI device that you want to assign to the guest Machine"
        pci_device = self.select(msg, pci_devices)

        # Get device ID and device details
        # pci_device = `0000:00:1f.6 Ethernet controller: Intel Corporation Ethernet Connection (6) I219-LM (rev 30)`
        # device_id = `0000:00:1f.6`
        device_id = pci_device.split()[0]
        if not self.update_pci_device_details(device_id=device_id):
            return False

        # Create device XML file
        if not self.create_device_xml(dev_type="pci"):
            return False

        # Attach device to the guest
        if not self.update_vm_device(cmd="attach"):
            return False

        return True

    def verify_passthrough(self):
        """
        Subtest to verify if PCIe Passthrough was successful
        """
        # Saving the host device kernel driver
        print("Fetching the passthrough device driver ...")
        device_driver_cmd = Command("lspci -ks %s:%s:%s.%s" % (self.domain, self.bus, self.slot, self.function))
        try:
            device_driver = device_driver_cmd.getString(regex="\sKernel driver in use: (?P<driver>.*)",
                                                        singleLine=False,
                                                        regexGroup="driver")
        except HwCertCommandException:
            print("Error: Unable to fetch host device driver !!!")
            self.update_vm_device(cmd="detach")
            return False
        print("Device driver : %s" % device_driver)

        # Start the guest VM
        if not self.start_vm_update_ip_address():
            self.stop_vm()
            self.update_vm_device(cmd="detach")
            return False

        # Check if password-less SSH configuration
        if not self.check_vm_ssh_config():
            self.stop_vm()
            self.update_vm_device(cmd="detach")
            return False

        # Log PCI devices of guest VM
        print("\nLogging guest VM PCI devices ...")
        guest_pci_devices_cmd = Command(self.get_ssh_cmd("lspci -D"))
        try:
            guest_pci_devices_cmd.run()
            guest_pci_devices_cmd.printOutput()
        except HwCertCommandException:
            pass

        # Log PCI device tree of guest VM
        print("\nLogging guest VM PCI device tree ...")
        guest_pci_device_tree_cmd = Command(self.get_ssh_cmd("lspci -tv"))
        try:
            guest_pci_device_tree_cmd.run()
            guest_pci_device_tree_cmd.printOutput()
        except HwCertCommandException:
            pass

        # Validate if device passthrough was successful
        print("Validating the device passthrough ...")

        # 1. Check if device is present inside the guest VM
        print("\nChecking if passthrough device is present inside the guest VM ...")
        guest_passthrough_device_cmd = Command(self.get_ssh_cmd("lspci -vd %s:%s" % (self.vendor_id, self.product_id)))
        try:
            guest_passthrough_device = guest_passthrough_device_cmd.getStringList(regex=".*")
            guest_passthrough_device_cmd.printOutput()
            print("Passthrough device was found inside the guest VM !!!")
        except HwCertCommandException:
            print("Error: Passthrough device is not present inside the guest VM !!!")
            self.stop_vm()
            self.update_vm_device(cmd="detach")
            return False

        # 2. Logging the driver initialization logs
        print("\nLogging driver initialization logs ...")
        passthrough_device_id = guest_passthrough_device[0].split()[0]
        driver_initialization_logs_cmd = Command(self.get_ssh_cmd("journalctl -k | grep %s" % passthrough_device_id))
        try:
            driver_initialization_logs_cmd.run()
            driver_initialization_logs_cmd.printOutput()
        except HwCertCommandException:
            # For RHEL6 image use `/var/log/message` as journalctl is not present
            var_log_message = Command(self.get_ssh_cmd("grep '%s' /var/log/messages" % passthrough_device_id))
            try:
                var_log_message.run()
                var_log_message.printOutput()
            except HwCertCommandException:
                pass

        # 3. Compare the host device driver while guest VM is running
        print("\nFetching the passthrough device driver when guest VM is running ...")
        device_driver_cmd = Command("lspci -ks %s:%s:%s.%s" % (self.domain, self.bus, self.slot, self.function))
        try:
            device_driver_with_guest_run = device_driver_cmd.getString(regex="\sKernel driver in use: (?P<driver>.*)",
                                                                       singleLine=False,
                                                                       regexGroup="driver")
        except HwCertCommandException:
            print("Error: Unable to fetch host device driver while guest VM was running !!!")
            self.stop_vm()
            self.update_vm_device(cmd="detach")
            return False
        print("Device driver : %s" % device_driver_with_guest_run)

        print("\nComparing the host device drivers ...")
        print("Host device driver before starting the guest VM: %s" % device_driver)
        print("Host device driver while guest VM was running: %s" % device_driver_with_guest_run)
        if device_driver == device_driver_with_guest_run:
            print("Error: Host driver has not changed while guest VM was running !!!")
            self.stop_vm()
            self.update_vm_device(cmd="detach")
            return False
        print("Host device driver updated successfully while guest VM was running !!!")

        # Stop the guest VM
        self.stop_vm()
        # Detach the PCI device
        self.update_vm_device(cmd="detach")
        return True

    def run(self):
        # run configure guest subtest
        if not self.runSubTest(self.configure_guest, "Guest Passthrough Setup",
                               "Configure the guest VM for PCIe passthrough"):
            return False

        # run verify passthrough subtest
        if not self.runSubTest(self.verify_passthrough, "Verify PCIe Passthrough",
                               "Verfiy PCIe passthrough for the guest VM"):
            return False

        return True
