#!/usr/bin/env python
# Copyright (c) 2020 Red Hat, Inc. All rights reserved. This copyrighted material
# is made available to anyone wishing to use, modify, copy, or
# redistribute it subject to the terms and conditions of the GNU General
# Public License v.2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# Author: Petersingh Anburaj
#

import time

from rhcert.command import Command, HwCertCommandException
from rhcert.suites.hwcert.hwcert.fvPassthrough import FvPassthrough


class FvUSBPassthrough(FvPassthrough):

    def __init__(self):
        FvPassthrough.__init__(self, "fv_usb_passthrough")

    def configure_guest(self):
        """
        Subtest to configure the guest VM for USB Passthrough
        """
        # Check hypervisor configuration
        self.configureHypervisor()

        # Verify guest and guest files
        if not self.verifyGuest():
            return False

        # Wait for few secs as above functions will restart the libvirtd service
        time.sleep(2)
        # Setup password-less ssh for the guest
        if not self.setup_password_less_ssh_connection():
            return False

        # Get USB devices
        usb_devices = self.get_passthrough_devices(dev_type="usb")
        if not usb_devices:
            print("Error: unable to get USB devices !!!")
            return False

        # Allow user to select a USB device
        msg = "Select the USB device that you want to assign to the guest Machine"
        usb_device = self.select(msg, usb_devices)

        # Update device details
        if not self.update_usb_device_details(device_name=usb_device):
            return False

        # Create device XML file
        if not self.create_device_xml(dev_type="usb"):
            return False

        # Attach device to the guest
        if not self.update_vm_device(cmd="attach"):
            return False

        return True

    def verify_passthrough(self):
        """
        Subtest to verify if USB Passthrough was successful
        """
        # Saving the host device driver
        print("Fetching the passthrough device driver ...")
        device_driver = self.get_usb_device_driver()
        if not device_driver:
            print("Error: Unable to fetch host device driver !!!")
            self.update_vm_device(cmd="detach")
            return False
        print("Device driver : %s" % device_driver)

        # Start the guest VM
        if not self.start_vm_update_ip_address():
            self.stop_vm()
            self.update_vm_device(cmd="detach")
            return False

        # Check if password-less SSH configuration
        if not self.check_vm_ssh_config():
            self.stop_vm()
            self.update_vm_device(cmd="detach")
            return False

        # Log USB devices of guest VM
        print("\nLogging guest VM USB devices ...")
        guest_usb_devices_cmd = Command(self.get_ssh_cmd("lsusb"))
        try:
            guest_usb_devices_cmd.run()
            guest_usb_devices_cmd.printOutput()
        except HwCertCommandException:
            pass

        # Log USB device tree of guest VM
        print("\nLogging guest VM USB device tree ...")
        guest_usb_device_tree_output = []
        guest_usb_device_tree_cmd = Command(self.get_ssh_cmd("lsusb -t"))
        try:
            guest_usb_device_tree_output = guest_usb_device_tree_cmd.getStringList(regex=".*")
            guest_usb_device_tree_cmd.printOutput()
        except HwCertCommandException:
            pass

        # Validate if device passthrough was successful
        print("\nValidating the device passthrough ...")

        # 1. Check if device is present inside the guest VM
        print("Checking if passthrough device is present inside the guest VM ...")
        guest_passthrough_device = None
        guest_passthrough_device_cmd = Command(self.get_ssh_cmd("lsusb -d %s:%s" % (self.vendor_id, self.product_id)))
        try:
            guest_passthrough_device = guest_passthrough_device_cmd.getString(regex=".*")
            guest_passthrough_device_cmd.printOutput()
            print("Passthrough device was found inside the guest VM !!!")
        except HwCertCommandException:
            print("Error: Passthrough device is not present inside the guest VM !!!")
            self.stop_vm()
            self.update_vm_device(cmd="detach")
            return False

        # 2. Logging the driver initialization logs
        print("\nFetching passthrough device ID ...")
        driver, passthrough_device_id = self.parse_usb_device_tree(output=guest_usb_device_tree_output,
                                                                   bus=int(guest_passthrough_device.split()[1]),
                                                                   device=int(guest_passthrough_device.split()[3][:-1]))
        if passthrough_device_id:
            print("Logging driver initialization logs ...")
            driver_initialization_logs_cmd = Command(self.get_ssh_cmd("journalctl -k | grep %s" %
                                                                      passthrough_device_id))
            try:
                driver_initialization_logs_cmd.run()
                driver_initialization_logs_cmd.printOutput()
            except HwCertCommandException:
                # For RHEL6 image use `/var/log/message` as journalctl is not present
                var_log_message = Command(self.get_ssh_cmd("grep '%s' /var/log/messages" % passthrough_device_id))
                try:
                    var_log_message.run()
                    var_log_message.printOutput()
                except HwCertCommandException:
                    pass
        else:
            print("Warning: Unable to fetch passthrough device ID. Skipping driver initialization logs !!!")

        # 3. Compare the host device driver while guest VM is running
        print("\nFetching the passthrough device driver when guest VM is running ...")
        device_driver_with_guest_run = self.get_usb_device_driver()
        if not device_driver_with_guest_run:
            print("Error: Unable to fetch host device driver while guest VM was running !!!")
            self.stop_vm()
            self.update_vm_device(cmd="detach")
            return False
        print("Device driver : %s" % device_driver_with_guest_run)

        print("\nComparing the host device drivers ...")
        print("Host device driver before starting the guest VM: %s" % device_driver)
        print("Host device driver while guest VM was running: %s" % device_driver_with_guest_run)
        if device_driver == device_driver_with_guest_run:
            print("Error: Host driver has not changed while guest VM was running !!!")
            self.stop_vm()
            self.update_vm_device(cmd="detach")
            return False
        print("Host device driver updated successfully while guest VM was running !!!")

        # Stop the guest VM
        self.stop_vm()
        # Detach the USB device
        self.update_vm_device(cmd="detach")
        return True

    def run(self):
        # run configure guest subtest
        if not self.runSubTest(self.configure_guest, "Guest Passthrough Setup",
                               "Configure the guest VM for USB passthrough"):
            return False

        # run verify passthrough subtest
        if not self.runSubTest(self.verify_passthrough, "Verify USB Passthrough",
                               "Verfiy USB passthrough for the guest VM"):
            return False

        return True
