#!/usr/bin/env python
# Copyright (c) 2006 Red Hat, Inc. All rights reserved. This copyrighted material
# is made available to anyone wishing to use, modify, copy, or
# redistribute it subject to the terms and conditions of the GNU General
# Public License v.2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# Author: Ben Rose
# Adapted from code by Daniel Bristot
#
import sys, time, os, shutil, random, math
from os.path import isfile

from rhcert.tags import Constants, DeviceClass, Tags, TestTag, RHEL, Attributes
from rhcert.command import Command, HwCertCommandException
from rhcert.environment import Environment
from rhcert.continuation import Continuation
from rhcert.redhatrelease import RedHatRelease

from rhcert.suites.hwcert.hwcert.procDeviceDetector import ProcDeviceDetector
from rhcert.suites.hwcert.hwcert.hardwareTest import HardwareTest
from rhcert.suites.rhcert.platform import CheckHardwareHealth
from xml.dom.minidom import parseString

class FvRealtimeTest(HardwareTest):

    def __init__(self):
        HardwareTest.__init__(self, "fv_realtime")
        self.deviceClass = DeviceClass.processor
        if os.uname()[3].find("RT ") == -1:
            self.realtime = False
        else:
            self.realtime = True
        self.environment = Environment()
        self.continuation = Continuation(self.environment)
        self.rebootTimeLimit = self.environment.getRebootTimelimit()
        self.proxyURLs = self.environment.getProxyURLs()
        self.waitTime = 5*60 #for reboot
        self.redHatRelease = RedHatRelease()

        self.tolerableLatency = 40
        self.guestTestDir = "/opt/kvm-rt"
        self.guestTestConf = "test.conf"
        self.libvirtImgDir = "/var/lib/libvirt/images/"
        self.testDuration = 12*60 #minutes
        self.baseKSFile = "/usr/share/kvm-rt-cert/kvm-rt-guest.ks"
        self.baseXmlTemplate = "base.xml"
        self.baseGuestName = "base"
        self.cloneGuestName = "clone"
        self.baseGuestIsolcpu = 1
        self.baseDiskSize = 10
        self.qcow2Image = "rhel-kvm-rt-image.qcow2"
        self.libvirtImgPath = "%s/%s" % (self.libvirtImgDir, self.qcow2Image)
        self.hugePagesUnit = ""
        self.hugePagesSize = 0
        self.guestConfigDirectory = self.environment.getKVMGuestConfigDirectory()
        self.guestConfigPath = "%s/%s.xml" % (self.guestConfigDirectory, self.baseGuestName)
        # kvm-rt team unsure how to pick, so went w/ 2 here
        self.baseVcpuNum = 2
        gigPerCpu = 1
        if RHEL.major == "RHEL7":
            gigPerCpu = 1
        elif RHEL.major == "RHEL8":
            gigPerCpu = 1.5
        self.baseMemorySize = 1024 * gigPerCpu * self.baseVcpuNum
        self.baseMemorySize = int(math.ceil(self.baseMemorySize))

        self.minimum_cpus = 3

    def tags(self):
        return [TestTag.noninteractive, TestTag.certification, TestTag.realtime]

    def getRequiredRPMs(self):
        return ["kernel-rt-kvm", "rt-tests", "libvirt", "virt-install", "virt-viewer", "tuned-profiles-realtime",
                "tuned", "tuned-profiles-nfv", "libguestfs-tools-c", "numactl", "bc", "yum-utils", "wget"]

    def __listToRange(self, numbers):
        """ Changes list of number strings to list of ranges as one string
            "2", "4", "5", "6", "7" => "2,4-7"
        """
        numbers = [int(x) for x in numbers]
        numbers.sort()
        new_list = []
        lower = numbers[0]
        current_compare = lower
        dash = False
        for number in numbers[1:]:
            if number - 1 == current_compare:
                dash = True
                current_compare = number
            elif dash:
                new_list.append("%s-%s" % (lower, current_compare))
                dash = False
                current_compare = number
                lower = number
            else:
                new_list.append(str(lower))
                current_compare = number
                lower = number
                dash = False
        #last number or range
        if dash:
            new_list.append("%s-%s" % (lower, number))
        else:
            new_list.append(str(number))
        return ",".join(new_list)

    def __setVmDisk(self):
        """ Place image file """
        # Close in virsh in case it's already running
        Command("virsh destroy %s" % self.baseGuestName).echoIgnoreErrors()
        # Undefine in case it already exists
        Command("virsh undefine %s" % self.baseGuestName).echoIgnoreErrors()
        # delete the file in case it's already in directory
        Command("rm %s/%s" % (self.libvirtImgDir, self.qcow2Image)).echoIgnoreErrors()
        # copy qcow2 to the directory from temp download location
        Command("cp /root/temp/%s %s" % (self.qcow2Image, self.libvirtImgDir)).echoIgnoreErrors()

    def __updateVmXml(self):
        """ Update xml to be optimized """
        # copy base xml to new location
        try:
            Command("pwd").echo()
            Command("cp %s %s" % (self.baseXmlTemplate, self.guestConfigPath)).echo()
        except HwCertCommandException:
            print("Error: Could not copy base xml template")
            return False
        # apply changes using values from host setup... TODO - refactor those calculations out into its own function that updates modular variables
        configAsString = ""
        with open(self.guestConfigPath, 'r') as f:
            configAsString = f.read()

        configAsString = configAsString.replace("DOMAIN_NAME", self.baseGuestName)
        configAsString = configAsString.replace("DOMAIN_NODE", self.baseNode)
        configAsString = configAsString.replace("DOMAIN_HOUSEKEEPING_LIST", str(self.baseHousekeepingCores[0]))
        configAsString = configAsString.replace("DOMAIN_MEMORY", str(self.baseMemorySize))
        configAsString = configAsString.replace("HW_HUGE_PAGES_SIZE", str(self.hugePagesSize))
        configAsString = configAsString.replace("HW_HUGE_PAGES_UNIT", self.hugePagesUnit)
        configAsString = configAsString.replace("DOMAIN_VCPUS_NR", str(self.baseVcpuNum))
        if self.baseVcpuNum == 1:
            configAsString = configAsString.replace("DOMAIN_VCPUS_RANGE", "0")
        else:
            configAsString = configAsString.replace("DOMAIN_VCPUS_RANGE", "0-%s" % (self.baseVcpuNum - 1))
        ###### From clone
        import uuid
        self.uuidgen = str(uuid.uuid1())
        configAsString = configAsString.replace("DOMAIN_UUID", self.uuidgen)
        configAsString = configAsString.replace("DOMAIN_DISK_PATH", self.libvirtImgPath)
        configAsString = configAsString.replace("DOMAIN_MAC_ADDR", "BA:BA:DA:16:%02x:%02x\n" % (random.randint(0, 255), random.randint(0, 255)))
        configAsString = configAsString.replace("DEFAULT_MACHINE", "pc-i440fx-rhel7.0.0")

        with open(self.guestConfigPath, 'w') as f:
            f.write(configAsString)


    def __defineVm(self):
        try:
            Command("virsh define %s" % self.guestConfigPath).echo()
        except HwCertCommandException:
            print("Error: Could not define VM via configuration file")
            return False

    def __copyScriptsToVm(self):
        """ Copy scripts for running the test on guest into VM """
        try:
            Command("guestfish --rw -d %s -i mkdir /root/bin/" % self.baseGuestName).echoIgnoreErrors()
            Command("guestfish --rw -d %s -i copy-in ./krcert-guest-service.sh /root/bin/" % self.baseGuestName).echo()
        except HwCertCommandException:
            print("Error: could not copy scripts to VM")
            return False

    def __copyServiceToVm(self):
        """ Copy service for running the test on guest into VM """
        with open("/tmp/kvm-rt-guest.service", "w") as f:
            f.write("[Unit]\n")
            f.write("Description=Automated cyclictest run\n")
            f.write("After=tuned.service\n\n")
            f.write("[Service]\n")
            f.write("ExecStart=/root/bin/krcert-guest-service.sh\n")
            f.write("TimeoutSec=0\n")
            f.write("RemainAfterExit=yes\n")
            f.write("WorkingDirectory=/root/bin\n")
            f.write("StandardOutput=tty\n")
            f.write("StandardError=tty\n")
            f.write("\n")
            f.write("[Install]\n")
            f.write("WantedBy=multi-user.target\n")

        try:
            #Command("virt-copy-in %s /tmp/kvm-rt-guest.service /usr/lib/systemd/system/" % self.baseGuestName).echo()
            Command("guestfish --rw -d %s -i copy-in /tmp/kvm-rt-guest.service /usr/lib/systemd/system/" % self.baseGuestName).echo()
            Command("rm /tmp/kvm-rt-guest.service").echo()
        except HwCertCommandException:
            print("Error: could not copy kvm-rt-guest.service into VM")


    def __allocateHugePages(self):
        """ Update huge pages file """
        hugepageKB = 0
        if "G" in self.hugePagesUnit:
            hugepageKB = self.hugePagesSize * 1024 * 1024
        elif "M" in self.hugePagesUnit:
            hugepageKB = self.hugePagesSize * 1024
        else:
            print("Error: unknown pages size unit: %s" % self.hugePagesUnit)
            return False
        self.hugePagesFile = "/sys/devices/system/node/node%s/hugepages/hugepages-%skB/nr_hugepages" % (self.baseNode, hugepageKB)
        if not os.path.isfile(self.hugePagesFile):
            print("Error: huge page file does not exist")
            print("Huge page is either not enabled or the config is wrong")
            return False
        self.hugePagesNeeded = int(self.baseMemorySize) * 1024 / int(hugepageKB)
        self.hugePagesCurrent = Command("cat %s" % self.hugePagesFile).getString()
        self.hugePagesCurrent = int(self.hugePagesCurrent)

        print("Attempting to allocate huge pages")
        size = int(self.hugePagesCurrent + self.hugePagesNeeded)
        print("value: %s" % size)
        try:
            Command("echo %s >> %s" % (size, self.hugePagesFile)).echo()
            self.hugePagesCurrent = Command("cat %s" % self.hugePagesFile).getString()
            self.hugePagesCurrent = int(self.hugePagesCurrent)
        except:
            print("Error: Could not write to %s" % self.hugePagesFile)
            print("Possibly we don't have enough memory")
            return False

    def __addStartServiceToRcLocal(self):
        shutil.rmtree("/tmp/rc.local", ignore_errors=True)
        with open("/tmp/rc.local", "w") as f:
            f.write("#!/bin/sh\n\n")
            f.write("chmod +x /root/bin/krcert-guest-service.sh\n\n")
            f.write("chmod +x /root/bin/tune_vm.sh\n\n")
            # allow pinning to happen
            f.write("sleep 30\n\n")
            f.write("systemctl enable kvm-rt-guest\n\n")
            f.write("sleep 20\n\n")
            f.write("chmod -x /etc/rc.local\n\n")
            #f.write("shutdown -h now\n\n")
        try:
            Command("chmod +x /tmp/rc.local").echo()
            Command("guestfish --rw -d %s -i chmod 777 /etc/rc.d/rc.local" % self.baseGuestName).echoIgnoreErrors()
            Command("virt-copy-in -d %s /tmp/rc.local /etc/rc.d/" % self.baseGuestName).echo()
            Command("virt-copy-in -d %s /tmp/rc.local /etc/" % self.baseGuestName).echo()
        except:
            print("Error: Issue updating rc.local to enable kvm-rt-guest service on %s" % self.baseGuestName)
            return False
        

    def __onlineGuestSetup(self):
        """ Enables services and installs required packages on guest """
        shutil.rmtree("/tmp/rc.local", ignore_errors=True)
        with open("/tmp/realtime-virtual-guest-variables.conf", "w") as f:
            f.write("isolated_cores=1")

        with open("/tmp/rc.local", "w") as f:
            f.write("#!/bin/sh\n\n")
            f.write("echo 'isolated_cores=1' > /etc/tuned/realtime-virtual-guest-variables.conf\n\n")
            f.write("ls /root/stress*.x86_64.rpm | sort | tail -n1 | xargs -I {} rpm -ihv {}\n\n")
            #f.write("tuna -c 1\n\n")
            f.write("tuned-adm profile realtime-virtual-guest\n\n")
            f.write("sleep 30\n\n")
            f.write("echo 'shutdown...'\n\n")
            f.write("shutdown -h now")

        try:
            Command("yumdownloader --destdir=/tmp stress").echoIgnoreWarnings()
            Command("virt-copy-in -d %s /tmp/stress*.rpm /root/" % self.baseGuestName).echo()
            Command("virt-copy-in -d %s /tmp/realtime-virtual-guest-variables.conf /etc/tuned/" % self.baseGuestName).echo()
            Command("chmod +x /tmp/rc.local").echo()
            Command("guestfish --rw -d %s -i chmod 777 /etc/rc.d/rc.local" % self.baseGuestName).echoIgnoreErrors()
            Command("virt-copy-in -d %s /tmp/rc.local /etc/rc.d/" % self.baseGuestName).echo()
            Command("virt-copy-in -d %s /tmp/rc.local /etc/" % self.baseGuestName).echo()
            Command("virsh start %s" % self.baseGuestName).echo()
        except HwCertCommandException as e:
            print("Error: Starting virsh session for %s" % self.baseGuestName)
            print(str(e))
            return False
        time.sleep(20)
        Command("virsh shutdown %s" % self.baseGuestName).echoIgnoreErrors()
        time.sleep(20)
        Command("virsh destroy %s" % self.baseGuestName).echoIgnoreErrors()

    def __restoreHugePagesFile(self):
        """ Restore Huge Pages File from backup """
        print("Attempting to restore huge pages")
        self.hugePagesCurrent = Command("cat %s" % self.hugePagesFile).getString()
        self.hugePagesCurrent = int(self.hugePagesCurrent)
        size = max(int(self.hugePagesCurrent - self.hugePagesNeeded), 0)
        print("value: %s" % size)
        try:
            Command("echo %s >> %s" % (size, self.hugePagesFile)).echo()
        except:
            print("Error: Could not write to %s" % self.hugePagesFile)
            return False

    def __backupHugePagesFile(self):
        """ Create backup of hugepages file to set back when test completes """

    def __presentInVirsh(self):
        """ Verify that the domain is present in virsh list """
        outputList = ""
        try:
            output = Command("virsh list --all").getString(regex=self.baseGuestName, ignoreErrors=True, singleLine=False)
        except HwCertCommandException:
            print("Error verifying domain is present in virsh list")
            return False
        print("%s found in virsh list, VM created" % self.baseGuestName)

    def __configureGuest(self):

        # verify qcow2 image

        # Set base vm disk
        self.__setVmDisk()
        # Set optimized xml
        self.__updateVmXml()
        # virsh define vm
        self.__defineVm()

        # copy scripts into vm
        self.__copyScriptsToVm()
        # copy service into vm
        self.__copyServiceToVm()

        # set root passwd for easier debugging, etc
        Command("virt-customize -a %s --root-password password:redhat" % self.libvirtImgPath).echoIgnoreErrors()

        # maybe this is innacurate# backup hugepages file
        ###self.__backupHugePagesFile()
        # allocate hugepages
        self.__allocateHugePages()
        # guest enable service and install stress
        self.__onlineGuestSetup()
        # restore hugepages file
        self.__restoreHugePagesFile()
        # verify it's present in output of virsh
        return self.__presentInVirsh()


    def __tuneGuest(self):
        """ Perform tuning on the guest VM """

        self.__allocateHugePages()
        print("Copying tune config into VM..")
        try:
            Command("guestfish --rw -d %s -i mkdir /etc/kvm-rt-cert/" % self.baseGuestName).echoIgnoreErrors()
            shutil.copy("./guest_tuning.conf", "/tmp/guest.conf")
            Command("guestfish --rw -d %s -i copy-in /tmp/guest.conf /etc/kvm-rt-cert/" % self.baseGuestName).echo()
        except HwCertCommandException:
            print("Error: could not copy tune config into VM")
            return False

        print("Copying tune script into VM..")
        try:
            Command("chmod +x ./tune_vm.sh").echo()
            Command("guestfish --rw -d %s -i copy-in ./tune_vm.sh /usr/bin" % self.baseGuestName).echo()
        except HwCertCommandException:
            print("Error: could not copy tune script into VM")
            return False

        try:
            Command("virsh start %s" % self.baseGuestName).echo()
        except HwCertCommandException as e:
            print("Error: Starting virsh session for %s" % self.baseGuestName)
            print(str(e))
            return False
        time.sleep(60)
        Command("virsh shutdown %s" % self.baseGuestName).echoIgnoreErrors()
        time.sleep(20)
        Command("virsh destroy %s" % self.baseGuestName).echoIgnoreErrors()

        self.__restoreHugePagesFile()



    def __pinned(self):
        vcpuinfoList = Command("virsh vcpuinfo %s" % self.baseGuestName).getStringList(regex="^.*")
        for line in vcpuinfoList:
            if 'yyyy' in line:
                return False
        return True

    def __testScriptRun(self):
        """ Run test script on guest VM """

        self.__allocateHugePages()
        print("Copying cyclictest script into VM..")
        try:
            Command("chmod +x ./run-cyclictest.sh").echo()
            Command("guestfish --rw -d %s -i copy-in ./run-cyclictest.sh /usr/bin" % self.baseGuestName).echo()
        except HwCertCommandException:
            print("Error: could not copy cyclictest script into VM")
            return False
        print("Copying cyclictest run config into VM..")
        try:
            shutil.copy("guest.conf", "/tmp/guest.conf")
            with open("/tmp/guest.conf", "r") as f:
                fileContents = f.read()
            durationString = "%sm" % self.testDuration
            fileContents = fileContents.replace("DURATION", durationString)
            with open("/tmp/guest.conf", "w") as f:
                f.write(fileContents)
            Command("guestfish --rw -d %s -i mkdir /etc/kvm-rt-cert/" % self.baseGuestName).echoIgnoreErrors()
            Command("guestfish --rw -d %s -i copy-in /tmp/guest.conf /etc/kvm-rt-cert/" % self.baseGuestName).echo()
        except HwCertCommandException:
            print("Error: could not cyclic test run config into VM")
            return False

        try:
            print("shutting down VM before starting back up again")
            Command("virsh shutdown %s" % self.baseGuestName).echoIgnoreErrors()
            time.sleep(10)
            print("starting VM for service enablement..")
            Command("virsh start %s" % self.baseGuestName).echo()
            time.sleep(30)

            Command("virsh shutdown %s" % self.baseGuestName).echoIgnoreErrors()
            time.sleep(20)
            Command("virsh destroy %s" % self.baseGuestName).echoIgnoreErrors()

            print("starting VM second time after setting enablement")
            Command("virsh start %s" % self.baseGuestName).echo()
            #Command("virsh console %s" % self.baseGuestName).echo()
            #time.sleep((self.testDuration+1) * 60) # m -> s
            time.sleep(120)

            Command("virsh shutdown %s" % self.baseGuestName).echoIgnoreErrors()
            time.sleep(20)
            Command("virsh destroy %s" % self.baseGuestName).echoIgnoreErrors()

            print("starting VM for running tests..")
            Command("virsh start %s" % self.baseGuestName).echo()
            # need to wait long enough for VM to start...
            attempts = 30
            wait = 1
            while not self.__pinned():
                Command("virsh vcpupin %s %s %s" % (self.baseGuestName, 0, self.isolated[-2])).echo()
                Command("virsh vcpupin %s %s %s" % (self.baseGuestName, 1, self.isolated[-1])).echo()
                time.sleep(wait)
                attempts = attempts - 1
                if not attempts:
                    print("Error: Was not able to pin guest VM")
                    return False
            Command("virsh vcpuinfo base").echo()
        except HwCertCommandException as e:
            print("Error: Starting virsh session for %s" % self.baseGuestName)
            print(str(e))
            return False


    def __copyOutResults(self):
        """ Remove any previously copied out files and then copy results out of guest VM """
        attempts = 5
        count = 0
        wait_seconds = 20
        print("Copying test run results out of VM..")
        while 1:
            if count < attempts:
                try:
                    Command("guestfish --ro -d %s -i copy-out /root/results ." % self.baseGuestName).echo()
                    break
                except HwCertCommandException as exception:
                    if count >= attempts:
                        print("Error: could not copy results out of VM")
                        print("Exception given: %s" % exception)
                        return False
                    else:
                        print("trying again to get results files from guest image")
                        time.sleep(wait_seconds)
                        count = count + 1
        return True

    def __evaluateResults(self):
        """ Evaluate results of file copied out from guest VM """
        try:
            maxLatency = str(Command("cat ./results/cyclictest.out").getString(regex=".*Max Latencies: (?P<maxLatency>[0-9]*)", singleLine=False, regexGroup="maxLatency"))
        except HwCertCommandException:
            print("Error: Gathering maximum detected latency")
            return False
        maxLatency = int(maxLatency)
        print("maximum latency: %s" % maxLatency)
        print("tolerable latency: %s" % self.tolerableLatency)
        if self.tolerableLatency >= maxLatency:
            return True
        else:
            print("Error: Maximum detected latency (%s) is higher than the tolerable latency (%s)" % (maxLatency, self.tolerableLatency))
            return False

    def __runTestOnGuest(self):
        """ perform test on VM """
        # enable script
        self.__addStartServiceToRcLocal()
        # tune guest - this is now done in onlineconfigure
        self.__tuneGuest()
        # run test on guest
        self.__testScriptRun()
        sys.stdout.flush()
        # check the guest's status in every minute, this test should not consume more than 1 hour
        count = 0
        while 1:
            if count == self.testDuration+30:
                try:
                    print("time out: destroying the guest.")
                    Command("virsh destroy %s" % self.baseGuestName).echo()
                except HwCertCommandException as exception:
                    print("Warning: %s" % exception)
                break
            else:
                time.sleep(60)
                try:
                    guestIsRunning = Command("virsh list").getString(regex=self.baseGuestName, singleLine=False)
                    print("Guest is still running")
                    count = count + 1
                except HwCertCommandException:
                    print("Guest has shutdown")
                    break
                except Exception as e:
                    print("Warning, virsh list failed.")
                    print(str(e))

        #give even more time for VM to settle
        #time.sleep(5*60)

        self.__restoreHugePagesFile()

        time.sleep(45)
        # copy output out of guest and evaluate results
        return self.__copyOutResults() and self.__evaluateResults()

    def __configureHost(self):
        Command("echo 'got here, starting to configure host!!!' > /root/quickout.txt").echoIgnoreErrors()
        try:
            nr_cpus = Command("lscpu | grep -E '^CPU\(s\):' | awk '{print $2}'").getString()
        except HwCertCommandException:
            print("Error: Could not determine cpus")
            return False
        # Check number of cpus
        if int(nr_cpus) < 3:
            print("Error: The system only has %s cpus, but a minimum of %s are required" % (nr_cpus, self.minimum_cpus))
            return False
        # Calculate number of numa nodes
        try:
            num_numa_nodes = Command("ls -d /sys/devices/system/node/node[0-9]* | wc -l 2>/dev/null").getString()
        except HwCertCommandException:
            print("Error: Could not calculate number of numa nodes")
            return False
        # Calculate total number of (online) cpus in the system
        try:
            num_total_cpus = Command("ls -d /sys/devices/system/cpu/cpu[0-9]* | wc -l 2>/dev/null").getString()
        except HwCertCommandException:
            print("Error: Could not calculate number of online cpus in the system")
            return False
        # Capture cpu flags
        try:
            flags = Command("grep flags /proc/cpuinfo").getStringList(regex="flags.*: (?P<flags>.*)", regexGroup="flags")[0]
        except HwCertCommandException:
            print("Error: Could not retrieve cpu flags")
            return False
        # Huge pages booleans
        has_2mb_huge_pages = "pse36" in flags
        if has_2mb_huge_pages:
            self.hugePagesUnit = "M"
            self.hugePagesSize = 2
        has_1gb_huge_pages = "pdpe1gb" in flags
        if has_1gb_huge_pages:
            self.hugePagesUnit = "G"
            self.hugePagesSize = 1

        supports_huge_pages = has_1gb_huge_pages or has_2mb_huge_pages
        # Dict for cpus for each numa node
        node_cpus = Command("numactl -H").getStringDict(regex="node (?P<node>[0-9]*) cpus: (?P<cpus>.*)", keyGroup="node", valueGroup="cpus")
        for key in node_cpus.keys():
            node_cpus[key] = node_cpus[key].split(" ")
        # base node is last in the dict
        base_node = list(node_cpus.keys())[-1]
        self.baseNode = base_node
        # Determine base rt cores
        base_rt_cores = [node_cpus[base_node][-2], node_cpus[base_node][-1]]
        # Determine base hosekeeping core
        base_housekeeping_cores = [node_cpus[base_node][-3]]
        self.baseHousekeepingCores = base_housekeeping_cores
        # Determine isolated cores
        isolated = base_rt_cores + base_housekeeping_cores
        self.isolated = isolated
        # Generate isolated cores string (examples: "2,4-7" or "2-23")
        isolated.sort()
        self.isolated_range = self.__listToRange(isolated)
        # Edit isolated cores file
        isolated_core_file_lines = Command("cat /etc/tuned/realtime-virtual-host-variables.conf").getStringList(regex=".*")
        desired_line = "isolated_cores=%s" % self.isolated_range
        if not desired_line in isolated_core_file_lines:
            Command("echo %s >> /etc/tuned/realtime-virtual-host-variables.conf" % desired_line).echo()
        # Set hugepages in grub if not already set
        cmdline_linux = Command("cat /etc/default/grub").getString(regex="GRUB_CMDLINE_LINUX.*", singleLine=False)
        if not "default_hugepagesz" in cmdline_linux and supports_huge_pages:
            size = 0
            unit = 'b'
            if has_1gb_huge_pages:
                size = 1
                unit = 'G'
            elif has_2mb_huge_pages:
                size = 2
                unit = 'M'
            Command('sed -i "/^GRUB_CMDLINE_LINUX=/ s/.$/ default_hugepagesz=%s%s\\"/" /etc/default/grub' % (size, unit)).echo()

        # Generate grub config file
        Command("grub2-mkconfig -o /boot/grub2/grub.cfg 2>/dev/null").echoIgnoreErrors()

        # Isolate cpus
        try:
            Command("tuna -c %s --isolate" % self.isolated_range).echo()
        except HwCertCommandException:
            print("Error: tuna isolation command failed")

        # Ensure tuned is active
        tuned_status = Command("systemctl is-active tuned").getString(regex="", ignoreErrors=True)
        if tuned_status != 'active':
            wait_seconds = 20
            iterations = 5 # try 5 times
            while 1:
                if iterations < 1:
                    print("Error: Could not run tuned on the host")
                    break
                try:
                    Command("systemctl start tuned").echo()
                    break
                except HwCertCommandException:
                    iterations = iterations - 1
                    print("Encountered an error starting tuned on the host.  Attempting {} more times.".format(iterations))
                    time.sleep(wait_seconds)

            # Recheck 'Tuned' after 5 re-attempts to start
            tuned_status = Command("systemctl is-active tuned").getString(regex="", ignoreErrors=True)
            if tuned_status != 'active':
                return False

        # Activate host profile
        try:
            Command("/sbin/tuned-adm active | grep realtime-virtual-host").getString()
        except HwCertCommandException:
            print("Host profile not active.  Activating profile...")
            try:
                Command("/sbin/tuned-adm profile realtime-virtual-host").echo()
                Command("/sbin/tuned-adm active").echo()
            except HwCertCommandException:
                pass

            wait_seconds = 30
            iterations = 20*60/30 # 20 minutes (BZ1259039)
            while 1:
                if iterations < 1:
                    print("Error: Could not activate host profile")
                    break
                try:
                    Command("/sbin/tuned-adm active | grep realtime-virtual-host").getString()
                    break
                except HwCertCommandException:
                    iterations = iterations - 1
                    time.sleep(wait_seconds)
        # Verify host changes
        try:
            Command("/sbin/tuned-adm active | grep realtime-virtual-host").getString()
        except HwCertCommandException:
            print("Error: realtime-virtual-host profile is not active")
            sys.stdout.flush()
            return False

        Command("echo 'all done configuring host!!!' > /root/quickout.txt").echoIgnoreErrors()
        sys.stdout.flush()
        return True

    def __isHVMCapableViaLibvirt(self):
        """ determine if system is FV/hvm capable, via virsh from libvirt """
        virsh = Command("virsh capabilities")
        try:
            virsh.run()
            parser = parseString("".join(virsh.output))
            elements = parser.getElementsByTagName("guest")
            for element in elements:
                osType = element.getElementsByTagName(Tags.os_type)[0].childNodes[0]
                if osType.data == "hvm":
                    return True
        except Exception as e:
            if self.getDebug() != Constants.off:
                print("Note: could not parse virsh capabilities")
                print(e)

        #otherwise
        return False

    def __isHVMInCapabilitiesFile(self):
        """ try /sys/../capabilities, looking for FV support"""
        try:
            capabilities = '/sys/hypervisor/properties/capabilities'
            capabilityFile = open(capabilities)
            self.hypervisorCapabilities = capabilityFile.readline()
            capabilityFile.close()
            if  Constants.hvm in self.hypervisorCapabilities:
                return True
        except IOError:
            if self.getDebug() != Constants.off:
                print("Note: could not read " + capabilities)

        #otherwise
        return False

    def __hasVirtualizationCPUFlags(self, verbose=False):
        """ check cpuinfo for virtualization flags """
        cpuflags = Command("fgrep flags /proc/cpuinfo")
        try:
            flags = cpuflags.getString(regex="(vmx|smx|svm)", singleLine=False)
            if verbose or self.getDebug() != Constants.off:
                print("/proc/cpuinfo has %s flag" % flags)
            return True
        except HwCertCommandException:
            if verbose or self.getDebug() != Constants.off:
                print("Note: /proc/cpuinfo does not have vmx, smx, or svm flags")

        return False

    def is_kvm_enabled(self):
        """
        This method identifies whether kvm is enabled
        in bios.
        Returns:
        bool: True if kvm is enabled else False.
        """
        print("INFO: Checking KVM")
        kvm_command = Command("lsmod | grep kvm")
        try:
            kvm = kvm_command.getStringList(regex="^.*")
            kvm_command_string = ' '.join(kvm)
            if 'kvm' in kvm_command_string and 'kvm_intel' in kvm_command_string:
                return True
            print("Error: KVM is disabled by bios")
        except HwCertCommandException:
            print("Error: Could not perform KVM check")
        return False

    def plan(self, planner):
        tests = list()
        nr_cpus = 0
        if not self.realtime or not self.is_kvm_enabled():
            print("INFO: Could not plan test {}".format(self.path))
            return tests

        try:
            nr_cpus = Command("lscpu | grep -E '^CPU\(s\):' | awk '{print $2}'").getString()
        except HwCertCommandException:
            print("Error: Could not determine cpus")
            return tests
        checkHardwareHealth = CheckHardwareHealth("ram_per_core")
        if int(nr_cpus) >= 3 and isfile("/sys/kernel/realtime"):
            with open("/sys/kernel/realtime", "r") as f:
                if f.readline().strip() == "1" \
                        and checkHardwareHealth.ram_per_core() \
                        and (self.__isHVMCapableViaLibvirt()
                            or self.__isHVMInCapabilitiesFile()
                            or self.__hasVirtualizationCPUFlags()):
                    tests.append(self.makeCopy())
        if not tests:
            print("INFO: Could not plan test {}".format(self.path))
        else:
            print("INFO: Successfully planned test {}".format(self.path))
        return tests

    def installFile(self, destinationDirectory, filename):
        """ TODO: copied from fvtest.py, needs to be consolidated into centralized helper class or something"""
        sys.stdout.flush()
        urlsToTry = list()
        tarredFilename = filename + ".tar.bz2"
        partnerServerBaseUrl = os.path.join(self.environment.getPartnerServerURL(), Constants.fv_images, RHEL.major)
        if self.getTestServer() and self.getTestServer() != "unknown":
            testServerBaseUrl ="http://%s" % os.path.join(self.getTestServer(), Constants.store, Constants.transfer, Constants.fv_images, RHEL.major)
            urlsToTry.append(os.path.join(testServerBaseUrl, self.redHatRelease.getArch(), tarredFilename))
            #the old store/transfer/fv-images/... path was removed.  This is the one partners use now.
            testServerNewBaseUrl = "http://%s" % os.path.join(self.getTestServer(), Constants.rhcert_transfer, Constants.fv_images, RHEL.major)
            urlsToTry.append(os.path.join(testServerNewBaseUrl, self.redHatRelease.getArch(), tarredFilename))
            #to avoid partners re-using existing LTS stored images w/ new arch SUT without accidentally taking old x86_64 if arch specific is available on partner ftp
            urlsToTry.append(os.path.join(partnerServerBaseUrl, self.redHatRelease.getArch(), tarredFilename))
            urlsToTry.append(os.path.join(testServerBaseUrl, tarredFilename))
            urlsToTry.append(os.path.join(testServerNewBaseUrl, tarredFilename))
        else:
            print("Warning: no hwcert test server set - can not look for guest files there")
        urlsToTry.append(os.path.join(partnerServerBaseUrl, self.redHatRelease.getArch(), tarredFilename))
        urlsToTry.append(os.path.join(partnerServerBaseUrl, tarredFilename))

        wget_cmd = "wget "
        if self.proxyURLs:
            # In proxy environment, use proxy of ftp protocol if that is set,
            # else show a warning message and use proxy of http or https protocol.
            # If none of the above protocols found, FAIL the test.
            supported_protocols = ["ftp", "http", "https"]
            proxy = None
            protocol_used = None
            for protocol in supported_protocols:
                if protocol in list(self.proxyURLs.keys()):
                    protocol_used = protocol
                    proxy = self.proxyURLs[protocol]
                    break
            if protocol_used is None:
                print("Error: Proxy settings are missing for ftp protocol, no other valid proxies found. "
                      "Please check settings in /etc/rhcert.xml.")
                print("Error: Failed to download the guest file %s" % tarredFilename)
                return False
            elif protocol_used != "ftp":
                print("Warning: Proxy settings are missing for ftp protocol. Please check settings in /etc/rhcert.xml. "
                      "Using proxy of '%s' protocol" % protocol_used)
            wget_cmd = "wget -e ftp_proxy='%s' " % proxy

        for url in urlsToTry:
            download_cmd = wget_cmd + url
            print("\nTrying to download the guest file using command: %s" % download_cmd)
            if os.system(download_cmd) == 0:
                print("success downloading from %s" % url)
                try:
                    if os.path.exists(tarredFilename):
                        unpackedFilename = Command("tar xmvfj " + tarredFilename).getString()
                        os.remove(tarredFilename)
                        # move it to the right place
                        shutil.move(unpackedFilename, os.path.join(destinationDirectory, filename))
                        print("Moved: " + unpackedFilename + " to " + os.path.join(destinationDirectory, filename))
                        restorecon = Command("/sbin/restorecon -v " + os.path.join(destinationDirectory, filename))
                        restorecon.echo()
                        print("Downloaded " + filename + " to " + destinationDirectory)
                        return True
                    else:
                        print("Note: could not download " + url)
                except Exception as e:
                    print("Warning: download " + url + " failed:")
                    print(e)
            else:
                print("Warning: could not download " + url) 

        # otherwise
        print("Error: could not find or download guest file " + tarredFilename)
        return False

    def run(self):

        #left for debugging purposes
        #self.__onlineGuestSetup()
        #return True

        result = False

        #run code from krcert-setup-host
        #TODO - put this back, commenting out for tighter feedback loop
        if not self.incomplete or not self.continuation.isInitialized():
        #if not True: #self.incomplete or not self.continuation.isInitialized():
            if not self.__configureHost():
                return False
            self.continuation.setInitConfig(self.getPath(), Constants.reboot)

            #command to restart
            Command("reboot now").echoIgnoreErrors()

            # wait here for reboot
            print("Waiting for reboot...")
            time.sleep(self.waitTime)
            print("Error: The reboot took too long")
            return False

        #TODO - put this back, commenting out for tighter feedback loop
        elif self.incomplete and self.continuation.isInitialized():
        #else:
            #after host machine is restarted
            result = self.continuation.verify(marker=self.getPath())
            self.continuation.removeInitConfig()

            #run configure host again to populate variables
            self.__configureHost()

            #Download vm
            #TODO - check if it's there first
            Command("mkdir /root/temp/").echoIgnoreErrors()
            if not self.installFile("/root/temp/", self.qcow2Image):
                return False
            #configure guest VM
            self.__configureGuest()
            #run test on guest
            result = self.runSubTest(self.__runTestOnGuest, name="Stress and Cyclic")
            return result
        else:
            print("Error: test was started outside of expected conditions")
            return False


if __name__ == "__main__":
    test = FvRealtimeTest()
    returnValue = test.do(sys.argv)
    sys.exit(returnValue)
