#!/usr/bin/bash
#
# run-cyclictest.sh: run cyclictest for KVM-RT testing.
#
# This script runs cyclictest with the command-line used for KVM-RT
# testing plus the following:
#
#    - Run stress on isolated CPUs so that we simulate DPDK behavior
#    - Supports running rteval on housekeeping CPUs for load test
#    - Saves important test run-time information
#    - Check dmesg for bugs
#
# Command-line example:
#
#  # run-cyclictest.sh -d 15m -c 1,2,3 -r
#
# Runs cyclictest for 15 minutes on CPUs 1,2,3 with rteval load on
# housekeeping CPUs.
#
# Luiz Capitulino <lcapitulino@redhat.com>

duration=
isolcpus=
run_rteval=n
kill_stress=n
cpu_list=
nr_cpus=
results_dir=/root/results
test_dir=
cyclic_out=
stress_out=
rteval_out=
test_out=
dmesg_out=

check_for_bug()
{
	if dmesg | egrep '(BUG|WARNING)' > /dev/null; then
		return 1
	fi
	return 0
}

die()
{
	echo "$*" > /dev/stderr
	exit 1
}

while getopts ":d:c:hrk" opt; do
	case $opt in
		d)
			duration=$OPTARG
			;;
		c)
			isolcpus=$OPTARG
			;;
		r)
			run_rteval=y
			;;
		k)
			kill_stress=y
			;;
		h)
			echo "usage: run-cyclictest.sh -d DURATION -c CPU-list -r -k"
			echo
			echo "Options:"
			echo "           -d test duration (eg. -d 1m for 1 minute)"
			echo "           -c CPU list (eg. -c 1,2,3)"
			echo "           -r run rteval (recommended)"
			echo "           -k kill running stress instances"
			echo
			exit 0
			;;
		:)
			die "ERROR: option -$OPTARG requires an argument"
			;;
	esac
done

if ! cyclictest --help > /dev/null; then
	exit 1
fi

if [ ! -d $results_dir ]; then
	if ! mkdir $results_dir; then
		exit 1
	fi
fi

# We cannot have more than more stress running
while true; do
	if ! pgrep stress > /dev/null; then
		break
	fi
	if [ "$kill_stress" = "y" ]; then
		kill -9 $(pgrep stress)
	else
		die ERROR: stress seems to be running and -k not specified
	fi
done

# Check test duration
if ! echo $duration | egrep '\b[1-9]+[0-9]*(m|h|d)\b' > /dev/null; then
	die ERROR: invalid test duration
fi

# Check isolcpus
if [ -z "$isolcpus" ]; then
	read -a cmdline <<< "$(cat /proc/cmdline)"
	for str in "${cmdline[@]}"; do
		if echo $str | grep isolcpus > /dev/null; then
			isolcpus=$(echo $str | cut -d= -f2)
			break
		fi
	done

	if [ -z $isolcpus ]; then
		die ERROR: CPU list not found in kernel command-line and not specified
	fi
fi

if echo $isolcpus | egrep '(-|[a-z]|[A-Z])' > /dev/null; then
	die ERROR: bad CPU list or isolcpus command-line
fi

cpu_list="$(echo -n $isolcpus | sed 's/\,/ /g')"
nr_cpus=$(echo -n $isolcpus | sed 's/\,//g' | wc -m)

for cpu in $cpu_list; do
	if ! grep processor /proc/cpuinfo | grep $cpu > /dev/null; then
		die ERROR: CPU$cpu does not exist
	fi
done

# Check rteval
if [ "$run_rteval" = "y" ]; then
	if ! rteval --help > /dev/null; then
			exit 1
	fi
fi

# Check for bug in dmesg
if ! check_for_bug; then
	die ERROR: bug found in dmesg, aborting test!
fi

# XXX: Assuming we have just one at time
# test_dir=$(mktemp -d -p $results_dir cyclictest-results.XXXXXX)
test_dir=$results_dir
cyclic_out=$test_dir/cyclictest.out
stress_out=$test_dir/stress.out
rteval_out=$test_dir/rteval.out
test_out=$test_dir/README
dmesg_out=$test_dir/dmesg.out

echo -e "Test started at $(date)\n"   >> $test_out
echo "Test duration:    $duration"    >> $test_out
echo "Run rteval:       $run_rteval"  >> $test_out
echo "Isolated CPUs:    $isolcpus"    >> $test_out
echo "Kernel:           $(uname -r)"  >> $test_out
echo "Kernel cmd-line:  $(cat /proc/cmdline)" >> $test_out
echo "Results dir:      $test_dir"    >> $test_out

clear
cat $test_out

echo
echo running stress
for cpu in $cpu_list; do
	cmdline="taskset -c $cpu stress --cpu 1"
	echo -e "   $cmdline"
	$cmdline 2>&1 > $stress_out &
done

if [ "$run_rteval" = "y" ]; then
	echo
	echo running rteval
	cmdline="rteval --onlyload --duration=$duration --verbose"
	echo -e "   $cmdline"
	$cmdline 2>&1 > $rteval_out &

	# Wait a bit for rteval setup (for short duration tests)
	# Yes, this is random
	sleep 4s
fi

# Wait a bit for stress
sleep 1s

echo
echo "starting $(date)"
cmdline="taskset -c $isolcpus cyclictest -m -n -q -p95 -D $duration -h60 -i 200 -t $nr_cpus -a $isolcpus"
echo -e "   $cmdline"
$cmdline > $cyclic_out
echo "ended $(date)"
echo

echo -e "\nTest ended at $(date)" >> $test_out

dmesg > $dmesg_out

echo output dir is $test_dir
echo

if ! check_for_bug; then
	echo ERROR: bug found in dmesg, test may be invalid! > /dev/stderr
fi

echo
egrep '(Min|Avg|Max) Latencies' $cyclic_out
echo

pkill stress 2>&1 > /dev/null
