#!/usr/bin/env python
# Copyright (c) 2019 Red Hat, Inc. All rights reserved. This copyrighted material
# is made available to anyone wishing to use, modify, copy, or
# redistribute it subject to the terms and conditions of the GNU General
# Public License v.2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# Author: Petersingh Anburaj
#
import os
import shutil
import sys
import time
from six.moves import urllib_parse

from requests import HTTPError
from rhcert.collabrequest import CollabRequest
from rhcert.command import Command, HwCertCommandException
from rhcert.continuation import Continuation
from rhcert.environment import Environment
from rhcert.redhatrelease import RedHatRelease
from rhcert.tags import DeviceClass, TestTag, Constants
from rhcert.compatability import rhcert_make_str
from rhcert.suites.hwcert.hwcert.hardwareTest import HardwareTest


class IPXETest(HardwareTest):

    def __init__(self):
        HardwareTest.__init__(self, "iPXE")
        self.interactive = True
        self.deviceClass = DeviceClass.network_interface
        self.redHatRelease = RedHatRelease()
        self.environment = Environment()
        self.boot_order = None
        self.boot_loader_name = "rhcert-ipxe"
        self.continuation = Continuation(self.environment)
        self.filename = "ipxe.json"
        self.rhcert_efi_file = "/usr/share/ipxe/ipxe-x86_64-rhcert.efi"

    def tags(self):
        return [TestTag.certification, TestTag.interactive]

    def getRequiredRPMs(self):
        return ["ipxe-rhcert"]

    def plan(self, planner):
        tests = list()
        # CERTPX-6202 : Disabling the auto-plan for iPXE test
        # # Check if system is UEFI booted
        # try:
        #     uefi_boot = Command("test -d /sys/firmware/efi")
        #     uefi_boot.run()
        # except HwCertCommandException as e:
        #     print("Warning: System is not booted in UEFI mode")
        #     return tests
        #
        # # Save the current boot order of the system
        # self.boot_order = self.get_boot_order()
        # if not self.boot_order:
        #     print("Warning: Unable to get current boot order")
        #     return tests
        #
        # # Check if OS is RHEL8 and above
        # if self.redHatRelease.getVersion() < 8:
        #     print("Warning: Unsupported Operating System")
        #     return tests
        #
        # # Check if system is x86_64 arch
        # if self.redHatRelease.getArch() == Constants.x86_64:
        #     test = self.makeCopy()
        #     tests.append(test)
        # else:
        #     print("Warning: %s is not supported architecture for iPXE test" % self.redHatRelease.getArch())
        return tests

    def get_sut_details(self):
        mac_addr = {}
        try:
            cmd = Command("/sbin/ip -o link | awk '$2 != \"lo:\" {print $2, $(NF-2)}'")
            mac_addr = cmd.getStringDict(regex="(?P<interface>[^:]+):\s(?P<mac>[^s]+)", keyGroup="interface", valueGroup="mac")
            print("MAC Address of SUT: %s" % str(mac_addr))
        except HwCertCommandException as e:
            print("Error: Failed to get MAC Address of SUT")
            print(str(e))
        return mac_addr

    @staticmethod
    def save_details(**kwargs):
        mac = kwargs.get("mac")
        filename = kwargs.get("filename")
        if not mac:
            error = "Error: MAC not found !!!"
            print(error)
            raise Exception(error)
        if filename not in [Constants.before_pxe, Constants.after_pxe]:
            error = "Error: Wrong file name passed !!!"
            print(error)
            raise Exception(error)
        try:
            store_data = "echo '%s' > /tmp/%s" % (mac, filename)
            Command(store_data).run()
        except HwCertCommandException as e:
            error = "Error: While storing data in LTS: %s" % (str(e))
            print(error)
            raise Exception(error)
        return "Success: Data saved on LTS !!"

    def save_details_to_lts(self, mac, filename):
        test_name = "IPXETest"
        test_package_trail = "hwcert.tests.ipxe.ipxe"
        function_name = "save_details"
        kwargs = {"mac": mac, "filename": filename}
        collab_request = CollabRequest(test_name, test_package_trail, function_name, kwargs=kwargs)
        try:
            collab_response = self.collabCall(collab_request)
            print(collab_response)
        except HTTPError as e:
            # print("Error: %s" % str(e))
            print(str(e.response.content))
            return False
        return True

    @staticmethod
    def get_details(**kwargs):
        filename = kwargs.get("filename")
        if filename not in [Constants.before_pxe, Constants.after_pxe]:
            error = "Error: Wrong file name passed !!!"
            print(error)
            raise Exception(error)
        try:
            get_data = "cat /tmp/%s" % filename
            # regex removes the lines that starts with '#'
            data = Command(get_data).getString(regex="^[^#].*", singleLine=False)
            print(data)
        except HwCertCommandException as e:
            error = "Error: While retrieving data from LTS: %s" % (str(e))
            print(error)
            raise Exception(error)
        return data

    def get_details_from_lts(self, filename):
        test_name = "IPXETest"
        test_package_trail = "hwcert.tests.ipxe.ipxe"
        function_name = "get_details"
        kwargs = {"filename": filename}
        collab_request = CollabRequest(test_name, test_package_trail, function_name, kwargs=kwargs)
        try:
            file_content = self.collabCall(collab_request)
        except HTTPError as e:
            # print("Error: %s" % str(e))
            print(str(e.response.content))
            return None
        return file_content

    @staticmethod
    def delete_files():
        for filename in [Constants.before_pxe, Constants.after_pxe]:
            try:
                cmd = "rm /tmp/%s" % filename
                Command(cmd).run(ignoreErrors=True)
            except HwCertCommandException as e:
                error = "Error: Failed to remove file: %s : %s" % (filename, str(e))
                print(error)
                raise Exception(error)
        return "Success: Temp files deleted on LTS !!"

    def delete_files_from_lts(self):
        test_name = "IPXETest"
        test_package_trail = "hwcert.tests.ipxe.ipxe"
        function_name = "delete_files"
        collab_request = CollabRequest(test_name, test_package_trail, function_name)
        try:
            collab_response = self.collabCall(collab_request)
            print(collab_response)
        except HTTPError as e:
            # print("Error: %s" % str(e))
            print(str(e.response.content))
            return False
        return True

    def get_boot_order(self):
        # Get the current boot order
        get_boot_order = Command("efibootmgr | awk '/BootOrder:/{print $2}'")
        boot_order = None
        try:
            boot_order = get_boot_order.getString(regex=".*")
            print("Current Boot Order: %s " % str(boot_order))
        except Exception as e:
            print("Warning: Failed to get Boot Order: %s" % str(e))
        return boot_order

    def reset_boot_order(self):
        # Check if self.boot_order is not None
        if not self.boot_order:
            print("Error: Original boot order is None")
            return False

        # Update the boot order to self.boot_order
        print("\nUpdating the boot order to original boot order")
        set_boot_order = Command("efibootmgr -o %s" % str(self.boot_order))
        try:
            set_boot_order.run()
        except HwCertCommandException as e:
            print("Error: Failed to set Boot Order: %s" % str(e))
            return False

        # Compare the boot order to confirm if that was set properly
        print("\nComparing the boot order to confirm if that was set properly")
        updated_boot_order = self.get_boot_order()
        if str(updated_boot_order) != str(self.boot_order):
            print("Error: Comparision failed after resetting the boot order.")
            return False
        print("Boot order has been updated successfully")
        return True

    def restore_boot_changes(self):
        print("\nRestoring the boot changes..")

        # Get the ID of the new boot loader if that was created
        print("\nFetching the ID of the new boot loader if that was created")
        get_boot_loader_ids = Command("efibootmgr | grep %s" % self.boot_loader_name)
        try:
            boot_loader_ids = get_boot_loader_ids.getStringList(regex="Boot(?P<id>[^*]+)",
                                                                regexGroup="id")
            print("ID of the %s boot loader: %s" % (self.boot_loader_name, str(boot_loader_ids)))
        except HwCertCommandException:
            print("Unable to fine any %s boot loader" % self.boot_loader_name)
            return True

        # Delete the new boot loader
        print("\nDeleting the new boot loader")
        for boot_loader_id in boot_loader_ids:
            delete_boot_loader = Command("efibootmgr -b %s -B" % str(boot_loader_id))
            try:
                delete_boot_loader.run()
                print("Successfully deleted the %s boot loader with id %s " % (self.boot_loader_name, boot_loader_id))
            except HwCertCommandException as e:
                print("Error: Failed to delete the %s boot loader with id %s" % (self.boot_loader_name, boot_loader_id))
                return False

        # Delete the boot loader directory if it exist
        print("\nDeleting the boot loader directory if it exist")
        boot_loader_dir = "/boot/efi/EFI/%s" % str(self.boot_loader_name)
        if os.path.isdir(boot_loader_dir):
            delete_old_dir = Command("rm -rf %s" % boot_loader_dir)
            delete_old_dir.run()
        return True

    def steps_before_pxe(self):
        # Check if test server is configured
        if self.getTestServer() is None or self.getTestServer() == "unknown":
            print("Error: No test server was set. iPXE test requires test server to run the test.")
            return False

        # Check if system is UEFI booted
        print("\nChecking if system is booted in UEFI mode")
        try:
            uefi_boot = Command("test -d /sys/firmware/efi")
            uefi_boot.run()
        except HwCertCommandException:
            print("Error: System is not booted in UEFI mode. "
                  "iPXE test can run only in UEFI booted system.")
            return False

        # Save the current boot order of the system
        self.boot_order = self.get_boot_order()
        if not self.boot_order:
            print("Error: Unable to get current boot order")
            return False

        # Delete the old temp files from LTS
        print("\nDeleting the old temp files from LTS")
        if not self.delete_files_from_lts():
            print("Failed to delete old temp files from LTS")
            return False

        # Fetch the MAC address of the SUT
        print("\nFetching the MAC address of the SUT")
        mac_addr = self.get_sut_details()
        if not mac_addr:
            print("Error: Unable to fetch the MAC address of the SUT")
            return False

        # Save the MAC address to the file
        try:
            self.continuation.dumpDictDataInFile(self.filename, mac_addr)
        except IOError as exception:
            print("Error: Failed to save MAC address : %s" % exception)

        # Check if efi partition is mounted
        print("\nChecking if efi partition is mounted")
        check_efi_mount = Command("df -h | awk '/\/boot\/efi/{print $1}'")
        try:
            partition = check_efi_mount.getString(regex=".*")
            print("Found partition %s mounted at /boot/efi" % partition)
        except HwCertCommandException:
            print("Error: Couldn't find /boot/efi partition mounted")
            return False

        # Get device name for the partition
        print("\nFetching the device name for the partition")
        get_device_name = Command("lsblk -no PKNAME %s" % partition)
        get_device_number = Command("lsblk -no MAJ:MIN %s" % partition)
        try:
            device_name = get_device_name.getString(regex=".*")
            device = "/dev/%s" % device_name
            device_number = get_device_number.getString(regex=".*:(?P<devno>[^\s]+)",
                                                          regexGroup="devno")
            print("Using device name as %s and device number as %s for partition %s " % (
                device, device_number, partition))
        except HwCertCommandException:
            print("Error: Unable to fine the device name or number for the given partition %s" % partition)
            return False

        # Delete the boot loader directory if it already exist
        print("\nDeleting the boot loader directory if it already exist")
        boot_loader_dir = "/boot/efi/EFI/%s" % str(self.boot_loader_name)
        if os.path.isdir(boot_loader_dir):
            delete_old_dir = Command("rm -rf %s" % boot_loader_dir)
            delete_old_dir.run()

        # Create a new directory in EFI system partition (ESP)
        print("\nCreating a new directory in EFI system partition (ESP)")
        try:
            create_new_dir = Command("mkdir %s" % boot_loader_dir)
            create_new_dir.run()
        except HwCertCommandException:
            print("Error: Failed to create directory in EFI system partition (ESP)")
            return False

        # Create rhcert-script.ipxe file in /tmp/ directory
        print("\nCreating the rhcert-script.ipxe file in /tmp/ directory")
        rhcert_script_file = "/tmp/rhcert-script.ipxe"
        if os.path.isfile(rhcert_script_file):
            os.remove(rhcert_script_file)
        rhcert_script = open(rhcert_script_file, "w")
        rhcert_script.write("#!ipxe\n")
        rhcert_script.write("dhcp\n")

        test_name = "IPXETest"
        test_package_trail = "hwcert.tests.ipxe.ipxe"
        function_name = "save_details"
        kwargs = '{"mac": "TemporaryMACPlaceHolder", "filename": "after-pxe"}'
        param = {
            "message_body": '{"test_name": "%s", "function_name": "%s", "test_package_trail": "%s", "kwargs": %s}' % (
                test_name, function_name, test_package_trail, kwargs)}
        encoded_param = urllib_parse.urlencode(param)
        # Replace plus '+' signs with %20
        # This can be done in python3 directly urllib_parse.urlencode(param, quote_via=urllib_parse.quote)
        # But not supported in python2
        encoded_param = encoded_param.replace('+', '%20')
        final_param = encoded_param.replace("TemporaryMACPlaceHolder", "${mac}")
        # print(final_param)

        test_server = self.getTestServer()
        chain_url = "http://%s/collab/?%s" % (test_server, final_param)
        rhcert_script.write("chain %s\n" % chain_url)
        rhcert_script.close()
        Command("cat %s" % rhcert_script_file).run(ignoreErrors=True)
        print("Successfully created the rhcert-script.ipxe file")

        # Copy rhcert.efi file and rhcert-script.ipxe into the newly created directory
        print("\nCopying rhcert.efi file and rhcert-script.ipxe into the newly created directory")
        try:
            shutil.copy(self.rhcert_efi_file, os.path.join(boot_loader_dir, "rhcert.efi"))
            shutil.copy(rhcert_script_file, os.path.join(boot_loader_dir, "rhcert-script.ipxe"))
            print("Successfully copied the required files")
        except Exception as e:
            print("Error: Failed to copy required files to EFI system partition: %s " % str(e))
            return False

        print("Using the original boot order as : %s " % self.boot_order)

        # Register a new boot loader
        print("\nRegistering a new boot loader")
        try:
            create_bootloader = "efibootmgr -c -d %s -p %s -L %s -l \\\\EFI\\\\%s\\\\rhcert.efi" % (
                device, device_number, self.boot_loader_name, self.boot_loader_name)
            Command(create_bootloader).run()
        except HwCertCommandException as e:
            print("Error: Failed to create the efi boot loader: %s" % str(e))
            return False

        # Check the boot order after registering the new boot loader
        new_boot_order = self.get_boot_order()
        if not new_boot_order:
            print("Error: Unable to check the updated boot order")
            return False
        print("Boot order after registering new boot loader: %s" % str(new_boot_order))

        # Reset the boot order to the old one
        print("\nResetting the boot order to the old one")
        if not self.reset_boot_order():
            print("Error: Failed to reset the boot order")
            return False

        # Get the ID of the new boot loader
        print("\nFetching the ID of the new boot loader")
        get_boot_loader_id = Command("efibootmgr | grep %s" % self.boot_loader_name)
        try:
            boot_loader_id = get_boot_loader_id.getString(regex="Boot(?P<id>[^*]+)",
                                                          regexGroup="id")
            print("ID of the new boot loader: %s" % boot_loader_id)
        except HwCertCommandException:
            print("Error: Unable to fine the ID of the new boot loader")
            return False

        # Set the new boot loader as next boot
        print("\nSetting the new boot loader as next boot")
        set_next_boot = Command("efibootmgr -n %s" % boot_loader_id)
        try:
            set_next_boot.run()
            print("Added %s boot loader as next boot" % self.boot_loader_name)
        except HwCertCommandException:
            print("Error: Failed to set new boot loader as next boot")
            return False

        # Log the final boot settings
        print("\nLogging final boot settings")
        boot_settings = Command("efibootmgr -v")
        boot_settings.run(ignoreErrors=True)

        # Reboot SUT to test iPXE
        print("\nSUT will be rebooted to complete the iPXE test")
        if not self.promptConfirm("SUT will be rebooted to complete the iPXE test. Ready to reboot?"):
            return False
        return True

    def steps_after_pxe(self):

        # Restore the boot related changes
        self.restore_boot_changes()

        # Get MAC stored in SUT before PXE
        print("\nFetching MAC stored in SUT before PXE")
        mac_addr = self.continuation.loadDictDataFromFile(self.filename)
        print("MAC stored in SUT before PXE: %s " % str(mac_addr))
        if not mac_addr:
            print("Error: Failed to retrieve stored MAC address of SUT")
            return False

        # Delete the dump file
        self.continuation.deleteDumpFile(self.filename)

        # Get MAC saved in LTS during iPXE
        print("\nFetching MAC saved in LTS during iPXE")
        ipxe_mac = self.get_details_from_lts(Constants.after_pxe)

        if not ipxe_mac:
            print("Error: Failed to retrieve MAC address from file")
            print("\nError: iPXE test failed !!!")
            return False
        print("MAC received during iPXE: %s" % rhcert_make_str(ipxe_mac))

        # Check if iPXE MAC match any of the MACs of SUT
        print("\nComparing iPXE MAC with the MACs of SUT")
        for key, value in mac_addr.items():
            # Python3 returns a byte hence converting it to string for comparision
            if rhcert_make_str(value) == rhcert_make_str(ipxe_mac):
                print("MAC address '%s' matches for interface: %s" % (rhcert_make_str(ipxe_mac), key))
                print("\nSuccess: iPXE test was successful !!")
                return True
        print("\nError: No match found for MAC address %s." % rhcert_make_str(ipxe_mac))
        print("\nError: iPXE test failed !!!")
        return False

    def run(self):

        if not self.incomplete or not self.continuation.isInitialized():
            if not self.runSubTest(self.steps_before_pxe, "Server Setup for iPXE",
                                   "Configure the server for iPXE test"):
                self.restore_boot_changes()
                return False
            self.continuation.setInitConfig(self.getPath(), Constants.reboot)
            sys.stdout.flush()
            Command("reboot now").run(ignoreErrors=True)

            # wait here for shutdown
            wait_time = 500  # seconds
            time.sleep(wait_time)
            print("Error: Shutdown took too long...")
            self.restore_boot_changes()
            self.continuation.verify(marker=self.getPath())
            self.continuation.removeInitConfig()
            return False

        elif self.incomplete and self.continuation.isInitialized():
            self.continuation.verify(marker=self.getPath())
            self.continuation.removeInitConfig()
            return self.runSubTest(self.steps_after_pxe, "Verify iPXE test",
                                   "Check if the iPXE test was successful")

        else:
            print("Error: test was started outside of expected conditions")
            return False


if __name__ == "__main__":
    test_ipxe = IPXETest()
    returnValue = test_ipxe.do(sys.argv)
    sys.exit(returnValue)
