#!/usr/bin/env python
# Copyright (c) 2008 Red Hat, Inc. All rights reserved. This copyrighted material
# is made available to anyone wishing to use, modify, copy, or
# redistribute it subject to the terms and conditions of the GNU General
# Public License v.2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# Author: Greg Nichols <gnichols@redhat.com>

import os, sys, time, re, socket, uuid

from rhcert.tags import Constants, DeviceClass, TestTag
from rhcert.command import Command, HwCertCommandException
from rhcert.environment import Environment
from rhcert.continuation import Continuation
from rhcert.configfile import ConfigFile
from rhcert.redhatrelease import RedHatRelease

from rhcert.suites.hwcert.hwcert.udevDeviceDetector import UdevDeviceDetector

from rhcert.suites.hwcert.hwcert.hardwareTest import HardwareTest

class KDumpTest(HardwareTest):

    def __init__(self):
        HardwareTest.__init__(self, "kdump")
        self.interactive = True
        self.deviceClass = DeviceClass.system
        self.environment = Environment()
        self.continuation = Continuation(self.environment)
        self.rebootTimeLimit = self.environment.getRebootTimelimit()
        self.priority = 1001 # run last
        self.coreCollector = "makedumpfile -c -d 31"
        self.kdumpConfigPath = "/etc/kdump.conf"
        self.redHatRelease = RedHatRelease()
        self.serverMode = True # nfs for kdump to nfs
        self.crashDirectory = "/var/crash/" + str(uuid.uuid1())
        self.filename = "kdumpconf.json"
        self.waitTime = 60

    def tags(self):
        return [TestTag.interactive, TestTag.certification]

    def getRequiredRPMs(self):
        rpms = ["kexec-tools", "crash", "kernel-debuginfo"]
        # version = Command("uname -r").getString()
        # rpms.append("kernel-debuginfo-%s" % version)
        return rpms

    def plan(self, planner):
        tests = list()
        # plan kdump twice: once on storage (local disk) and once over network
        for dumpMethod in [Constants.nfs, Constants.local]:
            test = self.makeCopy()
            test.setLogicalDeviceName(dumpMethod)
            tests.append(test)
        return tests

    def logDevices(self, planner):
        devices = list()
        if self.getLogicalDeviceName() == Constants.local:
            self.deviceClass = DeviceClass.hard_disk
        elif self.getLogicalDeviceName() == Constants.nfs:
            properties = dict()
            self.deviceClass = DeviceClass.network_interface
            try:
                serverIP = socket.gethostbyname(self.getTestServer())
                ifname = Command("/sbin/ip -o route get to %s" % serverIP).getString(regex=".* dev (?P<ifname>\w+) .*", regexGroup="ifname")
                properties["ifname"] = ifname
                properties["mac"] = Command("/sbin/ip -o link show %s" % ifname).getString(regex=".* link/ether (?P<mac>[a-f0-9:]+) .*", regexGroup="mac")
                properties["ipaddr"] = Command("/sbin/ip -o -4 addr show %s" % ifname).getString(regex=".* inet (?P<ipaddr>\d+\.\d+\.\d+\.\d+/\d+) .*", regexGroup="ipaddr")
                udevDevices = UdevDeviceDetector().getDevices()
                for dev in udevDevices:
                    if dev.getProperty("INTERFACE") == ifname:
                        dev.getProperties().update(properties)
                        break
            except HwCertCommandException as e:
                print("Warning: fail to get the network interface info in-use")
                print(e)

        devices.extend(HardwareTest.logDevices(self, planner))
        return devices

    def verify(self):
        if self.getLogicalDeviceName() == Constants.nfs and (self.getTestServer() == None or self.getTestServer() == "unknown"):
            print("kdump test for %s failed verification: No test server was set." % self.getLogicalDeviceName())
            return False
        return True

    def startServer(self):
        result = True
        try:
            start = Command("systemctl start nfs-server.service")
            start.echo()
        except HwCertCommandException as e:
            print("Error: could not start NFS service")
            print(e)
            result = False

        serverDirectory = self.environment.getServerDirectory()
        exportDirectory = os.path.join(serverDirectory, self.environment.getServerExport())
        for directory in (serverDirectory, exportDirectory):
            try:
                os.mkdir(directory)
                print("Created %s" % directory)
            except OSError: # file probably exists
                if not os.path.exists(directory):
                    print("Warning: unable to create directory %s." % directory)

        if os.system("exportfs -o rw,no_root_squash,insecure *:%s" % exportDirectory) != 0:
            print("Error: could not export %s" % exportDirectory)
            result = False

        return result

    def stopServer(self):
        # nfs
        try:
            stop = Command("systemctl stop nfs-server.service")
            stop.echo()
        except HwCertCommandException as e:
            print("Error: could not stop nfs")
            print(e)
            return False

        return True

    def statusServer(self):
        result = True
        try:
            # changed from using "exportfs" since this is called via httpd/cgi, and exportfs uses a
            # lock file not readable by non-root users.F
            exportfs = Command("cat /var/lib/nfs/etab")
            exportfs.getString(regex=".*%s/%s" % (self.environment.getServerDirectory(), self.environment.getServerExport()), singleLine=False)
            print("%s/%s is exported" % (self.environment.getServerDirectory(), self.environment.getServerExport()))
        except HwCertCommandException as exception:
            print("Error: %s/export is not available" % self.environment.getDataDirectory())
            result = False

        try:
            serviceNFSStatus = Command("systemctl status nfs-server.service")
            print("%s:" % serviceNFSStatus.command)
            serviceNFSStatus.echo()
        except HwCertCommandException as exception:
            print(exception)
            result = False

        return result

    def configureKDump(self):
        verified = True
        # log required rpms installation
        print("Checking required packages:")
        for rpm in self.getRequiredRPMs():
            try:
                print(Command("rpm -q %s" % rpm).getString())
            except:
                pass

        print("Checking kdump configuration")
        # check crashkernel on kernel command line
        # typical settings:
        #     ppc64:  512M@64M (/etc/yaboot.conf)
        #     x86_64: 128M@16M (/etc/grub.conf)
        kernelBootParams = Command("cat /proc/cmdline")
        kdump = None
        try:
            kernelBootParams = Command("cat /proc/cmdline")
            kdump = kernelBootParams.getString(regex=".+ crashkernel=(?P<kdump>[^\ ]+)", regexGroup="kdump")
            print("Found crashkernel=%s boot parameter" % kdump)
        except:
            print("Error: \"crashkernel\" is not set in boot parameters")
            return False

        # check kernel panic timeout, and set to 1sec if necessary
        try:
            procPanicTimeout = "/proc/sys/kernel/panic"
            getPanic = Command("cat " + procPanicTimeout)
            timeout = getPanic.getInteger()
            print("Kernel panic reboot timeout is %u" % timeout)
            if not timeout:
                print("Setting to 1 sec.")
                setPanic = Command("echo 1 > " + procPanicTimeout)
                setPanic.echo()
                if getPanic.getInteger() != 1:
                    print("Error: could not set panic timeout")
                    verifited = False
        except HwCertCommandException as e:
            print("Error: could not set " + procPanicTimeout)
            print(e)
            verified = False


        (success, restart) = self.setKdumpConfig()
        if not success:
            return False

        if restart:
            print("restarting kdump with new configuration...")
        # restart it every time
        try:
            Command("systemctl stop kdump").echo()
            Command("systemctl start kdump").echoIgnoreErrors()
        except HwCertCommandException as exception:
            print("Error: kdump restart failed")
            print(exception)
            exception.command.printErrors()
            print("kdump status:")
            Command("systemctl status -l kdump").echo()
            verified = False


        print("Checking kdump service")
        try:
            kdump = Command("systemctl status kdump")
            kdump.getString(regex="Active: active", singleLine=False)
            print("kdump is running")
        except HwCertCommandException as e:
            print("Error: kdump is not running - can not test it")
            print(e)
            verified = False

        return verified

    def setKdumpConfig(self):
        """ set kdump configuration, returns (success, restart) """

        devname = ""
        uuid = ""
        fstype = ""
        # get disk UUID and fstype from system
        if not os.path.exists(self.crashDirectory):
            os.makedirs(self.crashDirectory)
        try:
            devname = Command("df " + self.crashDirectory).getString(regex="/dev/(?P<devname>[^\s]+)",
                                                          regexGroup="devname",
                                                          singleLine=False)
            uuid = Command("blkid -s UUID /dev/%s" % devname).getString(regex="/dev/%s: (?P<uuid>.*$)" % devname,
                                                                        regexGroup="uuid")
            fstype = Command("mount").getString(regex="/dev/%s on.*type (?P<fstype>[^\s]+)" % devname,
                                                regexGroup="fstype",
                                                singleLine=False)
        except HwCertCommandException as exception:
            # if detect UUID/fstype fail, merely give a warning and continue with
            # default settings - blank local dump target means kdump will try to
            # mount the disk where self.crashDirectory locates at and dump.
            print("Warning: cannot get disk UUID or FS type of local disk!")
            print("         will try default configuration of local dump.")
            print(exception)


        # add crash directory and get all kdump.conf parameters
        try:
            configFile = ConfigFile(self.kdumpConfigPath)
            configFile.removeParameter("path")
            configFile.addParameter("path", self.crashDirectory)
            configFile.save()
            from_config = configFile.getParameter("path")
            print("kdump configuration:")
            print("--------------------")
            configFile.dump()
            print("--------------------\n")
            coreCollector = configFile.getParameter("core_collector")
            nfs = configFile.getParameter("nfs")
            local = configFile.getParameter(fstype) if fstype else ""
        except IOError as exception:
            print("Error: %s" % exception)
            return (False, False)

        # configure kdump to use makedumpfile -d 31
        restart = False
        if not coreCollector:
            print("Adding core_collector = %s to %s" % (self.coreCollector, self.kdumpConfigPath))
            configFile.addParameter("core_collector", self.coreCollector)
            restart = True
        else:
            print("core_collector currently set to \"" + coreCollector + "\"")

        # configure for local dump
        if self.getLogicalDeviceName() == Constants.local:
            if local:
                print("Local dump target set to %s" % local)
            else:
                if devname and uuid and fstype:
                    # remove any quotes around the uuid
                    uuid = uuid.replace('\"', '')
                    print("Adding %s %s to %s" % (fstype, uuid, self.kdumpConfigPath))
                    configFile.addParameter(fstype, uuid)
                    restart = True

            if nfs:
                print("Removing nfs parameter for local target dump")
                configFile.removeParameter("nfs")
                restart = True
        # configure for nfs dump
        elif self.getLogicalDeviceName() == Constants.nfs:
            if nfs:
                print("NFS dump target set to %s" % nfs)
            else:
                if self.getTestServer() and self.getTestServer() != "unknown":
                    # Note: the assumption is that the environment for the SUT is the same as
                    # the environment for the server being used - both have the same server
                    # directory setting.
                    nfs = "%s:%s/%s" % (self.getTestServer(),
                                        self.environment.getServerDirectory(),
                                        self.environment.getServerExport())
                    print("Adding nfs %s to %s" % (nfs, self.kdumpConfigPath))
                    configFile.addParameter("nfs", nfs)

                    # we need to mount it locally or kdump won't start
                    self.mountNFSDumpDirectory(configFile)
                    restart = True
                else:
                    print("Error: hwcert test server not set for NFS dump")
                    return (False, False)

            if local:
                print("Removing local setting for nfs target dump")
                configFile.removeParameter(fstype)
                restart = True

        if restart:
            configFile.save()
 
        try:
            # dump the changes made to kdump.conf to a file
            self.continuation.dumpDictDataInFile(self.filename, configFile.modifiedParams)
        except IOError as exception:
            print("Error: %s" % exception)

        try:
            configFile = ConfigFile(self.kdumpConfigPath)
            print("updated kdump configuration:")
            print("--------------------")
            configFile.dump()
            print("--------------------\n")
        except IOError as exception:
            print("Error: %s" % exception)
            return (False, False)

        return (True, restart)


    def panic(self):
        self.markOutput("initialize")
        if not self.configureKDump():
            return False

        # otherwise
        print("The test will now cause a kernel panic to exercise kdump")

        if self.getMode() != Constants.auto:
            if not self.promptConfirm("Ready to restart?"):
                return False

        # set up restart, and log start time
        self.continuation.setInitConfig(self.getPath(), Constants.panic)

        # we need a delay here to give hwcert time to finish writing results.xml
        self.closeOutput()
        sys.stdout.flush()
        self.waitForLull()

        try:
            Command("echo 1 > /proc/sys/kernel/sysrq").echo()
            Command("echo \"c\" > /proc/sysrq-trigger").echo()
        except Exception as e:
            print("Error: attempt to cause kernel panic failed")
            print(e)
            return False

        # wait here for reboot
        print("Waiting for kernel panic...")
        time.sleep(self.waitTime)
        print("Error: The kernel panic took too long")
        return False

    def verifyKDumpImage(self):
        imageDirectory = self.getImageDirectory()
        if not imageDirectory:
            return False
        def first_level_subdirs(path):
            return [x for x in os.listdir(path) if os.path.isdir(os.path.join(path, x))]

        # find the vmcore image file matching the timestamp
        # vmcore directories are like this: 127.0.0.1-2011-03-10-13:18:27
        #                but wait, it's now 127.0.0.1-2013.02.13-11:10:57 (not the date fields use . as a separator!
        vmcoreDirectoryPattern = re.compile("(?P<ipaddr>[0-9]+\.[0-9]+\.[0-9]+)-(?P<date>[0-9]+(-|\.)[0-9]+(-|\.)[0-9]+)-(?P<time>[0-9]+:[0-9]+:[0-9]+)")
        minimumDuration = None
        directoryMatch = None
        # if there's more than one directory, that's a problem
        dirs = first_level_subdirs(imageDirectory)
        if len(dirs) > 1:
            print("Error: multiple vmcores found.")
            return False
        for dir in dirs:
            match = vmcoreDirectoryPattern.search(dir)
            if match:
                timestamp = "%s %s" % (match.group("date").replace(".", "-"), match.group("time"))
                duration = self.continuation.getDuration(timestamp)
                print("%s took %s" % (dir, duration))
                if not directoryMatch or (duration.seconds < minimumDuration):
                    minimumDuration = duration.seconds
                    directoryMatch = dir
        if directoryMatch and (minimumDuration < self.rebootTimeLimit*60):
            vmcorePath = os.path.join(imageDirectory, directoryMatch, "vmcore")
            print("Found kdump image: " + vmcorePath)
        elif directoryMatch:
            print("Warning: could not locate vmcore file within the maximum reboot duration window")
            print("Using closest match " + os.path.join(imageDirectory, directoryMatch, "vmcore"))
            print("Reboot timestamp: " + self.continuation.getTimestamp())
            vmcorePath = os.path.join(imageDirectory, directoryMatch, "vmcore")
        else:
            print("Error: could not locate vmcore file")
            if directoryMatch:
                print("Closest match is " + os.path.join(imageDirectory, directoryMatch, "vmcore"))
                print("Reboot timestamp: " + self.continuation.getTimestamp())
            return False

        # run crash to see if it's a valid image
        try:
            # just run crash on it and quit, then look for the return value
            # using -s to supress startup output, then "sys" command for a summary, then "q" to quit the interpreter
            crash = Command("echo \"sys\nq\" | crash -s %s /usr/lib/debug/lib/modules/`uname -r`/vmlinux" % vmcorePath)
            crash.echo()
            # if we get here, it must be OK
            print("kdump image %s verified"% vmcorePath)
            return True
        except HwCertCommandException as exception:
            print("Error: could not verify kdump image %s" % vmcorePath)
            print(exception)
            crash.printErrors()

        return False

    def getImageDirectory(self):
        # parse /etc/kdump.conf to find the location of the image file
        configFile = ConfigFile(self.kdumpConfigPath)
        imageDirectory = configFile.getParameter("path")
        if not imageDirectory:
            imageDirectory = self.crashDirectory

        # if nfs network dump, mount the filesystem
        if configFile.getParameter("nfs"):
            imageDirectory = self.mountNFSDumpDirectory(configFile)

        if imageDirectory:
            print("Looking for vmcore image directories under " + imageDirectory)
        else:
            print("Error: could not find image directory.")
        return imageDirectory

    def mountNFSDumpDirectory(self, configFile):
        nfs = configFile.getParameter("nfs")
        imageDirectory = configFile.getParameter("path")
        if not imageDirectory:
            imageDirectory = self.crashDirectory
        tries = 5
        waitTime = 10 # sec.
        while True:
            print("Attempting to mount nfs setting %s as nfs." % nfs)
            # first, unmount it if necessary
            try:
                Command("umount %s" % (nfs)).echo()
            except:
                # ignore
                pass
            try:
                localPath = "hwcert-nfs"
                if not os.path.exists(localPath):
                    Command("mkdir %s" % localPath).echo()
                Command("mount %s %s" % (nfs, localPath)).echo()
                if os.path.isabs(imageDirectory):
                    imageDirectory = localPath + imageDirectory
                else:
                    imageDirectory = os.path.join(localPath, imageDirectory)
                # need to make the image directory because kdump complains if it's not there
                Command("mkdir -p %s" % imageDirectory).echo()
                return imageDirectory
            except HwCertCommandException as exception:
                if tries == 0:
                    print("Error: could not mount nfs image directory")
                    print(exception)
                    print(exception.command.printErrors())
                    return None
                else:
                    print("Warning: could not mount nfs image directory, retrying in %s sec." % waitTime)
                    tries = tries-1
                    time.sleep(waitTime)

        #otherwise
        return None

    def run(self):
        result = True
        if self.getLogicalDeviceName() == Constants.nfs:
            Command("ip addr").echoIgnoreErrors()
        if self.incomplete and self.continuation.isInitialized():
            time.sleep(self.waitTime)
            self.markOutput("verify")
            result = self.continuation.verify(marker=self.getPath())
            self.continuation.removeInitConfig()
            configFile = ConfigFile(self.kdumpConfigPath)
            self.crashDirectory = configFile.getParameter("path")
            if self.continuation.getMethod() == Constants.panic and not self.verifyKDumpImage():
                result = False
            self.closeOutput()

            #reset kdump config
            try:
                modifiedParams = self.continuation.loadDictDataFromFile(self.filename)
                for parameter in modifiedParams.get('added'):
                    configFile.removeParameter(parameter.get('name'))
                for parameter in modifiedParams.get('removed'):
                    configFile.addParameter(parameter.get('name'),parameter.get('value'))
                configFile.save()
            except (IOError,KeyError) as exception:
                print("Warning: Could not reset /etc/kdump.conf file")
                print(exception)

            try:
                Command("systemctl restart kdump").echoIgnoreErrors()
            except HwCertCommandException as exception:
                print("Warning: kdump restart failed")
                print(exception)
                exception.command.printErrors()
                print("kdump status:")
                Command("systemctl status -l kdump").echoIgnoreErrors()

            self.continuation.deleteDumpFile(self.filename)

        else:
            result = self.panic()

        return result
