#!/usr/bin/env python
# Copyright (c) 2006 Red Hat, Inc. All rights reserved. This copyrighted material
# is made available to anyone wishing to use, modify, copy, or
# redistribute it subject to the terms and conditions of the GNU General
# Public License v.2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# Author: Greg Nichols
#
import sys, time

from rhcert.tags import Constants, DeviceClass, TestTag
from rhcert.command import Command, HwCertCommandException

from rhcert.suites.hwcert.hwcert.procDeviceDetector import ProcDeviceDetector
from rhcert.suites.hwcert.hwcert.hardwareTest import HardwareTest

class LidTest(HardwareTest):

    def __init__(self):
        HardwareTest.__init__(self, "lid")
        self.deviceClass = DeviceClass.display

    def tags(self):
        return [TestTag.interactive, TestTag.certification, TestTag.portable]

    def plan(self, planner):
        tests = list()
        detector = planner.getProcDeviceDetector()
        devices = detector.loadDevices(ProcDeviceDetector.inputBusPath)
        for device in devices:
            if device.hasPropertyValue("Lid Switch"):
                test = self.makeCopy()
                test.setDevice(device)
                tests.append(test)
        return tests

    def getDevice(self):
        """ get the lid switch device """
        detector = ProcDeviceDetector()
        device = detector.getDevice(self.getUDI(), ProcDeviceDetector.inputBusPath)
        if device:
            self.setDevice(device)
        else:
            print("Error: could not find device udi %s" % self.getUDI())

        return device

    def checkLid(self, checkForOpen):
        lidState = Command("cat /proc/acpi/button/lid/*/state")
        try:
            lidState.getString("open")
            return checkForOpen
        except HwCertCommandException as e:
            if self.getDebug() != Constants.off:
                print(e)

        return not checkForOpen

    def run(self):

        SLEEPLIMIT=20 # sleep cycles
        SLEEPTIME=0.5 # sec

        if self.getMode() == Constants.auto:
            print("Error: the lid test can not be run in auto mode.")
            return False

        if not self.checkLid(checkForOpen=True):
            self.statusMessage("Error: lid must be open for this test")
            return False
        #otherwise
        if not self.promptConfirm("Ready to begin the lid/backlight test?"):
            return False


        self.statusMessage("Please close the lid, verify the backlight turns off when the lid is closed, then reopen it.")
        sys.stdout.flush()

        sleepLimit = SLEEPLIMIT
        #wait while lid is open and limit has not expired
        while self.checkLid(checkForOpen=True) and sleepLimit > 0:
            sys.stdout.write("%d..." % sleepLimit)
            sys.stdout.flush()
            time.sleep(SLEEPTIME)
            sleepLimit = sleepLimit -1
        print("")

        if sleepLimit <= 0:
            message = "Error: did not detect the lid was closed within %d sec." % (SLEEPLIMIT*SLEEPTIME)
            self.statusMessage(message)
            return False

        sleepLimit = SLEEPLIMIT
        # wait while lid is closed and time is not expired
        while self.checkLid(checkForOpen=False) and sleepLimit > 0:
            sys.stdout.write("%d..." % sleepLimit)
            sys.stdout.flush()
            time.sleep(SLEEPTIME)
            sleepLimit = sleepLimit -1
        print("")

        if sleepLimit <= 0:
            message = "Error: did not detect the lid was re-opened within %d sec." % (SLEEPLIMIT*SLEEPTIME)
            self.statusMessage("message")
            return False

        if not self.promptConfirm("Did the display backlight turn off when the lid was closed?"):
            self.statusMessage("Error: backlight must turn off when the lid is closed")
            return False

        return True


if __name__ == "__main__":
    test = LidTest()
    returnValue = test.do(sys.argv)
    sys.exit(returnValue)
