#!/usr/bin/env python
# Copyright (c) 2006 Red Hat, Inc. All rights reserved. This copyrighted material
# is made available to anyone wishing to use, modify, copy, or
# redistribute it subject to the terms and conditions of the GNU General
# Public License v.2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# Author: Greg Nichols
#
import os, sys

from rhcert.tags import Constants, DeviceClass, TestTag
from rhcert.command import Command, HwCertCommandException

from rhcert.suites.hwcert.hwcert.procDeviceDetector import ProcDeviceDetector
from rhcert.suites.hwcert.hwcert.hardwareTest import HardwareTest
from rhcert.suites.rhcert.platform import CheckHardwareHealth

class MemoryTest(HardwareTest):

    def __init__(self):
        HardwareTest.__init__(self, "memory")
        self.oldHandler = None

        self.owner = "Cerification Partner Experience Team <certpx@redhat.com>"
        self.description = "Memory Test"
        self.priority = 4
        self.deviceClass = DeviceClass.memory

    def tags(self):
        return [TestTag.noninteractive, TestTag.certification]

    def plan(self, planner):
        tests = list()
        checkHardwareHealth = CheckHardwareHealth("ram_per_core")
        if checkHardwareHealth.ram_per_core():
            tests.append(self.makeCopy())
        if not tests:
            print("INFO: Could not plan test {}".format(self.path))
        else:
            print("INFO: Successfully planned test {}".format(self.path))
        return tests

    def verify(self):
        success = self.getMemoryInfo()
        if success:
            # check the swap configuration
            largeMemory = 4 # TB
            minSwap = 0.025 # 2.5%
            if self.systemMemory*Constants.MB > largeMemory * Constants.TB:
                if self.swapMemory*Constants.MB < largeMemory * Constants.TB * minSwap:
                    print("Error: Swap of %s GB is too small for a %s TB system." % (
                        self.swapMemory*Constants.MB/Constants.GB, self.systemMemory*Constants.MB/Constants.TB))
                    success = False
        return success

    def logDevices(self, planner):
        devices = list()
        proc = ProcDeviceDetector()
        devices.extend(proc.getDevices(ProcDeviceDetector.procMeminfo))
        # call the default to log udev devices
        devices.extend(HardwareTest.logDevices(self, planner))
        return devices

    def run(self):

        passed = self.runSubTest(self.getMemoryInfo, name="Limits", description="Get test parameters based on hardware")
        if not passed:
            return False

        #  if process memory is limited, run multiple processes
        if self.isProcessLimited:
            passed = self.runSubTest(self.runMultipleProcessTest, name="Multi-process")
            if not passed:
                return False
        else:
            passed = self.runSubTest(self.runSingleProcessTest, name="Single-process")
            if not passed:
                return False

        #otherwised, passed
        return True


    def runSingleProcessTest(self):
        if not self.runThreadedMemoryTest():
            return False

        #otherwise
        return True

    def runMultipleProcessTest(self):
        numberOfProcesses = self.freeMemory/self.processMemory
        # if not swap-less, nfs root, add a process to hit swap
        if not (self.isNFSRootSystem and self.swapMemory == 0):
            numberOfProcesses += 1
            # check for zero swap
            if not self.swapMemory:
                print("Error: this test requires non-zero swap memory.")
                return False
            # check to make sure there's enough swap
            requiredMemory = self.processMemory*numberOfProcesses
            if requiredMemory > self.systemMemory + self.swapMemory:
                print("Error: this test requires a minimum of %u KB of swap memory (%u configured)" % (requiredMemory-self.systemMemory, self.swapMemory))
        print("Testing memory with %u processes" % numberOfProcesses)

        print("Running threaded memory test:")
        runTime = 60 # sec.
        if not self.runProcesses(numberOfProcesses, ("./threaded_memtest -qpv -m%um -t%u" % (self.processMemory, runTime))):
            print("Multi-process, threaded memory Test FAILED")
            return False

        return True

    def runThreadedMemoryTest(self):
        # single-process threaded test
        print("Starting Threaded Memory Test")

        # run for Free Memory plus the lesser of 5% or 1GB
        print("free memory: %s" % self.freeMemory)
        memory = (self.freeMemory * 5)/100
        print("memory: %s" % memory)
        if memory > 1024: # MB
            memory = 1024 # MB
        memory = memory + self.freeMemory

        # Correction - CERTPX-2679 - segmentation fault in threaded_memtest results from having memory less than 
        # number of threads (2*number of cpus) * pagesize.  Increasing the memory used in the test and keeping the
        # same number of threads resuls in the test exiting early for not having enough swap memory
        #  For that reason, we must modify the number of threads being used.
        # But wait, there's a warning about having less threads than cpus and sure enough, the threaded_mem c script is killed.
        # So, now we will see if the default threads's required memory is higher and if so, we will use the number of cpus as 
        # the number of threads and calculate a new memory ammount that will hopefully not violate the swap memory requirement.
        # Nevermind, I see that this is also a problem and chokes the system.  So, we calculate the threads by memory already calculated
        # (int) divided by page size and then to get the memory to pass in, we multiply threads by page size and divide by 8 to get to MB.

        threads = 0
        try:
            pageSize = Command("getconf PAGESIZE").getString()
            numCpus = Command("grep -c ^processor /proc/cpuinfo").getString()
            minimumMemory = int(pageSize) * 2 * int(numCpus)
            if memory < minimumMemory:
                #threads = numCpus
                newThreads = memory / int(pageSize)
                #subtract one thread so that rhcertd isn't killed
                if newThreads > 1:
                    newThreads = newThreads - 1
                # CERTPX-3874: Thread count should be min '1', if it is less than '1' eg: 0.684
                # the 'atoi' function in 'threaded_memtest.c' converts it to '0' and throws error.
                if newThreads < 1:
                    newThreads = 1
                newMemory = int(pageSize) * newThreads / 8 # bits to bytes
                if newMemory <= 0:
                    print("Warning: could not calculate new number of threads and memory values")
                    print("originally calculated memory value: %s" % memory)
                    print("page size: %s" % pageSize)
                    print("number of cpus: %s" % numCpus)
                else:
                    memory = newMemory
                    threads = newThreads
                    print("threads set to %s" % threads)
                    print("memory for test set to %s" % memory)
        except HwCertCommandException:
            print("Warning: Could not determine pagesize.")


        # run a test that will swap, except for nfs root systems without swap
        if not (self.isNFSRootSystem and self.swapMemory == 0):

            # check for zero swap
            if not self.swapMemory:
                print("Error: this test requires non-zero swap memory.")
                return False

            # is there enough swap memory for the test?
            # ignoring when there's a huge number of threads
            if threads:
                pass
            elif memory > self.systemMemory + self.swapMemory:
                print("Error: this test requires a minimum of %u KB of swap memory (%u configured)" % (memory-self.systemMemory, self.swapMemory))
                return False

            # otherwise
            runtime = 60 # sec.
            if threads:
                threadString = " -n%s " % threads
                threadStringMessage = " using %s threads" % threads
            else:
                threadString = threadStringMessage = ""
            print("running for more than free memory at %u MB for %u sec%s." % (memory, runtime, threadStringMessage))
            try:
                Command("./threaded_memtest -qpv -m%um -t%u %s" % (memory, runtime, threadString)).echo()
                print("done.")
            except HwCertCommandException:
                print("Error encountered when running command.")
                return False

        # run again for 15 minutes
        print("running for free memory")
        result = self.runCommand("./threaded_memtest -qpv")
        print("done.")
        sys.stdout.flush()
        return (result == 0)

    def runProcesses(self, number, command):
        processes = 0;
        passed = True
        process = []
        for i in range(number):
            processCommand = Command(command)
            processCommand.start()
            process.append(processCommand)
            print("Started: process %u pid %u: %s" % (i, process[i].getPID(), command))
        sys.stdout.flush()
        waiting = True
        while waiting:
            waiting = False
            for i in range(number):
                if process[i]:
                    found = True
                    line = process[i].readline()
                    if line and len(line) > 1:
                        print("process %u pid %u: %s" % (i, process[i].getPID(), line))
                        sys.stdout.flush()
                    returnValue = process[i].poll()
                    if returnValue is None or returnValue == -1:
                        waiting = True
                    else:
                        if returnValue > 0:
                            print("Error: process  %u pid %u returned %u" % (i, process[i].getPID(), returnValue))
                            passed = False
                        print("process %u pid %u returned success" % (i, process[i].getPID()))
                        process[i] = None
        sys.stdout.flush()
        return passed

    def runCommand(self, command):
        if self.mode == Constants.self:
            print("Self Test Mode, not running command")
            return 0
        else:
            sys.stdout.flush()
            return os.system(command)

if __name__ == "__main__":
    memoryTest = MemoryTest()
    returnValue = memoryTest.do(sys.argv)
    sys.exit(returnValue)
