#!/usr/bin/env python
# Copyright (c) 2006 Red Hat, Inc. All rights reserved. This copyrighted material 
# is made available to anyone wishing to use, modify, copy, or 
# redistribute it subject to the terms and conditions of the GNU General 
# Public License v.2.
# 
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# Author: Greg Nichols
#         Petersingh Anburaj
#
from rhcert.test import TestTag, Constants
from rhcert.command import Command, HwCertCommandException, HwCertCommandErrorOutput
from rhcert.tags import DeviceClass
from rhcert.suites.hwcert.hwcert.procDeviceDetector import ProcDeviceDetector
from rhcert.suites.hwcert.hwcert.hardwareTest import HardwareTest
from rhcert.redhatrelease import RedHatRelease


class ProfilerTest(HardwareTest):

    def __init__(self, path):
        HardwareTest.__init__(self, path)
        self.deviceClass = DeviceClass.processor
        self.forcedTimerConfig = False
        self.redHatRelease = RedHatRelease()
        self.profilerType = None

    def logDevices(self, planner):
        devices = list()
        proc = ProcDeviceDetector()
        devices.extend(proc.getDevices(ProcDeviceDetector.procCPUInfo))
        # call the default to log udev devices
        devices.extend(HardwareTest.logDevices(self, planner))
        return devices

    def perf_test(self):
        if not self.profilerType:
            print("Error: Type not found for the profiler test.")
            return False

        if self.profilerType == "hardware":
            event_name = "cycles"
        elif self.profilerType == "software":
            event_name = "cpu-clock"
        else:
            print("Error: Wrong type set for profiler test.")
            return False

        if not self.collect_samples(event_name):
            return False
        if not self.check_event(event_name):
            return False
        return self.check_samples()

    def collect_samples(self, event_name):
        perfRecord = "perf record -a -e %s -o hwcert-perf.data sleep 5" % event_name
        try:
            print("Collecting %s events using the perf command '%s'" % (self.profilerType, perfRecord))
            Command(perfRecord).echoIgnoreWarnings()
        except HwCertCommandException as exception:
            print("Error: Failed to collect %s events using the perf command '%s'" % (self.profilerType, perfRecord))
            print(str(exception))
            return False
        return True

    def check_event(self, event_name):
        perfEvlist = "perf evlist -i hwcert-perf.data"
        try:
            events = Command(perfEvlist).getStringList(regex=".*")
            if event_name in events:
                print("Required %s event found" % self.profilerType)
            else:
                print("Error: Required %s event not available" % self.profilerType)
                return False
        except HwCertCommandErrorOutput:
            pass  # Ignore the stderr if return value is 0
        except HwCertCommandException as exception:
            print("Error: Failed to check events for perf using command '%s'" % perfEvlist)
            print(str(exception))
            return False
        return True

    def check_samples(self):
        perfReport = "perf report -i hwcert-perf.data --stdio"
        try:
            Command(perfReport).getStringList(regex="^\\s+[\\d\\.]+%\\s+.*")
        except HwCertCommandErrorOutput:
            pass  # Ignore the stderr if return value is 0
        except HwCertCommandException as exception:
            print("Error: No samples found. Failed to fetch report using perf command '%s'" % perfReport)
            print(str(exception))
            return False

        print("Samples found for the %s event" % self.profilerType)
        return True

    def run(self):
        return self.runSubTest(self.perf_test, "Perf Test")


class HardwareProfilerTest(ProfilerTest):
    """
    This will be planned if cpu*cycles file is found in /sys/devices/ directory
    Exact path may be different for different archs
    For aarch64 : /sys/devices/armv8_pmuv3_0/events/cpu_cycles
    For other archs: /sys/devices/cpu/events/cpu-cycles
    """
    def __init__(self):
        ProfilerTest.__init__(self, "hw_profiler")
        self.profilerType = "hardware"

    def tags(self):
        return [TestTag.noninteractive, TestTag.certification, TestTag.profiler]

    def plan(self, planner):
        tests = list()
        check_sysfs = "find /sys/devices/* -type f -name 'cpu*cycles'"
        try:
            output = Command(check_sysfs).getString(regex=".*")
            if self.getDebug() != Constants.off:
                print("File found for hw event: %s" % output)
            test = self.makeCopy()
            tests.append(test)
        except HwCertCommandException as exception:
            if self.getDebug() != Constants.off:
                print("Kernel doesn't supports hardware profiling")
        return tests


class SoftwareProfilerTest(ProfilerTest):
    """
    This will be planned only if hardware profiler test was not planned
    """
    def __init__(self):
        ProfilerTest.__init__(self, "sw_profiler")
        self.profilerType = "software"

    def tags(self):
        return [TestTag.noninteractive, TestTag.certification, TestTag.profiler]

    def plan(self, planner):
        tests = list()
        # Check if no other profiler test is planned
        if hasattr(planner, "allTestsSoFar"):
            testsSoFar = planner.allTestsSoFar
            for planned_test in testsSoFar:
                if TestTag.profiler in planned_test.tags():
                    return tests
        test = self.makeCopy()
        tests.append(test)
        return tests
