#!/usr/bin/env python
# Copyright (c) 2008 Red Hat, Inc. All rights reserved. This copyrighted material
# is made available to anyone wishing to use, modify, copy, or
# redistribute it subject to the terms and conditions of the GNU General
# Public License v.2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# Author: Ben Rose <brose@redhat.com>

from functools import partial
from requests import HTTPError, ConnectionError
import json, os, string, sys, socket, re, time, decimal, filecmp, uuid, stat, time, inspect

from rhcert.environment import Environment
from rhcert.tags import Constants, DeviceClass, TestTag
from rhcert.command import Command, HwCertCommandException
from rhcert.collabrequest import CollabRequest
from rhcert.compatability import XmlrpcServerProxy
from rhcert.suites.hwcert.hwcert.hardwareTest import HardwareTest
from rhcert.suites.hwcert.hwcert.procDeviceDetector import ProcDeviceDetector
from rhcert.suites.hwcert.tests.rdmaConnection import rdmaHelper
from rhcert.suites.hwcert.tests.rdmaConnection.rdmaHelper import parseDevices, getNetDevices, matchLinkGidIp6, DevicePort
from rhcert.suites.hwcert.tests.network import network


class RdmaConnectionTest(HardwareTest):

    def __init__(self, path):
        path = "rdma/" + path
        HardwareTest.__init__(self, path)
        self.deviceClass = DeviceClass.network_interface
        self.planWithNetDevices = False
        self.abstract = True
        self.devicesPorts = []
        self.ibBandwidthWarning = 0.75
        self.ibBandwidthFail = 0.5
        self.deviceName = ""
        self.isEthernet = False
        self.serverDeviceName = ""
        self.serverDevicePort = ""
        self.linkLayer = "Ethernet"  # or it could be InfiniBand
        self.driverModules = []  # must be set in child classes
        self.ibstatusOutput = ""
        self.devicePortID = ""
        self.netDevice = ""
        self.baseLID = ""
        self.smLID = ""
        self.linkSpeed = ""
        self.potentialIPAddresses = []
        self.serverIP = ""
        self.serverDevice = ""
        self.subTests = [[self.pingTest, "ping test"],
                         [self.rpingTest, "rping test"],
                         [self.rcopyTest, "rcopy test"],
                         [self.ibWriteBwTest, "ib_write_bw test"],
                         [self.ibReadBwTest, "ib_read_bw test"],
                         [self.ibSendBwTest, "ib_send_bw test"]]
        self.requiredRPMs = ["perftest", "librdmacm-utils", "libibverbs-utils", "infiniband-diags"]  # more included in subclasses

    def tags(self):
        return [TestTag.interactive, TestTag.certification]

    def getRequiredRPMs(self):
        return self.requiredRPMs

    def gatherProperties(self):
        name = self.getLogicalDeviceName()
        if self.planWithNetDevices:
            m = re.match('(?P<device>.*?)_devicePort_(?P<num>\d+)_netDevice_(?P<netDevice>.*$)', name)
            if m:
                self.netDevice = m.group("netDevice")
        if not self.netDevice:
            m = re.match('(?P<device>.*?)_devicePort_(?P<num>\d+)', name)

        self.deviceName = m.group("device")
        self.devicePortID = m.group("num")
        self.devicePorts = parseDevices(self.driverModules)
        for devicePort in self.devicePorts:
            if devicePort.module \
                    and devicePort.deviceName == self.deviceName \
                    and devicePort.portNumber == self.devicePortID:
                self.potentialIPAddresses = devicePort.potentialIPAddresses
                self.smLID = devicePort.attributes.get('sm lid')
                self.linkSpeed = devicePort.attributes.get('rate')
                self.baseLID = devicePort.attributes.get('base lid')

    def verifyTransport(self, deviceName=""):
        """ Override when transport matters for determination."""
        return True

    def plan(self, planner):
        tests = list()
        self.devicePorts = parseDevices(self.driverModules)
        try:
            lsmodOutput = Command("lsmod").getStringList(regex=".*")
        except HwCertCommandException:
            print("Error running lsmod command, no rdma tests planned")
            return tests
        foundDrivers = [x for x in lsmodOutput if re.search("|".join(self.driverModules), x)]
        if not foundDrivers:
            return tests
        for devicePort in self.devicePorts:
            if not self.verifyTransport(devicePort.deviceName):
                continue
            if devicePort.module:
                print("device %s devicePort: %s is %s" % (
                    devicePort.deviceName, devicePort.portNumber, devicePort.attributes.get('state')))
                if "ACTIVE" in devicePort.attributes.get('state') and devicePort.attributes.get(
                        'link_layer') == self.linkLayer:
                    gid = devicePort.attributes.get('default gid')
                    netDevices = getNetDevices(devicePort.module, gid)
                    try:
                        regex = "^(?P<speed>[0-9]+)[^\(]+\((?P<rate_type>[^\)]+).*$"
                        pattern = re.compile(regex)
                        m = pattern.match(devicePort.attributes.get('rate'))
                    except:
                        print("Error: Failed to detect rate.")
                        return tests

                    if self.planWithNetDevices and netDevices:
                        for netDevice in netDevices:
                            self.deviceName = devicePort.deviceName
                            self.devicePortID = devicePort.portNumber
                            if (self.rateContains in m.group("rate_type")) or (self.rateContains == m.group("speed")):
                                test = self.makeCopy()
                                deviceDisplayName = "%s_devicePort_%s_netDevice_%s" % (
                                    test.deviceName, test.devicePortID, netDevice)
                                test.setLogicalDeviceName(deviceDisplayName)
                                tests.append(test)
                    else:
                        if (self.rateContains in m.group("rate_type")) or (self.rateContains == m.group("speed")):
                            self.deviceName = devicePort.deviceName
                            self.devicePortID = devicePort.portNumber
                            test = self.makeCopy()
                            deviceDisplayName = "%s_devicePort_%s" % (test.deviceName, test.devicePortID)
                            test.setLogicalDeviceName(deviceDisplayName)
                            tests.append(test)
        return tests

    def printDevices(self):
        for devicePort in self.devicePorts:
            print("Device name: %s" % devicePort.deviceName)
            print("Port number: %s" % devicePort.portNumber)
            print("Port state: %s" % devicePort.attributes.get('state'))
            print("Port link layer: %s" % devicePort.attributes.get('link_layer'))
            print("-----")
        print("_______________________")

    # ALL rcopy related functions below except getRcopymd5sum require root because that's the user running client

    @staticmethod
    def opapmaQuery(**kwargs):
        outputFilePath = "/tmp/queryoutput.txt"
        # check that opa-basic-tools is installed.  If not, write that out to output
        installedOutput = Command("rpm -q opa-basic-tools").getString(ignoreErrors=True)
        if "not installed" in installedOutput:
            errorString = "Error: Failed on LTS, \n%s" % installedOutput
            with open(outputFilePath, "w") as f:
                f.write(errorString)
            return json.dumps({"output": errorString, })

        lID = kwargs.get('lID')
        portNumber = kwargs.get('portNumber')
        if not lID:
            return None
        try:
            commandString = "opapmaquery -l %s -m %s 2>&1 > %s" % (lID, portNumber, outputFilePath)
            output = Command(commandString).getString(regex="^.*$", ignoreErrors=True)
        except HwCertCommandException:
            print("Error running opapmaquery")
            return False
        return json.dumps({"output": output, })

    @staticmethod
    def smpQuery(**kwargs):
        lID = kwargs.get('lID')
        portNumber = kwargs.get('portNumber')
        deviceName = kwargs.get('deviceName')
        if not lID:
            return None
        try:
            commandString = "smpquery ND -L %s -P%s -C %s 2>&1 > /tmp/queryoutput.txt" % (lID, portNumber, deviceName)
            print("SMPQuery command being used: %s" % commandString)
            output = Command(commandString).getString(regex="^.*$", ignoreErrors=True)
        except HwCertCommandException:
            print("Error running smpquery")
            return False
        return json.dumps({"output": output, })

    @staticmethod
    def getQueryResults():
        environment = Environment()
        server = XmlrpcServerProxy(('http://127.0.0.1:%s' % environment.getServerListenerPort()), allow_none=True)
        output = None
        for x in range(5):
            try:
                output = server.getFileContent("queryoutput.txt")
                break
            except Exception as e:
                time.sleep(2)
                pass
        return json.dumps({"output": output})

    @staticmethod
    def removeQueryFiles():
        try:
            Command("rm /tmp/queryoutput.txt").run()
        except HwCertCommandException:
            print("Error: Failed to remove query related files")
            return False
        return True

    @staticmethod
    def startIbBwCommand(command, **kwargs):
        device = kwargs.get("device", None)
        portNumber = kwargs.get("portNumber", None)
        ethernet = kwargs.get("ethernet", None)
        if device:
            command = "%s -d %s" % (command, device)
            if portNumber:
                command = "%s -i %s" % (command, portNumber)
        if ethernet:
            command = "%s -R" % command
        success = False
        try:
            for x in range(5):
                Command(command).start()
                time.sleep(5)
                if len(Command('ps aux | grep "%s" 2>&1' % command).getStringList(regex='.*', ignoreErrors=False)) > 1:
                    success = True
                    break
                else:
                    print("Attempt to run %s was not sucessful" % command)
        except HwCertCommandException as e:
            print("Error: Failed to run %s command" % command)
            print(e)
            return False
        if success:
            print("Successfully started %s on LTS" % command)
        else:
            print("Error: Failed to run %s command" % command)
        return success

    @staticmethod
    def startIbReadBw(**kwargs):
        return RdmaConnectionTest.startIbBwCommand("ib_read_bw", **kwargs)

    @staticmethod
    def startIbWriteBw(**kwargs):
        return RdmaConnectionTest.startIbBwCommand("ib_write_bw", **kwargs)

    @staticmethod
    def startIbSendBw(**kwargs):
        return RdmaConnectionTest.startIbBwCommand("ib_send_bw", **kwargs)

    @staticmethod
    def startRcopy():
        success = False
        try:
            for x in range(5):
                Command('rcopy').start()
                time.sleep(5)
                try:
                    Command('ps -C rcopy').run()
                    success = True
                    break
                except HwCertCommandException:
                    print("Attempt to run rcopy was not successful")
        except HwCertCommandException:
            print("Error: Failed to run rcopy command")
            return False
        if success:
            print("Successfully started rcopy on LTS")
        else:
            print("Error: Failed to run rcopy command")
        return success

    @staticmethod
    def stopRcopy():
        try:
            Command('pkill -9 -f rcopy').run()
        except HwCertCommandException:
            print("Error: Failed to stop rcopy command")
            return False
        return True

    @staticmethod
    def md5sumRcopy():
        isRcopyFilePresent = False
        isFileWritten = False
        isEmpty = True
        md5sumPath = "/tmp/md5sum.txt"
        rcopyFilePath = "/tmp/rcopiedFile.txt"
        for x in range(5):
            isRcopyFilePresent = os.path.isfile(rcopyFilePath)
            if not isRcopyFilePresent:
                time.sleep(2)
                continue
            isMd5FileWritten = os.path.isfile(md5sumPath)
            if isMd5FileWritten:
                isEmpty = os.path.getsize(md5sumPath) == 0
            if isMd5FileWritten and not isEmpty:
                break
            try:
                Command("md5sum %s 2>&1 > %s" % (rcopyFilePath, md5sumPath)).run()
            except HwCertCommandException as e:
                time.sleep(1)
                print("Warning - %s" % str(e))
                pass
        if not isRcopyFilePresent:
            print("Error: %s not found" % rcopyFilePath)
            Command("echo 'Error: %s not found' > %s" % (rcopyFilePath, md5sumPath)).echoIgnoreErrors()
            return False
        if not isMd5FileWritten:
            print("Error: could not save md5sum to file")
            return False
        if isEmpty:
            print("Error: md5sum not written to file")
            Command("echo 'Error: md5sum not written to file' > %s" % md5sumPath).echoIgnoreErrors()
            return False
        return True

    @staticmethod
    def getRcopymd5sum():
        environment = Environment()
        server = XmlrpcServerProxy(('http://127.0.0.1:%s' % environment.getServerListenerPort()), allow_none=True)
        md5sum = None
        # makes certain no new text is being added to the file by waiting a second and comparing values before accepting result
        for x in range(5):
            try:
                md5sum_new = server.getFileContent("md5sum.txt")
                print(md5sum_new)
                if md5sum_new and md5sum == md5sum_new:
                    break
                else:
                    md5sum = md5sum_new
                    time.sleep(1)
            except Exception as e:
                time.sleep(1)
                print("Warning - %s" % str(e))
                pass
        return json.dumps({"output": md5sum})

    @staticmethod
    def removeRcopyFiles():
        try:
            Command("rm /tmp/rcopiedFile.txt").echoIgnoreErrors()
            Command("rm /tmp/md5sum.txt").echoIgnoreErrors()
        except HwCertCommandException:
            print("Error: Failed to remove rcopy related files")
            return False
        return True

    @staticmethod
    def startRping(**kwargs):
        """ Starts rping command using serverIP and writing uuid on first line of output file provided in kwargs.  Requires root in order to run. """
        serverIP = kwargs.get('serverIP')
        print("SERVER IP: %s" % serverIP)
        uniqID = kwargs.get('uniqID')
        print("UNIQID: %s" % uniqID)
        try:
            Command('echo "%s" > /tmp/rping.log' % uniqID).run()
        except HwCertCommandException:
            print("Error: Failed to write uuid to test's rping log")
            return False
        try:
            rpingCommand = "rping -s -a %s -v -C 50 2>&1 >> /tmp/rping.log" % serverIP
            for x in range(5):
                Command(rpingCommand).start()
                time.sleep(5)
                try:
                    Command('ps -C rping').run()
                    success = True
                    break
                except HwCertCommandException:
                    print("ERROR: Attempt to run %s was not sucessful" % rpingCommand)
        except HwCertCommandException:
            print("Error starting rping")
            return False
        return os.path.isfile("/tmp/rping.log")

    @staticmethod
    def getRpingLog():
        environment = Environment()
        server = XmlrpcServerProxy(('http://127.0.0.1:%s' % environment.getServerListenerPort()), allow_none=True)
        rpingoutput = server.getFileContent("rping.log")
        print(rpingoutput)
        return json.dumps({"output": rpingoutput})

    @staticmethod
    def getAllDevices(driverModules=None):
        """ Returns list of strings for devices, each with ip4, name, rdma device name and port number"""
        try:
            results = []
            devices = rdmaHelper.parseDevices(driverModules)

            for netDevice in rdmaHelper.collectIpDevices():
                if not netDevice.ip4.strip():
                    continue
                for devicePort in devices:
                    gid = devicePort.attributes.get('default gid')
                    if matchLinkGidIp6(link=netDevice.link, gid=gid, ip6=netDevice.ip6):
                        deviceItem = "%s %s Device=%s Port=%s" % (netDevice.ip4, netDevice.name, devicePort.deviceName, devicePort.portNumber)
                        results.append(deviceItem)

        except HwCertCommandException:
            print("Error: Could not collect any devices")
            return None
        return json.dumps({"Devices": results, })

    @staticmethod
    def getAllIPs():
        ip_addrs = []
        try:
            ip_addrs = Command("ip addr").getStringList(regex="inet .*", ignoreErrors=True)
        except HwCertCommandException:
            print("Error collecting ip addresses")
        return json.dumps({"IPs": ip_addrs, })

    @staticmethod
    def getRdmaIPs(**kwargs):
        driverModules = kwargs.get('driverModules')
        body = {}
        ipAddresses = []
        for devicePort in parseDevices(driverModules):
            ipAddresses = ipAddresses + devicePort.potentialIPAddresses
        body["IPs"] = ipAddresses
        return json.dumps(body)

    def collabCall(self, collab_request):

        tries = 5
        for attempt in range(tries):
            try:
                print("attempt #%s to request from LTS - %s" % (attempt+1, collab_request.function_name))
                collab_response = super(HardwareTest, self).collabCall(collab_request)
                print("success")
                break
            except (HTTPError, ConnectionError) as e:
                print("no success")
                if attempt == tries-1:
                    print("Error: %s" % str(e))
                    return None
                else:
                    time.sleep(2)

        return collab_response

    def rcopyStart(self):
        test_name = "RdmaConnectionTest"
        test_package_trail = "hwcert.tests.rdmaConnection.rdmaConnection"
        function_name = "startRcopy"
        requires_root = True

        collab_request = CollabRequest(test_name, test_package_trail, function_name, requires_root)
        return self.collabCall(collab_request)

    def ibSendBwStart(self):
        test_name = "RdmaConnectionTest"
        test_package_trail = "hwcert.tests.rdmaConnection.rdmaConnection"
        function_name = "startIbSendBw"
        requires_root = True
        kwargs = {"device": self.serverDeviceName, "portNumber": self.serverDevicePort, "ethernet": self.isEthernet}

        collab_request = CollabRequest(test_name, test_package_trail, function_name, requires_root, kwargs=kwargs)
        return self.collabCall(collab_request)

    def ibWriteBwStart(self):
        test_name = "RdmaConnectionTest"
        test_package_trail = "hwcert.tests.rdmaConnection.rdmaConnection"
        function_name = "startIbWriteBw"
        requires_root = True
        kwargs = {"device": self.serverDeviceName, "portNumber": self.serverDevicePort, "ethernet": self.isEthernet}

        collab_request = CollabRequest(test_name, test_package_trail, function_name, requires_root, kwargs=kwargs)
        return self.collabCall(collab_request)

    def ibReadBwStart(self):
        test_name = "RdmaConnectionTest"
        test_package_trail = "hwcert.tests.rdmaConnection.rdmaConnection"
        function_name = "startIbReadBw"
        requires_root = True
        kwargs = {"device": self.serverDeviceName, "portNumber": self.serverDevicePort, "ethernet": self.isEthernet}

        collab_request = CollabRequest(test_name, test_package_trail, function_name, requires_root, kwargs=kwargs)
        return self.collabCall(collab_request)

    def rcopyStop(self):
        test_name = "RdmaConnectionTest"
        test_package_trail = "hwcert.tests.rdmaConnection.rdmaConnection"
        function_name = "stopRcopy"
        requires_root = True

        collab_request = CollabRequest(test_name, test_package_trail, function_name, requires_root)
        return self.collabCall(collab_request)

    def rcopyMd5sum(self):
        test_name = "RdmaConnectionTest"
        test_package_trail = "hwcert.tests.rdmaConnection.rdmaConnection"
        function_name = "md5sumRcopy"
        requires_root = True

        collab_request = CollabRequest(test_name, test_package_trail, function_name, requires_root)
        return self.collabCall(collab_request)

    def rcopyRemoveFiles(self):
        test_name = "RdmaConnectionTest"
        test_package_trail = "hwcert.tests.rdmaConnection.rdmaConnection"
        function_name = "removeRcopyFiles"
        requires_root = True

        collab_request = CollabRequest(test_name, test_package_trail, function_name, requires_root)
        return self.collabCall(collab_request)

    def queryGetResults(self):

        test_name = "RdmaConnectionTest"
        test_package_trail = "hwcert.tests.rdmaConnection.rdmaConnection"
        function_name = "getQueryResults"

        collab_request = CollabRequest(test_name, test_package_trail, function_name)
        collab_response = self.collabCall(collab_request)

        results = ""
        try:
            collab_body = json.loads(collab_response)
            results = collab_body["output"]
        except TypeError as e:
            print("Error: %s" % str(e))
            return None
        return results

    def queryRemoveFiles(self):
        test_name = "RdmaConnectionTest"
        test_package_trail = "hwcert.tests.rdmaConnection.rdmaConnection"
        function_name = "removeQueryFiles"
        requires_root = True

        collab_request = CollabRequest(test_name, test_package_trail, function_name, requires_root)
        return self.collabCall(collab_request)

    def rcopyGetMd5sum(self):
        test_name = "RdmaConnectionTest"
        test_package_trail = "hwcert.tests.rdmaConnection.rdmaConnection"
        function_name = "getRcopymd5sum"

        collab_request = CollabRequest(test_name, test_package_trail, function_name)
        collab_response = self.collabCall(collab_request)

        results = ""
        try:
            collab_body = json.loads(collab_response)
            results = collab_body["output"]
        except TypeError as e:
            print("Error: %s" % str(e))
            return None
        return results

    def getRpingOutput(self):
        test_name = "RdmaConnectionTest"
        test_package_trail = "hwcert.tests.rdmaConnection.rdmaConnection"
        function_name = "getRpingLog"

        collab_request = CollabRequest(test_name, test_package_trail, function_name)
        collab_response = self.collabCall(collab_request)

        results = ""
        try:
            collab_body = json.loads(collab_response)
            results = collab_body["output"]
        except TypeError as e:
            print("Error: %s" % str(e))
            return None
        return results

    def startServerRping(self, uniqID, serverIP):
        test_name = "RdmaConnectionTest"
        test_package_trail = "hwcert.tests.rdmaConnection.rdmaConnection"
        function_name = "startRping"
        requires_root = True

        collab_request = CollabRequest(test_name, test_package_trail, function_name, requires_root,
                                       kwargs={"uniqID": uniqID, "serverIP": serverIP})

        return self.collabCall(collab_request)

    def smpQueryLID(self, baseTenLID, portNumber, deviceName):
        return self.baseQueryLID(baseTenLID, portNumber, deviceName, "smpQuery")

    def opapmaQueryLID(self, baseTenLID, portNumber, deviceName):
        return self.baseQueryLID(baseTenLID, portNumber, deviceName, "opapmaQuery")

    def baseQueryLID(self, baseTenLID, portNumber, deviceName, function_name=""):
        queryLIDFunctions = ["smpQuery", "opapmaQuery"]
        if function_name not in queryLIDFunctions:
            print("Error: Wrong function name provided for baseQueryLID function")
            return False
        test_name = "RdmaConnectionTest"
        test_package_trail = "hwcert.tests.rdmaConnection.rdmaConnection"
        requires_root = True

        collab_request = CollabRequest(test_name, test_package_trail, function_name, requires_root,
                                       kwargs={"lID": baseTenLID, "portNumber": portNumber, "deviceName": deviceName})

        return self.collabCall(collab_request)

    def getServerAllDevices(self):
        test_name = "RdmaConnectionTest"
        test_package_trail = "hwcert.tests.rdmaConnection.rdmaConnection"
        function_name = "getAllDevices"

        collab_request = CollabRequest(test_name, test_package_trail, function_name,
                                       kwargs={"driverModules": self.driverModules})
        try:
            collab_response = self.collabCall(collab_request)
        except HTTPError as e:
            print("Error: %s" % str(e))
            return None

        devices = ""
        try:
            collab_body = json.loads(collab_response)
            devices = collab_body["Devices"]
        except TypeError as e:
            print("Error: %s" % str(e))
            return None
        return devices

    def getServerAllIPs(self):
        test_name = "RdmaConnectionTest"
        test_package_trail = "hwcert.tests.rdmaConnection.rdmaConnection"
        function_name = "getAllIPs"

        collab_request = CollabRequest(test_name, test_package_trail, function_name)
        collab_response = self.collabCall(collab_request)

        ipAddresses = None
        try:
            collab_body = json.loads(collab_response)
            ipAddresses = collab_body["IPs"]
        except TypeError as e:
            print("Error: %s" % str(e))
            return None

        return ipAddresses

    def verifyIpv4(self):
        if self.netDevice:
            try:
                deviceIP = Command("ip addr show %s" % self.netDevice).getString(
                    regex=".*inet (?P<ipAddress>(\d{1,3}\.){3}\d{1,3})", regexGroup="ipAddress", singleLine=False,
                    ignoreErrors=True)
            except HwCertCommandException:
                print("Error: ip addr show %s has no inet configured" % self.netDevice)
                return False
        return True

    def rpingTest(self):
        """ Run rping on LTS and SUT, compare results and uuid."""
        # check to make certain there is an ipv4 address on the device being tested before proceeding
        if not self.verifyIpv4():
            return False

        uniqID = str(uuid.uuid4())
        print("Testing rping with self given uuid %s with IP address: %s" % (uniqID, self.serverIP))
        self.startServerRping(uniqID, self.serverIP)
        time.sleep(5)
        print("rping started on LTS: rping -s -a %s -v -C 50" % self.serverIP)
        # call rping on SUT and get the results in list
        print("rping on SUT: timeout 10s rping -c -a %s -v -C 50" % self.serverIP)
        try:
            sut_rping_output = Command("timeout 10s rping -c -a %s -v -C 50 2>&1" % self.serverIP).getStringList(
                regex=".*", ignoreErrors=True)
        except HwCertCommandException:
            print("Error running rping on SUT")
            return False
        print("rping ran on SUT.  Verifying uuid and data.")
        # get output from LTS rping
        time.sleep(2)
        lts_rping_output = self.getRpingOutput()
        # compare uuid
        if not lts_rping_output:
            print("ERROR: could not find rping output from LTS")
            return False
        lts_uuid = lts_rping_output.split("\n")[0]
        lts_uuid.strip()
        print("uuid in test's rping log on LTS: %s" % lts_uuid)
        if lts_uuid == uniqID:
            print("uuid matches up.")
        else:
            print("ERROR: didn't find the corresponding rping output file for verification")
            return False
        for line in sut_rping_output:
            if "error" in line or "ERROR" in line:
                print("Error: rping failed in SUT.")
                return False
            elif "ping data" in line and not line in lts_rping_output:
                print("Error: rping output on LTS and SUT don't match.")
                print("This line is in SUT output, but not LTS:\n%s" % line)
                return False
        print("Data matches up as well.  rping ran successfully on lts and sut.")
        return True

    def pingTest(self):
        """ Run ping, and verify that expected ip address appears in route"""
        deviceIP = None
        if self.netDevice:
            try:
                deviceIP = Command("ip addr show %s" % self.netDevice).getString(
                    regex=".*inet (?P<ipAddress>(\d{1,3}\.){3}\d{1,3})", regexGroup="ipAddress", singleLine=False,
                    ignoreErrors=True)
            except HwCertCommandException:
                print("Warning - ip addr show %s has no inet configured" % self.netDevice)
        if not deviceIP:
            if self.potentialIPAddresses:
                deviceIP = self.potentialIPAddresses[0]
            else:
                print("Error, could not determine device IP address")
                return False
        pingCommandString = "ping -R -v -I %s -c 1 %s" % (deviceIP, self.serverIP)
        print("Running:\n%s" % pingCommandString)
        try:
            rawOutput = Command(pingCommandString).getIgnoreErrors()
            if not rawOutput:
                print("Error: Failed to run ping on SUT")
                return False
            pingOutput = '\n'.join(rawOutput)
        except HwCertCommandException:
            print("Error: Failed to run ping on SUT")
            return False
        if 'Error' in pingOutput or 'errors' in pingOutput:
            print("Ping resulted in Error")
            return False
        elif '0 received' in pingOutput:
            print("No packets received")
            return False
        else:
            print("Ping was successful, validating route to ensure source ip address was as expected.")
            return True

    def parseAverageBandwidth(self, ib_output):
        if not ib_output:
            print("Error: could not find output from command")
            return 0
        content = [x.strip() for x in ib_output]
        values = {}

        for index, line in enumerate(content):
            if "BW average[MB/sec]" in line:
                break

        try:
            values = dict(zip(re.split("  +", content[index]), re.split("  +", content[index + 1])))
        except IndexError:
            print("Error: could not parse output from command")
            return 0

        averageBandwidth = 0
        try:
            averageBandwidth = float(values["BW average[MB/sec]"].strip()) * 8 / 1000
        except KeyError:
            print("Error: could not find average bandwidth to compare with expected")
        return averageBandwidth

    def verifyLinkSpeed(self, averageBandwidth, comparing):
        """ Takes output from ib_*_bw command and compares the average bandwidth with the expected link speed """
        if not comparing:
            return True
        return averageBandwidth >= float(self.linkSpeed.split(" ")[0]) * comparing

    def performBwTest(self, command, serverFunction):
        # check to make certain there is an ipv4 address on the device being tested before proceeding
        if not self.verifyIpv4():
            return False

        time.sleep(15)
        print("starting %s command on SUT" % command)
        if self.serverDeviceName:
            print("using server side device named %s" % self.serverDeviceName)
        else:
            print(
                "Error: Server side device is not applicable.  Please review selected server side device and try again.")
            return False
        if self.serverDevicePort:
            print("using server side port id: %s" % self.serverDevicePort)
        serverFunction()
        time.sleep(30)
        serverIPNoPort = self.serverIP  # .split(":")[0]
        if self.deviceName:
            command = "%s -d %s" % (command, self.deviceName)
            if self.devicePortID:
                command = "%s -i %s" % (command, self.devicePortID)
        if self.isEthernet:
            command = "%s -R" % command
        command = "%s %s" % (command, serverIPNoPort)
        print("starting %s command on SUT" % command)
        print("Running %s on SUT" % command)
        bwCommand = Command(command)

        output = ""
        for x in range(5):
            try:
                output = bwCommand.getIgnoreErrors()
                break
            except HwCertCommandException:
                time.sleep(3)
        print(output)
        if not output:
            print("ERROR: could not gather output from %s" % command)
            return False
        averageBandwidth = self.parseAverageBandwidth(output)
        if averageBandwidth:
            if self.ibBandwidthWarning and self.verifyLinkSpeed(averageBandwidth, self.ibBandwidthWarning):
                print("Average bandwidth: %s Gbs falls within %s margin of %s" % (
                    averageBandwidth, self.ibBandwidthWarning, self.linkSpeed))
                return True
            elif self.ibBandwidthFail and self.verifyLinkSpeed(averageBandwidth, self.ibBandwidthFail):
                print("Average bandwidth: %s Gbs falls within %s margin of %s" % (
                    averageBandwidth, self.ibBandwidthFail, self.linkSpeed))
                return True
            elif not self.verifyLinkSpeed(averageBandwidth, self.ibBandwidthFail):
                print("Error: Average bandwidth of: %s Gbs did not fall within %s margin of %s" % (
                    averageBandwidth, self.ibBandwidthFail, self.linkSpeed))
                return False
            else:
                print("Warning - Average bandwidth of: %s Gbs did not fall within %s margin of %s" % (
                    averageBandwidth, self.ibBandwidthWarning, self.linkSpeed))
                return True
        else:
            return False

    def ibWriteBwTest(self):
        return self.performBwTest('ib_write_bw', self.ibWriteBwStart)

    def ibSendBwTest(self):
        return self.performBwTest('ib_send_bw', self.ibSendBwStart)

    def ibReadBwTest(self):
        return self.performBwTest('ib_read_bw', self.ibReadBwStart)

    def rcopyTest(self):
        # check to make certain there is an ipv4 address on the device being tested before proceeding
        if not self.verifyIpv4():
            return False

        print("removing any files from previous runs")
        self.rcopyRemoveFiles()
        # rcopy using random binary file, comparing md5sums
        # create random file (overwrite existing)
        print("Creating random binary file...")
        with open("/tmp/rcopiedFile.txt", "wb") as f:
            f.write(os.urandom(1024 * 50))
        try:
            sut_rcopy_md5 = Command("md5sum /tmp/rcopiedFile.txt 2>&1").getString()
            sut_rcopy_md5 = sut_rcopy_md5
        except HwCertCommandException:
            print("Error: Could not md5sum rcopy file in SUT")
            return False
        print("Starting rcopy on LTS")
        self.rcopyStart()
        time.sleep(5)
        print("Running rcopy, copying file from SUT")
        try:
            rcopyCommand = Command("rcopy /tmp/rcopiedFile.txt %s 2>&1" % self.serverIP)
            output = rcopyCommand.getIgnoreErrors()
            print(output)
        except HwCertCommandException as e:
            print("Error: Failed to run rcopy on SUT")
            print(str(e))
            return False
        time.sleep(5)
        print("stopping rcopy on LTS")
        self.rcopyStop()
        time.sleep(1)
        print("calculating md5sum on LTS")
        self.rcopyMd5sum()
        print("getting md5sum from LTS")
        time.sleep(1)
        lts_rcopy_md5 = self.rcopyGetMd5sum()
        print("Comparing md5sum from SUT and LTS")
        print("SUT md5: %s" % sut_rcopy_md5)
        print("LTS md5: %s" % lts_rcopy_md5)
        if not lts_rcopy_md5:
            print("Error in gathering rcopy file md5sum")
            return False
        if sut_rcopy_md5.strip() != lts_rcopy_md5.strip():
            print("ERROR: rcopy file on LTS does not match the one on SUT in md5sum comparison")
            return False
        return True

    def getServiceStatus(self, serviceName):
        statusCommand = Command("systemctl status %s" % serviceName)
        statusRegex = "^.*Active:\s+(?P<status>.*?)\s"
        try:
            currentStatus = statusCommand.getString(regex=statusRegex, regexGroup="status", ignoreErrors=True,
                                                    singleLine=False)
            return currentStatus
        except HwCertCommandException:
            print("Error: %s service could not be found or status parsed" % serviceName)
            return None

    def startService(self, serviceName):
        print("Starting %s service..." % serviceName)
        startCommand = Command("systemctl start %s" % serviceName)
        try:
            startCommand.run()
            return True
        except HwCertCommandException:
            print("Error: issue starting %s service" % serviceName)
            return False

    def stopService(self, serviceName):
        print("Stopping %s service..." % serviceName)
        stopCommand = Command("systemctl stop %s" % serviceName)
        try:
            stopCommand.run()
            return True
        except HwCertCommandException:
            print("Error: issue stopping %s service" % serviceName)
            return False

    def restartService(self, serviceName):
        print("Restarting %s service..." % serviceName)
        restartCommand = Command("systemctl restart %s" % serviceName)
        try:
            restartCommand.run()
            return True
        except HwCertCommandException:
            print("Error: issue restarting %s service" % serviceName)
            return False

    def serviceTest(self, serviceName, acceptFailAsStopped=False, versionReference="version needed here"):
        currentStatus = self.getServiceStatus(serviceName)
        if not currentStatus:
            return False
        print("Initial status of %s is: %s" % (serviceName, currentStatus))

        if not self.stopService(serviceName):
            return False

        currentStatus = self.getServiceStatus(serviceName)
        if not currentStatus:
            return False
        if currentStatus.upper() == "INACTIVE" or (currentStatus.upper() == "FAILED" and acceptFailAsStopped):
            pass
        else:
            print("Error: stopping service did not work.  Service Status: %s" % currentStatus)
            return False
        print("Status after stop: %s" % currentStatus)
        if currentStatus.upper() == "FAILED" and acceptFailAsStopped:
            print(
                "NOTE: Failed status reported after calling stop on the service is acceptable for versions prior to %s" % versionReference)

        if not self.startService(serviceName):
            print("Error: issue starting %s service" % serviceName)
            return False
        currentStatus = self.getServiceStatus(serviceName)
        if not currentStatus:
            return False
        if currentStatus.upper() != "ACTIVE":
            print("Error: starting service did not work.  Service Status: %s" % currentStatus)
            return False
        print("Status after start: %s" % currentStatus)

        if not self.restartService(serviceName):
            print("Error: issue restarting %s service" % serviceName)
            return False
        currentStatus = self.getServiceStatus(serviceName)
        if not currentStatus:
            return False
        if currentStatus.upper() != "ACTIVE":
            print("Error: restarting service did not work.  Service Status: %s" % currentStatus)
            return False
        print("Status after restart: %s" % currentStatus)
        return True

    def linkSpeedTest(self, expected="100"):
        print("Verifying link speed is %s" % expected)
        print("Link Speed: %s" % self.linkSpeed)
        if expected in self.linkSpeed:
            print("Valid - link speed is as expected")
            return True
        else:
            print("Error: link speed is not as expected")
            return False

    def lIDTest(self):
        result = True
        print("Verifying base LID is different than subnet manager LID")
        print("base LID: %s" % self.baseLID)
        print("subnet manager LID: %s" % self.smLID)
        if self.baseLID == self.smLID:
            print("Error: base LID is the same as the subnet manager LID")
            result = False
        print("Verifying base LID is not 0xFFFF")
        if self.baseLID == "0xFFFF":
            print("Error: baseLID is not set (having default value of 0xFFFF")
            result = False
        else:
            print("Valid - baseLID has been set")
        return result

    def smpQueryTest(self):
        return self.baseQueryTest(self.smpQueryLID, testname="smpquery")

    def opapmaQueryTest(self):
        return self.baseQueryTest(self.opapmaQueryLID, testname="opapmaquery")

    def baseQueryTest(self, queryFunction=None, testname="query"):
        """Test capability for LTS to query for node on SUT"""
        if not queryFunction:
            print("Error: no command string provided for query test")
            return False
        try:
            baseTenLID = int(self.baseLID, 16)
        except ValueError:
            print("Error: Failed to convert base LID: %s to int" % self.baseLID)
            return False
        print("Starting %s test" % testname)

        output = queryFunction(baseTenLID, self.devicePortID, self.serverDeviceName)
        time.sleep(5)
        output = self.queryGetResults()
        self.queryRemoveFiles()
        if output:
            # parse output and verify there was not any failure
            # note the output of smpquery and opapmquery include 'error' as part of the standard output
            if "failed" in output.lower():
                print("Error: %s failed with output:\n%s" % (testname, output))
                return False
            else:
                print("Success querying SUT node from LTS:")
                print("Output:\n%s" % output)
                return True
        else:
            print("Failed to query SUT node from LTS")
            return False

    def ethtoolTest(self):
        # run ethtool <device> and verify results do not contain 'No such device'
        if not self.netDevice:
            print("Error: no net device found")
            return False
        print("Running ethtool %s" % self.netDevice)
        ethtoolsCommand = Command("ethtool %s" % self.netDevice)
        output = None
        try:
            output = ethtoolsCommand.getStringList(regex=".*", ignoreErrors=True)
            print("\n".join(output))
        except HwCertCommandException:
            print("Error running command")
            return False
        if not output:
            return False
        for line in output:
            if 'No such device' in output:
                print("Error: ethtool did not find the device %s" % self.netDevice)
                return False
        return True

    def rdmaNddTest(self):
        try:
            status = Command("systemctl status rdma-ndd").getString(regex="^.*Active:\s+(?P<status>.*?)\s",
                                                                    regexGroup="status", ignoreErrors=True,
                                                                    singleLine=False)
        except HwCertCommandException:
            print("Error: failed to attain rdma-ndd status")
            return False
        if status.upper() != "ACTIVE":
            print("Error: rdma-ndd status is not active, but %s" % status)
            return False
        print("rdma-ndd status is %s" % status)
        return True

    def run(self):

        result = True
        self.gatherProperties()
        print("device name: %s" % self.deviceName)
        print("port ID: %s" % self.devicePortID)


        if not self.serverDevice:
            # call remote function to get server's possible ip addresses
            serverDevices = self.getServerAllDevices()
            if not serverDevices:
                print("Error, failed to gather ip choices from LTS")
                return False
            self.serverDevice = self.promptEdit(
                "Which ip address (LTS) would you like to test %s with: " % self.getLogicalDeviceName(), serverDevices[0],
                serverDevices)
            if not self.serverDevice:
                print("Error, selected serverDevice is %s" % serverDevice)
                return False
        serverDeviceSplit = self.serverDevice.split(" ")
        if len(serverDeviceSplit) < 2:
            print("Error, expected format: inet <ip address>, was %s" % serverDeviceSplit[0])
            return False
        serverIP = serverDeviceSplit[0]
        serverIP = serverIP.split("/")[0]

        print("Using IP address of: %s" % serverIP)
        self.serverIP = serverIP

        # run ethernet test
        if self.netDevice and self.ethernetTestClass:
            ethernetTest = self.ethernetTestClass()
            ethernetTest.setLogicalDeviceName(self.netDevice)
            #CERTPX-5953 - need to use ifup/ifdown for rdma tech, but not for all networking tests
            ethernetTest.interfaceConnect = "ifup"
            ethernetTest.interfaceDisconnect = "ifdown"
            ethernetTest.testParameters.set(Constants.testserver, self.serverIP)
            result = ethernetTest.run()
        else:
            if not self.netDevice:
                print("Warning - no net device found, not performing ethernet test")
            if not self.ethernetTestClass:
                print("Warning - no ethernet test class defined for this rdma test")

        # retrieve serverDevice and serverPort if present
        # these are already split by spaces
        for item in serverDeviceSplit:
            if "Device=" in item:
                itemSplit = item.split("=")
                if len(itemSplit) > 1:
                    self.serverDeviceName = itemSplit[-1]
            if "Port=" in item:
                itemSplit = item.split("=")
                if len(itemSplit) > 1:
                    self.serverDevicePort = itemSplit[-1]

        for test in self.subTests:
            if not self.runSubTest(test[0], test[1]):
                result = False
        return result


class InfinibandConnectionTest(RdmaConnectionTest):
    def __init__(self, path):
        RdmaConnectionTest.__init__(self, path)
        self.ethernetTestClass = None  # each subclass needs this, such as network.Ethernet1GigTest for example
        self.linkLayer = "InfiniBand"
        self.driverModules = ['Mlx4_ib', 'mlx4_en', 'mlx4_core', 'mlx5_ib', 'mlx5_core', 'ib_qib']
        self.subTests.append([partial(self.serviceTest, "rdma-ndd"), "rdma-ndd service test"])
        self.subTests.append([partial(self.serviceTest, "opensm"), "opensm service test"])
        self.subTests.append([self.lIDTest, "LID test"])
        self.subTests.append([self.smpQueryTest, "smpquery test"])
        self.planWithNetDevices = True
        self.requiredRPMs.append("opensm")
        self.isEthernet = False
        self.bandwidthTarget = 0.35  # 35% - override default for ethernet test
        self.ibBandwidthWarning = None
        self.ibBandwidthFail = 0.75


class InfinibandQDRTest(InfinibandConnectionTest):
    def __init__(self):
        InfinibandConnectionTest.__init__(self, "Infiniband_QDR")
        self.rateContains = "QDR"
        self.ethernetTestClass = network.Ethernet40GigTest


class InfinibandFDRTest(InfinibandConnectionTest):
    def __init__(self):
        InfinibandConnectionTest.__init__(self, "Infiniband_FDR")
        self.rateContains = "FDR"
        self.ethernetTestClass = network.Ethernet50GigTest


class InfinibandEDRTest(InfinibandConnectionTest):
    def __init__(self):
        InfinibandConnectionTest.__init__(self, "Infiniband_EDR")
        self.rateContains = "EDR"
        self.ethernetTestClass = network.Ethernet100GigTest

class InfinibandHDRTest(InfinibandConnectionTest):
    def __init__(self):
        InfinibandConnectionTest.__init__(self, "Infiniband_HDR")
        self.rateContains = "HDR"
        self.ethernetTestClass = network.Ethernet200GigTest

class InfinibandSocketDirectTest(InfinibandConnectionTest):
    """ At present, only Infiniband supports socket direct.  In the event of other RDMA technologies support, this 
        class can easily be modified into a base class.
        This test must not be planned automatically, but rather manually, as there is at present no discernable
        differentiation in two separate devices and one device connected via socket direct.
    """
    def __init__(self):
        InfinibandConnectionTest.__init__(self, "Infiniband_Socket_Direct")
        self.deviceClass = DeviceClass.network_interface
        self.firstNetDevice = None
        self.firstDeviceName = None
        self.firstDevicePortID = None
        self.secondNetDevice = None
        self.secondDeviceName = None
        self.secondDevicePortID = None

    def plan(self, planner):
        tests = list()
        # manual planning only
        return tests

    def add(self, planner):
        pass

    def matchDeviceToTest(self, portDevice, devices):
        """portDevice is partially populated, using attributes set to find fully populated device parsed
           then finds and returns appropriate test based off of the rate
        """
        for device in devices:
            if portDevice.deviceName == device.deviceName and portDevice.portNumber == device.portNumber:
                class_objects = []
                for name, obj in inspect.getmembers(sys.modules[__name__]):
                    if 'Infiniband' in name and inspect.isclass(obj):
                        class_objects.append(obj)
                for class_object in class_objects:
                    try:
                        obj = class_object()
                    except TypeError:
                        continue
                    try:
                        rate = obj.rateContains
                        if rate in device.attributes['rate']:
                            return obj
                    except AttributeError:
                        print("{} doesn't have rate data".format(name))
        print("Error, no appropriate test class determined for {}".format(portDevice.deviceName))
        return None

    def solicitSUTDevice(self, driverModules, message=""):
        """ Asks user what device they want to test with locally """
        if not message:
            message = "Please specify the socket direct device on this machine for testing"
        socketDirectDevices = RdmaConnectionTest.getAllDevices(driverModules=driverModules)
        socketDirectDevices = json.loads(socketDirectDevices)["Devices"]
        if not socketDirectDevices:
            print("Error, failed to gather socket direct devices on SUT")
            return False
        socketDirectDevice = self.promptEdit(
            message, socketDirectDevices[0], socketDirectDevices)
        if not socketDirectDevice:
            print("Error, selected device is %s" % socketDirectDevice)
            return None
        print("You selected: {}".format(socketDirectDevice))
        return socketDirectDevice

    def constructPortDevice(self, selectionString):
        """ Parse selectionString and return DevicePort object """

        match = re.search('(?P<ip4>[^\s]*?) (?P<netDeviceName>[^\s]*?) Device=(?P<portDeviceName>[^\s]*?) Port=(?P<port>[^\s]*$)', selectionString)
        ip4 = match.group("ip4")
        netDeviceName = re.sub(r"@.*$", "", match.group("netDeviceName"))
        portDeviceName = re.sub(r"@.*$", "", match.group("portDeviceName"))
        port = match.group("port")
        return DevicePort(portDeviceName, port, potentialIPAddresses=[ip4], netDeviceName=netDeviceName)
    

    def constructLogicalDeviceName(self, device):
        """ Formulate logical device name used by rdma connection test from values in DevicePort object """

        try:
            if device.netDeviceName:
                return "{}_devicePort_{}_netDevice_{}".format(device.deviceName, device.portNumber, device.netDeviceName)
            else:
                return "{}_devicePort_{}".format(device.deviceName, device.portNumber)
        except AttributeError:
            print("Error, failed to generate logical device name from device")

    def setupSocketDirectTestClass(self):
        """ Solicit input from user, determine and set up test to run """

        socketDirectDevice = self.solicitSUTDevice(driverModules=self.driverModules)

        portDevice = self.constructPortDevice(socketDirectDevice)
        #print("ip4: {}, netDeviceName: {}, portDeviceName: {}, port: {}".format(ip4, netDeviceName, portDeviceName, port))
        logicalDeviceName = self.constructLogicalDeviceName(portDevice)
        print("translated to logical device name: {}".format(logicalDeviceName))

        # call remote function to get server's possible ip addresses
        serverDevices = self.getServerAllDevices()
        if not serverDevices:
            print("Error, failed to gather ip choices from LTS")
            return None

        serverDevice = self.promptEdit(
            "Which ip address (LTS) would you like to test %s with: " % socketDirectDevice, serverDevices[0],
            serverDevices)
        if not serverDevice:
            print("Error, selected serverDevice is %s" % serverDevice)
            return None
        serverDeviceSplit = serverDevice.split(" ")
        if len(serverDeviceSplit) < 2:
            print("Error, expected format: inet <ip address>, was %s" % serverDeviceSplit[0])
            return None
        serverIP = serverDeviceSplit[0]
        serverIP = serverIP.split("/")[0]

        print("Using IP address of: %s" % serverIP)
        self.serverIP = serverIP

        print("finding appropriate test to run on device")
        classObj = self.matchDeviceToTest(portDevice, parseDevices(self.driverModules))
        if classObj:
            print("found {} as match!".format(classObj.__class__.__name__))
        else:
            return None
        classObj.setLogicalDeviceName(logicalDeviceName)
        classObj.serverDevice = serverDevice
        classObj.netDevice = portDevice.netDeviceName
        classObj.testParameters.set(Constants.testserver, self.serverIP)

        return classObj

    

    def run(self):
        print("Gathering selections for the two socket direct connections...")
        print("---------1st connection LTS<->SUT---------")
        firstConnectionTest = self.setupSocketDirectTestClass()
        if not firstConnectionTest:
            print("Error, could not set up first connection")
            return False

        print("---------2nd connection LTS<->SUT---------")
        secondConnectionTest = self.setupSocketDirectTestClass()
        if not secondConnectionTest:
            print("Error, could not set up second connection")
            return False

        # We want to run both tests, even if the first one fails
        firstTestResult = firstConnectionTest.run()
        secondTestResult = secondConnectionTest.run()

        return firstTestResult and secondTestResult
        


class OmnipathConnectionTest(RdmaConnectionTest):

    def __init__(self):
        RdmaConnectionTest.__init__(self, "Omnipath")
        self.linkLayer = "InfiniBand"
        self.driverModules = ['hfi1']
        self.opafmFailNotStopVersion = "opa-fm-10.5.1.0.1-1.el7"
        opafmVersion = Command("rpm -q opa-fm").getString(ignoreErrors=True)
        opafmAcceptFailAsStopped = opafmVersion < self.opafmFailNotStopVersion
        self.subTests.append([partial(self.serviceTest, "rdma-ndd"), "rdma-ndd service test"])
        self.subTests.append(
            [partial(self.serviceTest, "opafm", opafmAcceptFailAsStopped, self.opafmFailNotStopVersion),
             "opafm service test"])
        self.subTests.append([self.lIDTest, "LID test"])
        self.subTests.append([self.linkSpeedTest, "link speed test"])
        self.subTests.append([self.opapmaQueryTest, "opapmaquery test"])
        self.planWithNetDevices = True
        self.requiredRPMs.append("opa-fm")
        self.requiredRPMs.append("opa-basic-tools")

        # only 100Gig rated connection is of interest at this time
        self.rateContains = "100"
        self.ethernetTestClass = network.Ethernet100GigTest
        self.isEthernet = False
        self.bandwidthTarget = 0.35  # 35% - override default for ethernet test
        self.ibBandwidthWarning = None
        self.ibBandwidthFail = 0.2


class RoCEConnectionTest(RdmaConnectionTest):

    def __init__(self, path):
        RdmaConnectionTest.__init__(self, path)
        self.linkLayer = "Ethernet"
        self.driverModules = ['bnxt_re', 'Mlx4_ib', 'mlx4_en', 'mlx4_core', 'mlx5_ib', 'mlx5_core', 'ocrdma', 'Qed',
                              'qedr', 'vmw_pvrdma', 'qede']
        self.subTests.append([self.ethtoolTest, "ethtool test"])
        self.planWithNetDevices = True
        self.requiredRPMs.append("opensm")
        self.isEthernet = True
        self.bandwidthTarget = 0.75  # 75% - override default for ethernet test
        self.ibBandwidthWarning = None
        self.ibBandwidthFail = 0.5


class RoCE200Gig(RoCEConnectionTest):
    def __init__(self):
        RoCEConnectionTest.__init__(self, "200GigRoCE")
        self.rateContains = "200"
        self.ethernetTestClass = network.Ethernet200GigTest

class RoCE100Gig(RoCEConnectionTest):
    def __init__(self):
        RoCEConnectionTest.__init__(self, "100GigRoCE")
        self.rateContains = "100"
        self.ethernetTestClass = network.Ethernet100GigTest

class RoCE50Gig(RoCEConnectionTest):
    def __init__(self):
        RoCEConnectionTest.__init__(self, "50GigRoCE")
        self.rateContains = "50"
        self.ethernetTestClass = network.Ethernet50GigTest


class RoCE40Gig(RoCEConnectionTest):
    def __init__(self):
        RoCEConnectionTest.__init__(self, "40GigRoCE")
        self.rateContains = "40"
        self.ethernetTestClass = network.Ethernet40GigTest


class RoCE25Gig(RoCEConnectionTest):
    def __init__(self):
        RoCEConnectionTest.__init__(self, "25GigRoCE")
        self.rateContains = "25"
        self.ethernetTestClass = network.Ethernet25GigTest


class RoCE20Gig(RoCEConnectionTest):
    def __init__(self):
        RoCEConnectionTest.__init__(self, "20GigRoCE")
        self.rateContains = "20"
        self.ethernetTestClass = network.Ethernet20GigTest


class RoCE10Gig(RoCEConnectionTest):
    def __init__(self):
        RoCEConnectionTest.__init__(self, "10GigRoCE")
        self.rateContains = "10"
        self.ethernetTestClass = network.Ethernet10GigTest


class RoCE5Gig(RoCEConnectionTest):
    def __init__(self):
        RoCEConnectionTest.__init__(self, "5GigRoCE")
        self.rateContains = "5"
        self.ethernetTestClass = network.Ethernet5GigTest


class RoCE2_5Gig(RoCEConnectionTest):
    def __init__(self):
        RoCEConnectionTest.__init__(self, "2_5GigRoCE")
        self.rateContains = "2.5"
        self.ethernetTestClass = network.Ethernet2_5GigTest


class IWarpConnectionTest(RdmaConnectionTest):

    def __init__(self, path):
        RdmaConnectionTest.__init__(self, path)
        self.linkLayer = "Ethernet"
        self.driverModules = ['iw_cxgb3', 'iw_cxgb4', 'I40e', 'i40iw', 'Qed', 'qedr', 'qede']
        self.subTests.append([self.ethtoolTest, "ethtool test"])
        self.planWithNetDevices = True
        self.requiredRPMs.append("opensm")
        self.isEthernet = True
        self.bandwidthTarget = 0.75  # 75% - override default for ethernet test
        self.ibBandwidthWarning = None
        self.ibBandwidthFail = 0.4

    def verifyTransport(self, deviceName=""):
        command = Command("ibv_devinfo -d %s | grep -e transport" % deviceName)
        transport = ""
        try:
            transport = command.getString(regex=".*transport:.*", singleLine=False, ignoreErrors=True)
        except HwCertCommandException:
            print("Warning: Could not determine transport of %s" % deviceName)
            return False
        return "iWARP" in transport


class IWarp200Gig(IWarpConnectionTest):
    def __init__(self):
        IWarpConnectionTest.__init__(self, "200GigiWarp")
        self.rateContains = "200"
        self.ethernetTestClass = network.Ethernet200GigTest

class IWarp100Gig(IWarpConnectionTest):
    def __init__(self):
        IWarpConnectionTest.__init__(self, "100GigiWarp")
        self.rateContains = "100"
        self.ethernetTestClass = network.Ethernet100GigTest

class IWarp50Gig(IWarpConnectionTest):
    def __init__(self):
        IWarpConnectionTest.__init__(self, "50GigiWarp")
        self.rateContains = "50"
        self.ethernetTestClass = network.Ethernet50GigTest


class IWarp40Gig(IWarpConnectionTest):
    def __init__(self):
        IWarpConnectionTest.__init__(self, "40GigiWarp")
        self.rateContains = "40"
        self.ethernetTestClass = network.Ethernet40GigTest


class IWarp25Gig(IWarpConnectionTest):
    def __init__(self):
        IWarpConnectionTest.__init__(self, "25GigiWarp")
        self.rateContains = "25"
        self.ethernetTestClass = network.Ethernet25GigTest


class IWarp20Gig(IWarpConnectionTest):
    def __init__(self):
        IWarpConnectionTest.__init__(self, "20GigiWarp")
        self.rateContains = "20"
        self.ethernetTestClass = network.Ethernet20GigTest


class IWarp10Gig(IWarpConnectionTest):
    def __init__(self):
        IWarpConnectionTest.__init__(self, "10GigiWarp")
        self.rateContains = "10"
        self.ethernetTestClass = network.Ethernet10GigTest
