#!/usr/bin/env python
# Copyright (c) 2009-2012 Red Hat, Inc. All rights reserved.
#
# This copyrighted material
# is made available to anyone wishing to use, modify, copy, or
# redistribute it subject to the terms and conditions of the GNU General
# Public License v.2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# Author: Clark Williams <williams@redhat.com>
# Author: David Sommerseth <davids@redhat.com>
#

import os, sys
from os.path import isfile
from rhcert.suites.rhcert.platform import CheckHardwareHealth

if __name__ == "__main__":
    sys.path.insert(0, '/usr/share/hwcert/lib')


from rhcert.command import Command, HwCertCommandException, HwCertCommandErrorOutput
from rhcert.tags import Constants, TestTag
from rhcert.tags import DeviceClass

from rhcert.suites.hwcert.hwcert.hardwareTest import HardwareTest

###################################################
#
# default test parameters, may be overriden by
# variables in /etc/rteval.conf
#
###################################################
rteval_config     = "/etc/rteval.conf"

hwlat_threshold   =      15   # fifteen microseconds
hwlat_window      = 1000000   # one million microseconds  (1.0s)
hwlat_width       =  800000   # 800 thousand microseconds (0.8s)
hwlat_duration    =    7022   # 2 hours

rteval_duration   =  12*60*60 # 12 hours
rteval_maxlatency =     150.0 # 150us max latency
rteval_maxstddev  =       3.0 # 3us max standard deviation


class Realtime(HardwareTest):

    def __init__(self):
        HardwareTest.__init__(self, "realtime")
        self.interactive = False
        if os.uname()[3].find("RT ") == -1:
            self.realtime = False
        else:
            self.realtime = True
        self.data = {}
        self.__rteval = None
        self.deviceClass = DeviceClass.processor
        self.owner = "David Sommerseth <davids@redhat.com>"
        self.description = "Realtime tests (rteval/hwlatdetect)"
        self.requires = ["rt-setup",  "rteval", "rteval-loads", "rt-tests"]


    def tags(self):
        return [TestTag.noninteractive, TestTag.certification, TestTag.realtime]

    def getRequiredRPMs(self):
        return self.requires


    def __read_sys_dmi(self, name):
        try:
            f = open("/sys/class/dmi/id/%s" % name, "r")
            ret = f.readline().strip()
            f.close()
            return ret
        except IOError:
            print("Could not open /sys/class/dmi/id/%s - no DMI info available?" % name)
            return None


    def plan(self, planner):
        tests = list()
        if not self.realtime:
            return tests

        checkHardwareHealth = CheckHardwareHealth("ram_per_core")
        if isfile("/sys/kernel/realtime"):
            with open("/sys/kernel/realtime", "r") as f:
                if f.readline().strip() == "1" and checkHardwareHealth.ram_per_core():
                    tests.append(self.makeCopy())
        if not tests:
            print("INFO: Could not plan test {}".format(self.path))
        else:
            print("INFO: Successfully planned test {}".format(self.path))
        return tests

    def determine_sockets(self):
        """Determines the number of sockets machine has.  Returns None if unseuccessful."""
        try:
            sockets_command =  Command("lscpu | awk '/^Socket/{print $2}'")
            sockets = int(sockets_command.getString())
            return sockets
        except (HwCertCommandException, ValueError) as e:
            print("Warning.  Issue running command to determine number of sockets.  Using default requirements.")
            return None

    def adjust_reqs(self):
        # Update max latency and std deviation requirements based off number of sockets
        # <=3 stays the default
        # >=4 has max latency = 250us and std dev = 5us
        # Note: these values are still possible to overwrite in /etc/rteval.conf so that functionality does not change
        sockets = self.determine_sockets()
        if sockets >= 4:
            global rteval_maxlatency
            global rteval_maxstddev
            rteval_maxlatency =     250.0 # us max latency
            rteval_maxstddev  = 5.0 # us max standard deviation
            print("Adjustment to requirements due to number of detected sockets:")
            print("detected sockets: %s" % sockets)
            print("maximum latency: %s" % rteval_maxlatency)
            print("maximum standard deviation: %s" % rteval_maxstddev)

    def run(self):

        self.adjust_reqs()
        # Try to identify the rteval version available
        try:
            from rteval import RTEVAL_VERSION
            print("Found rteval v%s" % RTEVAL_VERSION)
            return self.__run_v2()
        except ImportError:
            # If this import fails, we presume rteval v1.x is installed
            print("Running rteval v1.x")
            return self.__run_old()


    def __run_old(self):
        if self.__run_rteval_v1() == True \
                and self.__run_hwlatdetect() == True:
            return True
        return False


    def __run_rteval_v1(self):
        try:
            # Prepare rteval v1.x
            import rteval.rteval

            args = ['-d', str(rteval_duration)]
            rteval = rteval = rteval.rteval.RtEval(args)

            # Run rteval
            if rteval.rteval() != 0:
                print("Error: could not run rteval")
                return False
        except Exception as e:
            print("Error: Failed to run rteval: %s" % str(e))
            return False

        # Validate the rteval results
        ok_values = {"/rteval/cyclictest/system/statistics/maximum": rteval_maxlatency,
                     "/rteval/cyclictest/system/statistics/standard_deviation": rteval_maxstddev}

        return self.__validate_rteval_results(rteval.XMLreport(), ok_values)


    def __run_hwlatdetect(self):
        sys.path.insert(0, '/usr/bin')
        # Prepare hwlatdetect
        try:
            import hwlatdetect
        except Exception as e:
            print("Error: could not import hwlatdetect")
            return False

        try:
            detect = hwlatdetect.Detector()
        except Exception as e:
            raise Exception("Error: could not initialize hwlatdetect: %s" % str(e))

        detect.set("threshold", hwlat_threshold)
        detect.set("window", hwlat_window)
        detect.set("width", hwlat_width)
        detect.testduration = hwlat_duration

        print("hwlatdetect:  test duration %d seconds" % detect.testduration)
        print("   parameters:")
        print("        Latency threshold:    %dus" % detect.get("threshold"))
        print("        Sample window:        %dus" % detect.get("window"))
        print("        Sample width:         %dus" % detect.get("width"))
        print("        Non-sampling period:  %dus" % (detect.get("window") - detect.get("width")))

        detect.detect()
        exceeding = detect.get("count")
        print("detected %d samples exceeding threshold %d\n" % (exceeding, detect.get("threshold")))

        for s in detect.samples:
            print(s)

        detect.cleanup()

        if exceeding:
            return False
        return True


    def __run_v2(self):
        try:
            # Prepare rteval v2.x
            # - this rteval generation will also run hwlatdetect and
            #   report the results via rteval.  In the future it might
            #   even run more loads and measurment modules too.
            #
            from rteval.Log import Log
            from rteval.modules.loads import LoadModules
            from rteval.modules.measurement import MeasurementModules
            from rteval import RtEval, rtevalConfig
    
            # Prepare logging
            log = Log()
            log.SetLogVerbosity(Log.INFO)

            cfg = rtevalConfig.rtevalConfig(logger=log)
            cfg.Load(rteval_config)

            rtevcfg = cfg.GetSection("rteval")

            #Set the run time for rteval profiles
            rtevcfg.duration = rteval_duration

            # Set the log verbosity, based on the config file
            rtevcfg = cfg.GetSection("rteval")
            loglev = Log.ERR | Log.WARN \
                | (rtevcfg.verbose and Log.INFO) \
                | (rtevcfg.debugging and Log.DEBUG)
            log.SetLogVerbosity(loglev)

            # Make sure we have the needed loads and measurement modules configured.
            # This may be overridden by /etc/rteval.conf
            if not cfg.HasSection("loads"):
                cfg.AppendConfig("loads", {
                        "kcompile":  "module",
                        "hackbench": "module"
                        })

            if not cfg.HasSection("measurement"):
                cfg.AppendConfig("measurement", {
                        "cyclictest":  "module",
                        "HWLatDetect": "module"
                        })

            # Set the hwlatdetect measurement settings - if not already configured
            if not cfg.HasSection("HWLatDetect"):
                cfg.AppendConfig("HWLatDetect", {
                        "threshold": hwlat_threshold,
                        "window":    hwlat_window,
                        "width":     hwlat_width,
                        "duration":  hwlat_duration
                        })

            # Prepare load and measurement modules
            loadmods = LoadModules(cfg, logger=log)
            measuremods = MeasurementModules(cfg, logger=log)

            # Prepare and start the core rteval run
            self.__rteval = RtEval(cfg, loadmods, measuremods, log)
            self.__rteval.Prepare(False)

            sys.stdout.flush()
            if self.__rteval.Measure() != 0:
                print("Error: Failed running rteval")
                return False
        except Exception as e:
            print("Error: could not intitialize rteval: %s" % str(e))
            return False

        # Check the result, make sure system latency is within accepted boundaries
        ok_values = {"/rteval/Measurements/Profile/cyclictest/system/statistics/maximum": rteval_maxlatency,
                     "/rteval/Measurements/Profile/cyclictest/system/statistics/standard_deviation": rteval_maxstddev,
                     "/rteval/Measurements/Profile/hwlatdetect/samples/@exceeding": 0}

        return self.__validate_rteval_results(self.__rteval.GetXMLreport(), ok_values)


    def __validate_rteval_results(self, report, ok_values):
        xpath = report.xpathNewContext()
        for k in ok_values.keys():
            val = xpath.xpathEval(k)
            if len(val) == 1:
                v = float(val[0].get_content())
                print("Checking %s: Must be < %f, measured to %f ..." % (k, ok_values[k], v))
                if ok_values[k] < v:
                    print("** FAILED **")
                    return False
                print("PASS")
        return True



if __name__ == "__main__":
    test = Realtime()
    test.plan(None)
    returnValue = test.do(sys.argv)
    sys.exit(returnValue)
