#!/usr/bin/env python
# Copyright (c) 2018 Red Hat, Inc. All rights reserved. This copyrighted material
# is made available to anyone wishing to use, modify, copy, or
# redistribute it subject to the terms and conditions of the GNU General
# Public License v.2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# Author: Rajasi Kulkarni
"""All storage tests need to be declared in this module"""

from rhcert.suites.hwcert.hwcert.storageTest import StorageTest
from rhcert.test import TestTag


class StorageSATATest(StorageTest):
    """
    This is a base class for all the SATA Test.
    """
    def __init__(self, path):
        StorageTest.__init__(self, path)
        self.hostPatternString = "host[0-9]+"
        self.controllerPatternString = "SATA"
        self.controllerPatternNotString = "SAS"
        self.lsscsiPatternString = "sata"
        self.planning_methods = [self.planByController, self.planByLsscsi]
        self.req_planning_methods = [self.planByDeviceUnique]

    def tags(self):
        return [TestTag.noninteractive, TestTag.certification, TestTag.storage]

    def plan(self, planner):
        return StorageTest.plan(self, planner)


class StorageU2SATATest(StorageSATATest):
    """
    This will be planned if storage is U2 SATA
    """
    def __init__(self):
        StorageSATATest.__init__(self, "U2_SATA")
        self.req_planning_methods.append(self.planByStorageU2)


class StorageM2SATATest(StorageSATATest):
    """
    This will be planned if storage is M2 SATA
    """
    def __init__(self):
        StorageSATATest.__init__(self, "M2_SATA")
        self.req_planning_methods.append(self.planByStorageM2)


class StorageSATASSDTest(StorageSATATest):
    """
    This will be planned if storage is SATA and SSD
    """
    def __init__(self):
        StorageSATATest.__init__(self, "SATA_SSD")
        self.req_planning_methods.append(self.planByStorageSSD)


class StorageDefaultSATATest(StorageSATATest):
    """
    This will be planned if no other SATA sub tests are planed.
    This test should be at the end of all the SATA subclasses
    """
    def __init__(self):
        StorageSATATest.__init__(self, "SATA")


class StorageSASTest(StorageTest):
    """
    This is a base class for all the SAS Test.
    """
    def __init__(self, path):
        StorageTest.__init__(self, path)
        self.hostPatternString = "host[0-9]+"
        self.controllerPatternString = "SAS"
        self.controllerPatternNotString = "SATA"
        self.lsscsiPatternString = "sas"
        self.planning_methods = [self.planByController, self.planByLsscsi]
        self.req_planning_methods = [self.planByDeviceUnique]

    def tags(self):
        return [TestTag.noninteractive, TestTag.certification, TestTag.storage]

    def plan(self, planner):
        return StorageTest.plan(self, planner)


class StorageSASSSDTest(StorageSASTest):
    """
    This will be planned if storage is SAS and SSD
    """
    def __init__(self):
        StorageSASTest.__init__(self, "SAS_SSD")
        self.req_planning_methods.append(self.planByStorageSSD)


class StorageDefaultSASTest(StorageSASTest):
    """
    This will be planned if no other SAS sub tests are planed.
    This test should be at the end of all the SAS subclasses
    """
    def __init__(self):
        StorageSASTest.__init__(self, "SAS")


class StorageNVMETest(StorageTest):
    """
    This is a base class for all NVMe Test.
    """
    def __init__(self, path):
        StorageTest.__init__(self, path)
        self.hostPatternString = "nvme[0-9]+"
        self.req_planning_methods = [self.planByDeviceUnique]

    def tags(self):
        return [TestTag.noninteractive, TestTag.certification, TestTag.storage]

    def plan(self, planner):
        return StorageTest.plan(self, planner)


class StorageU2NVMETest(StorageNVMETest):
    """
    This will be planned if storage is U2 NVMe
    """
    def __init__(self):
        StorageNVMETest.__init__(self, "U2_NVMe")
        self.req_planning_methods.append(self.planByStorageU2)


class StorageM2NVMETest(StorageNVMETest):
    """
    This will be planned if storage is M2 NVMe
    """
    def __init__(self):
        StorageNVMETest.__init__(self, "M2_NVMe")
        self.req_planning_methods.append(self.planByStorageM2)


class StorageDefaultNVMETest(StorageNVMETest):
    """
    This will be planned if no other NVMe sub tests are planed.
    This test should be at the end of all the NVMe subclasses
    """
    def __init__(self):
        StorageNVMETest.__init__(self, "PCIE_NVMe")


class StorageNVDIMMTest(StorageTest):
    """
    This is a base class for all NVDIMM Test.
    """
    def __init__(self, path):
        StorageTest.__init__(self, path)
        self.hostPatternString = "ndbus[0-9]+"
        self.req_planning_methods = [self.planByDeviceUnique]

    def tags(self):
        return [TestTag.noninteractive, TestTag.certification, TestTag.storage]

    def plan(self, planner):
        return StorageTest.plan(self, planner)


class StorageDefaultNVDIMMTest(StorageNVDIMMTest):
    """
    This will be planned if storage is NVDIMM
    """
    def __init__(self):
        StorageNVDIMMTest.__init__(self, "NVDIMM")
        self.req_planning_methods.append(self.planByStorageNVDIMM)


"""
class StorageVIRTIOTest(StorageTest):
    ''' This test is for testing assigned devices
        for fv_storage test'''

    def __init__(self):
        StorageTest.__init__(self, "virtio_storage")
        self.UDIPatternString = "virtio"
        self.planning_methods = [self.planByUDI]

    def tags(self):
        return [TestTag.noninteractive, TestTag.certification]

    def plan(self, planner):
        return StorageTest.plan(self, planner)
"""

"""
# To be completed in subsequent JIRAs

class SDStorageTest(StorageTest):

    def __init__(self):
        StorageTest.__init__(self, "Storage_SD")
        self.hostPatternString = "mmc[0-9]+"


class MMCStorageTest(StorageTest):

    def __init__(self):
        StorageTest.__init__(self, "Storage_MMC")
        self.hostPatternString = "mmc[0-9]+"


class eMMCStorageTest(StorageTest):

    def __init__(self):
        StorageTest.__init__(self, "Storage_eMMC")
        self.hostPatternString = "mmc[0-9]+"
"""


class DefaultStorageTest(StorageTest):
    """
    This is the default storage test. This would be planned only if none of the above
    tests are planned.
    This should always be the last class declared in this module, as the priority of
    planning this tests should be the least
    """
    def __init__(self):
        StorageTest.__init__(self, "STORAGE")
        self.planning_methods = [self.planByDeviceUnique]

    def tags(self):
        return [TestTag.noninteractive, TestTag.certification, TestTag.storage]

    def plan(self, planner):
        # Plan the default storage test only if no other test is planned
        return StorageTest.plan(self, planner)
