#!/usr/bin/env python
# Copyright (c) 2006 Red Hat, Inc. All rights reserved. This copyrighted material
# is made available to anyone wishing to use, modify, copy, or
# redistribute it subject to the terms and conditions of the GNU General
# Public License v.2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# Author: Greg Nichols
#
import os, sys, time, syslog

from rhcert.tags import Constants, DeviceClass, TestTag
from rhcert.controller import Controller
from rhcert.redhatrelease import RedHatRelease
from rhcert.suites.hwcert.hwcert.hardwareTest import HardwareTest

class SuspendTest(HardwareTest):

    def __init__(self):
        HardwareTest.__init__(self, "suspend")
        self.acpiListen = None
        self.hwcertController = Controller()
        self.interactive = True
        self.deviceClass = DeviceClass.battery
        self.priority = 1  # run suspend before any other test
        self.mandatory = Constants.pre
        #  451154 - failures on unknown suspend methods
        self.knownMethods = [Constants.memory, Constants.disk]
        self.suspendMessage = dict()
        self.suspendMessage[Constants.memory] = "Starting Suspend"
        self.suspendMessage[Constants.disk] = "Starting Hibernate"
        self.stateMessage = dict()
        self.stateMessage[Constants.memory] = ["PM: suspend entry (deep)", "PM: Entering mem sleep"]
        self.stateMessage[Constants.disk] = ["PM: hibernation entry", "PM: Creating hibernation image"]
        self.need_to_reset_mem_sleep_to_default = False
        self.mem_sleep_file_path = os.path.join("/sys/power/mem_sleep")
        self.state_file_path = os.path.join("/sys/power/state")
        self.release = RedHatRelease()
        # RHEL-8 and above specific sub-tests
        if int(self.release.getVersion()) >= 8:
            self.knownMethods.append(Constants.freeze)
            self.suspendMessage[Constants.freeze] = "Starting Freeze"
            self.stateMessage[Constants.freeze] = ["PM: suspend entry (s2idle)", "PM: Entering freeze sleep"]

    def tags(self):
        return [TestTag.interactive, TestTag.certification, TestTag.portable]

    def plan(self, planner):
        tests = list()
        acAdapterDevice = None
        devices = planner.getUdevDeviceDetector().getDevices()
        for device in devices:
            if "power_supply" in device.getProperty("SUBSYSTEM") and "BAT" in device.getProperty("POWER_SUPPLY_NAME") and self.getSuspendMethods():
                test = self.makeCopy()
                test.setDevice(device)
                tests.append(test)
                # 450297: suspend test planned once per battery
                # change to only plan this test once, no matter how many
                # batteries are reported
                break
        return tests

    def getSuspendMethods(self):
        methods = list()
        sysPowerState = open(self.state_file_path, "r")
        if sysPowerState:
            while True:
                line = sysPowerState.readline()
                if not line:
                    break
                states = line.split()
                #  451154 - failures on unknown suspend methods
                for state in states:
                    if state in self.knownMethods:
                        methods.append(state)
                if len(methods) > 0:
                    return methods

        # Check 'resume=' is in the kernel parameters which Specifies a swap device to use when waking from hibernation.
        if Constants.disk in methods and not 'resume=' in open("/proc/cmdline", "r").read():
            print("Warning: Removing suspend to disk option as it seems to be unsupported by configuration")
            methods.remove(Constants.disk)

        if not methods:
            print("Error: could not determine allowable suspend methods")

        return methods

    def suspendWithoutOSCommand(self, method):
        """ suspend using a write operation to /sys/power/state """
        if method == Constants.memory and os.path.exists(self.mem_sleep_file_path) \
                and '[s2idle]' in open(self.mem_sleep_file_path, "r").read():
            self.need_to_reset_mem_sleep_to_default = True if \
                os.system("echo deep > {}".format(self.mem_sleep_file_path)) is 0 else False
        sysPowerState = open(format(self.state_file_path), "w")
        if sysPowerState:
            sysPowerState.write(method)
            sysPowerState.close()
        else:
            print("Error: could not suspend the system")
            return False
        return True

    def suspend(self, source, method):
        print("This test will suspend the operating system.")
        print("Please resume by pressing the power button after suspend has completed.")
        sys.stdout.flush()
        # start acpi_listen
        # if self.acpiListen:
        #    self.acpiListen.close()
        # self.acpiListen = os.popen("acpi_listen")
        markerName = "%s-%s-%s" % (self.getPath(), source, method)
        syslog.syslog(self.hwcertController.getSystemLogMarker(markerName, "begin"))
        if source == Constants.OSCommand:
            if not self.promptConfirm(" suspend? "):
                return False
            waitTime = 5
            print("Suspending in %d sec" % waitTime)
            sys.stdout.flush()
            time.sleep(waitTime)
            return self.suspendWithoutOSCommand(method)

        # otherwise
        if source == Constants.FunctionKey:
            if not self.promptConfirm(
                    "Does this system have a function key (Fn) to suspend the system to %s?" % method):
                print("Warning: suspend test to %s not run from function key" % method)
                return False
            if not self.promptConfirm(
                    "Are you ready to press the function key (Fn) to suspend the system to %s. Answer the question and then press the function key (Fn) if you want to suspend the system" % method):
                return False
            return True

        # otherwise
        print("Error: unknown suspend source")
        return False

    def checkResume(self, source, method):
        self.verifiedResumeMethod = None
        sys.stdout.flush()
        self.promptConfirm("Has resume completed? ")
        markerName = "%s-%s-%s" % (self.getPath(), source, method)
        syslog.syslog(self.hwcertController.getSystemLogMarker(markerName, "end"))

        # check system log for acpi messages
        log = self.hwcertController.getSystemLog("%s-%s-%s" % (self.getPath(), source, method))

        if source == Constants.FunctionKey:
            # check Function Key message (given by systemd) first:
            try:
                suspendMessage = self.suspendMessage[method]
            except KeyError:
                print("Error: unknown suspend method %s" % method)
                return False

            if not suspendMessage in log:
                print("Error: could not verify suspend")
                return False

        # check suspend/hibernate/freeze message (given by kernel)
        suspendMessage = "Freezing user space processes"
        if not suspendMessage in log:
            print("Error: could not verify suspend")
            return False

        # otherwise
        print("Verified suspend")

        # verifying resuming
        resumeMessage = "Restarting tasks"
        if not resumeMessage in log:
            print("Error: could not verify resume")
            return False

        # check that they used the correct method
        for state in self.stateMessage.keys():
            methodMessages = self.stateMessage[state]
            for methodMessage in methodMessages:
                if methodMessage in log:
                    self.verifiedResumeMethod = state
                    print("Found '%s' in log messages" % methodMessage)
                    break
            if self.verifiedResumeMethod:
                break

        if not self.verifiedResumeMethod:
            print("Error: No valid suspend method found in logs")
            return False

        if method != self.verifiedResumeMethod:
            print("Error: The suspend method was %s and should have been %s" % (self.verifiedResumeMethod, method))
            return False

        # otherwise
        print("Verified resume from %s" % self.verifiedResumeMethod)

        return True

    def reset_mem_sleep_to_default(self):
        if self.need_to_reset_mem_sleep_to_default:
            self.need_to_reset_mem_sleep_to_default = False if \
                os.system("echo s2idle > {}".format(self.mem_sleep_file_path)) is 0 else True

    def runSuspendResume(self, source, method):
        if self.suspend(source, method):
            status = self.checkResume(source, method)
            self.reset_mem_sleep_to_default()
            return status
        elif source == Constants.FunctionKey:
            # allow user to opt-out of fn key tests, since system may not have them
            self.verifiedResumeMethod = method
            self.reset_mem_sleep_to_default()
            return True
        # otherwise
        return False

    def run(self):

        if self.getMode() == Constants.auto:
            print("Error: suspend test can not be run in auto mode.")
            return False

        methods = self.getSuspendMethods()
        if not methods:
            return False

        # forced mode - being run as preparations for other tests
        if self.getMode() == Constants.forced:
            result = True
            if not self.runSuspendResume(Constants.OSCommand, methods[0]):
                result = False
            elif self.verifiedResumeMethod != methods[0]:
                result = False
            return result

        # otherwise, full test
        result = True
        for source in [Constants.OSCommand, Constants.FunctionKey]:
            for method in methods:
                self.markOutput(name=source + "-" + method)
                if not self.runSuspendResume(source, method):
                    result = False
                    self.markSummary(Constants.FAIL)
                elif self.verifiedResumeMethod != method:
                    result = False
                    self.markSummary(Constants.FAIL)
                else:
                    self.markSummary(Constants.PASS)

        self.closeOutput()
        return result


if __name__ == "__main__":
    test = SuspendTest()
    returnValue = test.do(sys.argv)
    sys.exit(returnValue)
