#!/usr/bin/env python
# Copyright (c) 2006 Red Hat, Inc. All rights reserved. This copyrighted material
# is made available to anyone wishing to use, modify, copy, or
# redistribute it subject to the terms and conditions of the GNU General
# Public License v.2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# Author: Greg Nichols
#         YangKun <ykun@redhat.com>
#

import sys, re

from rhcert.command import Command, HwCertCommandException
from rhcert.tags import DeviceClass, TestTag

from rhcert.suites.hwcert.hwcert.udevDeviceDetector import UdevDeviceDetector
from rhcert.suites.hwcert.hwcert.hardwareTest import HardwareTest

class TapeTest(HardwareTest):

    def __init__(self):
        HardwareTest.__init__(self, "tape")
        self.deviceClass = DeviceClass.sequential

    def tags(self):
        return [TestTag.noninteractive, TestTag.certification]

    def plan(self, planner):
        tests = list()
        devices = planner.getUdevDeviceDetector().getDevices()
        for device in devices:
            if 'tape' in device.getProperty("SUBSYSTEM") and re.compile("\/dev\/st\d+$").match(device.getProperty("DEVNAME")):
                test = self.makeCopy()
                test.setDevice(device)
                tests.append(test)
        return tests

    def getRequiredRPMs(self):
        return ["mt-st"]

    def getTapeDeviceName(self):

        if self.getLogicalDeviceName():
            return self.getLogicalDeviceName()

        #otherwise
        if not self.getUDI():
            print("Error: no UDI or logical device name for tape")
            return None

        #otherwise
        print("Using UDI to find the correct device name .")
        try:
            device = UdevDeviceDetector().getDevice(self.getUDI())
            if device:
                return device.getProperty("DEVNAME")

        except HwCertCommandException as exception:
            print("Error: could not determine tape device name")
            print(exception)
            return None


    def run(self):
        tapeDevice = self.getTapeDeviceName()
        if not tapeDevice:
            return False

        print("Testing tape device %s" % tapeDevice)

        # set default block size to 32k (64 x 512byte = 32k)
        bs=64
        try:
            # rewind the tape incase the tape is not rewinded.
            Command("mt -f %s rewind 2>/dev/null" % tapeDevice).run()
            if Command("tar -Pcb %s -f %s /usr" % (bs, tapeDevice)).echo() == 0:
                print("Write data done. Start comparing ...")
                if Command("tar -Pdb %s -f %s /usr" % (bs, tapeDevice)).echo() == 0:
                    print("Tape test on device %s passed." % tapeDevice)
                    return True
                else:
                    print("Error: Data comparison failed !")
                    return False
            else:
                print("Error: Write data Error !")
                return False
        except HwCertCommandException as exception:
            print(exception)
            return False

if __name__ == "__main__":
    test = TapeTest()
    returnValue = test.do(sys.argv)
    sys.exit(returnValue)
