#!/usr/bin/env python
# Copyright (c) 2019 Red Hat, Inc. All rights reserved. This copyrighted material
# is made available to anyone wishing to use, modify, copy, or
# redistribute it subject to the terms and conditions of the GNU General
# Public License v.2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# Author: Aniket Khisti <akhisti@redhat.com>
#
import sys, ast, time

from rhcert.suites.hwcert.hwcert.hardwareTest import HardwareTest
from rhcert.command import Command, HwCertCommandException
from rhcert.tags import DeviceClass, TestTag, Constants, TestResult
from rhcert.redhatrelease import RedHatRelease

# To make it work for Python 2+3 and with Unicode
try:
    to_unicode = unicode
except NameError:
    to_unicode = str

class ThunderboltTest(HardwareTest):

    def __init__(self):
        HardwareTest.__init__(self, "thunderbolt3")
        self.interactive = True
        self.deviceClass = DeviceClass.usb
        self.test_device = None
        self.test_port = 1
        self.numberOfSockets = 1

    def tags(self):
        return [TestTag.certification, TestTag.interactive]

    def getRequiredRPMs(self):
        return ["bolt"]

    def plan(self, planner):
        tests = list()
        release = RedHatRelease()
        if release.getVersion() >= 8 and self.check_thunderbolt_controller():
            test = self.makeCopy()
            tests.append(test)
            return tests

    def check_thunderbolt_controller(self):
        """
        This function checks if the system has a thunderbolt controller
        """
        # Check the Thunderbolt controller in lspci output
        check_tb_cmd = "lspci | grep Thunderbolt"
        try:
            check_tb = Command(check_tb_cmd).getStringList(regex=".*")
            return True
        except HwCertCommandException:
            if self.getDebug() != Constants.off:
                print("Warning: Thunderbolt controller not found in lspci !! \nChecking for controller in /sys/bus/wmi ...")

        # Get controller ID eg: 86XXXX48-205E-XXXX-9C48-2021XXXXE341
        get_controller_id = "ls /sys/bus/wmi/drivers/intel-wmi-thunderbolt -1 | grep '.*-.*-.*-.*-.*'"
        try:
            controller_id = Command(get_controller_id).getString(regex=".*")
        except HwCertCommandException:
            if self.getDebug() != Constants.off:
                print("Warning: Failed to fetch the thunderbolt controller ID")
            return False

        # Force Power On the controller
        power_on_cmd = "echo 1 > /sys/bus/wmi/devices/{}/force_power".format(controller_id)
        try:
            Command(power_on_cmd).run()
        except HwCertCommandException:
            if self.getDebug() != Constants.off:
                print("Warning: Failed to power on the thunderbolt controller")
            return False

        # Check the Thunderbolt controller in lspci output
        check_tb_cmd = "lspci | grep Thunderbolt"
        try:
            check_tb = Command(check_tb_cmd).getStringList(regex=".*")
        except HwCertCommandException:
            if self.getDebug() != Constants.off:
                print("Warning: Thunderbolt controller not found in /sys/bus/wmi")
            return False

        # Power Off the controller
        power_on_cmd = "echo 0 > /sys/bus/wmi/devices/{}/force_power".format(controller_id)
        try:
            Command(power_on_cmd).run()
        except HwCertCommandException:
            pass

        return True

    def __divide_list_into_chunks(self, l, n):
        # looping till length l
        for i in range(0, len(l), n):
            yield l[i:i + n]

    def __get_thunderbolt_devices(self):
        """
        :return: list of all thunderbolt devices with status:disconnected/authorized
        """
        devices = list()
        command = Command("boltctl")
        try:
            command.run()
            devices_info = command.raw_output
        except HwCertCommandException:
            self.statusMessage("Error: Could not fetch the list of Thunderbolt devices")

        if devices_info:
            properties = ['uuid:', 'name:', 'vendor:', 'status:']
            for device_info in devices_info.split('type:'):
                for property in properties:
                    if property in device_info:
                        devices.append(device_info.split(property).pop().strip().splitlines()[0])

            devices = self.__divide_list_into_chunks(devices, len(properties))
        return devices

    def get_authorized_thunderbolt_devices(self):
        """
        :return: list of dictionaries of only active thunderbolt devices among all the listed devices
        """
        active_devices = list()
        devices = self.__get_thunderbolt_devices()
        properties = ['UUID', 'NAME', 'VENDOR']
        if devices:
            for device in devices:
                (uuid, name, vendor, status) = device
                if status == "authorized":
                    active_devices.append(dict(zip(properties,[uuid, name, vendor])))
        return active_devices

    def get_pci_usb_display_devices(self):
        """
        required to identify probable type of thunderbolt device connected
        :return: record of above mentioned devices at that instant
        """
        all_devices = dict()
        command = Command('lspci -nn')
        try:
            command.run()
            all_devices['PCI'] = command.raw_output
        except HwCertCommandException:
            self.statusMessage("Warning: Could not fetch the list of PCI devices")

        command = Command('lsusb')
        try:
            command.run()
            all_devices['USB'] = command.raw_output
        except HwCertCommandException:
            self.statusMessage("Warning: Could not fetch the list of USB devices")

        command = Command('ls -1F /sys/class/drm/*/modes') # hack in place to avoid xhost,X11,wayland issues
        try:
            command.run()
            all_devices['DISPLAY'] = command.raw_output
        except HwCertCommandException:
            self.statusMessage("Warning: Could not fetch the list of Display devices")
        return all_devices

    def run_hotplug_test(self):
        """
        Thunderbolt device hotplug test, executes plug & unplug consecutively for user-specified number of sockets
        """
        result = TestResult()
        self.statusMessage("\nThunderbolt hotplug Test")
        if self.getMode() == Constants.auto:
            self.statusMessage("Error: The Thunderbolt test can not be run in auto mode.")
            return False
        last_inserted_device_info= dict()
        self.numberOfSockets = self.promptInteger("\nHow many Thunderbolt sockets are required to be tested? ")
        if self.numberOfSockets < 1:
            self.statusMessage("No Thunderbolt sockets to test")
            return result.combine(TestResult.FAIL)

        devices = list()
        while (not devices) and (self.test_port <= self.numberOfSockets):
            if self.promptConfirm("\nProceed to test socket #{}?".format(self.test_port)):
                self.statusMessage("\nPlease unplug the Thunderbolt devices from all the available sockets to continue the test.")
                devices = self.get_authorized_thunderbolt_devices()
                while devices:
                    if self.promptConfirm("\nEncountered already plugged-in device, Please unplug it to continue test."
                                          " Reply 'Yes' to continue (recheck) or 'No' to abort the whole test run."):
                        devices = self.get_authorized_thunderbolt_devices()
                        continue
                    else:
                        return False
                self.statusMessage("\nCapturing snapshot of all the PCI, USB and DISPLAY devices connected at the moment...")
                all_devices_before_plug_in = self.get_pci_usb_display_devices()
                if self.promptConfirm("\nPlease insert the Thunderbolt device in the socket #{} "
                                      "and confirm once, if it is plugged-in?".format(self.test_port)):
                    wait = 15  # device detection time
                    self.statusMessage("\nPlease wait a moment ({} seconds) to detect the new device...".format(wait))
                    time.sleep(wait)  # new device enroll, authorization & listing process by gnome-shell takes time

                    devices = self.get_authorized_thunderbolt_devices()
                    if not devices:
                        self.statusMessage("\nError: Could not detect active thunderbolt device."
                                           "\nFailure: The hotplug test failed.")
                        result.combine(TestResult.FAIL)
                        if self.test_port <= self.numberOfSockets:
                            self.test_port += 1 # loop control
                            devices = list()  # clean up the session for next socket test
                            continue # skip testing this device/socket any further, test next one

                    self.test_device = self.prompt("\nSelect the inserted thunderbolt device to be subject to hotplug test.",
                                                   answers=devices)
                    self.statusMessage("\nCapturing snapshot of all the PCI, USB and DISPLAY devices connected at the moment...")
                    all_devices_after_plug_in = self.get_pci_usb_display_devices()
                    inserted_device_info = {k: all_devices_after_plug_in[k] for k, _ in # diff of list of devices
                                            set(all_devices_after_plug_in.items()) - set(all_devices_before_plug_in.items())}
                    if inserted_device_info.keys():
                        logs = str([''.join(i) for i in inserted_device_info.items()])
                        self.statusMessage("\nINFO: Inserted device can be probably {} device.".
                                           format(' or '.join(inserted_device_info.keys())))
                        if "Ridge" in logs: # Intel's Thunderbolt 3 controller (codenamed Alpine Ridge, or the new Titan Ridge)
                            self.statusMessage("INFO: It is found to be connected to a valid Thunderbolt 3 socket.")
                        else:
                            self.statusMessage("Warning: But it may not be connected to a Thunderbolt 3 socket.")

                    if (self.numberOfSockets > 1) and (inserted_device_info == last_inserted_device_info):
                        self.statusMessage("\nWarning: The inserted device is maybe using the same socket as the last time.".
                                           format(inserted_device_info.keys()))
                    last_inserted_device_info.update(inserted_device_info)

                    self.statusMessage("\nSuccess: The inserted device {} is found to be authorized and connected."
                                       .format(self.test_device))

                    self.statusMessage("\nPlease unplug the device to continue the hotplug test")
                    if self.promptConfirm("Please confirm if the device is unplugged?"):
                        if isinstance(self.test_device, to_unicode):
                            self.test_device = ast.literal_eval(self.test_device)
                        if self.test_device not in self.get_authorized_thunderbolt_devices():
                            self.statusMessage("\nSuccess: Confirmed that the device {} is unplugged.".
                                               format(self.test_device))
                            result.combine(TestResult.PASS)
                        else:
                            result.combine(TestResult.FAIL)
                            self.statusMessage("\nError: The test device {} found to be still connected."
                                               "\nFailure: The hotplug test failed.")
                    else:
                        result.combine(TestResult.FAIL)
                        self.statusMessage("\nError: User fault, The test device {} was not removed."
                                           "\nFailure: The hotplug test failed.")
                else:
                    result.combine(TestResult.FAIL)
                    self.statusMessage("\nError: User fault, The test device was not inserted."
                                       "\nFailure: The hotplug test failed.")
            else:
                result.combine(TestResult.FAIL)
                self.statusMessage("\nError: User fault, skipped testing the socket."
                                   "\nFailure: The hotplug test failed.")

            devices = list() # clean up the session for next socket test
            if self.test_port <= self.numberOfSockets:
                self.test_port += 1 # loop control
        return result

    def run(self):
        result = self.run_hotplug_test()
        return result

if __name__ == "__main__":
    test_thunderbolt = ThunderboltTest()
    returnValue = test_thunderbolt.do(sys.argv)
    sys.exit(returnValue)
