#!/usr/bin/env python
# Copyright (c) 2006 Red Hat, Inc. All rights reserved. This copyrighted material
# is made available to anyone wishing to use, modify, copy, or
# redistribute it subject to the terms and conditions of the GNU General
# Public License v.2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# Author: Greg Nichols
#
import os, sys, re

from rhcert.command import Command
from rhcert.tags import Constants

from rhcert.suites.hwcert.hwcert.device import LogDevice
from rhcert.suites.hwcert.hwcert.usbTest import USBTestBase
from rhcert.suites.hwcert.hwcert.hardwareTest import HardwareTest

class USBTest(USBTestBase):

    def __init__(self, path):
        path = "usbbase/" + path
        USBTestBase.__init__(self, path)
        self.requireVersion = 0 # no required version (unknown)
        self.gen = None
        self.genSpeeds = {1: "5000M", 2: "10000M"}

    def getRequiredRPMs(self):
        return ["usbutils"]

    def plan(self, planner):
        tests = list()
        devices = planner.getUdevDeviceDetector().getDevices()
        # run if there's a device with SUBSYSTEM == usb
        for device in devices:
            if device.getProperty("SUBSYSTEM") == "usb":
                (rootHubs, rootHubsByID) = self.getRootHubs()
                id  = "%s:%s" % (device.getProperty("ID_VENDOR_ID"), device.getProperty("ID_MODEL_ID"))
                if id in rootHubsByID and self.requireVersion == rootHubsByID[id].getProperty(Constants.major_version):
                        test = self.makeCopy()
                        test.setDevice(device)
                        tests.append(test)
                        # only plan one test no matter how many are found
                        break
        return tests

    def checkReportedSpeed(self, bus, devNum):
        return True

    def determineGen(self):
        pass

    def getLSUSBDevices(self):
        pipe = os.popen("lsusb")
        result = False
        devices = list()
        # lsusb output is expected to look like this
        # Bus 001 Device 001: ID 1d6b:0002 Linux Foundation 2.0 root hub
        pattern = re.compile("Bus (?P<bus>[0-9]+) Device (?P<device>[0-9]+): ID (?P<id>[^\ ]+) (?P<product>.+)$")
        while True:
           line = pipe.readline()
           if not line:
               break
           match = pattern.search(line)
           if match:
               properties = dict()
               for attribute in [Constants.bus, Constants.device, Constants.id, Constants.product]:
                   properties[attribute] = match.group(attribute)
               devices.append(LogDevice(properties))

        return devices

    def getRootHubs(self):
        devices = self.getLSUSBDevices()
        rootHubsByID = dict() # indexed by ID
        rootHubs = list()
        pattern = re.compile("Linux Foundation (?P<major>[0-9]+).(?P<minor>[0-9]+) root hub")
        for device in devices:
            match = pattern.search(device.getProperty(Constants.product))
            if match:
                try:
                    device.setProperty(Constants.major_version, int(match.group("major")))
                    device.setProperty(Constants.minor_version, int(match.group("minor")))
                    rootHubsByID[device.getProperty(Constants.id)] = device
                    rootHubs.append(device)
                except ValueError:
                    print("Warning: could not parse major/minor bus version")
        return (rootHubs, rootHubsByID)

    def getRootHub(self, bus):
        (rootHubs, rootHubsByID) = self.getRootHubs()
        for rootHub in rootHubs:
            try:
                if int(rootHub.getProperty(Constants.bus)) == int(bus):
                    return rootHub
            except ValueError:
                pass
        return None

    def logDevices(self, planner):
        devices = self.getLSUSBDevices()
        # call the default to log udev devices
        devices.extend(HardwareTest.logDevices(self, planner))
        return devices

    def runPlugTest(self):
        # ask user for the number of ports:
        print("USB test:")
        self.pluggedPorts = list()

        self.checkAllUSBPortsUnplugged()
        self.fixedDevices = len(self.pluggedInDevices)
        if self.getMode() == Constants.auto:
            print("Error: the usb test can not be run in auto mode.")
            self.result = Constants.FAIL
            return False

        versionString = ""
        if self.requireVersion:
            versionString = "%s" % self.requireVersion

        self.numberOfSockets = self.promptInteger("How many USB%s sockets are required to be tested? " % versionString)
        if self.numberOfSockets < 1:
            print("No USB sockets to test")
            self.result = Constants.PASS
            return True

        # otherwise
        while len(self.pluggedInDevices) - self.fixedDevices < self.numberOfSockets:
            print("testing socket %s of %s..." % (len(self.pluggedInDevices)-self.fixedDevices+1, self.numberOfSockets))
            if self.promptContinue("Please plug in a USB%s device" % versionString):
                validFoundItem = False
                # find out which port was plugged
                for found in self.addNewPluggedUSBPorts():
                    (busID, devNUM, name, deviceID) = found
                    # when not required checkReportedSpeed (like in usb2) simply returns True in every call
                    if self.checkRootHub(busID) and self.checkReportedSpeed(busID, devNUM):
                        print("found inserted device on expected hub version meeting speed criteria")
                        validFoundItem = True
                        break

                if not validFoundItem:
                    if self.promptContinue("Please unplug the device"):
                        continue
                    else:
                        return False

                if self.promptContinue("Please unplug the device"):
                    if self.confirmUSBPortUnplugged(searchID=deviceID):
                        print("confirmed device %s (%s) unplugged" % (name, deviceID))
                        self.pluggedInDevices.append(deviceID)
                    else:
                        print("Did not confirm the device - repeating test.")
            else:
                self.result = Constants.ABORT
                return False

        if len(self.pluggedInDevices) - self.fixedDevices == self.numberOfSockets:
            self.result = Constants.PASS
            return True

        # otherwise
        self.result = Constants.FAIL
        return False

    def checkRootHub(self, busID):
        rootHub = self.getRootHub(busID)
        if not rootHub:
            if self.requireVersion:
                print("Error: could not find version of connected hub for bus ID %s" % busID)
                return False
            # otherwise:
            print("    could not find connected root hub")
            return True

        # otherwise
        print("    connected to root hub %s %s" % (rootHub.getProperty(Constants.id), rootHub.getProperty(Constants.product)))
        if self.requireVersion:
            if rootHub.getProperty(Constants.major_version) == self.requireVersion:
                print("Success: Plugged device is connected to a version %s root hub." % rootHub.getProperty(Constants.major_version))
                print("test version: %s" % self.requireVersion)
                return True
            else:
                print("Warning: Plugged device is connected to a version %s root hub." % rootHub.getProperty(Constants.major_version))
                print("test version: %s" % self.requireVersion)
                return False

    def run(self):

        # start testing
        success = True

        # determine Gen (if appropriate)
        self.determineGen()

        Command("lsusb -t").echo()

        if not self.runSubTest(self.runPlugTest, "plug", "run hot-plug/unplug test"):
            success = False

        return success

class USB2Test(USBTest):
    def __init__(self):
        USBTest.__init__(self, "usb2")
        self.requireVersion = 2

class USB3Test(USBTest):
    def __init__(self):
        USBTest.__init__(self, "usb3")
        self.requireVersion = 3
        self.gen = 1

    def checkReportedSpeed(self, bus, devNum):
        if not self.gen:
            return False
        expectedSpeed = self.genSpeeds[self.gen]
        busRegex = ".*?Bus *(?P<lineBus>\d+)"
        restRegex = ".*?Dev *(?P<lineDev>\d+).*?(?P<lineSpeed>\d+M)"
        output = Command("lsusb -t").getStringList(regex=".*")
        currentBus = ""
        for line in output:
            m = re.match(busRegex, line)
            if m:
                lineBus = m.group("lineBus")
                if lineBus:
                    currentBus = lineBus
            if int(bus) != int(currentBus):
                continue
            m = re.match(restRegex, line)
            if int(devNum) != int(m.group("lineDev")):
                continue
            if expectedSpeed == m.group("lineSpeed"):
                print("Line:\n%s\nMatches expected reported speed of: %s" % (line, expectedSpeed))
                return True
            else:
                print("Line:\n%s\nDoes does not match expected reported speed of: %s" % (line, expectedSpeed))
        return False

    def determineGen(self):
        genOptions = ["1","2"]
        self.gen = self.promptEdit("Which gen?", genOptions[0], genOptions)
        self.gen = int(self.gen)

if __name__ == "__main__":
    
    usbTest = USBTest()
    # usbTest.udi = "/org/freedesktop/Hal/devices/usb_device_0_0_0000_00_1d_2_if0"
    # usbTest.run()
    returnValue = usbTest.do(sys.argv)
    sys.exit(returnValue)
