#!/usr/bin/env python
# Copyright (c) 2006 Red Hat, Inc. All rights reserved. This copyrighted material
# is made available to anyone wishing to use, modify, copy, or
# redistribute it subject to the terms and conditions of the GNU General
# Public License v.2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.#
# Author: Irina Boverman
#

import os, re, sys, platform, signal, time
import io

from rhcert.command import Command, HwCertCommandException
from rhcert.tags import DeviceClass, TestTag
from rhcert.redhatrelease import RedHatRelease
from rhcert.compatability import rhcert_get_status_output

from rhcert.suites.hwcert.hwcert.hardwareTest import HardwareTest

class VideoTest(HardwareTest):

    def __init__(self):
        HardwareTest.__init__(self, "video")

        self.Xconfig_flag = ""
        self.depth = 0
        self.interface = "video-card"
        self.display = "2"
        self.Xconfigfile = "/tmp/hwcertXconfig"
        self.deviceClass = DeviceClass.display
        self.redHatRelease = RedHatRelease()

    def tags(self):
        return [TestTag.noninteractive, TestTag.certification]

    def plan(self, planner):
        properties = dict()
        properties["PCI_CLASS"] = "30000"
        tests = self.planWith(planner.getUdevDeviceDetector(), properties, allDevices=False, strict=True)
        return tests

    def getRequiredRPMs(self):
        rpms = list()
        rpms.append("xorg-x11-utils") # for xdpyinfo and xvinfo
        rpms.append("glx-utils") # for xdriinfo and glxinfo, and glxgears
        rpms.append("xorg-x11-server-Xorg") # for X
        rpms.append("xorg-x11-xinit") # for xinit
        rpms.append("xterm") # for xterm
        rpms.append("mesa-dri-drivers") # for openGL

        return rpms

    def getConfigFile(self):
        self.Xconfigfile = None
        logPath = "/var/log/Xorg.0.log"
        if os.path.exists(logPath):
            log = io.open(logPath, encoding='utf-8', errors='ignore')
            pattern = re.compile("\(==\) Using config file: \"(?P<filepath>[^\"]+)\"")
            for line in log.readlines():
                match = pattern.search(line)
                if match:
                    self.Xconfigfile = match.group("filepath")
                    print("Using config file " + self.Xconfigfile)
                    break
            log.close()
        if not self.Xconfigfile:
            print("No config file found - using Default Config")
        return True

    def setDepth(self):
        if self.Xconfigfile:
            (status, self.depth) = rhcert_get_status_output("grep DefaultDepth " + self.Xconfigfile +
                                                        " | awk ' { print $2; } '")
            if status != 0:
                print("Error: could not obtain default depth from config file")
                return False
            try:
                self.depth = int(self.depth)
                print("Depth set to %u" % self.depth)
            except:
                print("Error: could not convert depth %s to integer" % self.depth)
                self.depth = None
                return False
        return True

    def setFlag(self):
        # Adapt for flag changes
        (status, self.Xconfig_flag) = rhcert_get_status_output("X --help 2>&1 | grep xf86config -q && " +
                                                               "echo -xf86config || echo -config")
        if status != 0:
            print("FAILED to obtain config flag")
            return False
        print("Depth flag: " + self.Xconfig_flag)
        return True

    def checkDepthAndResolution(self):
        success = True
        # try to get the current screen resolution and color depth first
        try:
            # inclue display flag to ensure it is set
            xdpyinfo = Command("xdpyinfo -display :%s" % self.display)
            xdpyinfo.run()
            for line in xdpyinfo.output:
                if "dimensions:" in line:
                    screenResolution = line.split(" ")[6].split("x")
                    horizontalResolution = screenResolution[0]
                    verticalResolution = screenResolution[1]
                if "depth of root window:" in line:
                    colorDepth = line.split(" ")[9]
            print("\nScreen resolution: %sx%s" % (horizontalResolution,verticalResolution))
            print("      Color depth: %s" % colorDepth)
            # minimum requirement is 1024x768@24bit, but on some wide-screen
            # systems, the vertical resolution maybe smaller than 768
            if int(horizontalResolution) < 1024:
                print("Error: Current screen resolution: %sx%s does not meet the minimum requirements !\n" % (horizontalResolution,verticalResolution))
                success = False
            if int(colorDepth) < 24:
                print("Error:  Current color depth: %s does not meet the minimum requirements !\n" % colorDepth)
                success = False
        except:
            # record error and continue
            print("Error: failed to obtain current screen resolution and color depth !")
            success = False

        return success

    def checkXModules(self, display):
        logPath = "/var/log/Xorg.%s.log" % display
        try:
            log = io.open(logPath, "r", encoding='utf-8', errors='ignore')
        except IOError:
            print("Error: could not open %s" % logPath)
            return False

        success = True
        print("Checking %s for loaded X modules -----------------------" % logPath)
        for line in log.readlines():
            # this presumes Loading lines that include a slash means a path
            if line.find("(II) Loading /") >= 0:
                 # is there a oneline way to do this?
                 xmoduleFile = line.split()
                 xmoduleFile = xmoduleFile[-1]

                 try:
                     Command("rpm -Vf %s" % xmoduleFile).run()
                     if not self.checkXModuleVendorAndBuildHost(xmoduleFile):
                         success = False
                 except Exception as e:
                     print("Error: RPM verification failed for X module %s" % xmoduleFile)
                     print(e)
                     success = False

        print("-------------------------------------------------\n")
        log.close()
        return success

    def checkXModuleVendorAndBuildHost(self, xmoduleFile):
        # Maybe this should be in a library; this function lifted from info.py
        result = True
        goodVendorList = ["Red Hat", "Red Hat, Inc."]
        warnVendorList = ["Fedora", "Fedora Project"]
        try:

            vendor = Command("rpm -qf %s --qf %%{VENDOR}" % xmoduleFile).getString()
            if vendor in goodVendorList:
                print("Found %s by %s" % (xmoduleFile, vendor))
            elif vendor in warnVendorList:
                print("Warning: %s module %s found" % (vendor, xmoduleFile))
            else:
                print("Error: Non-Red Hat vendor %s for module %s" % (vendor, xmoduleFile))
                result = False

            buildhost = Command("rpm -qf %s --qf %%{BUILDHOST}" % xmoduleFile).getString()
            if "redhat.com" not in buildhost:
                print("Error: X module %s was built on %s and not built at Red Hat." % (xmoduleFile, buildhost))
                result = False
        except HwCertCommandException as exception:
                print("Error: could not determine X module packager for  %s" % xmoduleFile)
                print(exception)
                result = False

        return result

    def logDriverInfo(self, display):
        libDir = "/usr/lib64" if platform.architecture()[0] == "64bit" else "/usr/lib"
        logPath = "/var/log/Xorg.%s.log" % display
        try:
            log = io.open(logPath, "r", encoding='utf-8', errors='ignore')
        except IOError:
            print("Error: could not open %s" % logPath)
            return False
        print("\nDriver Info from %s:\n-------------------------------------------------" % logPath)
        copying = False
        entryCount = 0
        while 1:
            line = log.readline()
            if line:
                if not copying:
                    if line.find("%s/xorg/modules/drivers" % libDir) >= 0:
                        copying = True
                # stop on second "(II)" or blank line
                else:
                    if line.find("(II)") >= 0:
                        entryCount += 1
                    if line.strip() == "" or entryCount > 1:
                        copying = False
                        entryCount = 0
                        print("")
                if copying:
                    sys.stdout.write(line)
            else:
                log.close()
                break
        print("-------------------------------------------------")

        result = True
        for command in ["xvinfo", "xdriinfo", "glxinfo"]:
            info_command = "%s -display :%s" % (command, display)
            print("\n---------------------- %s ------------------" % info_command)
            try:
                Command(info_command).echoIgnoreWarnings()
            except HwCertCommandException:
                if command != "xvinfo":  # xvinfo always returns 1
                    result = False

        return result

    def startX(self):
        """ start an X server with gnome-shell """
        # Note: "xinit /usr/bin/gnome-shell --replace -- :%s " won't work as root
        startxCommand = "xinit -- :%s " % self.display
        if self.Xconfigfile:
            startxCommand += "%s %s " % (self.Xconfig_flag, self.Xconfigfile)
        if self.depth:
            startxCommand +=  "-depth %u " % self.depth

        try:
            print("Running: " + startxCommand)
            self.xServer = Command(startxCommand)
            self.xServer.start()
            print("process %s" % self.xServer.getPID())
            time.sleep(3)
        except HwCertCommandException as e:
            print("Error: %s failed" % startxCommand)
            print(e)
            print(self.xServer.printErrors())
            return False
        # otherwise
        return True

    def runGLXGears(self):
        glxGears = Command("glxgears -display :%s" % self.display)
        print(glxGears.command)
        try:
            glxGears.start()
            lines = 0
            while lines < 5:
                returnCode = glxGears.poll()
                if returnCode:
                    print("Error: glxgears failed, return code: %s" % returnCode)
                    glxGears.printOutput()
                    glxGears.printErrors()
                    return False
                line = glxGears.readline()
                if line:
                    print(line)
                    lines = lines + 1
            glxGears.pipe.kill()
        except Exception as e:
            print("Error: glxgears failed:")
            print(e)
            return False
        return True

    def stopX(self):
        try:
            print("Killing xinit PID %s" % self.xServer.getPID())
            os.kill(self.xServer.getPID(), signal.SIGTERM)
        except Exception as e:
            print("Error: could not kill X:")
            print(e)
            return False
        return True


    def checkConnections(self):
        sys.stdout.write("Checking for displays...")
        xrandr = Command("xrandr")
        try:
            xrandr.echo()
            sys.stdout.flush()
            # remove the check, as it  is unrelable
            # xrandr.getString(" connected", singleLine=False)
            # print "   done."
            # sys.stdout.flush()
        except HwCertCommandException as e:
            # print "\nError: no displays connected"
            print(e)
            # return FAILED

        return True

    def setConfiguration(self):
        return (self.getConfigFile() and self.setDepth() and self.setFlag())

    def runXServerTest(self):
        success = True
        if not self.startX():
            success = False
        if not self.runGLXGears():
            success = False
        if not self.stopX():
            success = False

        return success

    def logModulesAndDrivers(self):
        time.sleep(3)
        self.startX()
        success = self.checkDepthAndResolution()
        if not self.checkXModules(self.display):
            success = False
        if not self.logDriverInfo(self.display):
            success = False
        self.stopX()

        return success

    def run(self):
        print("Starting video test")
        print("on device udi: %s" % self.getUDI())
        success = True
        if not self.runSubTest(self.checkConnections, name="Check Connections", description="Log xrandr output"):
            success = False
        if not self.runSubTest(self.setConfiguration, name="Set Configuration", description="Set up configuration for test"):
            return False

        if not self.runSubTest(self.runXServerTest, name="The X Server Test", description="Run a second X test server"):
            success = False
        if not self.runSubTest(self.logModulesAndDrivers, name="Log Module and Drivers", description="Log depth, module and driver information"):
            success = False

        if success:
            print("Video test on device %s passed." % (self.interface))
            return True
        # otherwise
        print("Video test on device %s FAILED." % (self.interface))
        return False

if __name__ == "__main__":
    videoTest = VideoTest()
    returnValue = videoTest.do(sys.argv)
    sys.exit(returnValue)
