# Copyright (c) 2006 Red Hat, Inc. All rights reserved. This copyrighted material
# is made available to anyone wishing to use, modify, copy, or
# redistribute it subject to the terms and conditions of the GNU General
# Public License v.2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# Author: Greg Nichols
#


from rhcert.tags import DeviceClass
from rhcert.test import TestTag
from rhcert.command import Command, HwCertCommandException

from rhcert.suites.hwcert.hwcert.networkTest import NetworkTest
from rhcert.suites.hwcert.hwcert.wireless import Wireless


class WLanTest(NetworkTest):

    def __init__(self, path=None):
        if path:
            path = "wlan/" + path
        else:
            path = "wlan"
        NetworkTest.__init__(self, path)

        self.requiredSupport = None
        self.procNetDev = "/proc/net/dev"
        self.deviceClass = DeviceClass.network_interface

    def getRequiredRPMs(self):
        return ["rfkill", "lshw"]

    def tags(self):
        return [TestTag.noninteractive, TestTag.certification, TestTag.network, TestTag.wlan]

    def plan(self, planner):
        # get the wireless device name
        wireless = Wireless()
        interfaces = wireless.getInterfaces()
        tests = dict()
        testList = list()
        if interfaces:
            devices = self.getNetworkDevices(planner)
            for (logicalDevice, device) in devices.items():
                if logicalDevice in interfaces and self.isBestType(wireless.getType(logicalDevice)):
                    test = self.makeCopy()
                    test.setDevice(device)
                    test.setLogicalDeviceName(logicalDevice)
                    tests[logicalDevice] = test
            testList = list()
            testList.extend(tests.values())
        return testList

    def isBestType(self, supportedTypes):
        """ return True if this type is the best in the list """
        if supportedTypes:
            for type in ["ax", "ac", "n", "g"]:
                if type in supportedTypes:
                    return type == self.requiredSupport
        # if ax, ac, n or g is not supported and we don't require support, we're the best
        return not self.requiredSupport

    def verify(self):
        try:
            log = Command("iw %s link" % self.getLogicalDeviceName())
            log.run()
            return True
        except HwCertCommandException as exception:
            print("Error: get configuration for %s" % self.procNetDev)
            print(exception)
            exception.command.printErrors()

        return False

    def bringUpDevice(self):
        try:
            command = Command("%s %s"  % (self.interfaceConnect, self.getLogicalDeviceName()))
            command.echo()
        except HwCertCommandException as exception:
            print("Warning: could not bring up %s" % self.getLogicalDeviceName())
            print(exception)
            return False

        return True


    # 1. output of /proc/net/Dev (consistent across RHEL versions) should provide wireless interfaces details.
    def logProcNetDev(self):
        try:
            log = Command("cat " + self.procNetDev)
            log.echo()
            return True
        except HwCertCommandException as exception:
            print("Error: could not log %s" % self.procNetDev)
            print(exception)
            exception.command.printErrors()

        return False

    # 2. following command should provide list of access points in scan range of
    # <dev> and properties of AP
    def scan(self):
        commands = ["iw %s link" % self.getLogicalDeviceName(), "iw %s scan" % self.getLogicalDeviceName()]
        result = True
        for command in commands:
            try:
                scan = Command(command)
                scan.echo()
            except   HwCertCommandException as exception:
                print("Error: %s:" % command)
                print(exception)
                exception.command.printErrors()
                result = False

        return result

    # 3. Following iw commands are useful for simple outputs
    def iw(self):
        commands = ["iw %s link" % self.getLogicalDeviceName(), "iw %s info" % self.getLogicalDeviceName()]
        result = True
        for command in commands:
            try:
                iw = Command(command)
                iw.echo()
            except   HwCertCommandException as exception:
                print("Error: %s:" % command)
                print(exception)
                exception.command.printErrors()
                result = False
        return result

    def checkSupport(self):
        wireless = Wireless()
        if self.requiredSupport in wireless.getType(self.getLogicalDeviceName()):
            print("Wireless %s is supported." % self.requiredSupport)
            return True
        # otherwise
        print("Error: wireless %s is not supported by %s" % (self.requiredSupport, wireless.getType(self.getLogicalDeviceName())))
        return False

    def run(self):
        """
        """
        result = True
        if not self.runSubTest(self.logProcNetDev, "Log proc", "log " + self.procNetDev):
            result = False

        if not self.runSubTest(self.scan, "Scan Wireless Networks"):
            result = False

        if not self.runSubTest(self.iw, "Get wireless connection information"):
            result = False

        if self.requiredSupport and not self.runSubTest(self.checkSupport, "Check Support"):
            result = False

        if NetworkTest.run(self) != True:
            result = False

        return result

class WirelessGTest(WLanTest):

    def __init__(self):
        WLanTest.__init__(self, path="WirelessG")
        self.requiredSupport = "g"
        self.enforceSupport = True
        self.interfaceSpeed = 22 # speed in Mb/sec
        self.enforceSpeed = True

class WirelessNTest(WLanTest):

    def __init__(self):
        WLanTest.__init__(self, path="WirelessN")
        self.requiredSupport = "n"
        self.enforceSupport = True
        self.interfaceSpeed = 100 # speed in Mb/sec
        self.enforceSpeed = True

class WirelessACTest(WLanTest):

    def __init__(self):
        WLanTest.__init__(self, path="WirelessAC")
        self.requiredSupport = "ac"
        self.enforceSupport = True
        self.interfaceSpeed = 300 # speed in Mb/sec
        self.enforceSpeed = True

class WirelessAXTest(WLanTest):

    def __init__(self):
        WLanTest.__init__(self, path="WirelessAX")
        self.requiredSupport = "ax"
        self.enforceSupport = True
        self.interfaceSpeed = 1200 # speed in Mb/sec
        self.enforceSpeed = True
